#!/bin/sh 

EXEC_NAME=klayout

PWD=`pwd`
USOURCE=$PWD/src

# On cygwin, provide a windows path in SOURCE 
if cygpath -m $PWD >/dev/null 2>/dev/null; then
  WIN_CYGWIN=1
  SOURCE=`cygpath -m $PWD`/src 
else
  WIN_CYGWIN=0
  SOURCE=$PWD/src
fi

QTDIR="/usr/lib/qt4"
PLATFORM="linux-32-gcc-release"

MAKE_OPT=""

while [ "$*" != "" ]; do
  a=$1
  shift
  
  case $a in 
  -qt) 
    QTDIR=$1
    shift
    ;;
  -platform) 
    PLATFORM=$1
    shift
    ;;
  -option) 
    MAKE_OPT="$MAKE_OPT $1"
    shift
    ;;
  -help)
    echo "usage: build [options]"
    echo ""
    echo "options:"
    echo "  -qt <dir>         Set Qt installation directory (default '/usr/lib/qt')"
    echo "  -platform <key>   Select a platform"
    echo "  -option <option>  'make' options (i.e. -j2)"
    echo ""
    echo "Available platforms are:"
    for pf in `ls -1 config/*.conf.*`; do
      echo $pf | sed "s/config\/Makefile\.conf\./  /"
    done
    exit 0
    ;;
  *)
    echo "*** ERROR: unknown option $a (use -help to show options)"
    exit 1
    ;;
  esac

done

echo "Qt installation directory is $QTDIR"
echo "Platform is $PLATFORM"

gcc_path=`type gcc 2>/dev/null`
if [ "x$gcc_path" = "x" ]; then
  echo "*** ERROR: unable to locate gcc in path"
  exit 1
fi

if [ ! -r config/Makefile.conf.$PLATFORM ]; then
  echo "*** ERROR: unknown platform $PLATFORM"
  exit 1
fi

if [ ! -r $QTDIR/include/QtCore/Qt ]; then
  echo "*** ERROR: no Qt headers in $QTDIR/include"
  exit 1
fi

if [ ! -r $QTDIR/bin/moc ]; then
  echo "*** ERROR: unable to locate Qt MOC in $QTDIR/bin"
  exit 1
fi

if [ ! -r $QTDIR/bin/uic ]; then
  echo "*** ERROR: unable to locate Qt UIC in $QTDIR/bin"
  exit 1
fi

BUILD=$PWD/build.$PLATFORM
BIN=$PWD/bin.$PLATFORM

if ( gmake -v >/dev/null 2>/dev/null ); then
  MAKE_PRG=gmake
else
  MAKE_PRG=make
fi

mkdir -p $BUILD
cp -f src/Makefile.top $BUILD/Makefile
cp -f config/Makefile.conf.$PLATFORM $BUILD/Makefile.conf

mkdir -p $BUILD/main
echo "TOP_SOURCE=$USOURCE" >$BUILD/main/Makefile
echo "SOURCE=$SOURCE" >>$BUILD/main/Makefile
echo "USOURCE=$USOURCE" >>$BUILD/main/Makefile
echo 'include ../Makefile.env' >>$BUILD/main/Makefile
echo 'include ../Makefile.conf' >>$BUILD/main/Makefile
echo 'include $(SOURCE)/Makefile.body' >>$BUILD/main/Makefile

PLUGINS=""
if [ -d src ] && [ -d src/plugins ]; then
  cd src/plugins
  PLUGINS=`echo *`
  for plugin in $PLUGINS; do
    mkdir -p $BUILD/$plugin
    echo "TOP_SOURCE=$USOURCE" >$BUILD/$plugin/Makefile
    echo "SOURCE=$SOURCE/plugins/$plugin" >>$BUILD/$plugin/Makefile
    echo "USOURCE=$USOURCE/plugins/$plugin" >>$BUILD/$plugin/Makefile
    echo 'include ../Makefile.env' >>$BUILD/$plugin/Makefile
    echo 'include ../Makefile.conf' >>$BUILD/$plugin/Makefile
    echo 'include $(SOURCE)/Makefile.body' >>$BUILD/$plugin/Makefile
  done
  cd ../..
fi

echo "QTDIR=$QTDIR" >$BUILD/Makefile.env
echo "PLUGINS=$PLUGINS" >>$BUILD/Makefile.env

echo "Building plugins: $PLUGINS"
echo ""

# -- Running build  --
echo "Running build .."
echo "$MAKE_PRG"
cd $BUILD
$MAKE_PRG $MAKE_OPT all || exit 1
echo ""

# -- Installing binaries  --

if [ $WIN_CYGWIN = 0 ]; then

  echo "Installing binaries .."
  mkdir -p $BIN
  cp $BUILD/main/$EXEC_NAME $BIN
  for plugin in $PLUGINS; do
    cp $BUILD/$plugin/*.so $BIN
  done
  chmod 755 $BIN/*
  echo ""
  
  echo "Build done."
  echo ""
  echo "Final binary is ready in $BIN/$EXEC_NAME."
  echo "To run it, you may need to add the following to your LD_LIBRARY_PATH:"
  echo "$QTDIR/lib"
  
  LD_LIBRARY_PATH=$QTDIR/lib:$LD_LIBRARY_PATH
  $BIN/$EXEC_NAME -v
  $BIN/$EXEC_NAME -h

else 

  echo "Installing binaries .."
  mkdir -p $BIN
  cp $BUILD/main/$EXEC_NAME{.exe,.dll} $BIN
  for plugin in $PLUGINS; do
    cp $BUILD/$plugin/*.dll $BIN
  done
  chmod 755 $BIN/*
  echo ""
  
  echo "Build done."

fi

exit 0

