
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "antPropertiesPage.h"
#include "layLayoutView.h"

namespace ant
{

// -------------------------------------------------------------------------
//  helper function to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

// -------------------------------------------------------------------------
//  PropertiesPage implementation

PropertiesPage::PropertiesPage (ant::Service *rulers, double dbu, QWidget *parent)
  : lay::PropertiesPage (parent, rulers), m_dbu (dbu), mp_rulers (rulers), m_enable_cb_callback (true)
{
  mp_rulers->get_selection (m_selection);
  m_pos = m_selection.begin ();

  setupUi (this);

  connect (dbu_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));

  mp_rulers->clear_highlights ();
}

PropertiesPage::~PropertiesPage ()
{
  mp_rulers->restore_highlights ();
}

void 
PropertiesPage::back ()
{
  m_pos = m_selection.end ();
}

void 
PropertiesPage::front ()
{
  m_pos = m_selection.begin ();
}

const ant::Object &
PropertiesPage::current () const
{
  const ant::Object *ruler = dynamic_cast <const ant::Object *> ((*m_pos)->ptr ());
  return *ruler;
}

bool 
PropertiesPage::at_begin () const
{
  return (m_pos == m_selection.begin ());
}

bool 
PropertiesPage::at_end () const
{
  return (m_pos == m_selection.end ());
}

void 
PropertiesPage::operator-- ()
{
  --m_pos;
}

void 
PropertiesPage::operator++ ()
{
  ++m_pos;
}

void
PropertiesPage::leave ()
{
  mp_rulers->clear_highlights ();
}

void
PropertiesPage::display_mode_changed (bool)
{
  if (! m_enable_cb_callback) {
    return;
  }

  mp_rulers->view ()->dbu_coordinates (dbu_cb->isChecked ());
  update ();
}

void 
PropertiesPage::update ()
{
  mp_rulers->highlight (std::distance (m_selection.begin (), m_pos));

  m_enable_cb_callback = false;
  dbu_cb->setChecked (mp_rulers->view ()->dbu_coordinates ());
  m_enable_cb_callback = true;

  bool du = dbu_cb->isChecked ();

  x1->setText (coord_to_string (current ().p1 ().x (), m_dbu, du).c_str ());
  x1->setCursorPosition (0);
  x2->setText (coord_to_string (current ().p2 ().x (), m_dbu, du).c_str ());
  x2->setCursorPosition (0);
  y1->setText (coord_to_string (current ().p1 ().y (), m_dbu, du).c_str ());
  y1->setCursorPosition (0);
  y2->setText (coord_to_string (current ().p2 ().y (), m_dbu, du).c_str ());
  y2->setCursorPosition (0);

  double sx = (current ().p2 ().x () - current ().p1 ().x ());
  double sy = (current ().p2 ().y () - current ().p1 ().y ());
  dx->setText (coord_to_string (sx, m_dbu, du).c_str ());
  dx->setCursorPosition (0);
  dy->setText (coord_to_string (sy, m_dbu, du).c_str ());
  dy->setCursorPosition (0);
  dd->setText (coord_to_string (sqrt (sx * sx + sy * sy), m_dbu, du).c_str ());
  dd->setCursorPosition (0);
}

bool 
PropertiesPage::readonly ()
{
  return false;
}

void 
PropertiesPage::apply ()
{
  double dx1, dy1, dx2, dy2;

  tl::from_string (x1->text ().ascii (), dx1);
  tl::from_string (x2->text ().ascii (), dx2);
  tl::from_string (y1->text ().ascii (), dy1);
  tl::from_string (y2->text ().ascii (), dy2);

  double f = dbu_cb->isChecked () ? 1.0 : 1.0 / m_dbu;

  ant::Object ruler (db::DPoint (dx1 * f, dy1 * f), db::DPoint (dx2 * f, dy2 * f), current ().id ());
  mp_rulers->change_ruler (*m_pos, ruler);
}

}

