
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbUserObject.h"
#include "dbClipboard.h"
#include "tlString.h"
#include "layPlugin.h"
#include "layRenderer.h"
#include "laySnap.h"
#include "layLayoutView.h"
#include "layLayoutCanvas.h"
#include "layProperties.h"
#include "antService.h"
#include "antPropertiesPage.h"
#include "antConfig.h"

namespace ant
{

// -------------------------------------------------------------
//  Functionality to draw a ruler object

static void
tick_spacings (double d, double min_d, int &minor_ticks, double &ticks)
{
  if (min_d > d) {

    minor_ticks = -1;
    ticks = -1;

  } else {

    const double log10 = log (10.0);

    //  as a safety measure, not too many ticks are created.
    min_d = std::max (min_d, 0.001 * d);

    double l1 = log (min_d) / log10;
    double l0 = floor (l1);
    l1 -= l0;

    double m = 10.0;
    if (l1 < 0.3) {
      minor_ticks = 5;
    } else if (l1 < 0.7) {
      minor_ticks = 2;
    } else {
      minor_ticks = 1;
    }

    ticks = exp (log10 * l0) * m;

  }
}

void 
draw_ruler (const ant::Object &ruler, const db::DMagTrans &trans, bool sel, lay::Bitmap *bitmap)
{
  if (ruler.p1 ().sq_double_distance (ruler.p2 ()) < 1e-10 /*sq. micron*/) {
    return;
  }

  lay::Renderer renderer;
  renderer.reserve_edges (100);

  //  round the starting point, shift both, and round the end point 
  std::pair <db::DPoint, db::DPoint> v = lay::snap (trans * ruler.p1 (), trans * ruler.p2 ());
  db::DPoint q1 = v.first;
  db::DPoint q2 = v.second;

  int min_tick_spc = 10;
  int tick_length = 8;

  double ticks;
  int minor_ticks;

  double dx = ruler.p1 ().x () - ruler.p2 ().x ();
  double dy = ruler.p1 ().y () - ruler.p2 ().y ();
  double vv = sqrt (dx * dx + dy * dy);
  tick_spacings (vv, double (min_tick_spc) / trans.ctrans (1.0), minor_ticks, ticks);

  //  normal vector
  db::DPoint qq (dy / vv, -dx / vv);

  if (! sel) {
    renderer.insert (db::DEdge (q1, q2));
  } else {
    db::DPoint qw = qq * 2.0;
    renderer.insert (db::DEdge (q1 + qw, q2 + qw));
    renderer.insert (db::DEdge (q2 + qw, q2 - qw));
    renderer.insert (db::DEdge (q2 - qw, q1 - qw));
    renderer.insert (db::DEdge (q1 - qw, q1 + qw));
    renderer.render_fill (*bitmap);
  }

  //  create three tick vectors in tv_text, tv_short and tv_long
  double tf = double (tick_length);
  if (qq.y () < 1e-6) {
    tf = -tf;
  }
  db::DPoint tv_text  = qq * tf * 1.2;
  db::DPoint tv_short = qq * tf * 0.5;
  db::DPoint tv_long  = qq * tf;

  bool text_left;
  if (ruler.p1 ().y () == ruler.p2 ().y ()) {
    text_left = ruler.p1 ().x () > ruler.p2 ().x () + 1e-6 /*pixels*/;
  } else {
    text_left = qq.x () * tf > 1e-6 /*pixels*/;
  }

  renderer.insert (db::DEdge (q1, q1 + tv_long));
  renderer.insert (db::DEdge (q2, q2 + tv_long));

  renderer.insert (db::DBox (q2 + tv_text, q2 + tv_text), 
                   tl::sprintf ("%.8g", vv), 
                   db::DefaultFont, 
                   (text_left ? db::HAlignLeft : db::HAlignRight), db::VAlignBottom,
                   db::fixpoint_trans::r0);

  if (minor_ticks > 0 && ticks > 0) {

    //  unit vector of snapped ruler
    db::DPoint q (q2.x () - q1.x (), q2.y () - q1.y ());
    if (q.distance () > 1e-6) {
      q *= 1.0 / q.distance ();
    }

    int nmax = int (floor (vv / (ticks / double (minor_ticks)) - 1e-6));
    for (int n = 1; n <= nmax; ++n) {

      double v = ticks * double (n) / double (minor_ticks);
      double d = trans.ctrans (v); 

      db::DPoint qq = q1 + q * d;
      qq = db::DPoint (floor (qq.x () + 0.5), floor (qq.y () + 0.5));

      if (n % minor_ticks == 0) {
        /* no tick label:
        if (n == minor_ticks && v < vv * 0.3) {
          renderer.insert (db::DBox (qq + tv_text, qq + tv_text), tl::sprintf ("%.4g", ticks), db::DefaultFont,
                           (text_left ? db::HAlignLeft : db::HAlignRight), db::VAlignBottom,
                           db::fixpoint_trans::r0);
        }
        */
        renderer.insert (db::DEdge (qq, qq + tv_long));
      } else {
        renderer.insert (db::DEdge (qq, qq + tv_short));
      }

    }

  }

  renderer.render_contour (*bitmap);
  renderer.render_texts (*bitmap);
}

void 
draw_ruler (const ant::Object &ruler, double dbu, const db::MagTrans &trans, bool sel, lay::Bitmap *bitmap)
{
  draw_ruler (ruler.transformed (db::DMag (dbu)), db::DMagTrans (trans.t1, db::DMag (trans.t2.factor () / dbu)), sel, bitmap);
}


// -------------------------------------------------------------

//  rulers always refer to the primary cell view 
unsigned int rulers_cvindex = 0;  // @@@ should be variable?

//  search range for select in pixels
unsigned int search_range = 5;   // @@@ make variable?

// -------------------------------------------------------------

class ContourFinder
{
public:
  ContourFinder (const db::DPoint &original, const std::vector <db::DEdge> &cutlines)
    : m_any (false), m_any_exact (false), 
      m_original (original), m_tests (100 /* max. number of tests, TODO: make variable? */),
      mp_layout (0), m_cutlines (cutlines), mp_prop_sel (0)
  {
    //  nothing yet.
  }
  
  void
  find (lay::LayoutView *view, int cv_index, const db::MagTrans &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, unsigned int l,
        const std::set<db::PropertiesRepository::properties_id_type> *prop_sel, bool inv_prop_sel)
  {
    m_region = region; 
    mp_layout = &layout;
    mp_prop_sel = prop_sel;
    m_inv_prop_sel = inv_prop_sel;

    do_find (view, cv_index, cell, l, view->min_hier_levels (), view->max_hier_levels (), trans);
  }

  const db::DPoint &get_found () const
  {
    if (m_any_exact) {
      return m_closest_exact;
    } else if (m_any) {
      return m_closest;
    } else {
      return m_original;
    }
  }

  bool any () const
  {
    return (m_any || m_any_exact) && m_tests > 0;
  }

private:
  void 
  find_closest_exact (const db::DPoint &p)
  {
    db::DPoint dp (p * mp_layout->dbu ());
    if (! m_any_exact || m_original.distance (dp) < m_original.distance (m_closest_exact)) {
      m_closest_exact = dp;
      m_any_exact = true;
    }
  }

  void 
  find_closest (const db::DPoint &p)
  {
    db::DPoint dp (p * mp_layout->dbu ());
    if (! m_any || m_original.distance (dp) < m_original.distance (m_closest)) {
      m_closest = dp;
      m_any = true;
    }
  }

  void closest (const db::DPoint &p) 
  {
    if (m_cutlines.empty ()) {

      find_closest_exact (p);

    } else {

      //  Test, if there is a crossing between a cut line and the vertical
      //  or horizonal lines through the given point. Take this as the
      //  test point.
      for (std::vector <db::DEdge>::const_iterator cl = m_cutlines.begin (); cl != m_cutlines.end (); ++cl) {
        std::pair<bool, db::DPoint> ret;
        ret = db::DEdge (p, p + db::DPoint (1.0, 0.0)).cut_point (db::Mag (1.0 / mp_layout->dbu ()) * *cl);
        if (ret.first) {
          find_closest_exact (ret.second);
        }
        ret = db::DEdge (p, p + db::DPoint (0.0, 1.0)).cut_point (db::Mag (1.0 / mp_layout->dbu ()) * *cl);
        if (ret.first) {
          find_closest_exact (ret.second);
        }
      }

    }
  }

  void closest (const db::DEdge &e)
  {
    bool any_point = false;

    //  do the checks in dbu space rather than micron space because
    //  the tolerances are set up for this.
    for (std::vector <db::DEdge>::const_iterator cl = m_cutlines.begin (); cl != m_cutlines.end (); ++cl) {
      std::pair<bool, db::DPoint> ret = e.cut_point (db::Mag (1.0 / mp_layout->dbu ()) * *cl);
      if (ret.first) {
        //  if the projection exactly hits the edge and the point
        //  of crossing is inside the search region, take this as
        //  the exact hit. Exact hits have priority over projected
        //  hits.
        if (e.contains (ret.second) && m_region.contains (ret.second)) {
          find_closest_exact (ret.second);
        } else {
          find_closest (ret.second);
        }
        any_point = true;
      }
    }

    if (! any_point) {
    
      //  no certain direction to look into:
      //  compute the projection of the point and if within a 
      //  certain range, take the projected point.

      db::DCoord d = std::max (m_region.width (), m_region.height ());
      db::DPoint v (e.p2 () - e.p1 ());
      db::DPoint n (-v.y (), v.x ());
      double f = d / n.double_distance ();

      db::DPoint e1 (m_original.x () / mp_layout->dbu () - n.x () * f,
                     m_original.y () / mp_layout->dbu () - n.y () * f);
      db::DPoint e2 (m_original.x () / mp_layout->dbu () + n.x () * f,
                     m_original.y () / mp_layout->dbu () + n.y () * f);
      
      double dmin (std::numeric_limits <db::Coord>::min ());
      double dmax (std::numeric_limits <db::Coord>::max ());
      db::DBox dworld (dmin, dmin, dmax, dmax);

      if (dworld.contains (e1) && dworld.contains (e2)) {

        std::pair<bool, db::DPoint> ip = e.intersect_point (db::DEdge (e1, e2));
        if (ip.first) { 
          find_closest (ip.second);
        }

      }

    }

  }

  void
  do_find (lay::LayoutView *view, int cv_index, const db::Cell &cell, unsigned int l, int min_level, int max_level, const db::MagTrans &t)
  {
    if (max_level <= 0 || ! (t * cell.bbox (l)).touches (m_region) || view->is_cell_hidden (cell.cell_index (), cv_index)) {
      return;
    }

    if (min_level <= 0) {

      const db::Shapes &shapes = cell.shapes (l);

      db::ShapeIterator shape = shapes.begin_touching (t.inverted () * m_region, db::ShapeIterator::Polygons | db::ShapeIterator::Paths | db::ShapeIterator::Boxes, mp_prop_sel, m_inv_prop_sel);
      while (! shape.at_end () && m_tests > 0) {

        --m_tests;

        if (shape->is_polygon ()) {

          for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e);
          }

        } else if (shape->is_path ()) {

          //  test the "spine"
          db::Shape::point_iterator pt = shape->begin_point (); 
          if (pt != shape->end_point ()) {
            db::Point p (*pt);
            ++pt;
            for (; pt != shape->end_point (); ++pt) {
              test_edge (t * db::Edge (p, *pt));
              p = *pt;
            }
          }

          //  convert to polygon and test those edges
          db::Polygon poly;
          shape->polygon (poly);
          for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e);
          }

        } else if (shape->is_box ()) {

          const db::Box &box = shape->box ();

          test_edge (t * db::Edge (box.p1 (), db::Point (box.left (), box.top ())));
          test_edge (t * db::Edge (db::Point (box.left (), box.top ()), box.p2 ()));
          test_edge (t * db::Edge (box.p2 (), db::Point (box.right (), box.bottom ())));
          test_edge (t * db::Edge (db::Point (box.right (), box.bottom ()), box.p1 ()));

        }

        ++shape;

      }

    }

    db::Box touch_box (t.inverted () * m_region); 
    db::box_convert <db::CellInst> bc (*mp_layout, l);

    db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
    while (! inst.at_end () && m_tests > 0) {

      const db::CellInstArray &cell_inst = *inst;
      for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, bc); ! p.at_end () && m_tests > 0; ++p) {

        --m_tests;

        do_find (view, cv_index, 
                 mp_layout->cell (cell_inst.object ().cell_index ()), l,
                 min_level - 1, max_level - 1,
                 t * db::MagTrans (*p, cell_inst.mag ()));

      }

      ++inst;

    }

  }

  void
  test_edge (const db::DEdge &edg)
  {
    //  we hit the region with this edge: vertex snap.
    if (m_region.contains (edg.p1 ())) {
      closest (edg.p1 ());
    }
    if (m_region.contains (edg.p2 ())) {
      closest (edg.p2 ());
    }
    
    //  if the edge cuts through the active region: test the
    //  edge as a whole
    if (edg.clipped (m_region).first) {
      closest (edg);
    }
  }

  bool m_any, m_any_exact;
  db::DPoint m_closest, m_closest_exact;
  db::DPoint m_original;
  int m_tests;
  const db::Layout *mp_layout;
  db::DBox m_region;
  std::vector <db::DEdge> m_cutlines;
  const std::set<db::PropertiesRepository::properties_id_type> *mp_prop_sel;
  bool m_inv_prop_sel;
};


// -------------------------------------------------------------

View::View (ant::Service *rulers, const ant::Object *ruler, double dbu, bool selected)
  : lay::ViewObject (rulers->canvas ()), 
    mp_rulers (rulers), m_selected (selected), mp_ruler (ruler), m_dbu (dbu)
{
  //  .. nothing else ..
}

View::~View ()
{
  //  .. nothing else ..
}

void 
View::displace (const db::DPoint &p)
{
  if (m_displacement != p) {
    m_displacement = p;
    redraw ();
  }
}

void 
View::ruler (const ant::Object *r)
{
  mp_ruler = r;
  redraw ();
}

void 
View::render () 
{ 
  if (! mp_ruler) {
    return;
  }

  QColor c (mp_rulers->color ());
  if (! c.isValid ()) {
    c = QColor (canvas ()->foreground_color ().rgb ());
  }

  //  obtain bitmap to render on
  lay::Bitmap *bitmap = canvas ()->bitmap (lay::ViewOp (c.rgb ()));
  if (! bitmap) {
    return;
  }

  draw_ruler (mp_ruler->transformed (db::Mag (m_dbu)).moved (m_displacement), canvas ()->trans (), m_selected, bitmap);
}

// -------------------------------------------------------------

Service::Service (db::Manager *manager, lay::LayoutView *view)
  : lay::ViewService (view->canvas ()), 
    lay::Editable (view),
    lay::Plugin (view),
    lay::Drawing (1/*number of bitmaps*/, view->drawings ()),
    db::Object (manager),
    m_snap_mode (Any),
    m_grid (0.001),
    m_grid_snap (false), m_obj_snap (false), m_snap_range (1),
    m_max_number_of_rulers (-1 /*unlimited*/),
    mp_view (view),
    mp_active_ruler (0),
    m_drawing (false), m_current (),
    m_move_mode (MoveNone)
{ 
  //  .. nothing yet ..
}

Service::~Service ()
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    delete *r;
  }
  m_rulers.clear ();
}

bool 
Service::configure (const std::string &name, const std::string &value)
{
  bool taken = true;

  if (name == cfg_ruler_color) {

    QColor color;
    lay::ColorConverter ().from_string (value, color);

    //  make the color available for the dynamic view objects too.
    if (lay::test_and_set (m_color, color)) {
      canvas ()->touch ();
    }

  } else if (name == cfg_ruler_grid_micron) {

    double g = 0;
    tl::from_string (value, g);
    m_grid = g;
    taken = false; // to let others use the grid too.

  } else if (name == cfg_max_number_of_rulers) {

    int n = -1;
    tl::from_string (value, n);
    if (n != m_max_number_of_rulers) {
      m_max_number_of_rulers = n;
      reduce_rulers (n);
    }

  } else if (name == cfg_ruler_snap_range) {

    int n = 0;
    tl::from_string (value, n);
    m_snap_range = n;

  } else if (name == cfg_ruler_obj_snap) {
    tl::from_string (value, m_obj_snap);
  } else if (name == cfg_ruler_grid_snap) {
    tl::from_string (value, m_grid_snap);
  } else if (name == cfg_ruler_snap_mode) {

    Service::SnapMode sm = Service::Any;
    SnapModeConverter ().from_string (value, sm);
    m_snap_mode = sm;

  } else {
    taken = false;
  }

  return taken;

}

void 
Service::config_finalize ()
{
  //  the changing of the view ops is done here since it may depend on the 
  //  background color which might be changed by another configure call later.
  std::vector <lay::ViewOp> view_ops;
  if (m_color.isValid ()) {
    view_ops.push_back (lay::ViewOp (m_color.rgb (), lay::ViewOp::Copy, 0));
  } else {
    view_ops.push_back (lay::ViewOp (canvas ()->foreground_color ().rgb (), lay::ViewOp::Copy, 0));
  }
  set_view_ops (view_ops);
}

void 
Service::clear_highlights ()
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    (*r)->visible (false);
  }
}

void 
Service::restore_highlights ()
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    (*r)->visible (true);
  }
}

void 
Service::highlight (unsigned int n)
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    (*r)->visible (n-- == 0);
  }
}

void 
Service::clear_rulers ()
{
  drag_cancel ();
  reduce_rulers (0);
}

void 
Service::drag_cancel () 
{
  if (m_drawing) {

    canvas ()->ungrab_mouse (this);

    if (mp_active_ruler) {
      delete mp_active_ruler;
      mp_active_ruler = 0;
    }

    m_drawing = false;

  }
}

void 
Service::insert_ruler (const db::DPoint &p1, const db::DPoint &p2, const lay::CellView &cv)
{
  //  determine the last id
  int idmax = -1;
  for (db::UserObjectLayer::iterator r = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()); r != cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).end (db::UserObject::tag ()); ++r) {
    const ant::Object *robj = dynamic_cast <const ant::Object *> (r->ptr ());
    if (robj) {
      if (robj->id () > idmax) {
        idmax = robj->id ();
      }
    }
  }

  //  create the ruler
  cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).insert (db::UserObject (new ant::Object (p1, p2, idmax + 1)));

  //  delete surplus rulers
  reduce_rulers (m_max_number_of_rulers);

  mp_view->redraw_deco_layer ();
}

bool 
Service::begin_move (bool sel, const db::DPoint &p)
{
  //  cancel any pending move or drag operations, reset mp_active_ruler
  canvas ()->drag_cancel (); // KLUDGE: every service does this to the same service manager

  if (mp_view->cellviews () <= rulers_cvindex) {
    return false;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return false;
  }
  double dbu = cv->layout ().dbu ();

  //  choose move mode
  if (sel) {

    m_move_mode = MoveSelected;
    m_p1 = p * (1.0 / dbu);
    m_dist = db::DPoint (0.0, 0.0);

    for (std::vector <ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
      (*r)->thaw ();
    }
    return false;

  } else {
  
    m_move_mode = MoveNone;

    //  compute search box
    double l = double (search_range) / canvas ()->trans ().t2.factor ();
    db::DBox search_dbox = db::DBox (p, p).enlarged (db::DPoint (l, l));

    //  box-selection

    db::UserObjectLayer::touching_iterator r = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin_touching (db::Box::from_double (db::DMag (1.0 / dbu) * search_dbox), db::UserObject::tag ());

    while (m_move_mode == MoveNone && ! r.at_end ()) {
      const ant::Object *robj = dynamic_cast <const ant::Object *> ((*r).ptr ());
      if (robj) {
        if (search_dbox.contains (robj->p1 () * dbu)) {
          m_move_mode = MoveP1;
          m_p1 = robj->p2 ();
        } else if (search_dbox.contains (robj->p2 () * dbu)) {
          m_move_mode = MoveP2;
          m_p1 = robj->p1 ();
        } else if (db::DEdge (robj->p1 () * dbu, robj->p2 () * dbu).distance_abs (p) <= l) {
          m_move_mode = MoveRuler;
          m_p1 = p * (1.0 / dbu);
        } else {
          ++r;
        }
      }
    }

    //  if anything was found
    if (m_move_mode != MoveNone) {

      //  make the moved ruler the selection
      clear_selection ();
      m_selected.insert (std::make_pair (cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()) + r.index (), 0));
      m_current = *(dynamic_cast <const ant::Object *> ((*r).ptr ()));
      m_rulers.push_back (new ant::View (this, &m_current, dbu, true));
      m_rulers.back ()->thaw ();
      return true;

    } else {
      return false;
    }

  }
}

void 
Service::move (const db::DPoint &p)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  double dbu = cv->layout ().dbu ();

  if (m_rulers.empty () || m_selected.empty ()) {
    return;
  }

  if (m_move_mode == MoveP1 || m_move_mode == MoveP2) {
    
    if (m_move_mode == MoveP1) {
      m_current.p1 (snap2 (m_p1 * dbu, p).second * (1.0 / dbu));
    } else {
      m_current.p2 (snap2 (m_p1 * dbu, p).second * (1.0 / dbu));
    }
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveRuler) {

    //  try two ways of snapping
    db::DPoint dp = p * (1.0 / dbu) - m_p1;
    ant::Object original = *(dynamic_cast <const ant::Object *> ((*m_selected.begin ()->first).ptr ()));

    db::DPoint p1 = dp + original.p1 ();
    db::DPoint p2 = dp + original.p2 ();

    std::pair<bool, db::DPoint> r1 = snap1 (p1 * dbu);
    db::DPoint q1 = r1.second * (1.0 / dbu);
    std::pair<bool, db::DPoint> r2 = snap1 (p2 * dbu);
    db::DPoint q2 = r2.second * (1.0 / dbu);

    if ((!r2.first && r1.first) || ((r1.first || (!r1.first && !r2.first)) && q1.distance (p1) < q2.distance (p2))) {
      q2 = q1 + (original.p2 () - original.p1 ());
    } else {
      q1 = q2 + (original.p1 () - original.p2 ());
    }

    m_current.p1 (q1);
    m_current.p2 (q2);

    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveSelected) {

    db::DPoint dp = p - m_p1 * dbu;
    //  round the drag distance to grid if required: this is the least we can do in this case
    if (m_grid_snap) {
      dp = db::DPoint (lay::snap (dp.x (), m_grid), lay::snap (dp.y (), m_grid));
    } 

    m_dist = dp * (1.0 / dbu);

    for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
      (*r)->displace (dp);
    }

  }
}

void 
Service::end_move (const db::DPoint &p)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);

  if (! m_rulers.empty () && ! m_selected.empty ()) {

    if (m_move_mode == MoveP1 || m_move_mode == MoveP2 || m_move_mode == MoveRuler) {

      //  begin the transaction
      tl_assert (! manager ()->transacting ());
      manager ()->transaction (tl::translate ("Move ruler")); 
      //  this dummy operation will update the screen:
      manager ()->queue (this, new db::Op ());

      //  replace the ruler that was moved
      cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).replace (m_selected.begin ()->first, db::UserObject (new ant::Object (m_current)));

      //  end the transaction
      manager ()->commit ();

      //  redraw 
      mp_view->redraw_deco_layer ();

      //  clear the selection (that was artifically created before)
      clear_selection ();

    } else if (m_move_mode == MoveSelected) {

      //  begin the transaction
      tl_assert (! manager ()->transacting ());
      manager ()->transaction (tl::translate ("Move ruler")); 
      //  this dummy operation will update the screen:
      manager ()->queue (this, new db::Op ());

      //  replace the rulers that were moved:
      for (std::map<obj_iterator, unsigned int>::const_iterator s = m_selected.begin (); s != m_selected.end (); ++s) {

        const ant::Object *robj = dynamic_cast<const ant::Object *> (s->first->ptr ());

        //  compute moved object and replace
        ant::Object rnew = robj->moved (m_dist);
        cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).replace (s->first, db::UserObject (new ant::Object (rnew)));

      }

      //  end the transaction
      manager ()->commit ();

      //  redraw 
      mp_view->redraw_deco_layer ();

      //  and make selection "visible"
      selection_to_view ();

    }

  }

  //  termine the operation
  m_move_mode = MoveNone;

}

void
Service::selection_to_view ()
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  double dbu = cv->layout ().dbu ();

  //  the selection objects need to be recreated since we destroyed the old rulers
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    delete *r;
  }
  m_rulers.clear ();
  m_rulers.reserve (m_selected.size ());
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    r->second = m_rulers.size ();
    const ant::Object *robj = dynamic_cast<const ant::Object *> (r->first->ptr ());
    m_rulers.push_back (new ant::View (this, robj, dbu, true /*selected*/));
  }
}

void 
Service::edit_cancel () 
{
  //  Cancel any move operation
  if (m_move_mode != MoveNone) {

    clear_selection ();
    m_move_mode = MoveNone;

  }
}

bool 
Service::mouse_press_event (const db::DPoint &p, unsigned int buttons, bool prio) 
{
  if (prio && (buttons & lay::LeftButton) != 0 && mp_view->cellviews () > rulers_cvindex) {

    if (! m_drawing) {

      const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
      if (cv.is_valid ()) {

        //  cancel any edit operations so far
        m_move_mode = MoveNone;

        //  reset selection
        clear_selection ();
           
        //  set the maximum number of rulers minus 1 to account for the new ruler
        //  and clear surplus rulers
        reduce_rulers (m_max_number_of_rulers - 1);

        double dbu = cv->layout ().dbu ();
        m_p1 = snap1 (p).second * (1.0 / dbu);

        //  create and start dragging the ruler
        m_current = ant::Object (m_p1, m_p1, 0);
        if (mp_active_ruler) {
          delete mp_active_ruler;
        }
        mp_active_ruler = new ant::View (this, &m_current, dbu, false /*not selected*/);
        mp_active_ruler->thaw ();
        m_drawing = true;

        canvas ()->grab_mouse (this, false);

      }

    } else {

      //  create the ruler object

      //  begin the transaction
      tl_assert (! manager ()->transacting ());
      manager ()->transaction (tl::translate ("Create ruler")); 
      //  this dummy operation will update the screen:
      manager ()->queue (this, new db::Op ());

      //  rulers always refer to the primary cell view 
      mp_view->create_deco_layer (rulers_cvindex);
      const lay::CellView &cv = mp_view->cellview (rulers_cvindex);

      insert_ruler (m_current.p1 (), m_current.p2 (), cv);

      //  stop dragging
      drag_cancel ();

      //  end the transaction
      manager ()->commit ();
      
    }

    return true;

  }

  return false;
}

bool 
Service::mouse_move_event (const db::DPoint &p, bool prio) 
{
  if (m_drawing && prio && mp_view->cellviews () > rulers_cvindex) {
    const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
    m_current.p2 (snap2 (m_p1 * cv->layout ().dbu (), p).second * (1.0 / cv->layout ().dbu ()));
    mp_active_ruler->redraw ();
  }
  return false;
}

void
Service::redo (db::Op *op) throw ()
{
  //  redraw the deco layer (delayed) and clear the selection
  mp_view->redraw_deco_layer ();
}

void 
Service::undo (db::Op *op) throw ()
{
  //  redraw the deco layer (delayed) and clear the selection
  mp_view->redraw_deco_layer ();
}

void 
Service::deactivated ()
{
  drag_cancel ();
}

std::pair <bool, db::DPoint>
Service::snap1 (const db::DPoint &pt, const std::vector <db::DEdge> &cutlines)
{
  db::DPoint dp (pt);

  //  in grid snap mode, first snap to grid ..
  if (m_grid_snap) {
    dp = db::DPoint (lay::snap (dp.x (), m_grid), lay::snap (dp.y (), m_grid));
  }

  ContourFinder finder (dp, cutlines);

  if (m_obj_snap && mp_view) {

    for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); ! l.at_end (); ++l) {

      if (l->visible (true /*real*/)) {

        const lay::CellView &cv = mp_view->cellview (l->cellview_index ());
        if (cv.is_valid ()) {

          double px = dp.x () / cv->layout ().dbu ();
          double py = dp.y () / cv->layout ().dbu ();
          double dd = std::max (0.0, canvas ()->trans ().inverted ().ctrans (m_snap_range) / cv->layout ().dbu ());
          
          double dmin (std::numeric_limits <db::Coord>::min ());
          double dmax (std::numeric_limits <db::Coord>::max ());
          db::DBox dworld (dmin, dmin, dmax, dmax);

          db::DBox dregion = dworld & db::DBox (px - dd, py - dd, px + dd, py + dd);

          if (! dregion.empty ()) {
            finder.find (mp_view, l->cellview_index (), l->trans (), cv->layout (), 
                         db::Box (db::Point::from_double (dregion.p1 ()), db::Point::from_double (dregion.p2 ())), 
                         *cv.cell (), 
                         l->layer_index (), 
                         &l->prop_sel (), l->inverse_prop_sel ()); 
          }

        }

      }

    }

  } 

  if (finder.any ()) {
    return std::make_pair (true, finder.get_found ());
  } 

  //  try to project the point onto the cutlines
  db::DPoint closest;
  bool anyp = false;

  for (std::vector<db::DEdge>::const_iterator c = cutlines.begin (); c != cutlines.end (); ++c) {
    std::pair<bool, db::DPoint> r = c->cut_point (db::DEdge (dp, dp + db::DPoint (-c->dy (), c->dx ())));
    if (r.first) {
      if (! anyp || dp.sq_distance (r.second) < dp.sq_distance (closest)) {
        anyp = true;
        closest = r.second;
      }
    }
  }

  if (anyp) {
    return std::make_pair (false, closest);
  } else {
    return std::make_pair (false, dp);
  }
}

std::pair <bool, db::DPoint>
Service::snap2 (const db::DPoint &p1, const db::DPoint &p2)
{
  std::vector <db::DEdge> cutlines;

  //  set up cutlines depending on the mode
  if (m_snap_mode == Ortho) {
    cutlines.reserve (2);
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (0.0, 1.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 0.0)));
  } else if (m_snap_mode == Horizontal) {
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 0.0)));
  } else if (m_snap_mode == Vertical) {
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (0.0, 1.0)));
  } else if (m_snap_mode == Diagonal) {
    cutlines.reserve (4);
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (0.0, 1.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 0.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 1.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, -1.0)));
  } 

  return snap1 (p2, cutlines); 
}

struct RulerIdComp 
{
  bool operator() (const db::UserObjectLayer::iterator &a, const db::UserObjectLayer::iterator &b) const
  {
    return dynamic_cast<const ant::Object *>(a->ptr ())->id () < dynamic_cast<const ant::Object *>(b->ptr ())->id ();
  }
};

void
Service::reduce_rulers (int num)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return;  //  no deco layer - no rulers.
  }

  db::UserObjectLayer::iterator rfrom = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()); 
  db::UserObjectLayer::iterator rto = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).end (db::UserObject::tag ());

  size_t n = std::distance (rfrom, rto);
  if (num >= 0 && int (n) > num) {

    //  clear selection
    clear_selection ();

    //  extract all rulers and other objects

    std::vector <db::UserObjectLayer::iterator> positions;
    positions.reserve (n);

    for (db::UserObjectLayer::iterator r = rfrom; r != rto; ++r) {
      const ant::Object *robj = dynamic_cast <const ant::Object *> (r->ptr ());
      if (robj) {
        positions.push_back (r);
      }
    }

    //  sort so we find the ones that are too old, remove them and sort the 
    //  remaining positions
    std::sort (positions.begin (), positions.end (), RulerIdComp ());
    positions.erase (positions.begin () + (positions.size () - num), positions.end ());
    std::sort (positions.begin (), positions.end ());

    //  now we can erase these positions
    cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).erase_positions (db::UserObject::tag (), positions.begin (), positions.end ());

    //  issue a redraw request.
    mp_view->redraw_deco_layer ();

  }

}

void 
Service::cut ()
{
  if (has_selection ()) {

    //  if required insert a dummy operation to receive an undo event
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::Op ());
    }

    //  copy & delete the selected rulers
    copy_selected ();
    del_selected ();

    //  issue a redraw request.
    mp_view->redraw_deco_layer ();

  }
}

void 
Service::copy ()
{
  //  copy the selected rulers
  copy_selected ();
}

void
Service::copy_selected ()
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);

  //  extract all selected rulers and paste in "micron" space
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    r->second = m_rulers.size ();
    const ant::Object *robj = dynamic_cast<const ant::Object *> (r->first->ptr ());
    db::Clipboard::instance () += new db::ClipboardValue<ant::Object> (robj->transformed (db::DMag (cv->layout ().dbu ())));
  }
}

void 
Service::paste ()
{
  if (mp_view->cellviews () > rulers_cvindex && db::Clipboard::instance ().begin () != db::Clipboard::instance ().end ()) {

    //  if required insert a dummy operation to receive an undo event
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::Op ());
    }

    const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
    double dbu = cv->layout ().dbu ();

    //  determine the last id
    int idmax = -1;
    for (db::UserObjectLayer::iterator r = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()); r != cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).end (db::UserObject::tag ()); ++r) {
      const ant::Object *robj = dynamic_cast <const ant::Object *> (r->ptr ());
      if (robj && robj->id () > idmax) {
        idmax = robj->id ();
      }
    }

    bool any_added = false;
    for (db::Clipboard::iterator c = db::Clipboard::instance ().begin (); c != db::Clipboard::instance ().end (); ++c) {
      const db::ClipboardValue<ant::Object> *value = dynamic_cast<const db::ClipboardValue<ant::Object> *> (*c);
      if (value) {
        mp_view->create_deco_layer (rulers_cvindex);
        cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).insert (db::UserObject (new ant::Object (value->get ().p1 () * (1.0 / dbu), value->get ().p2 () * (1.0 / dbu), ++idmax)));
        any_added = true;
      }
    }

    //  issue a redraw request.
    if (any_added) {
      mp_view->redraw_deco_layer ();
    }

  }

}

void 
Service::del ()
{
  if (has_selection ()) {

    //  if required insert a dummy operation to receive an undo event
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::Op ());
    }

    //  delete the selected rulers
    del_selected ();

    //  issue a redraw request.
    mp_view->redraw_deco_layer ();

  }
}

void
Service::del_selected ()
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return;
  }

  //  positions will hold a set of iterators that are to be erased
  std::vector <db::UserObjectLayer::iterator> positions;
  positions.reserve (m_selected.size ());
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    positions.push_back (r->first);
  }

  //  clear selection
  clear_selection ();

  //  erase all and insert the ones that we want to keep 
  std::sort (positions.begin (), positions.end ());
  cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).erase_positions (db::UserObject::tag (), positions.begin (), positions.end ());
}

bool 
Service::has_selection ()
{
  return ! m_selected.empty ();
}

bool 
Service::select (obj_iterator obj, lay::Editable::SelectionMode mode)
{
  if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
    //  select
    if (m_selected.find (obj) == m_selected.end ()) {
      m_selected.insert (std::make_pair (obj, 0));
      return true;
    }
  } else if (mode == lay::Editable::Reset) {
    //  unselect
    if (m_selected.find (obj) != m_selected.end ()) {
      m_selected.erase (obj);
      return true;
    }
  } else {
    //  invert selection
    if (m_selected.find (obj) != m_selected.end ()) {
      m_selected.erase (obj);
    } else {
      m_selected.insert (std::make_pair (obj, 0));
    }
    return true;
  }
  return false;
}

void
Service::clear_selection ()
{
  select (db::DBox (), lay::Editable::Reset);
}

static bool 
is_selected (const ant::Object &ruler, const db::DPoint &pos, double enl, double &distance) 
{
  db::DBox b (ruler.p1 (), ruler.p2 ());

  //  enlarge this box by some pixels
  b.enlarge (db::DPoint (enl, enl));

  if (! b.contains (pos)) {
    return false;
  }

  double d = db::DEdge (ruler.p1 (), ruler.p2 ()).distance_abs (pos);
  if (d <= enl) {
    distance = d;
    return true;
  } else {
    return false;
  }
}

static bool 
is_selected (const ant::Object &ruler, const db::DBox &box, double enl)
{
  return (box.contains (ruler.p1 ()) && box.contains (ruler.p2 ()));
}

double
Service::click_proximity (const db::DPoint &pos)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return lay::Editable::click_proximity (pos);  //  cellview not available - may happen on initialisation
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return lay::Editable::click_proximity (pos);  //  no deco layer - no rulers.
  }

  double dbu = cv->layout ().dbu ();

  //  compute search box
  double l = double (search_range) / canvas ()->trans ().t2.factor ();
  db::Box search_box = db::Box::from_double (db::DMag (1.0 / dbu) * db::DBox (pos, pos).enlarged (db::DPoint (l, l)));

  //  point selection: look for the "closest" ruler
  double dmin = std::numeric_limits <double>::max ();
  bool any_found = false;

  db::UserObjectLayer::touching_iterator r = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin_touching (search_box, db::UserObject::tag ());
  while (! r.at_end ()) {
    const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
    if (robj) {
      double d;
      if (is_selected (robj->transformed (db::Mag (dbu)), pos, l, d)) {
        if (! any_found || d < dmin) {
          dmin = d;
        }
        any_found = true;
      }
    }
    ++r;
  }

  //  return the proximity value
  if (any_found) {
    return dmin;
  } else {
    return lay::Editable::click_proximity (pos); 
  } 
}

bool
Service::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return false;  //  cellview not available - may happen on initialisation
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return false;  //  no deco layer - no rulers.
  }

  double dbu = cv->layout ().dbu ();

  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selected.empty ()) {
      m_selected.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selected.empty ()) {
        m_selected.clear ();
        needs_update = true;
      }
    } else {

      db::UserObjectLayer::iterator rfrom = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()); 
      db::UserObjectLayer::iterator rto = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).end (db::UserObject::tag ());

      //  extract all rulers
      for (db::UserObjectLayer::iterator r = rfrom; r != rto; ++r) {
        any_selected = true;
        if (select (r, mode)) {
          needs_update = true;
        }
      }
    }

  } else {

    //  compute search box
    double l = double (search_range) / canvas ()->trans ().t2.factor ();
    db::Box search_box = db::Box::from_double (db::DMag (1.0 / dbu) * box.enlarged (db::DPoint (l, l)));

    if (! box.is_point ()) {

      //  box-selection
      db::UserObjectLayer::touching_iterator r = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin_touching (search_box, db::UserObject::tag ());
      while (! r.at_end ()) {
        const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
        if (robj) {
          if (is_selected (robj->transformed (db::Mag (dbu)), box, l)) {
            any_selected = true;
            if (select (cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()) + r.index (), mode)) {
              needs_update = true;
            }
          }
        }
        ++r;
      }

    } else {

      //  point selection: look for the "closest" ruler
      double dmin = std::numeric_limits <double>::max ();

      db::UserObjectLayer::touching_iterator r = cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin_touching (search_box, db::UserObject::tag ());
      db::UserObjectLayer::touching_iterator rmin (r);
      while (! r.at_end ()) {
        const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
        if (robj) {
          double d;
          if (is_selected (robj->transformed (db::Mag (dbu)), box.p1 (), l, d)) {
            if (! any_selected || d < dmin) {
              rmin = r;
              dmin = d;
            }
            any_selected = true;
          }
        }
        ++r;
      }

      //  select the one that was found
      if (any_selected) {
        select (cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).begin (db::UserObject::tag ()) + rmin.index (), mode);
        needs_update = true;
      }

    }

  }

  //  if required, update the list of ruler objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  //  return true if at least one element was selected
  return any_selected;
}

lay::PropertiesPage *
Service::properties_page (QWidget *parent)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return 0;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return 0;  //  no deco layer - no rulers.
  }

  return new PropertiesPage (this, cv->layout ().dbu (), parent);
}

void 
Service::get_selection (std::vector <obj_iterator> &sel) const
{
  sel.clear ();
  sel.reserve (m_selected.size ());

  //  positions will hold a set of iterators that are to be erased
  for (std::map<obj_iterator, unsigned int>::const_iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    sel.push_back (r->first);
  }
}

void
Service::change_ruler (obj_iterator pos, const ant::Object &to)
{
  if (mp_view->cellviews () <= rulers_cvindex) {
    return;
  }
  const lay::CellView &cv = mp_view->cellview (rulers_cvindex);
  if (cv.deco_layer () < 0) {
    return;  //  no deco layer - no rulers.
  }

  if (manager ()->transacting ()) {
    //  this dummy operation will update the screen:
    manager ()->queue (this, new db::Op ());
  }

  //  replace the object
  cv.cell ()->shapes ((unsigned int) cv.deco_layer ()).replace (pos, db::UserObject (new ant::Object (to)));

  //  we need to redraw
  mp_view->redraw_deco_layer ();

  //  and make selection "visible"
  selection_to_view ();
}

void 
Service::paint_on_bitmaps (const lay::CellView &cv, 
                           const db::MagTrans &trans,
                           const std::vector <lay::Bitmap *> &bitmaps)
{
  if (bitmaps.empty () || cv.deco_layer () < 0) {
    return;
  }

  db::Coord lim = std::numeric_limits<db::Coord>::max ();
  db::DBox world (trans * db::Box (db::Point (-lim, -lim), db::Point (lim, lim)));
  db::Box vp = trans.inverted () * (world & db::DBox (db::DPoint (0.0, 0.0), db::DPoint (bitmaps.front ()->width (), bitmaps.front ()->height ())));

  double dbu = cv->layout ().dbu ();

  lay::Renderer renderer;

  db::UserObject::tag user_object_tag = db::UserObject::tag ();
  db::UserObjectLayer::touching_iterator user_object = cv.cell ()->shapes (cv.deco_layer ()).begin_touching (vp, user_object_tag);
  while (! user_object.at_end ()) {
    const ant::Object *ruler = dynamic_cast <const ant::Object *> ((*user_object).ptr ());
    draw_ruler (*ruler, dbu, trans, false /*not selected*/, bitmaps.front ());
    ++user_object;
  }
}

void 
Service::menu_activated (const std::string &symbol)
{
  if (symbol == "ant::clear_all_rulers") {
    clear_rulers ();
  } else {
    lay::Plugin::menu_activated (symbol);
  }
}

// -------------------------------------------------------------

} // namespace ant


