
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_antService
#define HDR_antService

#include "layViewObject.h"
#include "layEditable.h"
#include "layPlugin.h"
#include "layDrawing.h"
#include "dbLayout.h"

#include <map>
#include <vector>

namespace ant {

class LayoutView;
class LayoutCanvas;
class Service;
class CellView;

// -------------------------------------------------------------

class Object 
  : public db::UserObjectBase
{
public:
  typedef db::coord_traits<coord_type> coord_traits;

  Object ()
    : m_p1 (), m_p2 (), m_id (0)
  {
    //  .. nothing yet ..
  }

  Object (const db::DPoint &p1, const db::DPoint &p2, int id)
    : m_p1 (p1), m_p2 (p2), m_id (id)
  {
    //  .. nothing else ..
  }

  Object (const ant::Object &d)
    : m_p1 (d.m_p1), m_p2 (d.m_p2), m_id (d.m_id)
  {
    //  .. nothing else ..
  }

  Object &operator= (const ant::Object &d)
  {
    if (this != &d) {
      m_p1 = d.m_p1;
      m_p2 = d.m_p2;
      m_id = d.m_id;
    }
    return *this;
  }

  virtual bool equals (const db::UserObjectBase *d) const
  {
    const ant::Object *ruler = dynamic_cast<const ant::Object *> (d);
    if (ruler) {
      return m_id == ruler->m_id && m_p1 == ruler->m_p1 && m_p2 == ruler->m_p2;
    } else {
      return false;
    }
  }

  virtual db::UserObjectBase *clone () const
  {
    return new ant::Object (*this);
  }

  virtual db::Box box () const
  {
    return db::Box (coord_traits::rounded_down (std::min (m_p1.x (), m_p2.x ())), coord_traits::rounded_down (std::min (m_p1.y (), m_p2.y ())),
                    coord_traits::rounded_up (std::max (m_p1.x (), m_p2.x ())), coord_traits::rounded_down (std::max (m_p1.y (), m_p2.y ())));
  }

  template <class Tr>
  Object &transform (const Tr &t)
  {
    *this = ant::Object (t * m_p1, t * m_p2, m_id);
    return *this;
  }

  template <class Tr>
  Object transformed (const Tr &t) const
  {
    return ant::Object (t * m_p1, t * m_p2, m_id);
  }

  Object &transform (const db::fixpoint_trans &t)
  {
    *this = ant::Object (t * m_p1, t * m_p2, m_id);
    return *this;
  }

  Object transformed (const db::fixpoint_trans &t) const
  {
    return ant::Object (t * m_p1, t * m_p2, m_id);
  }

  Object &move (const db::DPoint &p)
  {
    m_p1 += p;
    m_p2 += p;
    return *this;
  }

  Object moved (const db::DPoint &p) const
  {
    ant::Object d (*this);
    d.move (p);
    return d;
  }

  const db::DPoint &p1 () const
  {
    return m_p1;
  }

  const db::DPoint &p2 () const
  {
    return m_p2;
  }

  void p1 (const db::DPoint &p)
  {
    m_p1 = p;
  }

  void p2 (const db::DPoint &p)
  {
    m_p2 = p;
  }

  int id () const
  {
    return m_id;
  }

  void id (int _id) 
  {
    m_id = _id;
  }

  bool operator< (const ant::Object &b) const
  {
    return m_id < b.m_id || (m_id == b.m_id && (m_p1 < b.m_p1 || (m_p1 == b.m_p1 && m_p2 < b.m_p2)));
  }

private:
  db::DPoint m_p1, m_p2;
  int m_id;
};

// -------------------------------------------------------------

class View
  : public lay::ViewObject
{
public: 
  /**
   *  @brief Constructor attaching to a certain object
   */
  View (ant::Service *rulers, const ant::Object *ruler, double dbu, bool selected);

  /**
   *  @brief The destructor
   */
  ~View ();

  /**
   *  @brief Set a displacement
   *
   *  The displacement determines if the ruler is painted "shifted".
   *  The displacement must be specified in database coordinates.
   */
  void displace (const db::DPoint &p);

  /**
   *  @brief set the Ruler object
   */
  void ruler (const ant::Object *r);

  /**
   *  @brief Get the db::Ruler object that this view object is presenting
   */
  const ant::Object *ruler () const
  {
    return mp_ruler;
  }

private:
  ant::Service *mp_rulers;
  bool m_selected;
  const ant::Object *mp_ruler;
  double m_dbu;
  db::DPoint m_displacement;

  virtual void render ();

  //  no copying nor default construction
  View (const View &d);
  View &operator= (const View &d);
  View ();
};

// -------------------------------------------------------------

class Service
  : public lay::ViewService,
    public lay::Editable,
    public lay::Plugin,
    public lay::Drawing,
    public db::Object
{
public: 
  typedef db::UserObjectLayer::iterator obj_iterator;

  enum SnapMode { Any = 0, Diagonal, Ortho, Horizontal, Vertical, NumSnapModes };

  Service (db::Manager *manager, lay::LayoutView *view);

  ~Service ();

  /** 
   *  @brief Clear all highlights (for current object highlighting)
   */
  void clear_highlights ();

  /** 
   *  @brief Restore all highlights (for current object highlighting)
   */
  void restore_highlights ();

  /** 
   *  @brief Highlight a certain object
   */
  void highlight (unsigned int n);

  /**
   *  @brief Cancel dragging the ruler. The ruler is erased.
   */
  void drag_cancel ();

  /**
   *  @brief Cancel any edit operations (such as move)
   */
  void edit_cancel ();

  /** 
   *  @brief Clear all rulers
   */
  void clear_rulers ();

  /** 
   *  @brief "delete" operation
   */
  virtual void del ();

  /** 
   *  @brief "cut" operation
   */
  virtual void cut ();

  /** 
   *  @brief "copy" operation
   */
  virtual void copy ();

  /** 
   *  @brief "paste" operation
   */
  virtual void paste ();

  /**
   *  @brief Tell if anything is selected
   */
  virtual bool has_selection ();

  /** 
   *  @brief point selection proximity predicate
   */
  virtual double click_proximity (const db::DPoint &pos);

  /** 
   *  @brief "select" operation
   */
  virtual bool select (const db::DBox &box, lay::Editable::SelectionMode mode);

  /**
   *  @brief Implementation of the undo operations
   */
  virtual void undo (db::Op *op) throw ();

  /**
   *  @brief Implementation of the redo operations
   */
  virtual void redo (db::Op *op) throw ();

  /**
   *  @brief Insert a ruler at the given cv
   */
  void insert_ruler (const db::DPoint &p1, const db::DPoint &p2, const lay::CellView &cv);

  /**
   *  @brief Begin a "move" operation
   */
  virtual bool begin_move (bool sel, const db::DPoint &p);

  /**
   *  @brief Continue a "move" operation
   */
  virtual void move (const db::DPoint &p);

  /**
   *  @brief Terminate a "move" operation
   */
  virtual void end_move (const db::DPoint &p);

  /**
   *  @brief Create the properties page
   */
  virtual lay::PropertiesPage *properties_page (QWidget * /*parent*/);

  /**
   *  @brief Get the selection for the properties page
   */
  void get_selection (std::vector <obj_iterator> &selection) const;

  /**
   *  @brief Change a specific ruler 
   */
  void change_ruler (obj_iterator pos, const ant::Object &to);

  /**
   *  @brief Implementation of "Plugin" interface: configuration setup
   */
  bool configure (const std::string &name, const std::string &value);

  /**
   *  @brief Implementation of "Plugin" interface: configuration finalization
   */
  void config_finalize ();

  /**
   *  @brief Color accessor
   */
  QColor color () const
  {
    return m_color;
  }

  /**
   *  @brief Obtain the lay::ViewService interface
   */
  lay::ViewService *view_service_interface ()
  {
    return this;
  }

  /**
   *  @brief Obtain the lay::Drawing interface
   */
  lay::Drawing *drawing_interface ()
  {
    return this;
  }

  /**
   *  @brief Obtain the lay::Editable interface
   */
  lay::Editable *editable_interface ()
  {
    return this;
  }

  /**
   *  @brief Access to the view object
   */
  lay::LayoutView *view () const
  {
    return mp_view;
  }

  /**
   *  @brief Implement the menu response function
   */
  void menu_activated (const std::string &symbol);
  
private:
  //  Ruler display and snapping configuration
  QColor m_color;
  SnapMode m_snap_mode;
  double m_grid;
  bool m_grid_snap;
  bool m_obj_snap;
  int m_snap_range;

  //  Configuration parameter: maximum number of rulers
  int m_max_number_of_rulers;

  //  The layout view that the ruler service is attached to
  lay::LayoutView *mp_view;

  //  The ruler view objects representing the selection
  //  and the moved rules in move mode
  std::vector<ant::View *> m_rulers;
  //  The selection
  std::map<obj_iterator, unsigned int> m_selected;
  //  The reference point in move mode
  db::DPoint m_p1;
  //  The move distance in MoveSelection mode
  db::DPoint m_dist;
  //  The ruler representing the dragged ruler in "create ruler" mode
  ant::View *mp_active_ruler;
  //  True, if creating a ruler (dragging)
  bool m_drawing;
  //  The ruler object representing the ruler being created
  ant::Object m_current;
  //  The current move mode:
  //    MoveNone - not moving
  //    MoveP1 - dragging the first point
  //    MoveP2 - dragging the second point
  //    MoveRuler - dragging a whole ruler (one)
  //    MoveSelection - dragging a whole ruler (many)
  enum { MoveNone, MoveP1, MoveP2, MoveRuler, MoveSelected } m_move_mode;

  std::pair<bool, db::DPoint> snap1 (const db::DPoint &p, const std::vector <db::DEdge> &cutlines = std::vector <db::DEdge> ());
  std::pair<bool, db::DPoint> snap2 (const db::DPoint &p1, const db::DPoint &p2);

  virtual bool mouse_move_event (const db::DPoint &p, bool prio);
  virtual bool mouse_press_event (const db::DPoint &p, unsigned int buttons, bool prio);
  virtual void deactivated ();

  /**
   *  @brief Select a certain ruler
   *
   *  @return true, if the selection has changed
   */
  bool select (obj_iterator obj, lay::Editable::SelectionMode mode);

  /**
   *  @brief Clear the selection
   */
  void clear_selection ();

  /**
   *  @brief Limit the number of rulers to this number
   */
  void reduce_rulers (int num);

  /**
   *  @brief Delete the selected rulers
   *
   *  Used as implementation for "del" and "cut"
   */
  void del_selected ();

  /**
   *  @brief Copy the selected rulers to the clipboard
   *
   *  Used as implementation for "copy" and "cut"
   */
  void copy_selected ();

  /**
   *  @brief implementation of the "Drawing" interface
   */
  void paint_on_bitmaps (const lay::CellView &cellview, 
                         const db::MagTrans &trans,
                         const std::vector <lay::Bitmap *> &bitmaps);

  /**
   *  @brief Update m_rulers to reflect the selection
   */
  void selection_to_view ();
};

}

#endif

