
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_dbArray
#define HDR_dbArray

#include <algorithm> 

#include "dbTypes.h"
#include "dbMemStatistics.h"
#include "dbPoint.h"
#include "dbBox.h"
#include "dbCellInst.h"
#include "dbTrans.h"
#include "tlException.h"

namespace db
{

template <class Obj> struct array_iterator;

/**
 *  @brief The array iterator base class (internal)
 */

template <class Obj>
struct basic_array_iterator 
{
  typedef typename Obj::coord_type coord_type;
  typedef db::point <coord_type> point_type;
  typedef db::simple_trans <coord_type> trans_type;
  typedef db::box <coord_type> box_type;

  virtual ~basic_array_iterator ()
  { 
    // .. nothing yet ..
  }

  virtual trans_type get (const trans_type &tr) const = 0;

  virtual void inc () = 0;

  virtual bool at_end () const = 0;
  
  virtual basic_array_iterator<Obj> *clone () const = 0;
};

/**
 *  @brief The array base class (internal)
 */

template <class Obj>
struct basic_array
{
  typedef typename Obj::coord_type coord_type;
  typedef db::box <coord_type> box_type;
  typedef db::point <coord_type> point_type;
  typedef db::simple_trans <coord_type> trans_type;

  virtual ~basic_array ()
  {
    //  .. nothing yet ..
  }

  virtual array_iterator <Obj> begin_touching (const trans_type &disp, const box_type &b) const = 0;
  
  virtual array_iterator <Obj> begin (const trans_type &disp) const = 0;
  
  virtual basic_array <Obj> *clone () const = 0;

  virtual box_type bbox (const box_type &obox) const = 0;

  virtual size_t size () const = 0;

  virtual void invert (trans_type &t) = 0;

  virtual bool equal (const basic_array<Obj> *) const = 0;

  virtual bool less (const basic_array<Obj> *) const = 0;

  virtual bool is_regular_array (point_type & /*a*/, point_type & /*b*/, unsigned long & /*amax*/, unsigned long & /*bmax*/) const 
  {
    return false;
  }

  virtual size_t mem_used () const = 0;

  virtual size_t mem_reqd () const = 0;

  virtual bool is_mag () const 
  { 
    return false; 
  }

  virtual double mag () const 
  { 
    return 1.0; 
  }

  virtual unsigned int type () const
  {
    return 0;
  }
};

/**
 *  @brief The array iterator specialization for the regular case (internal)
 *
 *  The regular case is that the displacements are described as
 *    P=a*A+b*B (a=[amin..amax),b=[bmin..bmax)).
 */

template <class Obj>
struct regular_array_iterator
  : public basic_array_iterator <Obj>
{
  typedef typename basic_array_iterator<Obj>::point_type point_type;
  typedef typename basic_array_iterator<Obj>::box_type box_type;
  typedef typename basic_array_iterator<Obj>::trans_type trans_type;

  regular_array_iterator (const point_type &a, const point_type &b, unsigned long amin, unsigned long amax, unsigned long bmin, unsigned long bmax)
    : m_a (a), m_b (b),
      m_amin (amin), m_amax (amax),
      m_bmin (bmin), m_bmax (bmax),
      m_ai (amin), m_bi (bmin)
  { 
    //  if there is no way to iterate for ai, just finish:
    if (m_amin >= m_amax) {
      m_bi = m_bmax;
    }
  }

  virtual ~regular_array_iterator ()
  {
    //  .. nothing yet .. 
  }

  virtual trans_type 
  get (const trans_type &t) const 
  {
    return trans_type (t.rot (), t.disp () + point_type (m_ai * m_a.x () + m_bi * m_b.x (),
                                                         m_ai * m_a.y () + m_bi * m_b.y ()));
  }

  virtual void inc () 
  {
    ++m_ai;
    if (m_ai >= m_amax) {
      m_ai = m_amin;
      ++m_bi;
    }
  }

  virtual bool at_end () const 
  {
    return m_bi >= m_bmax;
  }

  virtual basic_array_iterator<Obj> *clone () const 
  {
    return new regular_array_iterator <Obj> (*this);
  }

private:
  point_type m_a, m_b;
  unsigned long m_amin, m_amax, m_bmin, m_bmax;
  unsigned long m_ai, m_bi;
};

/**
 *  @brief The array specialization for the regular case (internal)
 *
 *  The regular case is that the displacements are described as
 *    P=O+a*A+b*B (a=[amin..amax),b=[bmin..bmax)).
 */

template <class Obj>
struct regular_array 
  : public basic_array <Obj>
{
  typedef typename basic_array<Obj>::point_type point_type;
  typedef typename basic_array<Obj>::box_type box_type;
  typedef typename basic_array<Obj>::coord_type coord_type;
  typedef typename basic_array<Obj>::trans_type trans_type;

  regular_array (const point_type &a, const point_type &b, unsigned long amax, unsigned long bmax)
    : m_a (a), m_b (b), m_amax (amax), m_bmax (bmax)
  { 
    compute_det ();
  }

  virtual ~regular_array ()
  {
    //  .. nothing yet .. 
  }

  virtual array_iterator <Obj> 
  begin_touching (const trans_type &trans, const box_type &b) const
  {
    if (b.empty ()) {

      return array_iterator <Obj> (trans, new regular_array_iterator <Obj> (m_a, m_b, 0, 0, 0, 0));

    } else if (fabs (m_det) < 0.5) {

      return begin (trans);

    } else {

      std::pair <double, double> ab [4] = {
        ab_coord (b.p1 ()),
        ab_coord (point_type (b.left (), b.top ())),
        ab_coord (point_type (b.right (), b.bottom ())),
        ab_coord (b.p2 ())
      };

      double amin = ab [0].first;
      double amax = ab [0].first;
      double bmin = ab [0].second;
      double bmax = ab [0].second;

      for (int i = 1; i < 4; ++i) {
        amin = std::min (amin, ab [i].first);
        amax = std::max (amax, ab [i].first);
        bmin = std::min (bmin, ab [i].second);
        bmax = std::max (bmax, ab [i].second);
      }

      const double epsilon = 1e-3;

      unsigned long amini = 0;
      if (amin >= epsilon) {
        if (amin > double (std::numeric_limits <unsigned long>::max () - 1)) {
          amini = std::numeric_limits <unsigned long>::max () - 1;
        } else {
          amini = (unsigned long) (amin + 1.0 - epsilon);
        }
        if (amini > m_amax) {
          amini = m_amax;
        }
      }
         
      unsigned long amaxi = 0;
      if (amax >= -epsilon) {
        if (amax > double (std::numeric_limits <unsigned long>::max () - 1)) {
          amaxi = std::numeric_limits <unsigned long>::max () - 1;
        } else {
          amaxi = (unsigned long) (amax + epsilon) + 1;
        }
        if (amaxi > m_amax) {
          amaxi = m_amax;
        }
      }

      unsigned long bmini = 0;
      if (bmin >= epsilon) {
        if (bmin > double (std::numeric_limits <unsigned long>::max () - 1)) {
          bmini = std::numeric_limits <unsigned long>::max () - 1;
        } else {
          bmini = (unsigned long) (bmin + 1.0 - epsilon);
        }
        if (bmini > m_bmax) {
          bmini = m_bmax;
        }
      }
         
      unsigned long bmaxi = 0;
      if (bmax >= -epsilon) {
        if (bmax > double (std::numeric_limits <unsigned long>::max () - 1)) {
          bmaxi = std::numeric_limits <unsigned long>::max () - 1;
        } else {
          bmaxi = (unsigned long) (bmax + epsilon) + 1;
        }
        if (bmaxi > m_bmax) {
          bmaxi = m_bmax;
        }
      }
      
      return array_iterator <Obj> (trans, new regular_array_iterator <Obj> (m_a, m_b, amini, amaxi, bmini, bmaxi));
    }
  }
  
  virtual array_iterator <Obj> 
  begin (const trans_type &trans) const
  {
    return array_iterator <Obj> (trans, new regular_array_iterator <Obj> (m_a, m_b, 0, m_amax, 0, m_bmax));
  }

  virtual basic_array <Obj> *clone () const 
  {
    return new regular_array <Obj> (*this);
  }

  virtual box_type bbox (const box_type &obox) const
  {
    if (obox.empty ()) {
      return obox;
    } else {

      point_type ma (m_a.x () * (m_amax - 1), m_a.y () * (m_amax - 1));
      point_type mb (m_b.x () * (m_bmax - 1), m_b.y () * (m_bmax - 1));

      box_type b;

      if (m_amax > 0 && m_bmax > 0) {
        b += point_type (0, 0);
        b += ma;
        b += mb;
        b += ma + mb;
      }

      return box_type (obox.p1 () + b.p1 (), obox.p2 () + b.p2 ());

    }
  }

  virtual size_t size () const 
  {
    return m_amax * m_bmax;
  }

  virtual void invert (trans_type &t) 
  {
    t.invert ();
    db::fixpoint_trans f (t.rot ());
    m_a = -f (m_a);
    m_b = -f (m_b);
    compute_det ();
  }

  virtual bool less (const basic_array<Obj> *b) const
  {
    const regular_array<Obj> *d = static_cast<const regular_array<Obj> *> (b);
    return m_a < d->m_a || (m_a == d->m_a && (
           m_b < d->m_b || (m_b == d->m_b && (
           m_amax < d->m_amax || (m_amax == d->m_amax && m_bmax < d->m_bmax)))));
  }

  virtual bool equal (const basic_array<Obj> *b) const
  {
    const regular_array<Obj> *d = static_cast<const regular_array<Obj> *> (b);
    return (m_a == d->m_a && m_b == d->m_b && m_amax == d->m_amax && m_bmax == d->m_bmax);
  }

  virtual bool is_regular_array (point_type &a, point_type &b, unsigned long &amax, unsigned long &bmax) const 
  {
    a = m_a;
    b = m_b;
    amax = m_amax;
    bmax = m_bmax;
    return true;
  }

  virtual size_t mem_used () const
  {
    return sizeof (*this);
  }

  virtual size_t mem_reqd () const
  {
    return sizeof (*this);
  }

  virtual unsigned int type () const
  {
    return 1;
  }

protected:
  point_type m_a, m_b;
  unsigned long m_amax, m_bmax;
  double m_det;

  std::pair <double, double> ab_coord (const point_type &p) const
  {
    double a = (double (p.x ()) * double (m_b.y ()) - double (p.y ()) * double (m_b.x ())) / m_det;
    double b = (double (m_a.x ()) * double (p.y ()) - double (m_a.y ()) * double (p.x ())) / m_det;
    return std::make_pair (a, b);
  }

  void compute_det () 
  {
    m_det = double (m_a.x ()) * double (m_b.y ()) - double (m_a.y ()) * double (m_b.x ());
  }
};

/**
 *  @brief The array specialization for the regular, magnified case (internal)
 *
 *  The regular case is that the displacements are described as
 *    P=O+a*A+b*B (a=[amin..amax),b=[bmin..bmax)).
 *  Each instance is magnified with the given magnification.
 */

template <class Obj>
struct regular_mag_array 
  : public regular_array<Obj>
{
  typedef typename basic_array<Obj>::point_type point_type;
  typedef typename basic_array<Obj>::box_type box_type;
  typedef typename basic_array<Obj>::coord_type coord_type;
  typedef typename basic_array<Obj>::trans_type trans_type;

  regular_mag_array (double mag, const point_type &a, const point_type &b, unsigned long amax, unsigned long bmax)
    : regular_array<Obj> (a, b, amax, bmax), m_mag (mag)
  { 
    //  .. nothing yet ..
  }

  virtual basic_array <Obj> *clone () const 
  {
    return new regular_mag_array <Obj> (*this);
  }

  virtual void invert (trans_type &t) 
  {
    db::fixpoint_trans f (t.rot ());
    f.invert ();
    m_mag = 1.0 / m_mag;
    t = trans_type (f.rot (), f.ctrans (t.disp ()) * -m_mag);
    regular_array<Obj>::m_a = f (regular_array<Obj>::m_a) * -m_mag;
    regular_array<Obj>::m_b = f (regular_array<Obj>::m_b) * -m_mag;
    regular_array<Obj>::compute_det ();
  }

  virtual double mag () const
  {
    return m_mag;
  }

  virtual bool is_mag () const
  {
    return true;
  }

  virtual unsigned int type () const
  {
    return 2;
  }

private:
  double m_mag;
};

/**
 *  @brief The array specialization for the single magnified case (internal)
 *
 *  In the single case the displacement is 0.
 */

template <class Obj>
struct single_mag_inst 
  : public basic_array <Obj>
{
  typedef typename basic_array<Obj>::point_type point_type;
  typedef typename basic_array<Obj>::box_type box_type;
  typedef typename basic_array<Obj>::coord_type coord_type;
  typedef typename basic_array<Obj>::trans_type trans_type;

  single_mag_inst (double mag)
    : m_mag (mag)
  { 
    //  .. nothing yet ..
  }

  virtual array_iterator <Obj> 
  begin_touching (const trans_type &trans, const box_type &b) const
  {
    return array_iterator <Obj> (trans, ! b.contains (point_type (0, 0))); 
  }
  
  virtual array_iterator <Obj> 
  begin (const trans_type &trans) const
  {
    return array_iterator <Obj> (trans, false);
  }

  virtual basic_array <Obj> *clone () const 
  {
    return new single_mag_inst <Obj> (*this);
  }

  virtual box_type bbox (const box_type &obox) const
  {
    return obox;
  }

  virtual size_t size () const 
  {
    return 1;
  }

  virtual void invert (trans_type &t) 
  {
    db::fixpoint_trans f (t.rot ());
    f.invert ();
    m_mag = 1.0 / m_mag;
    t = trans_type (f.rot (), f.ctrans (t.disp ()) * -m_mag);
  }

  virtual bool equal (const basic_array<Obj> *d) const 
  {
    return true;
  }

  virtual bool less (const basic_array<Obj> *d) const 
  {
    return false;
  }

  virtual size_t mem_used () const
  {
    return sizeof (*this);
  }

  virtual size_t mem_reqd () const
  {
    return sizeof (*this);
  }

  virtual double mag () const
  {
    return m_mag;
  }

  virtual bool is_mag () const
  {
    return true;
  }

  virtual unsigned int type () const
  {
    return 3;
  }

private:
  double m_mag;
};

/** 
 *  @brief The array iterator 
 *
 *  This facade object wraps the actual basic_array iterator
 *  object. The iterator is not end-tested through comparing
 *  with end() but rather querying the at_end() property.
 *  This iterator as well acts as a single or zero instance
 *  iterator if the base pointer is 0.
 */

template <class Obj>
struct array_iterator
{
  typedef typename Obj::coord_type coord_type;
  typedef db::point <coord_type> point_type;
  typedef db::simple_trans <coord_type> trans_type;

  /**
   *  @brief The default constructor
   */
  array_iterator ()
    : m_trans (), mp_base (0), m_done (true)
  { 
    // .. nothing yet ..
  }

  /**
   *  @brief The iterator constructor
   *
   *  The constructor will receive a basic_array_iterator object
   *  and take over ownership over this.
   */
  array_iterator (const trans_type &trans, db::basic_array_iterator <Obj> *base)
    : m_trans (trans), mp_base (base), m_done (false)
  { 
    // .. nothing yet ..
  }

  /**
   *  @brief The single iterator constructor
   *
   *  The constructor takes a flag indicating whether to represent
   *  zero (true) or one (false) instance.
   */
  array_iterator (const trans_type &trans, bool done)
    : m_trans (trans), mp_base (0), m_done (done)
  { 
    // .. nothing yet ..
  }

  /**
   *  @brief The copy constructor
   */
  array_iterator (const array_iterator &d)
    : m_trans (d.m_trans), mp_base (0)
  {
    if (mp_base) {
      delete mp_base;
    }
    mp_base = d.mp_base ? d.mp_base->clone () : 0;
  }

  /**
   *  @brief The assignment operator
   */
  array_iterator &operator= (const array_iterator &d)
  {
    if (&d != this) {
      m_trans = d.m_trans;
      if (mp_base) {
        delete mp_base;
      }
      mp_base = d.mp_base ? d.mp_base->clone () : 0;
    }
    return *this;
  }

  /**
   *  @brief The destructor
   *
   *  This will delete the basic_array_iterator object passed
   *  in the constructor
   */
  ~array_iterator ()
  {
    if (mp_base) {
      delete mp_base;
    }
    mp_base = 0;
  }

  /**
   *  @brief The access operator
   *
   *  This delivers a displacement vector that is associated with
   *  the iterator's "position".
   */
  trans_type operator* () const 
  {
    if (mp_base) {
      return mp_base->get (m_trans);
    } else {
      return m_trans;
    }
  }

  /**
   *  @brief Increment operator
   */
  array_iterator &operator++ () 
  {
    if (mp_base) {
      mp_base->inc ();
    } else {
      m_done = true;
    }
    return *this;
  }

  /**
   *  @brief End test
   *
   *  @return true, if the iterator is at the end.
   */
  bool at_end () const 
  {
    return mp_base ? mp_base->at_end () : m_done;
  }

private:
  trans_type m_trans;
  basic_array_iterator <Obj> *mp_base;
  bool m_done;
};

/**
 *  @brief The array container
 *
 *  This is the facade object that wraps the array implementation
 *  which is either a regular_array or another object derived
 *  from a basic_array object.
 *  Access to the object instances is through the iterators which
 *  deliver the displacement vectors of the various instances, 
 *  possibly filtered by a box region.
 */

template <class Obj>
struct array
{
  typedef typename Obj::coord_type coord_type;
  typedef db::box <coord_type> box_type;
  typedef db::point <coord_type> point_type;
  typedef db::simple_trans <coord_type> trans_type;
  typedef db::array_iterator <Obj> iterator;
  typedef Obj object_type;

  /**
   *  @brief The default constructor
   */
  array ()
    : m_obj (), m_trans (), mp_base (0)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief The array constructor
   *
   *  The array object takes over the ownership over the
   *  basic_array object.
   *
   *  @param obj The object to put into an array
   *  @param base The basic_array object that describes the array further
   */
  array (const Obj &obj, const trans_type &trans, basic_array <Obj> *base)
    : m_obj (obj), m_trans (trans), mp_base (base)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief The regular array constructor
   *
   *  This is basically a convenience function that creates
   *  an appropriate basic_array object.
   */
  array (const Obj &obj, const trans_type &trans, const point_type &a, const point_type &b, unsigned long amax, unsigned long bmax)
    : m_obj (obj), m_trans (trans), mp_base (new regular_array <Obj> (a, b, amax, bmax))
  {
    //  .. nothing yet ..
  }
  
  /**
   *  @brief The singular array constructor
   *
   *  This is basically a convenience function that creates
   *  an appropriate basic_array object.
   *  mp_base == 0 is implicitly a single-object array.
   */
  explicit 
  array (const Obj &obj, const trans_type &trans)
    : m_obj (obj), m_trans (trans), mp_base (0)
  {
    //  .. nothing yet ..
  }
  
  /**
   *  @brief The regular magnified array constructor
   *
   *  This is basically a convenience function that creates
   *  an appropriate basic_array object.
   */
  array (const Obj &obj, const trans_type &trans, double mag, const point_type &a, const point_type &b, unsigned long amax, unsigned long bmax)
    : m_obj (obj), m_trans (trans), mp_base (new regular_mag_array <Obj> (mag, a, b, amax, bmax))
  {
    //  .. nothing yet ..
  }
  
  /**
   *  @brief The singular magnified instance constructor
   *
   *  This is basically a convenience function that creates
   *  an appropriate basic_array object.
   */
  explicit 
  array (const Obj &obj, const trans_type &trans, double mag)
    : m_obj (obj), m_trans (trans), mp_base (new single_mag_inst <Obj> (mag))
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief The destructor
   *
   *  This will delete the basic_array object that was passed
   *  to the constructor.
   */
  ~array ()
  {
    if (mp_base) {
      delete mp_base;
    }
    mp_base = 0;
  }

  /**
   *  @brief The copy constructor
   */
  array (const array &d)
    : m_obj (d.m_obj), m_trans (d.m_trans), mp_base (0)
  {
    if (d.mp_base) {
      mp_base = d.mp_base->clone ();
    }
  }

  /**
   *  @brief The assignment operator
   */
  array &operator= (const array &d)
  {
    if (&d != this) {
      m_trans = d.m_trans;
      m_obj = d.m_obj;
      if (mp_base) {
        delete mp_base;
      }
      if (d.mp_base) {
        mp_base = d.mp_base->clone ();
      } else {
        mp_base = 0;
      }
    }
    return *this;
  }

  /**
   *  @brief The region-selective iterator
   *
   *  There is no end() method. Rather use the at_end() method
   *  of the iterator to determine the end of the sequence.
   *  The iterator returned will report all object displacement 
   *  vectors of the objects that touch the given box and 
   *  possibly some more.
   */
  template <class BoxConv>
  array_iterator <Obj> begin_touching (const box_type &b, const BoxConv &bc) const 
  {
    if (b.empty ()) {
      if (mp_base) {
        return mp_base->begin_touching (m_trans, box_type ());
      } else {
        return array_iterator <Obj> (m_trans, true); 
      }
    } else if (mp_base) {
      box_type ob (bc (m_obj));
      if (ob.empty ()) {
        return mp_base->begin_touching (m_trans, box_type ());
      } else {
        if (mp_base->is_mag ()) {
          ob = ob * mp_base->mag ();
        }
        ob.transform (db::fixpoint_trans (m_trans.rot ()));
        return mp_base->begin_touching (m_trans, box_type (b.p1 () - ob.p2 () - m_trans.disp (), b.p2 () - ob.p1 () - m_trans.disp ()));
      }
    } else {
      box_type ob (bc (m_obj));
      if (ob.empty ()) {
        return array_iterator <Obj> (m_trans, true); 
      } else {
        ob.transform (db::fixpoint_trans (m_trans.rot ()));
        return array_iterator <Obj> (m_trans, ! box_type (b.p1 () - ob.p2 (), b.p2 () - ob.p1 ()).contains (m_trans.disp ()));
      }
    }
  }
  
  /**
   *  @brief The non-selective iterator
   *
   *  There is no end() method. Rather use the at_end() method
   *  of the iterator to determine the end of the sequence.
   */
  array_iterator <Obj> begin () const 
  {
    if (mp_base) {
      return mp_base->begin (m_trans);
    } else {
      return array_iterator <Obj> (m_trans, false);
    }
  }

  /**
   *  @brief The bbox of the array
   */
  template <class BoxConv>
  box_type bbox (const BoxConv &bc) const
  {
    if (mp_base) {
      if (mp_base->is_mag ()) {
        return mp_base->bbox (m_trans * (bc (m_obj) * mp_base->mag ()));
      } else {
        return mp_base->bbox (m_trans * bc (m_obj));
      }
    } else {
      return m_trans * bc (m_obj);
    }
  }

  /**
   *  @brief The raw bounding box of the array
   *
   *  The raw bounding box is the box that encloses the array instantiation
   *  points, ignoring the extensions of the object.
   *  The raw bounding boxes can be accumulated for all arrays having the same
   *  orientation, magnification and object. 
   *  bbox_from_raw_bbox can be used to compute the total bbox from such an
   *  accumulated raw bounding box. This is for example exploitet in the 
   *  update_bbox method of db::cell.
   */
  box_type raw_bbox () const
  {
    if (mp_base) {
      return mp_base->bbox (box_type (m_trans.disp (), m_trans.disp ()));
    } else {
      return box_type (m_trans.disp (), m_trans.disp ());
    }
  }

  /**
   *  @brief The bbox of the array computed from the raw bounding box
   *
   *  See "raw_bbox" for a description.
   */
  template <class BoxConv>
  box_type bbox_from_raw_bbox (const box_type &rb, const BoxConv &bc) const
  {
    if (mp_base) {
      if (mp_base->is_mag ()) {
        return rb * (m_trans.fp_trans () * (bc (m_obj) * mp_base->mag ()));
      } else {
        return rb * (m_trans.fp_trans () * bc (m_obj));
      }
    } else {
      return rb * (m_trans.fp_trans () * bc (m_obj));
    }
  }

  /**
   *  @brief The global transformation
   */
  const trans_type &global_trans () const
  {
    return m_trans;
  }

  /**
   *  @brief The number of single instances in the array
   */
  size_t size () const
  {
    if (mp_base) {
      return mp_base->size ();
    } else {
      return 1;
    }
  }

  /**
   *  @brief Fast access to the first instance
   */
  const trans_type &front () const
  {
    return m_trans;
  }

  /**
   *  @brief Access to the object in the instance
   */
  Obj &object ()
  {
    return m_obj;
  }

  /**
   *  @brief Access to the object in the instance (const version)
   */
  const Obj &object () const
  {
    return m_obj;
  }

  /**
   *  @brief Invert an array reference
   */
  void invert ()
  {
    if (mp_base) {
      mp_base->invert (m_trans);
    } else {
      m_trans.invert ();
    }
  }

  /**
   *  @brief Compare operator for equality
   */
  bool operator== (const array<Obj> &d) const
  {
    if (! mp_base) {
      return (m_trans == d.m_trans && m_obj == d.m_obj && ! d.mp_base);
    } else {
      if (m_trans != d.m_trans || ! (m_obj == d.m_obj) || type () != d.type ()) {
        return false;
      }
      return mp_base && mp_base->equal (d.mp_base);
    }
  }

  /**
   *  @brief Compare operator
   *
   *  Implementation of a comparison operator that just
   *  compares the object and the matrix part of the transformation.
   *  Needed for sorting of cell instances for example.
   */
  bool raw_less (const array<Obj> &d) const
  {
    if (! (m_obj == d.m_obj)) {
      return (m_obj < d.m_obj);
    }
    if (m_trans.rot () != d.m_trans.rot ()) {
      return (m_trans.rot () < d.m_trans.rot ());
    }
    if (mag () != d.mag ()) {
      return mag () < d.mag ();
    }
    return false;
  }

  /**
   *  @brief Functor derived from the raw_less compare operator
   */
  struct raw_less_f 
  {
    bool operator() (const array<Obj> &a, const array<Obj> &b) const
    {
      return a.raw_less (b);
    }
  };

  /**
   *  @brief Compare operator (equal)
   *
   *  Implementation of a comparison operator that just
   *  compares the object and the matrix part of the transformation.
   *  Needed for sorting of cell instances for example.
   */
  bool raw_equal (const array<Obj> &d) const
  {
    if (! (m_obj == d.m_obj)) {
      return false;
    }
    if (m_trans.rot () != d.m_trans.rot ()) {
      return false;
    }
    if (mag () != d.mag ()) {
      return false;
    }
    return true;
  }

  /**
   *  @brief Test, if the array is magnified
   *
   *  Returns true if the array represents magnified instances.
   */
  bool is_mag () const
  {
    return mp_base && mp_base->is_mag ();
  }

  /**
   *  @brief Return the type code 
   *
   *  Returns the type code of the array. Mainly used internally.
   */
  int type () const
  {
    return mp_base ? mp_base->type () : 0;
  }

  /**
   *  @brief Return the magnification factor
   *
   *  Returns the magnification to be applied on the instances.
   */
  double mag () const
  {
    return mp_base ? mp_base->mag () : 1.0;
  }

  /**
   *  @brief Test, if the array is a regular one
   *
   *  Returns true and the array parameters if the array is a regular
   *  one. 
   */
  bool is_regular_array (point_type &a, point_type &b, unsigned long &amax, unsigned long &bmax) const 
  {
    return mp_base && mp_base->is_regular_array (a, b, amax, bmax);
  }

  /**
   *  @brief Collect memory usage statistics
   */
  size_t mem_used () const
  {
    return sizeof (m_trans) + sizeof (mp_base) + db::mem_used (m_obj) + (mp_base ? mp_base->mem_used () : 0);
  }

  /**
   *  @brief Collect memory requirement statistics
   */
  size_t mem_reqd () const
  {
    return sizeof (m_trans) + sizeof (mp_base) + db::mem_reqd (m_obj) + (mp_base ? mp_base->mem_reqd () : 0);
  }

  /**
   *  @brief Swap with another object
   *
   *  This implementation does not require to reallocate the base pointer
   *  which might speed up somewhat the sorting of cell instances.
   */
  void
  swap (db::array<Obj> &b)
  {
    std::swap (m_trans, b.m_trans);
    std::swap (m_obj, b.m_obj);
    std::swap (mp_base, b.mp_base);
  }

private:
  Obj m_obj;
  trans_type m_trans;
  basic_array <Obj> *mp_base;
};

template <class Obj>
size_t mem_used (const array<Obj> &x)
{
  return x.mem_used ();
}

template <class Obj>
size_t mem_reqd (const array<Obj> &x)
{
  return x.mem_reqd ();
}

/**
 *  @brief A standard typedef for a box array
 */
typedef db::array <db::Box> BoxArray;

/**
 *  @brief A standard typedef for a box array in double coordinates
 */
typedef db::array <db::DBox> DBoxArray;

/**
 *  @brief A standard typedef for a cell instance array
 */
typedef db::array <db::CellInst> CellInstArray;

}

namespace std
{

//  injecting a global std::swap for arrays into the 
//  std namespace
template <class C>
void swap (db::array<C> &a, db::array<C> &b)
{
  a.swap (b);
}

} // namespace std

#endif

