
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbBox
#define HDR_dbBox

#include <dbPoint.h>
#include <dbTrans.h>
#include <dbShapeTag.h>
#include <string>

namespace db {

template <class Coord> class generic_repository;

/**
 *  @brief A box class
 *
 *  This object represents a box (a rectangular shape).
 *  Notation is: p1 is the lower left point, p2 the 
 *  upper right one. 
 *  A box can be empty. An empty box represents no area
 *  (not even a point). A box can be a point or a single
 *  line. In this case, the area is zero but the box still
 *  can overlap other boxes. 
 *  The template parameter C is the type to use for coordinate
 *  values.
 */

template <class C>
struct box
{
  typedef C coord_type;
  typedef box<C> box_type;
  typedef point<C> point_type;
  typedef typename coord_traits<C>::area_type area_type;
  typedef shape_tag< box<C> > tag;

  /**
   *  @brief Empty box constructor
   */
  box ()
    : m_p1 (1, 1), m_p2 (-1, -1)
  {
    //  .. nothing else ..
  }

  /**
   *  @brief Standard constructor with four coordinates
   *  
   *  Creates a box from four coordinates (left, bottom,
   *  right, top). The coordinates are sorted, so left and
   *  right can be swapped as well as top and bottom.
   *
   *  @param x1 The first x coordinate 
   *  @param y1 The first y coordinate
   *  @param x2 The second x coordinate
   *  @param y2 The second y coordinate
   */
  box (C x1, C y1, C x2, C y2)
    : m_p1 (x1 < x2 ? x1 : x2, y1 < y2 ? y1 : y2), 
      m_p2 (x2 > x1 ? x2 : x1, y2 > y1 ? y2 : y1)
  {
    //  .. nothing else ..
  }

  /**
   *  @brief The standard constructor taking two point objects
   *
   *  As the four coordinate constructor but accepting two
   *  point objects. The coordinates are sorted so the points
   *  do not neccessarily need to be lower/left or upper/right.
   *
   *  @param p1 The first point
   *  @param p2 The second point
   */
  box (const point<C> &p1, const point<C> &p2)
    : m_p1 (p1.x () < p2.x () ? p1.x () : p2.x (), p1.y () < p2.y () ? p1.y () : p2.y ()), 
      m_p2 (p2.x () > p1.x () ? p2.x () : p1.x (), p2.y () > p1.y () ? p2.y () : p1.y ())
  {
    //  .. nothing else ..
  }

  /**
   *  @brief The copy constructor that also does type conversions
   *
   *  The implementation relies on the ability of the point constructor
   *  to convert between types. It assumes that the conversion is 
   *  maintaining the order of the coordinates and the emptyness condition.
   */
  template <class D>
  box (const box<D> &b)
    : m_p1 (b.p1 ()), m_p2 (b.p2 ())
  {
    //  .. nothing else ..
  }

  /**
   *  @brief Standard conversion from DBox 
   */
  static box<C> from_double (const box<double> &b)
  {
    return box<C> (point<C>::from_double (b.p1 ()), point<C>::from_double (b.p2 ()));
  }

  /**
   *  @brief The (dummy) translation operator
   */
  void translate (const box<C> &d, db::generic_repository<C> &)
  {
    *this = d;
  }

  /**
   *  @brief The "less" operator to establish a sorting order.
   */
  bool operator< (const box<C> &b) const
  {
    return m_p1 < b.m_p1 || (m_p1 == b.m_p1 && m_p2 < b.m_p2);
  }

  /**
   *  @brief Equality test
   */
  bool operator== (const box<C> &b) const
  {
    if (empty () && b.empty ()) {
      return true;
    } else if (! empty () && ! b.empty ()) {
      return m_p1 == b.m_p1 && m_p2 == b.m_p2;
    } else {
      return false;
    }
  }

  /**
   *  @brief Inequality test
   */
  bool operator!= (const box<C> &b) const
  {
    return !operator== (b);
  }

  /**
   *  @brief Folding of boxes.
   *
   *  The *= operator folds the box with the one given as 
   *  the argument. The box resulting from "folding" is the
   *  outer boundary of the union set formed by placing 
   *  the second box at every point of the first.
   * 
   *  @param b The box to fold with *this.
   *
   *  @return The folded box.
   */
  box<C> &operator*= (const box<C> &b)
  {
    if (! b.empty ()) {
      if (! empty ()) {
        m_p1 += b.m_p1;
        m_p2 += b.m_p2;
      }
    } else {
      *this = b;
    }
    return *this;
  }

  /**
   *  @brief Joining of boxes.
   *
   *  The += operator joins the box with the one given as 
   *  the argument. Joining constructs a box that encloses
   *  both boxes given. Empty boxes are neutral: they do not
   *  change another box when joining. Overwrites *this
   *  with the result.
   * 
   *  @param b The box to join with *this.
   *
   *  @return The joined box.
   */
  box<C> &operator+= (const box<C> &b)
  {
    if (! b.empty ()) {
      if (empty ()) {
        *this = b;
      } else {
        db::point<C> p1 (m_p1.x () < b.m_p1.x () ? m_p1.x () : b.m_p1.x (),
                         m_p1.y () < b.m_p1.y () ? m_p1.y () : b.m_p1.y ());
        db::point<C> p2 (m_p2.x () > b.m_p2.x () ? m_p2.x () : b.m_p2.x (),
                         m_p2.y () > b.m_p2.y () ? m_p2.y () : b.m_p2.y ());
        m_p1 = p1;
        m_p2 = p2;
      }
    }
    return *this;
  }

  /**
   *  @brief Joining of a box with a point.
   *
   *  The += operator joins the box with a point such that
   *  the new box encloses the point and the old box.
   *  Overwrites *this with the result.
   * 
   *  @param p The point to join with *this.
   *
   *  @return The joined box.
   */
  box<C> &operator+= (const point<C> &p)
  {
    if (empty ()) {
      m_p1 = p;
      m_p2 = p;
    } else {
      db::point<C> p1 (m_p1.x () < p.x () ? m_p1.x () : p.x (),
                       m_p1.y () < p.y () ? m_p1.y () : p.y ());
      db::point<C> p2 (m_p2.x () > p.x () ? m_p2.x () : p.x (),
                       m_p2.y () > p.y () ? m_p2.y () : p.y ());
      m_p1 = p1;
      m_p2 = p2;
    }
    return *this;
  }

  /**
   *  @brief Intersection of boxes.
   *
   *  The intersection of two boxes is the largest
   *  box common to both boxes. The intersection may be 
   *  empty if both boxes to not touch. If the boxes do
   *  not overlap but touch the result may be a single
   *  line or point with an area of zero. Overwrites *this
   *  with the result.
   *
   *  @param c The box to take the intersection with
   *
   *  @return The intersection box.
   */

  box<C> &operator&= (const box<C> &b)
  {
    if (b.empty ()) {
      *this = box<C> ();
    } else if (! empty ()) {
      point<C> p1 (m_p1.x () > b.m_p1.x () ? m_p1.x () : b.m_p1.x (),
                   m_p1.y () > b.m_p1.y () ? m_p1.y () : b.m_p1.y ());
      point<C> p2 (m_p2.x () < b.m_p2.x () ? m_p2.x () : b.m_p2.x (),
                   m_p2.y () < b.m_p2.y () ? m_p2.y () : b.m_p2.y ());
      m_p1 = p1;
      m_p2 = p2;
    } 
    return *this;
  }

  /**
   *  @brief Returns the box moved by a certain distance
   *
   *  Moves the box by a given offset and returns the moved
   *  box. Does not modify *this. Does not check for coordinate
   *  overflows.
   *
   *  @param p The offset to move the box.
   *
   *  @return The moved box.
   */
  box<C> moved (const point<C> &p) const
  {
    box<C> b (*this);
    b.move (p);
    return b;
  }

  /**
   *  @brief Enlarges the box by a certain amount.
   *
   *  Enlarges the box by x and y value specified in the vector
   *  passed. Positive values with grow the box, negative ones
   *  will shrink the box. The result may be an empty box if the
   *  box disappears. The amount specifies the grow or shrink
   *  per edge. The width and height will change by twice the
   *  amount.
   *  Does not modify *this. Does not check for coordinate
   *  overflows.
   *
   *  @param p The grow or shrink amount in x and y direction
   *
   *  @return The enlarged box.
   */
  box<C> enlarged (const point<C> &p) const
  {
    box<C> b (*this);
    b.enlarge (p);
    return b;
  }

  /**
   *  @brief Transformation of the box with a fixpoint transformation
   * 
   *  Transforms the box with a given transformation and
   *  writes the result to *this. 
   *
   *  @param t The transformation to apply.
   *
   *  @return The transformed box.
   */
  box<C> &transform (const fixpoint_trans &t)
  {
    if (! empty ()) {
      *this = box<C> (t * m_p1, t * m_p2);
    }
    return *this;
  }

  /**
   *  @brief Returns the box transformed with a fixpoint transformation
   *
   *  Transforms the box and returns the result without changing
   *  the box.
   *
   *  @param t The transformation to apply.
   *
   *  @return The transformed box.
   */
  box<C> transformed (const fixpoint_trans &t) const
  {
    if (! empty ()) {
      return box<C> (t * m_p1, t * m_p2);
    } else {
      return box<C> ();
    }
  }

  /**
   *  @brief Transformation of the box
   * 
   *  Transforms the box with a given transformation and
   *  writes the result to *this. 
   *
   *  @param t The transformation to apply.
   *
   *  @return The transformed box.
   */
  template <class Tr>
  box<C> &transform (const Tr &t)
  {
    if (! empty ()) {
      *this = box<C> (t * m_p1, t * m_p2);
    }
    return *this;
  }

  /**
   *  @brief Returns the transformed box.
   *
   *  Transforms the box and returns the result without changing
   *  the box.
   *
   *  @param t The transformation to apply.
   *
   *  @return The transformed box.
   */
  template <class Tr>
  box<typename Tr::target_coord_type> transformed (const Tr &t) const
  {
    if (! empty ()) {
      return box<typename Tr::target_coord_type> (t * m_p1, t * m_p2);
    } else {
      return box<typename Tr::target_coord_type> ();
    }
  }

  /**
   *  @brief Moves the box
   *
   *  Like %moved but modifies the box so it becomes the moved
   *  box.
   *
   *  @param p The distance to move
   *
   *  @return The moved box.
   */
  box<C> &move (const point<C> &p)
  {
    m_p1 += p;
    m_p2 += p;
    return *this;
  }

  /**
   *  @brief Enlarges the box
   *
   *  Like %enlarged but modifies the box so it becomes the enlarged
   *  box.
   *
   *  @param p The grow/shrink to apply.
   *
   *  @return The enlarged box.
   */
  box<C> &enlarge (const point<C> &p)
  {
    m_p1 -= p;
    m_p2 += p;
    return *this;
  }

  /**
   *  @brief Accessor to the lower left point.
   *
   *  @return The lower left point
   */
  const point<C> &p1 () const
  {
    return m_p1;
  }

  /**
   *  @brief Accessor to the upper right point.
   *
   *  @return The upper right point
   */
  const point<C> &p2 () const
  {
    return m_p2;
  }

  /**
   *  @brief is_point predicate
   *
   *  A box has this predicate if it consists of exactly one point
   */
  bool is_point () const
  {
    return m_p1 == m_p2;
  }

  /**
   *  @brief Accessor to the center point
   *
   *  Due to rounding in integer space, the center coordinate may not
   *  be the exact center unless double coordinates as used.
   *  For very large boxes, width and height might overflow and the
   *  result may not be correct.
   *
   *  @return The center
   */
  point<C> center () const
  {
    return point<C> (m_p1.x () + width () / 2, m_p1.y () + height () / 2);
  }

  /**
   *  @brief The width of the box.
   *
   *  Returns the width of the box. May cause coordinate space
   *  overflow for very large boxes.
   *  Returns a negative value for empty boxes.
   *
   *  @return The width of the box.
   */
  C width () const
  {
    return m_p2.x () - m_p1.x ();
  }

  /**
   *  @brief The height of the box.
   *
   *  Returns the height of the box. May cause coordinate space
   *  overflow for very large boxes.
   *  Returns a negative value for empty boxes.
   *
   *  @return The height of the box.
   */
  C height () const
  {
    return m_p2.y () - m_p1.y ();
  }

  /**
   *  @brief The left boundary of the box.
   */
  C left () const
  {
    return m_p1.x ();
  }

  /**
   *  @brief The right boundary of the box.
   */
  C right () const
  {
    return m_p2.x ();
  }

  /**
   *  @brief The top boundary of the box.
   */
  C top () const
  {
    return m_p2.y ();
  }

  /**
   *  @brief The bottom boundary of the box.
   */
  C bottom () const
  {
    return m_p1.y ();
  }

  /**
   *  @brief Empty test of the box.
   *
   *  @return True if the box is empty (not if the area is zero)
   */
  bool empty () const
  {
    return m_p1.x () > m_p2.x () || m_p1.y () > m_p2.y ();
  }

  /**
   *  @brief Contains test.
   *
   *  Tests whether a point is inside the box.
   *  This includes if the point is on the box contour.
   *
   *  @param p The point to test against.
   *
   *  @return true if the point is inside p.
   */
  bool contains (const point<C> &p) const
  {
    if (empty ()) {
      return false;
    } else {
      return (m_p2.x () >= p.x () && m_p1.x () <= p.x ()) &&
             (m_p2.y () >= p.y () && m_p1.y () <= p.y ());
    }
  }

  /**
   *  @brief Inside test.
   *
   *  Tests whether the box is inside the given one. Returns false
   *  if either the box of the test box is empty.
   *  Inside implies also that the edges coincide.
   *
   *  @param b The box to test against.
   *
   *  @return true if the box is inside b.
   */
  bool inside (const box<C> &b) const
  {
    if (b.empty () || empty ()) {
      return false;
    } else {
      return (m_p1.x () >= b.m_p1.x () && m_p2.x () <= b.m_p2.x ()) &&
             (m_p1.y () >= b.m_p1.y () && m_p2.y () <= b.m_p2.y ());
    }
  }

  /**
   *  @brief Touching test.
   *
   *  Tests whether the box is touching the given one. Returns false
   *  if either the box of the test box is empty.
   *  Touching implies also overlap.
   *
   *  @param b The box to test against.
   *
   *  @return true if the box is touching b (has as least one common
   *  point with b)
   */
  bool touches (const box<C> &b) const
  {
    if (b.empty () || empty ()) {
      return false;
    } else {
      return (m_p1.x () <= b.m_p2.x () && b.m_p1.x () <= m_p2.x ()) &&
             (m_p1.y () <= b.m_p2.y () && b.m_p1.y () <= m_p2.y ());
    }
  }

  /**
   *  @brief Overlap test.
   *
   *  Tests whether the box is overlapping the given one. Returns false
   *  if either the box of the test box is empty.
   *
   *  @param b The box to test against.
   *
   *  @return true if the box is overlapping b (the area of the 
   *  intersection box is non-empty)
   */
  bool overlaps (const box<C> &b) const
  {
    if (b.empty () || empty ()) {
      return false;
    } else {
      return (m_p1.x () < b.m_p2.x () && b.m_p1.x () < m_p2.x ()) &&
             (m_p1.y () < b.m_p2.y () && b.m_p1.y () < m_p2.y ());
    }
  }

  /**
   *  @brief Computation of the area of a box
   *
   *  @return The area of the box. 0 if empty.
   */
  area_type area () const
  {
    if (empty ()) {
      return area_type (0);
    } else {
      return area_type (height ()) * area_type (width ());
    }
  }
  
  /** 
   *  @brief Conversion to string
   */
  std::string to_string () const
  {
    if (empty ()) {
      return "()";
    } else {
      return "(" + m_p1.to_string () + "),(" + m_p2.to_string () + ")";
    }
  }


private:
  point<C> m_p1, m_p2;
};

/**
 *  @brief "overlap" binary predicate
 */
template <class Box>
struct boxes_overlap
{
  bool operator() (const Box &b1, const Box &b2) const
  {
    return b1.overlaps (b2);
  }
};

/**
 *  @brief "touch" binary predicate
 */
template <class Box>
struct boxes_touch
{
  bool operator() (const Box &b1, const Box &b2) const
  {
    return b1.touches (b2);
  }
};

/**
 *  @brief Transformation with fixpoint transformation with the * operator
 *
 *  @param t Transformation to apply (first parameter)
 *  @param b The box to transform (second parameter)
 *
 *  @return The box b transformed by t.
 */
template <class C>
inline box<C>
operator* (const fixpoint_trans &t, const box<C> &b)
{
  return b.transformed (t);
}

/**
 *  @brief Transformation with the * operator
 *
 *  @param t Transformation to apply (first parameter)
 *  @param b The box to transform (second parameter)
 *
 *  @return The box b transformed by t.
 */
template <class C, class Tr>
inline box<typename Tr::target_coord_type> 
operator* (const Tr &t, const box<C> &b)
{
  return b.transformed (t);
}

/**
 *  @brief Intersection mapped on the & operator.
 *
 *  @param b1 The first box.
 *  @param b2 The second box.
 *
 *  @return The intersection of b1 and b2.
 */ 
template <class C>
inline box<C> 
operator& (const box<C> &b1, const box<C> &b2)
{
  box<C> bb (b1);
  bb &= b2;
  return bb;
}

/**
 *  @brief Scaling of a box
 *
 *  @param b The box to scale.
 *  @param s The scaling factor
 *
 *  @return The scaled box
 */ 
template <class C>
inline box<C> 
operator* (const box<C> &b, double s)
{
  if (b.empty ()) {
    return box<C> ();
  } else {
    return box<C> (b.p1 () * s, b.p2 () * s);
  }
}

/**
 *  @brief Box joining mapped on the + operator.
 *
 *  @param b1 The first box.
 *  @param b2 The second box.
 *
 *  @return b1 joined with b2.
 */ 
template <class C>
inline box<C> 
operator+ (const box<C> &b1, const box<C> &b2)
{
  box<C> bb (b1);
  bb += b2;
  return bb;
}

/**
 *  @brief "Folding" of two boxes
 *
 *  @param b1 The first box.
 *  @param b2 The second box.
 *
 *  @return b1 folded with b2 (see db::box::operator*= ()).
 */ 
template <class C>
inline box<C> 
operator* (const box<C> &b1, const box<C> &b2)
{
  box<C> bb (b1);
  bb *= b2;
  return bb;
}

/**
 *  @brief Stream output operator
 */
template <class C>
inline std::ostream &
operator<< (std::ostream &os, const box<C> &b)
{
  return (os << b.to_string ());
}

/**
 *  @brief The standard integer coordinate box
 */
typedef box<db::Coord>  Box;

/**
 *  @brief The double coordinate box
 */
typedef box<db::DCoord> DBox;

} // namespace db

#endif

