
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbBoxTree
#define HDR_dbBoxTree

#include <tlKDTree.h>
#include <tlVector.h>
#include <dbBox.h>
#include <dbMemStatistics.h>

#include <limits>

namespace db
{

// #define KDTREE 

#ifdef KDTREE

//  HINT:
//  A considerable speedup can be acchieved by rewriting the 
//  linear structure of the box tree and use the median object
//  in place and use left and right halves for lower and upper
//  branches. This would however, mean to change a lot of code 
//  in the iterators etc.

//  forward definitions for db::BoxConvert
struct simple_bbox_tag;
struct complex_bbox_tag;

/// @brief a helper class required for the box_tree implementation

template <class Box, class Obj, class BoxConv>
class box_tree_picker 
{
public:
  typedef typename Box::coord_type coord_type;
  typedef tl::vector<Obj> object_vector_type;
  typedef typename object_vector_type::const_iterator const_iterator;

  box_tree_picker ()
  { }

  box_tree_picker (const BoxConv &box_conv)
    : m_box_conv (box_conv)
  { }

  coord_type operator() (unsigned int l, const_iterator o) const 
  {
    switch (l % 4) {
    case 0:
    default:
      return m_box_conv (*o).left ();
    case 1:
      return m_box_conv (*o).bottom ();
    case 2:
      return -m_box_conv (*o).right ();
    case 3:
      return -m_box_conv (*o).top ();
    }
  }

private:
  BoxConv m_box_conv;
};

/// @brief a helper class required for the box_tree implementation

template <class Obj, class Box, class BoxConv>
class box_tree_cached_picker 
{
public:
  typedef typename Box::coord_type coord_type;
  typedef tl::vector<Obj> object_vector_type;
  typedef typename object_vector_type::const_iterator const_iterator;

  box_tree_cached_picker ()
  { }

  box_tree_cached_picker (const BoxConv &box_conv, const_iterator from, const_iterator to)
    : m_from (from)
  { 
    m_boxes.reserve (std::distance (from, to));

    for (const_iterator o = from; o != to; ++o) {
      m_boxes.push_back (box_conv (*o));
    }
  }

  coord_type operator() (unsigned int l, const_iterator o) const 
  {
    const Box &box = m_boxes [std::distance (m_from, o)];
    switch (l % 4) {
    case 0:
    default:
      return box.left ();
    case 1:
      return box.bottom ();
    case 2:
      return -box.right ();
    case 3:
      return -box.top ();
    }
  }

private:
  const_iterator m_from;
  tl::vector<Box> m_boxes;
};

/**
 *  @brief The iterator class
 *
 *  The iterator provides a region-select iterator for the box tree.
 *  It requires a compare operator (type Cmp) that provides coordinate-wise
 *  and box-wise comparison.
 */

template <class Tree, class Cmp>
class box_tree_it
{
public:
  typedef typename Tree::object_type object_type;
  typedef typename Tree::element element;
  typedef typename Tree::const_iterator const_iterator;
  typedef typename Tree::box_tree_picker_type value_picker_type;
  typedef typename Tree::difference_type difference_type;
  typedef typename Tree::size_type size_type;

  box_tree_it ()
    : m_j (0), m_n (0), m_l (0), m_path (0),
      mp_tree (0), m_picker (), m_compare () 
  { }

  box_tree_it (const Tree &t, value_picker_type p, const Cmp &c) 
    : m_j (t.size () / 2), m_n (t.size ()), m_l (0), m_path (0),
      mp_tree (&t), m_picker (p), m_compare (c) 
  { 
    if (m_n > 0 && need_visit ()) {
      traverse ();
      while (! at_end ()) {
        if (check ()) {
          break;
        }
        inc ();
      }
    } else {
      finish ();
    }
  }

  box_tree_it<Tree, Cmp> &operator++ () 
  {
    while (true) {
      inc ();
      if (at_end () || check ()) {
        break;
      }
    }
    return *this;
  }

  const object_type &operator* () const 
  {
    return *(mp_tree->elements () [m_j].ptr);
  }
  
  const_iterator position () const 
  {
    return mp_tree->elements () [m_j].ptr;
  }
  
  size_type internal_index () const 
  {
    return m_j;
  }

  /**
   *  @brief Return the index of the position indicated by this iterator
   */
  difference_type index () const
  {
    return std::distance (mp_tree->objects ().begin (), position ());
  }

  bool operator== (const box_tree_it<Tree, Cmp> &i) 
  {
    return m_j == i.m_j;
  }

  bool operator!= (const box_tree_it<Tree, Cmp> &i) 
  {
    return m_j != i.m_j;
  }

  bool operator== (size_type j)
  {
    return m_j == j;
  }

  bool operator!= (size_type j)
  {
    return m_j != j;
  }

  /**
   *  @brief returns true if at end
   */
  bool at_end () const
  {
    return m_j == m_n;
  }

private:
  size_type m_j;
  size_type m_n;
  int m_l; 
  size_type m_path;
  const Tree *mp_tree;
  value_picker_type m_picker;
  Cmp m_compare;

  //  return true if the node is within the search range
  bool check () const
  {
    return m_compare (*(mp_tree->elements () [m_j].ptr));
  }
  
  //  one level up. Returns true if from right
  bool up () 
  {
    size_type n = ((m_n + 1) >> m_l) - 1 + (m_path < ((m_n + 1) & ((1 << m_l) - 1)));

    --m_l;

    if (m_path & (1 << m_l)) {

      //  from right
      m_path &= ~(1 << m_l);
      m_j -= n / 2 + 1;
      return true;

    } else {

      //  from left
      m_j += n - n / 2;
      return false;

    }

  }

  //  one level down (right). Returns false if not possible
  bool down_right ()
  {
    m_path |= (1 << m_l);
    ++m_l;
    size_type n = ((m_n + 1) >> m_l) - 1 + (m_path < ((m_n + 1) & ((1 << m_l) - 1)));
    if (n == 0) {
      --m_l;
      m_path &= ~(1 << m_l);
      return false;
    } else {
      m_j += n / 2 + 1;
      return true;
    }
  }

  //  one level down (left). Returns false if not possible
  bool down_left ()
  {
    ++m_l;
    size_type n = ((m_n + 1) >> m_l) - 1 + (m_path < ((m_n + 1) & ((1 << m_l) - 1)));
    if (n == 0) {
      --m_l;
      return false;
    } else {
      m_j -= n - n / 2;
      return true;
    }
  }

  //  place at end
  void finish () 
  {
    m_j = m_n;
    m_l = -1;
  }
    
  //  return true if the current node needs a visit
  bool need_visit () const
  {
    return m_compare ((unsigned int) m_l, mp_tree->elements () [m_j].coord);
  }

  //  traverse as far as right down possible
  void traverse ()
  {
    bool next;

    do {
      next = false;
      if (down_right ()) {
        if (need_visit ()) {
          next = true;
        } else {
          up ();
        }
      } 
      if (! next && down_left ()) {
        if (need_visit ()) {
          next = true;
        } else {
          up ();
        }
      }
    } while (next);  
  }

  //  increment iterator
  void inc ()
  {
    if (! at_end ()) {
      if (up ()) { 
        if (! down_left ()) {
          // no left node. just stay here.
        } else if (! need_visit ()) {
          up ();
        } else {
          traverse ();
        }
      }
    }
  }
};

/// @brief a helper class required for the box_tree implementation

template <class Box, class Obj, class BoxConv, class BoxPred, class CoordPred>
class box_tree_sel
{
public:
  typedef typename Box::coord_type coord_type;

  box_tree_sel ()
  {
    m_c [0] = 0;
    m_c [1] = 0;
    m_c [2] = 0;
    m_c [3] = 0;
  }

  box_tree_sel (const Box &b, const BoxConv &conv) 
    : m_b (b), m_bpred (), m_cpred (), m_conv (conv)
  { 
    m_c [0] = m_b.right ();
    m_c [1] = m_b.top ();
    m_c [2] = -m_b.left ();
    m_c [3] = -m_b.bottom ();
  }

  bool operator() (const Obj &o) const
  {
#if 0 // this version requires no box predicate but may be somewhat slower:
    db::Box bx = m_conv(o);
    
    coord_type b[4];
    b[0] = bx.left ();
    b[1] = bx.bottom ();
    b[2] = -bx.right ();
    b[3] = -bx.top ();

    for (unsigned int i = 0; i < 4; ++i) {
      if (! m_cpred (b [i], m_c [i])) {
        return false;
      }
    }
    return true;
#else
    return m_bpred (m_conv (o), m_b);
#endif
  }

  bool operator() (unsigned int l, coord_type c) const
  {
    return m_cpred (c, m_c [l % 4]);
  }

private:
  Box m_b;
  coord_type m_c[4];
  BoxPred m_bpred;
  CoordPred m_cpred;
  BoxConv m_conv;
};

/** 
 *  @brief box tree object
 *
 *  A box tree is a vector with special sorting and
 *  query capabilities. It contains objects of type
 *  Obj that can be converted to db::box<C> objects with
 *  the BoxConv function. 
 *  A box tree can be in state "inserting", in which the
 *  insertion of new objects is supported, or can be 
 *  sorted, after which it can be queried for objects 
 *  whose box overlaps or touches a specified test box.
 */

template <class Box, class Obj, class BoxConv>
class box_tree 
{
public:
  typedef box_tree_picker<Box, Obj, BoxConv> box_tree_picker_type;
  typedef box_tree_sel<Box, Obj, BoxConv, db::boxes_overlap<Box>, std::less<typename Box::coord_type> > box_tree_sel_overlap_type;
  typedef box_tree_sel<Box, Obj, BoxConv, db::boxes_touch<Box>, std::less_equal<typename Box::coord_type> > box_tree_sel_touch_type;

  typedef Obj object_type;
  typedef tl::vector<Obj> obj_vector_type;
  typedef size_t size_type;
  typedef typename obj_vector_type::difference_type difference_type;
  typedef typename obj_vector_type::const_iterator const_iterator;
  typedef typename obj_vector_type::iterator iterator;
  typedef typename Box::coord_type coord_type;

  /**
   *  @brief A helper class that encapsulates a minimum coordinate and a pointer
   */
  struct element
  {
    coord_type coord;
    const_iterator ptr;

    bool operator< (const element &b) const
    {
      return coord < b.coord;
    }
  };

  typedef tl::vector<element> element_vector_type;
  typedef typename element_vector_type::iterator element_iterator;
  typedef box_tree<Box,Obj,BoxConv> box_tree_type;
  typedef box_tree_it<box_tree_type, box_tree_sel_touch_type> touching_iterator;
  typedef box_tree_it<box_tree_type, box_tree_sel_overlap_type> overlapping_iterator;

  /**
   *  @brief Creates a empty box tree object 
   */
  box_tree ()
  {
    // .. nothing else ..
  }

  /**
   *  @brief Insert a new object into the box tree
   *
   *  Inserts a new object into the box tree. The object
   *  is appended at the end of the vector. Sorting is 
   *  invalidated through this operation.
   *
   *  @param o The object to insert
   */
  iterator insert (const Obj &o)
  {
    m_objects.push_back (o);
    return m_objects.end () - 1;
  }

  /**
   *  @brief reserve space for a given number of elements
   */
  void reserve (size_t n)
  {
    m_objects.reserve (n);
  }

  /**
   *  @brief resize to a given number of elements
   */
  void resize (size_t n)
  {
    m_objects.resize (n);
  }

  /** 
   *  @brief Insert a range of objects
   *
   *  Analogeous to the other insert method, but accepting
   *  and range of objects [start,end).
   */
  template <class I> 
  void insert (I from, I to)
  {
    m_objects.insert (m_objects.end (), from, to);
  }

  /**
   *  @brief Replace an object
   *  
   *  Replace the object at the given (const) iterator position
   *  with the object given. Replacing an object invalidates
   *  the sorting state.
   *
   *  @param pos The position of the object to replace
   *  @param obj The object to replace the current one with
   */
  void replace (const_iterator pos, const Obj &obj)
  {
    m_objects [std::distance (((const box_tree_type *) this)->begin (), pos)] = obj;
  }

  /**
   *  @brief Erase an object
   *  
   *  Erase the object pointed to by the given iterator.
   *  Erasing an object will invalidate sorting. Erasing is
   *  not a cheap operation and its complexity is O(n).
   *
   *  @param pos The position of the object to erase
   */
  void erase (iterator pos)
  {
    m_objects.erase (pos);
  }

  /**
   *  @brief Erase an object sequence
   *  
   *  Erase the objects [from,to) to by the given iterator.
   *  Erasing an object will invalidate sorting. Erasing is
   *  not a cheap operation and its complexity is O(n).
   *
   *  @param pos The position of the object to erase
   */
  void erase (iterator from, iterator to)
  {
    m_objects.erase (from, to);
  }

  /**
   *  @brief Returns the size of the vector.
   *
   *  @return The number of elements in the vector.
   */
  size_type size () const
  {
    return m_objects.size ();
  }

  /** 
   *  @brief Empty the vector.
   */
  void clear ()
  {
    m_objects.clear ();
    m_elements.clear ();
  }

  /**
   *  @brief Sort the vector
   *
   *  Only after sorting the query iterators are available.
   *  Sorting complexity is approx O(N*log(N)).
   */
  void sort (const BoxConv &conv)
  {
    typename BoxConv::complexity complexity_tag;
    sort (conv, complexity_tag);
  }

  /**
   *  @brief Direct access to the underlying vector
   *
   *  @return A reference to the underlying object vector
   */
  const obj_vector_type &objects () const
  {
    return m_objects;
  }
  
  /**
   *  @brief Direct access to the underlying element vector
   *
   *  @return A reference to the underlying element vector
   */
  const element_vector_type &elements () const
  {
    return m_elements;
  }
  
  /**
   *  @brief Sequential access begin iterator
   *
   *  The normal sequential iterator delivers object by
   *  object in the sequence the objects have been stored
   *  in the vector. Using the sequential access iterator
   *  does not require a sorted vector.
   *
   *  @return A const iterator pointing to the first
   *  element in the vector
   */
  const_iterator begin () const
  {
    return m_objects.begin ();
  }

  /**
   *  @brief Sequential access end iterator
   *
   *  See %begin for a detailed description
   *
   *  @return A const iterator pointing past the end
   *  of the vector
   */
  const_iterator end () const
  {
    return m_objects.end ();
  }
  
  /**
   *  @brief Sequential access begin iterator
   *
   *  The normal sequential iterator delivers object by
   *  object in the sequence the objects have been stored
   *  in the vector. Using the sequential access iterator
   *  does not require a sorted vector.
   *
   *  @return A non-const iterator pointing to the first
   *  element in the vector
   */
  iterator begin () 
  {
    return m_objects.begin ();
  }

  /**
   *  @brief Sequential access end iterator
   *
   *  See %begin for a detailed description
   *
   *  @return A non-const iterator pointing past the end
   *  of the vector
   */
  iterator end () 
  {
    return m_objects.end ();
  }
  
  /** 
   *  @brief selection of objects touching the test box
   *
   *  The const iterator returned by this method will
   *  iterator over all objects touching the test box
   *  with their own box (as converted by BoxConv).
   *  The sequence the objects are delivered is arbitrary.
   *
   *  @param b The test box
   *  @param conv The box converter instance
   *
   *  @return The const touching iterator pointing to the first Object.
   */
  touching_iterator begin_touching (const Box &b, BoxConv conv) const 
  {
    box_tree_picker_type p (conv);
    box_tree_sel_touch_type s (b, conv);
    return box_tree_it<box_tree_type, box_tree_sel_touch_type> (*this, p, s);
  }
  
  /**
   *  @brief The past-the-end value for iteration of touching boxes
   *
   *  Although this value is not a iterator, it can be compared against
   *  overlapping_iterator objects which will render true if the
   *  iterator is past the end of the elements to consider.
   *
   *  @return A value suitable for comparison with a touching_iterator
   *  object.
   */
  size_type end_touching () const 
  {
    return size ();
  }

  /** 
   *  @brief selection of objects overlapping the test box
   *
   *  The const iterator returned by this method will
   *  iterator over all objects touching the test box
   *  with their own box (as converted by BoxConv).
   *  The sequence the objects are delivered is arbitrary.
   *
   *  @param b The test box
   *  @param conv The box converter instance
   *
   *  @return The const overlapping iterator pointing to the first Object.
   */
  overlapping_iterator begin_overlapping (const Box &b, BoxConv conv) const 
  {
    box_tree_picker_type p (conv);
    box_tree_sel_overlap_type s (b, conv);
    return box_tree_it<box_tree_type, box_tree_sel_overlap_type> (*this, p, s);
  }
  
  /**
   *  @brief The past-the-end value for iteration of overlapping boxes
   *
   *  Although this value is not a iterator, it can be compared against
   *  overlapping_iterator objects which will render true if the
   *  iterator is past the end of the elements to consider.
   *
   *  @return A value suitable for comparison with a overlapping_iterator
   *  object.
   */
  size_type end_overlapping () const 
  {
    return size ();
  }

private:
  /// The basic object and element vector
  obj_vector_type m_objects;
  element_vector_type m_elements;

  /// Sort implementation for simple bboxes - no caching
  void sort (const BoxConv &conv, const db::simple_bbox_tag &/*complexity*/)
  {
    if (m_objects.empty ()) {
      return;
    }

    box_tree_picker_type picker (conv);

    m_elements.clear ();
    m_elements.resize (m_objects.size ());

    typename element_vector_type::iterator e = m_elements.begin ();
    typename obj_vector_type::const_iterator o = m_objects.begin (); 
    for ( ; o != m_objects.end (); ++o, ++e) {
      e->coord = 0;
      e->ptr = o;
    }

    tree_sort (0, m_elements.begin (), m_elements.end (), picker);
  }

  /// Sort implementation for complex bboxes - with caching
  void sort (const BoxConv &conv, const db::complex_bbox_tag &/*complexity*/)
  {
    if (m_objects.empty ()) {
      return;
    }

    box_tree_cached_picker<Obj, Box, BoxConv> picker (conv, m_objects.begin (), m_objects.end ());

    m_elements.clear ();
    m_elements.resize (m_objects.size ());

    typename element_vector_type::iterator e = m_elements.begin ();
    typename obj_vector_type::const_iterator o = m_objects.begin (); 
    for ( ; o != m_objects.end (); ++o, ++e) {
      e->coord = 0;
      e->ptr = o;
    }

    tree_sort (0, m_elements.begin (), m_elements.end (), picker);
  }

  template <class CoordPicker>
  void tree_sort (unsigned int l, element_iterator from, element_iterator to, const CoordPicker &picker)
  {
    typename element_iterator::difference_type n = to - from;

    if (n == 1) {
      
      //  already "sorted"
      (*from).coord = picker (l, (*from).ptr);

    } else if (n == 2) {

      //  some simple optimization for the trivial case 2
      coord_type v1 = picker (l, from [0].ptr);
      coord_type v2 = picker (l, from [1].ptr);

      if (v1 > v2) {
        std::swap (v1, v2);
        std::swap (from [0], from [1]);
      }

      from [1].coord = v1;
      from [0].coord = picker (l + 1, from [0].ptr);

    } else {

      typename element_iterator::difference_type n1 = n / 2;
      element_iterator med = from + n1;

      //  obtain min and max coordinates

      coord_type min_coord = picker (l, (*from).ptr);
      (*from).coord = min_coord;

      for (element_iterator i = from + 1; i != to; ++i) {
        coord_type c = picker (l, (*i).ptr);
        (*i).coord = c;
        if (c < min_coord) {
          min_coord = c;
        }
      }
      
      std::nth_element (from, med, to);
          
      (*med).coord = min_coord;

      tree_sort (l + 1, from, med, picker);
      tree_sort (l + 1, med + 1, to, picker);

    }

  }

};

template <class Box, class Obj, class BoxConv>
size_t mem_used (const db::box_tree<Box, Obj, BoxConv> &bt)
{
  return mem_used (bt.elements ()) + 
         mem_used (bt.objects ());
}

template <class Box, class Obj, class BoxConv>
size_t mem_reqd (const db::box_tree<Box, Obj, BoxConv> &bt)
{
  return mem_reqd (bt.elements ()) + 
         mem_reqd (bt.objects ());
}

#else

//  forward definitions for db::BoxConvert
struct simple_bbox_tag;
struct complex_bbox_tag;

/// @brief a helper class required for the box_tree implementation

template <class Box, class Obj, class BoxConv>
class box_tree_picker 
{
public:
  typedef Box box_type;
  typedef tl::vector<Obj> object_vector_type;
  typedef typename object_vector_type::const_iterator const_iterator;

  box_tree_picker ()
  { }

  box_tree_picker (const BoxConv &box_conv)
    : m_box_conv (box_conv)
  { }

  box_type operator() (const_iterator o) const 
  {
    return m_box_conv (*o);
  }

private:
  BoxConv m_box_conv;
};

/// @brief a helper class required for the box_tree implementation

template <class Obj, class Box, class BoxConv>
class box_tree_cached_picker 
{
public:
  typedef Box box_type;
  typedef tl::vector<Obj> object_vector_type;
  typedef typename object_vector_type::const_iterator const_iterator;

  box_tree_cached_picker ()
  { }

  box_tree_cached_picker (const BoxConv &box_conv, const_iterator from, const_iterator to)
    : m_from (from)
  { 
    m_boxes.reserve (std::distance (from, to));

    for (const_iterator o = from; o != to; ++o) {
      m_boxes.push_back (box_conv (*o));
      m_bbox += m_boxes.back ();
    }
  }

  const box_type &operator() (const_iterator o) const 
  {
    return m_boxes [std::distance (m_from, o)];
  }

  const box_type &bbox () const
  {
    return m_bbox;
  }

private:
  const_iterator m_from;
  box_type m_bbox;
  tl::vector<box_type> m_boxes;
};

/**
 *  @brief The node object
 */

template <class Tree>
class box_tree_node
{
public:
  typedef typename Tree::point_type point_type;
  typedef typename Tree::coord_type coord_type;
  typedef typename Tree::box_type box_type;

  box_tree_node (box_tree_node *parent, const point_type &center, unsigned int quad) 
    : m_center (center)
  {
    for (int i = 0; i < 5; ++i) {
      m_lenq[i] = 0;
    }
    for (int i = 0; i < 4; ++i) {
      mp_children[i] = 0;
    }
    mp_parent = (box_tree_node *)((char *) parent + quad);
    if (parent) {
      parent->mp_children [quad] = this;
    }
  }

  ~box_tree_node ()
  {
    for (int i = 0; i < 4; ++i) {
      if (mp_children [i]) {
        delete mp_children [i];
        mp_children [i] = 0;
      }
    }
  }

  box_tree_node *child (int i) const
  {
    return mp_children [i];
  }

  void lenq (int i, size_t l) 
  {
    m_lenq[i + 1] = l;
  }

  size_t lenq (int i) const
  {
    return m_lenq[i + 1];
  }

  box_tree_node *parent () const
  {
    return (box_tree_node *)((char *) mp_parent - (size_t (mp_parent) & 3));
  }

  int quad () const
  {
    return (int)(size_t (mp_parent) & 3);
  }

  size_t mem_used () const
  {
    size_t m = sizeof (*this);
    for (int i = 0; i < 4; ++i) {
      if (mp_children [i]) {
        m += mp_children [i]->mem_used ();
      }
    }
    return m;
  }

  const point_type &center () const
  {
    return m_center;
  }

private:
  box_tree_node *mp_parent;
  size_t m_lenq [5];
  box_tree_node *mp_children [4];
  point_type m_center;
};

/**
 *  @brief The iterator class
 *
 *  The iterator provides a region-select iterator for the box tree.
 *  It requires a compare operator (type Cmp) that provides coordinate-wise
 *  and box-wise comparison.
 */

template <class Tree, class Cmp>
class box_tree_it
{
public:
  typedef typename Tree::object_type object_type;
  typedef typename Tree::point_type point_type;
  typedef typename Tree::coord_type coord_type;
  typedef typename Tree::box_type box_type;
  typedef typename Tree::element element;
  typedef typename Tree::const_iterator const_iterator;
  typedef typename Tree::box_tree_picker_type value_picker_type;
  typedef typename Tree::size_type size_type;
  typedef db::box_tree_node<Tree> box_tree_node;

  box_tree_it ()
    : mp_tree (0), m_picker (), m_compare () 
  { 
    mp_node = 0;
    m_index = 0;
    m_offset = 0;
    m_quad = -1;
  }

  box_tree_it (const Tree &t, value_picker_type p, const Cmp &c) 
    : mp_tree (&t), m_picker (p), m_compare (c) 
  { 
    mp_node = t.root ();
    m_index = 0;
    m_offset = 0;
    m_quad = -1;

    while (mp_node && mp_node->lenq (m_quad) == 0) {
      if (! next ()) {
        mp_node = 0;
      } else {
        down ();
      }
    }
    while (! at_end ()) {
      if (check ()) {
        break;
      }
      inc ();
    }
  }

  box_tree_it<Tree, Cmp> &operator++ () 
  {
    while (true) {
      inc ();
      if (at_end () || check ()) {
        break;
      }
    }
    return *this;
  }

  const object_type &operator* () const 
  {
    return *(mp_tree->elements () [m_index + m_offset]);
  }
  
  const object_type *operator-> () const 
  {
    return &*(mp_tree->elements () [m_index + m_offset]);
  }
  
  const_iterator position () const 
  {
    return mp_tree->elements () [m_index + m_offset];
  }
  
  /**
   *  @brief Return the index of the position indicated by this iterator
   */
  size_t index () const
  {
    return size_t (std::distance (mp_tree->objects ().begin (), position ()));
  }

  bool operator== (const box_tree_it<Tree, Cmp> &i) 
  {
    return mp_node == i.mp_node;
  }

  bool operator!= (const box_tree_it<Tree, Cmp> &i) 
  {
    return mp_node != i.mp_node;
  }

  /**
   *  @brief returns true if at end
   */
  bool at_end () const
  {
    return m_index + m_offset == mp_tree->elements ().size ();
  }

private:
  box_tree_node *mp_node;
  size_t m_index;
  size_t m_offset;
  int m_quad;
  const Tree *mp_tree;
  value_picker_type m_picker;
  Cmp m_compare;

  //  return true if the node is within the search range
  bool check () const
  {
    bool ret = m_compare.matches_obj (*(mp_tree->elements () [m_index + m_offset]));
    return ret;
  }
  
  //  check if the current quad needs visit
  bool need_visit () const
  {
    if (mp_node->lenq (m_quad) == 0) {
      return false;
    }
    if (m_quad < 0) {
      return true;
    } else {
      //  @@@ simplify this, i.e. by providing a special method on the compare function ..
      coord_type m = std::numeric_limits<coord_type>::max ();
      point_type c = mp_node->center ();
      if (m_quad == 0) {
        return m_compare.matches_box (box_type (c, point_type (m, m)));
      } else if (m_quad == 1) {
        return m_compare.matches_box (box_type (-m, c.y (), c.x (), m));
      } else if (m_quad == 2) {
        return m_compare.matches_box (box_type (point_type (-m, -m), c));
      } else {
        return m_compare.matches_box (box_type (c.x (), -m, m, c.y ()));
      }
    }
  }

  //  one level up. 
  bool up () 
  {
    box_tree_node *p = mp_node->parent ();
    if (p) {

      //  move the position to the beginning of the child node
      for (int q = -1; q < m_quad; ++q) {
        m_index -= mp_node->lenq (q);
      }

      m_quad = mp_node->quad ();
      mp_node = p;

      return true;

    } else {
      mp_node = 0;
      return false;
    }
  }

  //  move to next quad
  //  returns true if this is possible
  bool next () 
  {
    m_index += mp_node->lenq (m_quad);
    ++m_quad;
    while (m_quad < 4 && ! need_visit ()) {
      m_index += mp_node->lenq (m_quad);
      ++m_quad;
    }
    return m_quad < 4;
  }

  //  down one level
  //  returns true if this is possible
  bool down ()
  {
    box_tree_node *c = mp_node->child (m_quad);
    if (c) {

      mp_node = c;
      m_quad = -1;
      while (m_quad < 4 && ! need_visit ()) {
        m_index += mp_node->lenq (m_quad);
        ++m_quad;
      }

      if (m_quad == 4) {
        //  nothing to visit: up again
        up ();
        return false;
      } else {
        return true;
      }

    } else {
      return false;
    }
  }

  //  increment iterator
  void inc ()
  {
    if (! mp_node || m_offset + 1 < mp_node->lenq (m_quad)) {
      ++m_offset;
    } else {
      m_offset = 0;
      while (! next ()) {
        if (! up ()) {
          return;
        }
      }
      down ();
    }
  }
};

/// @brief a helper class required for the box_tree implementation

template <class Box, class Obj, class BoxConv, class BoxPred>
class box_tree_sel
{
public:
  typedef typename Box::coord_type coord_type;

  box_tree_sel ()
    : m_b (), m_bpred (), m_conv ()
  { 
    //  .. nothing yet ..
  }

  box_tree_sel (const Box &b, const BoxConv &conv) 
    : m_b (b), m_bpred (), m_conv (conv)
  { 
    //  .. nothing yet ..
  }

  bool matches_obj (const Obj &o) const
  {
    return m_bpred (m_conv (o), m_b);
  }

  bool matches_box (const Box &b) const
  {
    return m_bpred (b, m_b);
  }

private:
  Box m_b;
  BoxPred m_bpred;
  BoxConv m_conv;
};

/** 
 *  @brief box tree object
 *
 *  A box tree is a vector with special sorting and
 *  query capabilities. It contains objects of type
 *  Obj that can be converted to db::box<C> objects with
 *  the BoxConv function. 
 *  A box tree can be in state "inserting", in which the
 *  insertion of new objects is supported, or can be 
 *  sorted, after which it can be queried for objects 
 *  whose box overlaps or touches a specified test box.
 */

template <class Box, class Obj, class BoxConv, size_t min_bin = 100, size_t min_quads = 100>
class box_tree 
{
public:
  typedef box_tree_picker<Box, Obj, BoxConv> box_tree_picker_type;
  typedef box_tree_sel<Box, Obj, BoxConv, db::boxes_overlap<Box> > box_tree_sel_overlap_type;
  typedef box_tree_sel<Box, Obj, BoxConv, db::boxes_touch<Box> > box_tree_sel_touch_type;

  typedef Box box_type;
  typedef typename Box::coord_type coord_type;
  typedef db::point<coord_type> point_type;
  typedef Obj object_type;
  typedef tl::vector<Obj> obj_vector_type;
  typedef size_t size_type;
  typedef typename obj_vector_type::const_iterator const_iterator;
  typedef typename obj_vector_type::iterator iterator;
  typedef const_iterator element;
  typedef tl::vector<element> element_vector_type;
  typedef typename element_vector_type::iterator element_iterator;
  typedef box_tree<Box, Obj, BoxConv, min_bin, min_quads> box_tree_type;
  typedef db::box_tree_node<box_tree_type> box_tree_node;
  typedef box_tree_it<box_tree_type, box_tree_sel_touch_type> touching_iterator;
  typedef box_tree_it<box_tree_type, box_tree_sel_overlap_type> overlapping_iterator;

  /**
   *  @brief Creates a empty box tree object 
   */
  box_tree ()
    : mp_root (0)
  {
    // .. nothing else ..
  }

  /**
   *  @brief The destructor
   */
  ~box_tree ()
  {
    if (mp_root) {
      delete mp_root;
    }
    mp_root = 0;
  }

  /**
   *  @brief Insert a new object into the box tree
   *
   *  Inserts a new object into the box tree. The object
   *  is appended at the end of the vector. Sorting is 
   *  invalidated through this operation.
   *
   *  @param o The object to insert
   */
  iterator insert (const Obj &o)
  {
    m_objects.push_back (o);
    return m_objects.end () - 1;
  }

  /**
   *  @brief reserve space for a given number of elements
   */
  void reserve (size_t n)
  {
    m_objects.reserve (n);
  }

  /**
   *  @brief resize to a given number of elements
   */
  void resize (size_t n)
  {
    m_objects.resize (n);
  }

  /** 
   *  @brief Insert a range of objects
   *
   *  Analogeous to the other insert method, but accepting
   *  and range of objects [start,end).
   */
  template <class I> 
  void insert (I from, I to)
  {
    m_objects.insert (m_objects.end (), from, to);
  }

  /**
   *  @brief Replace an object
   *  
   *  Replace the object at the given (const) iterator position
   *  with the object given. Replacing an object invalidates
   *  the sorting state.
   *
   *  @param pos The position of the object to replace
   *  @param obj The object to replace the current one with
   */
  void replace (const_iterator pos, const Obj &obj)
  {
    m_objects [std::distance (((const box_tree_type *) this)->begin (), pos)] = obj;
  }

  /**
   *  @brief Erase an object
   *  
   *  Erase the object pointed to by the given iterator.
   *  Erasing an object will invalidate sorting. Erasing is
   *  not a cheap operation and its complexity is O(n).
   *
   *  @param pos The position of the object to erase
   */
  void erase (iterator pos)
  {
    m_objects.erase (pos);
  }

  /**
   *  @brief Erase an object sequence
   *  
   *  Erase the objects [from,to) to by the given iterator.
   *  Erasing an object will invalidate sorting. Erasing is
   *  not a cheap operation and its complexity is O(n).
   *
   *  @param pos The position of the object to erase
   */
  void erase (iterator from, iterator to)
  {
    m_objects.erase (from, to);
  }

  /**
   *  @brief Returns the size of the vector.
   *
   *  @return The number of elements in the vector.
   */
  size_type size () const
  {
    return m_objects.size ();
  }

  /** 
   *  @brief Empty the vector.
   */
  void clear ()
  {
    m_objects.clear ();
    m_elements.clear ();
  }

  /**
   *  @brief Sort the vector
   *
   *  Only after sorting the query iterators are available.
   *  Sorting complexity is approx O(N*log(N)).
   */
  void sort (const BoxConv &conv)
  {
    typename BoxConv::complexity complexity_tag;
    sort (conv, complexity_tag);
  }

  /**
   *  @brief Direct access to the underlying vector
   *
   *  @return A reference to the underlying object vector
   */
  const obj_vector_type &objects () const
  {
    return m_objects;
  }
  
  /**
   *  @brief Direct access to the underlying element vector
   *
   *  @return A reference to the underlying element vector
   */
  const element_vector_type &elements () const
  {
    return m_elements;
  }
  
  /**
   *  @brief Sequential access begin iterator
   *
   *  The normal sequential iterator delivers object by
   *  object in the sequence the objects have been stored
   *  in the vector. Using the sequential access iterator
   *  does not require a sorted vector.
   *
   *  @return A const iterator pointing to the first
   *  element in the vector
   */
  const_iterator begin () const
  {
    return m_objects.begin ();
  }

  /**
   *  @brief Sequential access end iterator
   *
   *  See %begin for a detailed description
   *
   *  @return A const iterator pointing past the end
   *  of the vector
   */
  const_iterator end () const
  {
    return m_objects.end ();
  }
  
  /**
   *  @brief Sequential access begin iterator
   *
   *  The normal sequential iterator delivers object by
   *  object in the sequence the objects have been stored
   *  in the vector. Using the sequential access iterator
   *  does not require a sorted vector.
   *
   *  @return A non-const iterator pointing to the first
   *  element in the vector
   */
  iterator begin () 
  {
    return m_objects.begin ();
  }

  /**
   *  @brief Sequential access end iterator
   *
   *  See %begin for a detailed description
   *
   *  @return A non-const iterator pointing past the end
   *  of the vector
   */
  iterator end () 
  {
    return m_objects.end ();
  }
  
  /** 
   *  @brief selection of objects touching the test box
   *
   *  The const iterator returned by this method will
   *  iterator over all objects touching the test box
   *  with their own box (as converted by BoxConv).
   *  The sequence the objects are delivered is arbitrary.
   *
   *  @param b The test box
   *  @param conv The box converter instance
   *
   *  @return The const touching iterator pointing to the first Object.
   */
  touching_iterator begin_touching (const Box &b, BoxConv conv) const 
  {
    box_tree_picker_type p (conv);
    box_tree_sel_touch_type s (b, conv);
    return box_tree_it<box_tree_type, box_tree_sel_touch_type> (*this, p, s);
  }
  
  /** 
   *  @brief selection of objects overlapping the test box
   *
   *  The const iterator returned by this method will
   *  iterator over all objects touching the test box
   *  with their own box (as converted by BoxConv).
   *  The sequence the objects are delivered is arbitrary.
   *
   *  @param b The test box
   *  @param conv The box converter instance
   *
   *  @return The const overlapping iterator pointing to the first Object.
   */
  overlapping_iterator begin_overlapping (const Box &b, BoxConv conv) const 
  {
    box_tree_picker_type p (conv);
    box_tree_sel_overlap_type s (b, conv);
    return box_tree_it<box_tree_type, box_tree_sel_overlap_type> (*this, p, s);
  }
  
  /**
   *  @brief Access to the root node object
   *
   *  This is mainly used by the iterator implementation
   */
  box_tree_node *root () const
  {
    return mp_root;
  }

private:
  /// The basic object and element vector
  obj_vector_type m_objects;
  element_vector_type m_elements;
  box_tree_node *mp_root;

  /// Sort implementation for simple bboxes - no caching
  void sort (const BoxConv &conv, const db::simple_bbox_tag &/*complexity*/)
  {
    if (m_objects.empty ()) {
      return;
    }

    box_tree_picker_type picker (conv);

    m_elements.clear ();
    m_elements.reserve (m_objects.size ());

    box_type bbox;
    for (typename obj_vector_type::const_iterator o = m_objects.begin (); o != m_objects.end (); ++o) {
      box_type b = conv (*o);
      if (! b.empty ()) {
        m_elements.push_back (o);
        bbox += b;
      }
    }

    //  @@@ resize m_elements to actual size ?

    if (mp_root) {
      delete mp_root;
    }
    mp_root = 0;

    tree_sort (0, m_elements.begin (), m_elements.end (), picker, bbox, 0);
  }

  /// Sort implementation for complex bboxes - with caching
  void sort (const BoxConv &conv, const db::complex_bbox_tag &/*complexity*/)
  {
    if (m_objects.empty ()) {
      return;
    }

    box_tree_cached_picker<Obj, Box, BoxConv> picker (conv, m_objects.begin (), m_objects.end ());

    m_elements.clear ();
    m_elements.reserve (m_objects.size ());

    for (typename obj_vector_type::const_iterator o = m_objects.begin (); o != m_objects.end (); ++o) {
      if (! picker (o).empty ()) {
        m_elements.push_back (o);
      }
    }

    //  @@@ resize m_elements to actual size ?

    if (mp_root) {
      delete mp_root;
    }
    mp_root = 0;

    tree_sort (0, m_elements.begin (), m_elements.end (), picker, picker.bbox (), 0);
  }

  template <class CoordPicker>
  void tree_sort (box_tree_node *parent, element_iterator from, element_iterator to, const CoordPicker &picker, const box_type &bbox, int quad)
  {
    size_t ntot = size_t (to - from);
    if (ntot <= min_bin) {
      return; //  not worth splitting
    } 

    element_iterator qloc [5] = { from, from, from, from, from };
    point_type center (bbox.center ());

    for (element_iterator e = from; e != to; ++e) {

      box_type b = picker (*e);

      int q = 0;
      if (b.right () <= center.x ()) {
        if (b.top () <= center.y ()) {
          q = 3;
        } else if (b.bottom () >= center.y ()) {
          q = 2;
        }
      } else if (b.left () >= center.x ()) {
        if (b.top () <= center.y ()) {
          q = 4;
        } else if (b.bottom () >= center.y ()) {
          q = 1;
        }
      }

      //  make space for the element and swap the new element into position 
      if (q < 4) {
        element el = *e;
        for (int i = 4; i > q; --i) {
          *qloc [i] = *qloc [i - 1];
          ++qloc [i];
        }
        *qloc [q] = el;
      }
      ++qloc [q];
      
    }

    //  compute sizes of quad fields
    size_t nx, n[4];
    size_t nn = 0;
    nx = std::distance (from, qloc[0]);
    for (int i = 0; i < 4; ++i) {
      n[i] = std::distance (qloc [i], qloc[i + 1]);
      nn += n[i];
    }

    //  is it worth to split into sub-quads?
    if (nn >= min_quads) {

      //  create a new node representing this tree
      box_tree_node *node = new box_tree_node (parent, center, quad);
      if (parent == 0) {
        mp_root = node;
      }

      //  tell the parent the length of the "overall" bin
      node->lenq (-1, nx);

      //  yes: create sub-quads
      box_type qboxes [4];
      qboxes [0] = box_type (center, bbox.p2 ()); 
      qboxes [1] = box_type (bbox.left (), center.y (), center.x (), bbox.top ());
      qboxes [2] = box_type (bbox.p1 (), center);
      qboxes [3] = box_type (center.x (), bbox.bottom (), bbox.right (), center.y ());
      for (unsigned int q = 0; q < 4; ++q) {
        if (n[q] > 0) {
          node->lenq (q, n[q]);
          tree_sort (node, qloc[q], qloc[q + 1], picker, qboxes [q], int (q));
        }
      }

    } 

  }

};

template <class Box, class Obj, class BoxConv>
size_t mem_used (const db::box_tree<Box, Obj, BoxConv> &bt)
{
  return mem_used (bt.elements ()) + 
         mem_used (bt.objects ()) +
         sizeof (void *) + 
         (bt.root () ? bt.root ()->mem_used () : 0);
}

template <class Box, class Obj, class BoxConv>
size_t mem_reqd (const db::box_tree<Box, Obj, BoxConv> &bt)
{
  return mem_reqd (bt.elements ()) + 
         mem_reqd (bt.objects ()) +
         sizeof (box_tree_node<db::box_tree<Box, Obj, BoxConv> > *) +
         sizeof (void *) + 
         (bt.root () ? bt.root ()->mem_used () : 0);
}

#endif


}

#endif

