
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_dbCell
#define HDR_dbCell

#include <dbObject.h>
#include <dbMemStatistics.h>
#include <dbBox.h>
#include <dbBoxTree.h>
#include <dbBoxConvert.h>
#include <dbShape.h>
#include <dbShapes.h>
#include <dbCellInst.h>
#include <dbArray.h>
#include <tlVector.h>
#include <tlAlgorithm.h>

#include <map>

namespace db
{

class CellGraph;

template <class CellInst> class cell_inst_index_box_convert;
template <class Graph> class cell;
template <class Coord> class generic_repository;

/**
 *  @brief The cell inst array per-layer box converter
 *
 *  This class is used as a function to convert a cell instance array
 *  to a box for a given layer. This requires that the per-layer 
 *  bounding boxes of the cell have been computed already.
 */

template <class CellInst>
struct cell_inst_index_box_convert 
{
  typedef db::array <CellInst> cell_inst_array_type;
  typedef typename CellInst::graph_type graph_type;
  typedef typename CellInst::box_type box_type;
  typedef cell<graph_type> cell_type;

  typedef db::complex_bbox_tag complexity;

  cell_inst_index_box_convert (const graph_type &g, const cell_type &c, unsigned int l)
    : m_bc (g, l), mp_cell (&c)
  { }
  
  cell_inst_index_box_convert (const graph_type &g, const cell_type &c)
    : m_bc (g), mp_cell (&c)
  { }
  
  box_type operator() (size_t t) const
  {
    return box_type (mp_cell->cell_inst (t).bbox (m_bc));
  }

private:
  box_convert <CellInst> m_bc;
  const cell_type *mp_cell;
};

/**
 *  @brief A parent instance
 *
 *  A parent instance is basically an inverse instance: instead of pointing
 *  to the child cell, it is pointing to the parent cell and the transformation
 *  is representing the shift of the parent cell relative to the child cell.
 *  For memory performance, a parent instance is not stored as a instance but
 *  rather as a reference to a child instance and a reference to the cell which
 *  is the parent.
 *  The parent instance itself is computed on the fly. It is representative for
 *  a set of instances belonging to the same cell index. The special parent instance
 *  iterator takes care of producing the right sequence.
 */

template <class Graph>
class parent_inst
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef typename Graph::box_type box_type;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::array<cell_inst_type> cell_inst_array_type;
  typedef db::cell<graph_type> cell_type;

  /**
   *  @brief The default ctor
   *
   *  This creates an invalid instance.
   */
  parent_inst ()
  {
    m_parent_cell_index = (cell_index_type) -1;
  }

  /**
   *  @brief Create a parent instance using the parent cell and instance
   */
  parent_inst (cell_index_type parent_cell_index, typename tl::vector<cell_inst_array_type>::const_iterator iter)
    : m_parent_cell_index (parent_cell_index), m_iter (iter)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Retrieve the reference to the parent cell 
   */
  cell_index_type parent_cell_index () const
  {
    return m_parent_cell_index;
  }

  /**
   *  @brief Retrieve the child instance associated with this parent instance
   */
  const cell_inst_array_type &child_inst () const
  {
    return *m_iter;
  }

  /**
   *  @brief Compute the inverse instance by which the parent is seen from the child
   */
  cell_inst_array_type inst () const
  {
    //  create a new parent instance by cloning and inverting the array
    cell_inst_array_type ci (*m_iter);
    ci.object () = cell_inst_type (m_parent_cell_index);
    ci.invert ();
    return ci;
  }

  /**
   *  @brief Access to the raw iterator
   *
   *  This method is used mainly by the parent instance iterator
   */
  typename tl::vector<cell_inst_array_type>::const_iterator iter () const
  {
    return m_iter;
  }

  /**
   *  @brief Increment the raw iterator
   *
   *  This method is used mainly by the parent instance iterator
   */
  void inc_iter () 
  {
    ++m_iter;
  }

private:
  cell_index_type m_parent_cell_index;
  typename tl::vector<cell_inst_array_type>::const_iterator m_iter;
};

/**
 *  @brief A parent instance iterator
 *
 *  The iterator is first an iterator over a vector of parent instances but then
 *  takes care of decomposing a single instance into a set of such.
 */

template <class Graph>
class parent_inst_iterator
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::cell<graph_type> cell_type;
  typedef parent_inst<Graph> parent_inst_type;

  /**
   *  @brief Default ctor
   */
  parent_inst_iterator ()
    : mp_graph (0), m_iter (), m_end ()
  { }

  /**
   *  @brief Constructor
   *
   *  This creates an iterator pointing to the start of the sequence represented
   *  by the parent instance object at *iter or at the end of the parent instance
   *  vector.
   */
  parent_inst_iterator (Graph *graph, typename tl::vector<parent_inst_type>::const_iterator iter, typename tl::vector<parent_inst_type>::const_iterator end)
    : mp_graph (graph), m_iter (iter), m_end (end)
  {
    if (m_iter != m_end) {
      m_rep = *m_iter;
    }
  }

  /** 
   *  @brief Equality operator
   */
  bool operator== (const parent_inst_iterator &b) const
  {
    return m_iter == b.m_iter && m_rep.iter () == b.m_rep.iter ();
  }

  /** 
   *  @brief Inequality operator
   */
  bool operator!= (const parent_inst_iterator &b) const
  {
    return ! (m_iter == b.m_iter && m_rep.iter () == b.m_rep.iter ());
  }

  /**
   *  @brief Access to the actual instance
   *
   *  The actual instance is created on the fly by the update_rep method.
   *  This will create an artificial instance which takes into account the
   *  offset from the representative.
   */
  const parent_inst_type &operator* () const
  {
    return m_rep;
  }

  /**
   *  @brief Access to the actual object
   * 
   *  See above.
   */
  const parent_inst_type *operator-> () const
  {
    return &m_rep;
  }

  /**
   *  @brief Increment operator
   */
  parent_inst_iterator &operator++() 
  {
    cell_index_type ci = m_rep.iter ()->object ().cell_index ();
    m_rep.inc_iter ();

    if (m_rep.iter () == mp_graph->cell (m_rep.parent_cell_index ()).end () ||
        m_rep.iter ()->object ().cell_index () != ci) {
      ++m_iter;
      if (m_iter != m_end) {
        m_rep = *m_iter;
      } else {
        m_rep = parent_inst_type ();
      }
    }

    return *this;
  }

  /**
   *  @brief Returns true, if the iterator is at the end
   */
  bool at_end () const
  {
    return m_iter == m_end;
  }
  
private:
  Graph *mp_graph;
  typename tl::vector<parent_inst_type>::const_iterator m_iter, m_end;
  parent_inst_type m_rep;
};

/**
 *  @brief A parent cell iterator
 *
 *  The iterator will report just cell indices, not instances.
 */

template <class Graph>
class parent_cell_iterator
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef parent_inst<Graph> parent_inst_type;

  /**
   *  @brief Default ctor
   */
  parent_cell_iterator ()
    : m_iter ()
  { }

  /**
   *  @brief Constructor
   */
  parent_cell_iterator (typename tl::vector<parent_inst_type>::const_iterator iter)
    : m_iter (iter)
  { }

  /** 
   *  @brief Equality operator
   */
  bool operator== (const parent_cell_iterator &b) const
  {
    return m_iter == b.m_iter;
  }

  /** 
   *  @brief Inequality operator
   */
  bool operator!= (const parent_cell_iterator &b) const
  {
    return m_iter != b.m_iter;
  }

  /**
   *  @brief Access to the actual instance
   *
   *  The actual instance is created on the fly by the update_rep method.
   *  This will create an artificial instance which takes into account the
   *  offset from the representative.
   */
  cell_index_type operator* () const
  {
    return m_iter->parent_cell_index ();
  }

  /**
   *  @brief Increment operator
   */
  parent_cell_iterator &operator++() 
  {
    ++m_iter;
    return *this;
  }

private:
  typename tl::vector<parent_inst_type>::const_iterator m_iter;
};

/**
 *  @brief A child cell iterator
 *
 *  The iterator will report just cell indices, not instances.
 */

template <class Graph>
class child_cell_iterator
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::array<cell_inst_type> cell_inst_array_type;

  /**
   *  @brief Default ctor
   */
  child_cell_iterator ()
    : m_iter (), m_end ()
  { }

  /**
   *  @brief Constructor
   */
  child_cell_iterator (typename tl::vector<cell_inst_array_type>::const_iterator iter, typename tl::vector<cell_inst_array_type>::const_iterator end)
    : m_iter (iter), m_end (end)
  { }

  /** 
   *  @brief Equality operator
   */
  bool operator== (const child_cell_iterator &b) const
  {
    return m_iter == b.m_iter;
  }

  /** 
   *  @brief Inequality operator
   */
  bool operator!= (const child_cell_iterator &b) const
  {
    return m_iter != b.m_iter;
  }

  /**
   *  @brief Access to the actual instance
   *
   *  The actual instance is created on the fly by the update_rep method.
   *  This will create an artificial instance which takes into account the
   *  offset from the representative.
   */
  cell_index_type operator* () const
  {
    return m_iter->object ().cell_index ();
  }

  /**
   *  @brief Increment operator
   */
  child_cell_iterator &operator++() 
  {
    cell_index_type ci = operator* ();
    do {
      ++m_iter;
    } while (m_iter != m_end && ci == operator* ());
    return *this;
  }

  /**
   *  @brief Report the end of the iterator (in non-STL schemes)
   */
  bool at_end () const
  {
    return m_iter == m_end;
  }

private:
  typename tl::vector<cell_inst_array_type>::const_iterator m_iter, m_end;
};

/**
 *  @brief The cell object
 * 
 *  A cell object consists of a set of shape containers (called layers),
 *  a set of child cell instances and auxiliary informations such as
 *  the parent instance list.
 *  A cell is identified through an index given to the cell upon instantiation.
 *  The cell index is valid in the context of a cell graph object which
 *  must issue the cell index.
 */

template <class Graph>
class cell
  : public db::Object
{
public:
  typedef Graph graph_type;
  typedef typename Graph::box_type box_type;
  typedef tl::vector<box_type> box_vector;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::array<cell_inst_type> cell_inst_array_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef tl::vector<db::shapes<coord_type, graph_type> *> shapes_vector;
  typedef db::box_convert<cell_inst_array_type> cell_inst_array_box_converter;
  typedef cell_inst_index_box_convert<cell_inst_type> cell_inst_index_box_convert;
  typedef db::box_tree<box_type, cell_inst_array_type, cell_inst_array_box_converter> cell_inst_tree_type;
  typedef typename cell_inst_tree_type::touching_iterator touching_iterator;
  typedef typename cell_inst_tree_type::overlapping_iterator overlapping_iterator;
  typedef db::box_tree<box_type, size_t, cell_inst_index_box_convert> cell_inst_layer_tree_type;
  typedef typename cell_inst_layer_tree_type::touching_iterator touching_layer_iterator;
  typedef typename cell_inst_layer_tree_type::overlapping_iterator overlapping_layer_iterator;
  typedef tl::vector<cell_inst_layer_tree_type> cell_inst_layer_tree_vector;
  typedef typename cell_inst_tree_type::iterator iterator;
  typedef typename cell_inst_tree_type::const_iterator const_iterator;
  typedef db::parent_inst<graph_type> parent_inst_type;
  typedef tl::vector<parent_inst_type> parent_inst_vector;
  typedef db::parent_inst_iterator<graph_type> parent_inst_iterator;
  typedef db::parent_cell_iterator<graph_type> parent_cell_iterator;
  typedef db::child_cell_iterator<graph_type> child_cell_iterator;

  /**
   *  @brief Standard constructor: create an empty cell object
   * 
   *  Takes the manager object from the graph object.
   *
   *  @param ci The index of the cell
   *  @param g A reference to the graph object that owns the cell
   */
  cell (cell_index_type ci, graph_type &g) 
    : db::Object (g.manager ()), 
      m_cell_index (ci), mp_graph (&g), m_hier_levels (0), m_bbox_needs_update (false)
  {
    //  .. nothing yet 
  }

  /**
   *  @brief The assignment operator 
   */
  cell &operator= (const cell &d)
  {
    if (this != &d) {

      m_cell_index = d.m_cell_index;

      //  because the shapes objects do not have a good 
      //  copy ctor which allows to attach them to a graph, 
      //  do the copy in this somewhat strange way ..
      clear_shapes_no_invalidate ();
      m_shapes_vector.reserve (d.m_shapes_vector.size ());
      for (unsigned int l = 0; l < d.m_shapes_vector.size (); ++l) {
        m_shapes_vector.push_back (new db::shapes<coord_type, graph_type> (*mp_graph));
        *(m_shapes_vector.back ()) = *(d.m_shapes_vector [l]);
      }

      m_cell_inst_trees = d.m_cell_inst_trees;
      m_cell_inst_tree = d.m_cell_inst_tree;
      m_parent_insts = d.m_parent_insts;
      m_bbox = d.m_bbox;
      m_bboxes = d.m_bboxes;
      m_hier_levels = d.m_hier_levels;
      m_bbox_needs_update = d.m_bbox_needs_update;

    }
    return *this;
  }

  /**
   *  @brief The copy ctor
   *
   *  In contrast to the assignment operator the copy ctor also
   *  copies the owning graph reference.
   */
  cell (const cell &d)
    : db::Object (d),
      mp_graph (d.mp_graph), m_hier_levels (d.m_hier_levels)
  {
    operator= (d);
  }

  /**
   *  @brief The destructor
   */
  ~cell ()
  {
    clear_shapes ();
  }

  /**
   *  @brief The translation operator 
   */
  void translate (const cell &d, generic_repository<coord_type> &rep)
  {
    tl_assert (this != &d);

    m_cell_index = d.m_cell_index;

    //  because the shapes objects do not have a good 
    //  copy ctor which allows to attach them to a graph, 
    //  do the copy in this somewhat strange way ..
    clear_shapes_no_invalidate ();
    m_shapes_vector.reserve (d.m_shapes_vector.size ());
    for (unsigned int l = 0; l < d.m_shapes_vector.size (); ++l) {
      m_shapes_vector.push_back (new db::shapes<coord_type, graph_type> (*mp_graph));
      m_shapes_vector.back ()->translate (*(d.m_shapes_vector [l]), rep);
    }

    m_cell_inst_trees = d.m_cell_inst_trees;
    m_cell_inst_tree = d.m_cell_inst_tree;
    m_parent_insts = d.m_parent_insts;
    m_bbox = d.m_bbox;
    m_bboxes = d.m_bboxes;
    m_hier_levels = d.m_hier_levels;
    m_bbox_needs_update = d.m_bbox_needs_update;
  }

  /**
   *  @brief Return the const shapes list of the given layer
   *
   *  This method allows to access the shapes list on a certain layer.
   *  If the layer does not exist yet, a reference to an empty list is
   *  returned.
   *
   *  @param index The layer index of the shapes list to retrieve
   *
   *  @return A const reference to the shapes list
   */
  db::shapes<coord_type, graph_type> &shapes (unsigned int index) 
  {
    if (m_shapes_vector.capacity () <= index) {
      m_shapes_vector.reserve (index + 1);
    }
    while (m_shapes_vector.size () <= index) {
      m_shapes_vector.push_back (new db::shapes<coord_type, graph_type> (*mp_graph));
    }
    return *(m_shapes_vector [index]);
  }

  /**
   *  @brief Return the shapes list of the given layer
   *
   *  This method allows to access the shapes list on a certain layer.
   *  If the layer does not exist yet, it is created.
   *
   *  @param index The layer index of the shapes list to retrieve
   *
   *  @return A reference to the shapes list
   */
  const db::shapes<coord_type, graph_type> &shapes (unsigned int index) const
  {
    if (m_shapes_vector.size () > index) {
      return *(m_shapes_vector [index]);
    } else {
      //  Because of a gcc bug it seems to be not possible
      //  to instantiate a simple static object here:
      static const db::shapes<coord_type, graph_type> *empty_shapes = 0;
      if (! empty_shapes) {
        empty_shapes = new db::shapes<coord_type, graph_type> ();
      }
      return *empty_shapes;
    }
  }

  /**
   *  @brief Clear all shapes in the cell
   */
  void clear_shapes ()
  {
    clear_shapes_no_invalidate ();
    mp_graph->invalidate_bboxes ();
  }

  /**
   *  @brief Clear the instance list
   */
  void clear_insts ()
  {
    if (m_cell_inst_tree.size () > 0) {
      m_cell_inst_tree.clear ();
      m_cell_inst_trees.clear ();
      mp_graph->invalidate_graph ();
      m_bbox_needs_update = true;
    }
  }

  /**
   *  @brief Swap the layers given
   */
  void swap (unsigned int i1, unsigned int i2)
  {
    if (i1 != i2) {
      shapes (i1).swap (shapes (i2));
    }
  }

  /**
   *  @brief Clear the shapes on the given layer
   */
  void clear (unsigned int index)
  {
    shapes (index).clear ();
    mp_graph->invalidate_bboxes ();
  }

  /**
   *  @brief Return the number of layers in this cell
   *
   *  The number of layer is the number of layers actually 
   *  created. There may be more layers but those are definitly
   *  empty.
   */
  unsigned int layers () const
  {
    return m_shapes_vector.size ();
  }

  /**
   *  @brief Reserve space for the given number of layers
   */
  void reserve_layers (unsigned int s)
  {
    m_shapes_vector.reserve (s);
  }

  /**
   *  @brief Erase a cell instance
   *  
   *  Erasing a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. update_bbox() must be called 
   *  explicitly before the bounding box can be updated.
   */
  void erase (iterator e)
  {
    m_cell_inst_tree.erase (e);
    mp_graph->invalidate_bboxes ();
    mp_graph->invalidate_graph ();
    m_bbox_needs_update = true;
  }

  /**
   *  @brief Erase a sequence of cell instances
   *  
   *  Erasing cell instances will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. update_bbox() must be called 
   *  explicitly before the bounding box can be updated.
   */
  void erase (iterator from, iterator to)
  {
    m_cell_inst_tree.erase (from, to);
    mp_graph->invalidate_bboxes ();
    mp_graph->invalidate_graph ();
    m_bbox_needs_update = true;
  }

  /**
   *  @brief Insert a single cell instance
   *  
   *  Inserting a single cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. update_bbox() must be called 
   *  explicitly before the bounding box can be updated.
   */
  cell_inst_array_type &insert (const cell_inst_type &inst)
  {
    mp_graph->invalidate_bboxes ();
    mp_graph->invalidate_graph ();
    iterator obj = m_cell_inst_tree.insert (cell_inst_array_type (inst));
    m_bbox_needs_update = true;
    return *obj;
  }

  /**
   *  @brief Insert a cell instance
   *  
   *  Inserting a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. update_bbox() must be called 
   *  explicitly before the bounding box can be updated.
   */
  cell_inst_array_type &insert (const cell_inst_array_type &inst)
  {
    mp_graph->invalidate_bboxes ();
    mp_graph->invalidate_graph ();
    iterator obj = m_cell_inst_tree.insert (inst);
    m_bbox_needs_update = true;
    return *obj;
  }

  /**
   *  @brief Insert a sequence [from,to) of cell instances
   *  
   *  @param from The start of the sequence
   *  @param to The past-the-end pointer of the sequence
   */
  template <class I>
  void insert (I from, I to)
  {
    m_cell_inst_tree.insert (from, to);
    mp_graph->invalidate_graph ();
    mp_graph->invalidate_bboxes ();
    m_bbox_needs_update = true;
  }

  /**
   *  @brief The cell index accessor method
   *
   *  @return The cell index of the cell
   */
  const cell_index_type &cell_index () const
  {
    return m_cell_index;
  }

  /**
   *  @brief The cell inst accessor by index
   *
   *  @return Returns the cell instance by index
   */
  const cell_inst_array_type &cell_inst (size_t i) const
  {
    return *(m_cell_inst_tree.begin () + i);
  }

  /**
   *  @brief Clear the parent instance list
   *
   *  Reserve the given number of entries.
   */
  void clear_parent_insts (size_t sz = 0)
  {
    m_parent_insts.clear ();
    m_parent_insts.reserve (sz);
  }

  /**
   *  @brief Count the number of parent instances
   * 
   *  This will accumulate the number of parent instances in the vector
   *  provided. Later, this count is used to resize the parent instance
   *  vector.
   */
  void count_parent_insts (std::vector <size_t> &count) const
  {
    cell_index_type last_ci = (cell_index_type) -1l;
    
    for (const_iterator c = begin (); c != end (); ++c) {
      cell_index_type ci = c->object ().cell_index ();
      if (ci != last_ci) {
        last_ci = ci;
        ++count [ci];
      }
    }
  }

  /**
   *  @brief Sort the child instance list
   * 
   *  The child instances are first sorted by cell index, then by raw transformation
   *  (excluding displacement). This allows to simplify the bbox computation by 
   *  convolution of the displacements bboxes with the object bboxes.
   */
  void sort_child_insts ()
  {
    tl::sort (begin (), end (), typename cell_inst_array_type::raw_less_f ());
  }

  /**
   *  @brief Update the child-parent relationships
   * 
   *  This will update the child-parent relationships. Basically
   *  this means entering the cell as a parent into all it's child
   *  cells.
   */
  void update_relations ()
  {
    cell_index_type last_ci = (cell_index_type) -1l;

    for (iterator c = begin (); c != end (); ++c) {
      cell_index_type ci = c->object ().cell_index ();
      if (ci != last_ci) {
        last_ci = ci;
        mp_graph->cell (ci).m_parent_insts.push_back (parent_inst_type (cell_index (), c));
      }
    }
  }

  /** 
   *  @brief Check if the bbox needs update because the shapes have changed
   *
   *  Returns true if the bounding box needs update because the
   *  shape bounding box (not one of the child bounding boxes) has changed.
   */
  bool is_shape_bbox_dirty () const
  {
    if (m_bbox_needs_update) {
      return true;
    }
    for (typename shapes_vector::const_iterator s = m_shapes_vector.begin (); s != m_shapes_vector.end (); ++s) {
      if ((*s)->is_bbox_dirty ()) {
        return true;
      }
    }
    return false;
  }

  /** 
   *  @brief Update the bbox 
   *
   *  This will update the bbox from the shapes and instances.
   *  This requires the bboxes of the child cells to be computed
   *  before. Practically this will be done by computing the
   *  bboxes bottom-up in the hierarchy.
   *  In addition, the number of hierarchy leves below is also
   *  updated.
   *
   *  @param layers The max. number of layers in the child cells
   *  @return true, if the bounding box has changed.
   */
  bool update_bbox (unsigned int layers)
  {
    unsigned int l;

    //  determine the bounding box
    box_type org_bbox = m_bbox;
    m_bbox = box_type ();

    //  prepare the box vector
    unsigned int n = std::max (layers, (unsigned int) m_shapes_vector.size ());
    if (n == 0) {
      return false;
    }

    //  save the original boxes for simple compare
    box_vector org_bboxes;
    org_bboxes.resize (n);
    org_bboxes.swap (m_bboxes);
    
    //  compute the per-layer bboxes of the cell instances
    //  exploit the fact that these are sorted by instance,
    //  rotation and magnification.
    for (const_iterator o = begin (); o != end (); ) {

      const_iterator o1 = o;

      const_iterator oo = o;
      while (++oo != end () && oo->raw_equal (*o)) 
        ;

      box_type raw_box;
      while (o != oo) {
        raw_box += o->raw_bbox ();
        ++o;
      }

      for (l = 0; l < layers; ++l) {

        //  the per-layer bounding boxes
        db::box_convert <cell_inst_type> bc (*mp_graph, l);
        box_type lbox = o1->bbox_from_raw_bbox (raw_box, bc);

        if (! lbox.empty ()) {
          m_bbox += lbox;
          m_bboxes[l] += lbox;
        }

      }
      
    }

    //  update the bboxes of the shapes lists
    for (l = 0; l < m_shapes_vector.size (); ++l) {

      m_shapes_vector [l]->update_bbox ();
      box_type sbox (m_shapes_vector [l]->bbox ());

      if (! sbox.empty ()) {
        m_bbox += sbox;
        m_bboxes[l] += sbox;
      }
     
    }

    //  reset "dirty child instances" flag
    m_bbox_needs_update = false;

    //  return true, if anything has changed with the box 
    bool changed = (org_bbox != m_bbox);
    for (l = 0; ! changed && l < n; ++l) {
      if (l >= org_bboxes.size ()) {
        changed = ! m_bboxes [l].empty ();
      } else {
        changed = (m_bboxes [l] != org_bboxes [l]);
      }
    }
    return changed;

  }

  /**
   *  @brief Sort the shapes lists 
   *
   *  This will sort the shapes lists for query of regions 
   *  on a per-shape basis. Since sorting of the shapes is
   *  guarded against redundant sorting (db::Layer::sort),
   *  we can safely call the sort method in any case.
   */
  void sort_shapes ()
  {
    for (typename shapes_vector::iterator s = m_shapes_vector.begin ();
         s != m_shapes_vector.end (); ++s) {
      (*s)->sort ();
    }
  }

  /**
   *  @brief Sort the cell instance list
   *
   *  This will sort the cell instance list. As a prerequesite
   *  the cell's bounding boxes must have been computed.
   *
   *  @param layers The maximum number of layers in the child cells
   */
  void sort_inst_tree (unsigned int layers)
  {
    //  sort the overall box instance tree
    m_cell_inst_tree.sort (db::box_convert<cell_inst_array_type> (*mp_graph));

    //  create the per-layer instance trees: 
    //  fill the per-layer cell trees with numbers from 0 to n
    //  corresponding to the cell instances in the main instance tree
    //  and sort.
    m_cell_inst_trees.clear ();

    if (mp_graph->per_layer_cell_inst_trees ()) {
      m_cell_inst_trees.resize (layers);
      for (unsigned int l = 0; l < layers; ++l) {
        cell_inst_layer_tree_type &t (m_cell_inst_trees [l]);
        t.reserve (m_cell_inst_tree.size ());
        for (size_t n = 0; n < m_cell_inst_tree.size (); ++n) {
          t.insert (n);
        }
        cell_inst_index_box_convert bc (*mp_graph, *this, l);
        m_cell_inst_trees [l].sort (bc); 
      }
    }

    //  update the number of hierarchy levels @@@ why here???
    m_hier_levels = count_hier_levels ();
  }

  /**
   *  @brief Retrieve the bounding box of the cell
   *
   *  Before the bounding box can be retrieved, it must have
   *  been computed using update_bbox. This is performed by 
   *  requesting an update from the graph.
   *
   *  @return The bounding box that was computed by update_bbox
   */
  const box_type &bbox () const
  {
    mp_graph->update ();
    return m_bbox;
  }

  /**
   *  @brief Retrieve the per-layer bounding box of the cell
   *
   *  Before the bounding box can be retrieved, it must have
   *  been computed using update_bbox. This is performed by 
   *  requesting an update from the graph.
   *
   *  @return The bounding box that was computed by update_bbox
   */
  const box_type &bbox (unsigned int l) const
  {
    mp_graph->update ();
    if (l < m_bboxes.size ()) {
      return m_bboxes [l];
    } else {
      static box_type empty_box;
      return empty_box;
    }
  }

  /**
   *  @brief Per-layer region query for the instances in "overlapping" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances overlapping with the given region b for the given layer.
   *  The end iterator is given by end_overlapping ().
   *
   *  @param b The region to query
   *  @param l The layer to do the query for
   *
   *  @return The region iterator for the child cell instances
   */
  overlapping_layer_iterator begin_overlapping (const box_type &b, unsigned int l) const
  {
    tl_assert (mp_graph->per_layer_cell_inst_trees ());

    mp_graph->update ();
    cell_inst_index_box_convert bc (*mp_graph, *this, l);
    if (l < m_cell_inst_trees.size ()) {
      return m_cell_inst_trees [l].begin_overlapping (b, bc);
    } else {
      static const cell_inst_layer_tree_type empty_inst_tree;
      return empty_inst_tree.begin_overlapping (b, bc);
    }
  }

  /**
   *  @brief Per-layer region query for the instances in "touching" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances touching the given region b for the given layer.
   *  The end iterator is given by end_touching ().
   *
   *  @param b The region to query
   *  @param l The layer to do the query for
   *
   *  @return The region iterator for the child cell instances
   */
  touching_layer_iterator begin_touching (const box_type &b, unsigned int l) const
  {
    tl_assert (mp_graph->per_layer_cell_inst_trees ());

    mp_graph->update ();
    cell_inst_index_box_convert bc (*mp_graph, *this, l);
    if (l < m_cell_inst_trees.size ()) {
      return m_cell_inst_trees [l].begin_touching (b, bc);
    } else {
      static const cell_inst_layer_tree_type empty_inst_tree;
      return empty_inst_tree.begin_touching (b, bc);
    }
  }

  /**
   *  @brief Region query for the instances in "overlapping" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances overlapping with the given region b. 
   *
   *  @param b The region to query
   *
   *  @return The region iterator for the child cell instances
   */
  overlapping_iterator begin_overlapping (const box_type &b) const
  {
    mp_graph->update ();
    return m_cell_inst_tree.begin_overlapping (b, db::box_convert<cell_inst_array_type> (*mp_graph));
  }

  /**
   *  @brief The "overlapping" mode region query end token
   *
   *  Basically this method is provided to maintain the begin..end
   *  semantics. Practically, the "at_end" method of the region
   *  iterator is faster to use.
   */
  size_t end_overlapping () const
  {
    return m_cell_inst_tree.end_overlapping ();
  }

  /**
   *  @brief Region query for the instances in "touching" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances touching the given region b. 
   *
   *  @param b The region to query
   *
   *  @return The region iterator for the child cell instances
   */
  touching_iterator begin_touching (const box_type &b) const
  {
    mp_graph->update ();
    return m_cell_inst_tree.begin_touching (b, db::box_convert<cell_inst_array_type> (*mp_graph));
  }

  /**
   *  @brief The "touching" mode region query end token
   *
   *  Basically this method is provided to maintain the begin..end
   *  semantics. Practically, the "at_end" method of the region
   *  iterator is faster to use.
   */
  size_t end_touching () const
  {
    return m_cell_inst_tree.end_touching ();
  }

  /**
   *  @brief The child cell iterator
   *
   *  This iterator will report the child cell indices, not every instance.
   *  Basically this is a filter.
   */
  child_cell_iterator begin_child_cells () const
  {
    return child_cell_iterator (m_cell_inst_tree.begin (), m_cell_inst_tree.end ());
  }

  /**
   *  @brief The child cell end iterator
   */
  child_cell_iterator end_child_cells () const
  {
    return child_cell_iterator (m_cell_inst_tree.end (), m_cell_inst_tree.end ());
  }

  /** 
   *  @brief Report the number of child cells
   *
   *  Just the number of child cells with distinct cell indices are
   *  reported.
   *  CAUTION: this method is SLOW!
   */
  size_t child_cells () const
  {
    size_t n = 0;
    child_cell_iterator i (m_cell_inst_tree.begin (), m_cell_inst_tree.end ());
    while (! i.at_end ()) {
      ++n;
      ++i;
    }
    return n;
  }

  /**
   *  @brief The cell instance access begin iterator
   *
   *  The begin() method together with the end() method
   *  allows to access the cell instance list as a whole
   *  with the usual begin..end semantics. This is the
   *  const implementation.
   */
  const_iterator begin () const
  {
    return m_cell_inst_tree.begin ();
  }

  /**
   *  @brief The cell instance access end iterator
   */
  const_iterator end () const
  {
    return m_cell_inst_tree.end ();
  }

  /**
   *  @brief The cell instance access begin iterator
   *
   *  The begin() method together with the end() method
   *  allows to access the cell instance list as a whole
   *  with the usual begin..end semantics.
   */
  iterator begin ()
  {
    return m_cell_inst_tree.begin ();
  }

  /**
   *  @brief The cell instance access end iterator
   */
  iterator end ()
  {
    return m_cell_inst_tree.end ();
  }

  /**
   *  @brief The parent instance list begin iterator
   *
   *  The begin_parent_insts() allows to access to the parent instance list.
   */
  parent_inst_iterator begin_parent_insts () const
  {
    mp_graph->update ();
    return parent_inst_iterator (mp_graph, m_parent_insts.begin (), m_parent_insts.end ());
  }

  /**
   *  @brief Report the number of parent cells 
   *
   *  Since the m_parent_insts vector just stores references to those cells
   *  that have distinct cell indices, we can simply report their length.
   */
  size_t parent_cells () const
  {
    return m_parent_insts.size ();
  }

  /**
   *  @brief The parent cell iterator
   * 
   *  This iterator will iterate over the parent cells, just returning their
   *  cell index.
   */
  parent_cell_iterator begin_parent_cells () const
  {
    return parent_cell_iterator (m_parent_insts.begin ());
  }

  /**
   *  @brief The parent cell end iterator
   */
  parent_cell_iterator end_parent_cells () const
  {
    return parent_cell_iterator (m_parent_insts.end ());
  }

  /**
   *  @brief Tell if the cell is a top-level cell
   *
   *  A cell is a top-level cell if there are no parent instantiations.
   */
  bool is_top () const
  {
    mp_graph->update ();
    return m_parent_insts.end () == m_parent_insts.begin ();
  }

  /**
   *  @brief Tell if the cell is a leaf cell
   *
   *  A cell is a leaf cell if there are no child instantiations.
   */
  bool is_leaf () const
  {
    return m_cell_inst_tree.end () == m_cell_inst_tree.begin ();
  }

  /**
   *  @brief Return the number of hierarchy levels below (expensive)
   */
  unsigned int hierarchy_levels () const
  {
    mp_graph->update ();
    return m_hier_levels;
  }

  /**
   *  @brief Collect memory usage statistics
   */
  void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.cell_info (m_cell_index);
    m.cell_info (mp_graph);
    m.cell_info (m_shapes_vector);                                         //  need more: see below
    m.cell_info (sizeof (m_cell_inst_tree), sizeof (m_cell_inst_tree));    //  need more: see below
    m.cell_info (sizeof (m_cell_inst_trees), sizeof (m_cell_inst_trees));  //  need more: see below
    m.cell_info (sizeof (m_parent_insts), sizeof (m_parent_insts));        //  need more: see below
    m.cell_info (m_bbox);
    m.cell_info (m_bboxes); 
    m.cell_info (m_hier_levels);
    m.cell_info (m_bbox_needs_update);

    m.inst_trees (m_cell_inst_tree);
    m.inst_trees (-sizeof (m_cell_inst_tree), -sizeof (m_cell_inst_tree));   //  the actual object is counted for cell, not for instances (see above)
    m.inst_trees (m_cell_inst_trees); 
    m.inst_trees (-sizeof (m_cell_inst_trees), -sizeof (m_cell_inst_trees)); //  the actual object is counted for cell, not for instances (see above)
    m.instances (m_parent_insts);
    m.instances (-sizeof (m_parent_insts), -sizeof (m_parent_insts));        //  the actual object is counted for cell, not for instances (see above)

    for (typename shapes_vector::const_iterator s = m_shapes_vector.begin (); s != m_shapes_vector.end (); ++s) {
      (*s)->collect_mem_stat (m);
    }
  }

private:
  cell_index_type m_cell_index;
  mutable graph_type *mp_graph;
  shapes_vector m_shapes_vector;
  cell_inst_tree_type m_cell_inst_tree;
  parent_inst_vector m_parent_insts;
  cell_inst_layer_tree_vector m_cell_inst_trees;
  box_type m_bbox;
  box_vector m_bboxes;
  unsigned int m_hier_levels;
  bool m_bbox_needs_update;

  //  clear the shapes without telling the graph
  void clear_shapes_no_invalidate ()
  {
    for (typename shapes_vector::iterator s = m_shapes_vector.begin ();
         s != m_shapes_vector.end (); ++s) {
      delete *s;
    }
    m_shapes_vector.clear ();
    m_bbox_needs_update = true;
  }

  //  helper function for computing the number of hierarchy levels
  //  must be called bottom-up
  unsigned int 
  count_hier_levels () const
  {
    unsigned int l = 0;

    for (const_iterator c = begin (); c != end (); ++c) {
      l = std::max (l, mp_graph->cell (c->object ().cell_index ()).m_hier_levels + 1);
    }

    return l;
  }
};

/** 
 *  @brief A typedef for a standard cell
 */
typedef db::cell<db::CellGraph> Cell;

} // namespace db

#endif

