
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <dbLayout.h>

#include <dbShape.h>
#include <dbStream.h>
#include <dbGDS2Writer.h>
#include <dbGDS2.h>

#include <stdio.h>
#include <errno.h>

namespace db
{

// ------------------------------------------------------------------
//  GDS2Writer implementation

GDS2Writer::GDS2Writer (db::Stream &stream)
  : mp_stream (&stream)
{
  //  .. nothing yet ..
}

void
GDS2Writer::write (const db::Layout &layout)
{
  //  write header

  write_short (6);
  write_short (sHEADER);
  write_short (600);

  write_short (4 + 12 * 2);
  write_short (sBGNLIB);
  write_time ();
  write_time ();

  std::string libname ("LIB");
  write_string_record (sLIBNAME, libname);

  write_short (4 + 8 * 2);
  write_short (sUNITS);
  write_double (layout.dbu ());
  write_double (layout.dbu () * 1e-6);

  //  body

  for (db::Layout::bottom_up_const_iterator cell = layout.begin_bottom_up (); cell != layout.end_bottom_up (); ++cell) {

    //  cell header 

    write_short (4 + 12 * 2);
    write_short (sBGNSTR);
    write_time ();
    write_time ();

    write_string_record (sSTRNAME, layout.cell_name (*cell));

    //  cell body 

    const db::Cell &cref (layout.cell (*cell));

    //  instances
    
    for (db::Cell::const_iterator inst = cref.begin (); inst != cref.end (); ++inst) {

      db::Point a, b;
      unsigned long amax, bmax;

      bool is_reg = inst->is_regular_array (a, b, amax, bmax);

      write_short (4);
      write_short (is_reg ? sAREF : sSREF);

      write_string_record (sSNAME, layout.cell_name (inst->object ().cell_index ()));

      db::Trans t = inst->front ();

      if (t.rot () != 0 || inst->is_mag ()) {

        write_short (6);
        write_short (sSTRANS);
        write_short (t.is_mirror () ? 0x8000 : 0);

        if (inst->is_mag ()) {
          write_short (4 + 8);
          write_short (sMAG);
          write_double (inst->mag ());
        }

        if ((t.rot () % 4) != 0) {
          write_short (4 + 8);
          write_short (sANGLE);
          write_double ((t.rot () % 4) * 90.0);
        }

      }

      if (is_reg) {
        write_short (4 + 2 * 2);
        write_short (sCOLROW);
        write_short (std::max ((unsigned long) 1, amax));
        write_short (std::max ((unsigned long) 1, bmax));
      }

      write_short (4 + (is_reg ? 3 : 1) * 2 * 4);
      write_short (sXY);
      write_int (t.disp ().x ());
      write_int (t.disp ().y ());

      if (is_reg) {
        write_int (t.disp ().x () + a.x () * amax);
        write_int (t.disp ().y () + a.y () * amax);
        write_int (t.disp ().x () + b.x () * bmax);
        write_int (t.disp ().y () + b.y () * bmax);
      }

      write_short (4);
      write_short (sENDEL);

    }

    //  shapes

    for (unsigned int l = 0; l < layout.layers (); ++l) {
 
      if (layout.is_valid_layer (l)) {

        const LayerProperties &prop = layout.get_properties (l);
        int layer = prop.layer;
        int datatype = prop.datatype;

        db::ShapeIterator shape (cref.shapes (l).begin (db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Texts));
        while (! shape.at_end ()) {

          bool any = true;

          if (shape->is_text ()) {
            
            db::Trans trans = shape->text_trans ();
          
            write_short (4);
            write_short (sTEXT);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sTEXTTYPE);
            write_short (datatype);

            if (trans.rot () != 0) {

              write_short (6);
              write_short (sSTRANS);
              write_short (trans.is_mirror () ? 0x8000 : 0);

              if ((trans.rot () % 4) != 0) {
                write_short (4 + 8);
                write_short (sANGLE);
                write_double ((trans.rot () % 4) * 90.0);
              }

            }

            write_short (4 + 2 * 4);
            write_short (sXY);
            write_int (trans.disp ().x ());
            write_int (trans.disp ().y ());

            write_string_record (sSTRING, shape->text_string ().c_str ());

          } else if (shape->is_polygon ()) {

            //  @@@ how to deal with holes, max vertex count?

            size_t n = 0;
            for (db::Shape::point_iterator e = shape->begin_hull (); e != shape->end_hull (); ++e) {
              ++n;
            }

            write_short (4);
            write_short (sBOUNDARY);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            write_short (4 + n * 2 * 4);
            write_short (sXY);
            for (db::Shape::point_iterator e = shape->begin_hull (); e != shape->end_hull (); ++e) {
              write_int ((*e).x ());
              write_int ((*e).y ());
            }

          } else if (shape->is_edge ()) {

            db::Edge e (shape->edge ());

            write_short (4);
            write_short (sPATH);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            write_short (4 + 2);
            write_short (sPATHTYPE);
            write_short (0);

            write_short (4 + 4);
            write_short (sWIDTH);
            write_int (0);

            write_short (4 + 2 * 2 * 4);
            write_short (sXY);
            write_int (e.p1 ().x ());
            write_int (e.p1 ().y ());
            write_int (e.p2 ().x ());
            write_int (e.p2 ().y ());

          } else if (shape->is_path ()) {

            //  instantiate the path and draw
            db::Path path;
            shape->path (path);

            write_short (4);
            write_short (sPATH);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            write_short (4 + 2);
            write_short (sPATHTYPE);
            write_short (4);

            write_short (4 + 4);
            write_short (sWIDTH);
            write_int (path.width ());

            write_short (4 + 4);
            write_short (sBGNEXTN);
            write_int (path.extensions ().first);

            write_short (4 + 4);
            write_short (sENDEXTN);
            write_int (path.extensions ().second);

            write_short (4 + path.points () * 2 * 4);
            write_short (sXY);
            for (db::Path::iterator p = path.begin (); p != path.end (); ++p) {
              write_int ((*p).x ());
              write_int ((*p).y ());
            }

          } else if (shape->is_box ()) {

            db::Box box (shape->box ());

            write_short (4);
            write_short (sBOUNDARY);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            write_short (4 + 5 * 2 * 4);
            write_short (sXY);
            write_int (box.left ());
            write_int (box.bottom ());
            write_int (box.left ());
            write_int (box.top ());
            write_int (box.right ());
            write_int (box.top ());
            write_int (box.right ());
            write_int (box.bottom ());
            write_int (box.left ());
            write_int (box.bottom ());

          } else {
            any = false;
          }

          if (any) {

            if (shape->has_prop_id ()) {

              const db::PropertiesRepository::properties_set &props = layout.properties_repository ().properties (shape->prop_id ());
              for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

                const db::PropertyValue &name = layout.properties_repository ().prop_name (p->first);

                long attr = -1;
                if (name.is_long ()) {
                  attr = name.get_long ();
                }

                if (attr >= 0 && attr < 65535) {

                  write_short (6);
                  write_short (sPROPATTR);
                  write_short (attr);

                  write_string_record (sPROPVALUE, p->second.get_string ().c_str ());

                }

              }

            }

            write_short (4);
            write_short (sENDEL);

          }

          ++shape;

        }

      }

    }

    //  end of cell

    write_short (4);
    write_short (sENDSTR);

  }

  write_short (4);
  write_short (sENDLIB);
}

void 
GDS2Writer::write_byte (unsigned char b)
{
  mp_stream->put ((const char *) &b, 1);
}

void 
GDS2Writer::write_short (int16_t i)
{
  char b[2];
  //  @@@ this can be done more efficiently?
  b[0] = ((i >> 8) & 0xff);
  b[1] = (i & 0xff);
  mp_stream->put (b, sizeof (b));
}

void 
GDS2Writer::write_int (int32_t l)
{
  char b[4];
  //  @@@ this can be done more efficiently?
  b[0] = ((l >> 24) & 0xff);
  b[1] = ((l >> 16) & 0xff);
  b[2] = ((l >> 8) & 0xff);
  b[3] = (l & 0xff);
  mp_stream->put (b, sizeof (b));
}

void 
GDS2Writer::write_double (double d)
{
  char b[8];
  
  b[0] = 0;
  if (d < 0) {
    b[0] = char (0x80);
    d = -d;
  }

  int e = int (ceil (log (d) / log (16.0)));
  d /= pow (16.0, e - 14);

  tl_assert (e >= -64 && e < 64);
  b[0] |= ((e + 64) & 0x7f);

  uint64_t m = uint64_t (d);
  for (int i = 7; i > 0; --i) {
    b[i] = (m & 0xff);
    m >>= 8;
  }

  mp_stream->put (b, sizeof (b));
}

void 
GDS2Writer::write_time ()
{
  //  @@@ not implemented yet
  for (unsigned int i = 0; i < 6; ++i) {
    write_short (0);
  }
}

void 
GDS2Writer::write_string (const char *t)
{
  size_t l = strlen (t);
  mp_stream->put (t, l);
  if ((l & 1) != 0) {
    write_byte (0);
  }
}

void 
GDS2Writer::write_string (const std::string &t)
{
  size_t l = t.size ();
  mp_stream->put (t.c_str (), l);
  if ((l & 1) != 0) {
    write_byte (0);
  }
}

void 
GDS2Writer::write_string_record (short record, const std::string &t)
{
  write_short (4 + ((t.size () + 1) / 2) * 2);
  write_short (record);
  write_string (t);
}


} // namespace db

