
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/




#ifndef HDR_dbLayout
#define HDR_dbLayout

#include <dbCellGraph.h>
#include <dbPropertiesRepository.h>
#include <tlObserver.h>

#include <string.h>
#include <map>
#include <string>

namespace db
{

class Manager;

/**
 *  @brief A layer property
 *
 *  The layer properties are basically to be used for storing of layer name and 
 *  layer/datatype information.
 */

struct LayerProperties
{
  LayerProperties () 
    : name (), layer (-1), datatype (-1)
  { }
  
  bool operator< (const LayerProperties &b) const
  {
    if (layer != b.layer) { return layer < b.layer; }
	return datatype < b.datatype;
  }

  std::string name;
  int layer;
  int datatype;
};

/**
 *  @brief The layout object
 *
 *  The layout object basically wraps the cell graphs and
 *  adds functionality for managing cell names and layer names.
 */

class Layout 
  : public db::CellGraph
{
public:
  /**
   *  @brief A helper functor to compare "const char *" by the content
   */
  struct name_cmp_f 
  {
    bool operator() (const char *a, const char *b) const {
      return strcmp (a, b) < 0;
    }
  };

  typedef db::CellGraph::cell_index_type cell_index_type;
  typedef std::map<const char *, cell_index_type, name_cmp_f> cell_map_type;

  /**
   *  @brief Standard constructor
   */
  Layout (db::Manager *manager = 0);

  /**
   *  @brief Destructor
   */
  ~Layout ();

  /**
   *  @brief Assignment operator
   */
  Layout &operator= (const Layout &d);

  /**
   *  @brief Accessor to the shape repository
   */
  GenericRepository &shape_repository ()
  {
    return m_shape_repository;
  }

  /**
   *  @brief Accessor to the properties repository
   */
  PropertiesRepository &properties_repository ()
  {
    return m_properties_repository;
  }

  /**
   *  @brief Accessor to the properties repository
   */
  const PropertiesRepository &properties_repository () const
  {
    return m_properties_repository;
  }

  /**
   *  @brief Collect memory statistics
   */
  void collect_mem_stat (db::MemStatistics &m) const;

  /**
   *  @brief Query, if a cell with a given name is present
   *
   *  Since internally just a "const char *" is required, the interface
   *  does not use std::string for the name to avoid unnecessary string
   *  allocations.
   *
   *  @param s The name of the cell
   *  @return true, if the cell is present
   */
  bool has_cell (const char *name);

  /**
   *  @brief Obtain the index of a cell with the given name
   *
   *  Since internally just a "const char *" is required, the interface
   *  does not use std::string for the name to avoid unnecessary string
   *  allocations.
   *
   *  @param name The name of the cell
   *  @return A pair, telling if the cell was present (first) and
   *          it's index if it is present (second)
   */ 
  std::pair<bool, cell_index_type> cell_by_name (const char *name);

  /** 
   *  @brief Tell the name of a cell with the given index
   *  
   *  @param index The index of the cell
   *  @return The name of the cell
   */
  const char *cell_name (cell_index_type index) const;

  /** 
   *  @brief Add a cell with a given name
   *
   *  Since internally just a "const char *" is required, the interface
   *  does not use std::string for the name to avoid unnecessary string
   *  allocations.
   *
   *  @param s The name of the new cell
   *  @return The index of the new cell
   */
  cell_index_type add_cell (const char *name);

  /**
   *  @brief Implementation of the undo operations
   */
  virtual void undo (db::Op *op) throw ();

  /**
   *  @brief Implementation of the redo operations
   */
  virtual void redo (db::Op *op) throw ();

  /**
   *  @brief Signal the start of an operation bringing the layout into invalid state
   *
   *  This method should be called whenever the layout is
   *  about to be brought into an invalid state. After calling
   *  this method, "under_construction" returns false which 
   *  tells foreign code (such as update which might be called
   *  asynchroneously for example because of a repaint event)
   *  not to use this layout object.
   *  This state is cancelled by the end_changes () method.
   *  The start_changes () method can be called multiple times
   *  and must be cancelled the same number of times.
   *  If a transaction is used to bracket the start and end of
   *  an operation, the start/end_changes method pair does not
   *  need to be called.
   */
  void start_changes ()
  {
    ++m_invalid;
  }

  /**
   *  @brief Cancel the "in changes" state (see "start_changes")
   */
  void end_changes ()
  {
    if (m_invalid > 0) {
      --m_invalid;
    }
  }

  /**
   *  @brief Tell if the layout object is under construction
   *
   *  A layout object is either under construction if a transaction
   *  is ongoing or the layout is brought into invalid state by
   *  "start_changes".
   */
  bool under_construction () const
  {
    return m_invalid > 0 || manager ()->transacting ();
  }

  /** 
   *  @brief Database unit read accessor
   */
  double dbu () const
  {
    return m_dbu;
  }

  /**
   *  @brief Database unit write accessor
   */
  void dbu (double d) 
  {
    m_dbu = d;
  }

  /**
   *  @brief Insert a new layer with the given properties
   */
  unsigned int insert_layer (const LayerProperties &props = LayerProperties ())
  {
    unsigned int i = db::CellGraph::insert_layer ();
    while (m_layer_props.size () <= i) {
      m_layer_props.push_back (LayerProperties ());
    }
    m_layer_props [i] = props;
    return i;
  }

  /**
   *  @brief Insert a new layer with the given properties at the given index
   */
  void insert_layer (unsigned int index, const LayerProperties &props = LayerProperties ())
  {
    db::CellGraph::insert_layer (index);
    while (m_layer_props.size () <= index) {
      m_layer_props.push_back (LayerProperties ());
    }
    m_layer_props [index] = props;
  }

  /**
   *  @brief Insert a new special layer with the given properties
   *
   *  A special layers is used for example to represent rulers.
   */
  unsigned int insert_special_layer (const LayerProperties &props = LayerProperties ())
  {
    unsigned int i = db::CellGraph::insert_layer (true /*special*/);
    while (m_layer_props.size () <= i) {
      m_layer_props.push_back (LayerProperties ());
    }
    m_layer_props [i] = props;
    return i;
  }

  /**
   *  @brief Insert a new layer with the given properties at the given index
   *
   *  A special layers is used for example to represent rulers.
   */
  void insert_special_layer (unsigned int index, const LayerProperties &props = LayerProperties ())
  {
    db::CellGraph::insert_layer (index, true /*special*/);
    while (m_layer_props.size () <= index) {
      m_layer_props.push_back (LayerProperties ());
    }
    m_layer_props [index] = props;
  }

  /**
   *  @brief Set the properties for a specified layer
   */
  void set_properties (unsigned int i, const LayerProperties &props)
  {
    m_layer_props [i] = props;
  }

  /**
   *  @brief Get the properties for a specified layer
   */
  const LayerProperties &get_properties (unsigned int i) const
  {
    return m_layer_props [i];
  }

  //  make public some of the methods of CellGraph that
  //  are supposed to be made public.
  using db::CellGraph::cells;
  using db::CellGraph::cell;
  using db::CellGraph::reserve;
  using db::CellGraph::update_relations;
  using db::CellGraph::begin;
  using db::CellGraph::end;
  using db::CellGraph::begin_bottom_up;
  using db::CellGraph::end_bottom_up;
  using db::CellGraph::begin_top_down;
  using db::CellGraph::end_top_down;
  using db::CellGraph::end_top_cells;
  using db::CellGraph::update;
  using db::CellGraph::invalidate_graph;
  using db::CellGraph::invalidate_bboxes;
  using db::CellGraph::layers;
  using db::CellGraph::is_valid_layer;
  using db::CellGraph::is_special_layer;
  using db::CellGraph::delete_layer;
  using db::CellGraph::reserve_layers;
  using db::CellGraph::per_layer_cell_inst_trees;

private:
  std::vector<const char *> m_cell_names;
  cell_map_type m_cell_map;
  std::vector<LayerProperties> m_layer_props;
  double m_dbu;
  unsigned int m_invalid;
  GenericRepository m_shape_repository;
  PropertiesRepository m_properties_repository;

  //  no public copy ctor because of attachment to a manager object
  Layout (const Layout &d);
};

}

#endif


