
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_dbOASISReader
#define HDR_dbOASISReader

#include "tlException.h"
#include "tlInternational.h"
#include "tlProgress.h"
#include "tlString.h"

#include "dbLayout.h"
#include "dbReader.h"
#include "dbStream.h"
#include "dbStreamLayers.h"
#include "dbPropertiesRepository.h"

#include <map>
#include <set>

namespace db
{

/**
 *  @brief Generic base class of OASIS reader exceptions
 */
class OASISReaderException
  : public ReaderException
{
public:
  OASISReaderException (const std::string &msg, size_t p, const std::string &cell)
    : ReaderException (tl::sprintf (tl::translate ("%s (position=%ld, cell=%s)").c_str (), msg.c_str (), p, cell.c_str ()))
  { }
};

class RepetitionBase;
class RepetitionIteratorBase;

/**
 *  @brief A repetition iterator
 */
class RepetitionIterator
{
public:
  /**
   *  @brief Create a repetition with the given implementation
   */
  RepetitionIterator (RepetitionIteratorBase *base);

  /**
   *  @brief Destructor
   */
  ~RepetitionIterator ();

  /**
   *  @brief Copy constructor
   */
  RepetitionIterator (const RepetitionIterator &d);

  /**
   *  @brief Assignment 
   */
  RepetitionIterator &operator= (const RepetitionIterator &d);

  /**
   *  @brief Comparison
   */
  bool operator== (const RepetitionIterator &d) const;

  /**
   *  @brief Inequality
   */
  bool operator!= (const RepetitionIterator &d) const
  {
    return !operator== (d);
  }

  /**
   *  @brief Tell, if the iterator is at the end
   */
  bool at_end () const;

  /**
   *  @brief Increment
   */
  RepetitionIterator &operator++ ();

  /**
   *  @brief Access
   */
  db::Point operator* () const;

private:
  RepetitionIteratorBase *mp_base; 
};

/**
 *  @brief A class representing a repetition
 */
class Repetition
{
public:
  /**
   *  @brief Create a repetition with the given implementation
   */
  Repetition (RepetitionBase *base = 0);

  /**
   *  @brief Destructor
   */
  ~Repetition ();

  /**
   *  @brief Copy constructor
   */
  Repetition (const Repetition &d);

  /**
   *  @brief Replace the base pointer
   */
  Repetition &operator= (RepetitionBase *base);

  /**
   *  @brief Assignment 
   */
  Repetition &operator= (const Repetition &d);

  /**
   *  @brief Comparison
   */
  bool operator== (const Repetition &d) const;

  /**
   *  @brief Inequality
   */
  bool operator!= (const Repetition &d) const
  {
    return !operator== (d);
  }

  /**
   *  @brief Check, if the repetition is a regular one
   *
   *  This method returns true, if the repetition is regular. It 
   *  returns true, if the repetition can be represented as a 
   *  set of points 0..i*a+j*b (i=0..n-1,j=0..m-1).
   *  This method does not only return a flag but the parameters as
   *  well.
   */
  bool is_regular (db::Point &a, db::Point &b, size_t &n, size_t &m) const;

  /**
   *  @brief Get the iterator
   */
  RepetitionIterator begin () const;

private:
  RepetitionBase *mp_base; 
};

class OASISReader;

/**
 *  @brief A template class representing a modal variable
 */
template <class T>
class modal_variable
{
public:
  /**
   *  @brief Set up the modal variable
   *
   *  The reader is required since errors are reported to the reader.
   *  The name is reported in the error message. 
   *  This constructor creates an uninitialized variable.
   */
  modal_variable (OASISReader *reader, const char *name)
    : mp_reader (reader), m_name (name), m_t (), m_initialized (false)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Get the value of the modal variable
   * 
   *  An error will be reported if the value is not initialized.
   */
  const T &get () const;

  /**
   *  @brief Get the value of the modal variable
   * 
   *  This method does not report an error if the value is not initialized.
   *  After modifying the object, set_initialized is supposed to be called.
   */
  T &get_non_const ()
  {
    return m_t;
  }

  /**
   *  @brief Assignment operator
   */
  modal_variable &operator= (const modal_variable<T> &d)
  {
    m_t = d.m_t;
    m_initialized = d.m_initialized;
    return *this;
  }

  /**
   *  @brief Assign a value to the modal variable
   *
   *  This sets the variable into the initialized state.
   */
  template <class X>
  modal_variable &operator= (const X &x)
  {
    m_t = x;
    m_initialized = true;
    return *this;
  }

  /**
   *  @brief Reset the initialized state of the variable
   */
  void reset () 
  {
    m_initialized = false;
  }

  /**
   *  @brief Set the initialized state of the variable
   *
   *  This method is supposed to be used after a modification has been performed
   *  with the non-const accessor.
   */
  void set_initialized () 
  {
    m_initialized = true;
  }

private:
  OASISReader *mp_reader;
  std::string m_name;
  T m_t;
  bool m_initialized;
};

/**
 *  @brief The OASIS format stream reader
 */
class OASISReader
  : public ReaderBase
{
public: 
  typedef std::vector<db::PropertyValue> property_value_list;

  /**
   *  @brief Construct a stream reader object
   *
   *  @param s The stream object from which to read stream data from
   */
  OASISReader (Stream &s);

  /**  
   *  @brief Destructor
   */
  ~OASISReader ();

  /** 
   *  @brief The basic read method 
   *
   *  This method will read the stream data and translate this to
   *  insert calls into the layout object. This will not do much
   *  on the layout object beside inserting the objects.
   *  It can be passed a layer map which tells which
   *  OASIS layer(s) to read on which logical layers.
   *  In addition, a flag can be passed that tells wheter to create 
   *  new layers. The returned map will contain all layers, the passed
   *  ones and the newly created ones.
   *
   *  @param layout The layout object to write to
   *  @param map The LayerMap object
   *  @param create true, if new layers should be created
   *  @return The LayerMap object that tells where which layer was loaded
   */
  virtual const LayerMap &read (db::Layout &layout, const LayerMap &map, bool create) throw (tl::Exception);

  /** 
   *  @brief The basic read method (without mapping)
   *
   *  This method will read the stream data and translate this to
   *  insert calls into the layout object. This will not do much
   *  on the layout object beside inserting the objects.
   *  This version will read all input layers and return a map
   *  which tells which OASIS layer has been read into which logical
   *  layer.
   *
   *  @param layout The layout object to write to
   *  @return The LayerMap object
   */
  virtual const LayerMap &read (db::Layout &layout) throw (tl::Exception);

  /**
   *  @brief Accessor method to the database unit
   */
  virtual double dbu () const { return m_dbu; }

  /**
   *  @brief Format
   */
  virtual const char *format () const { return "OASIS"; }

  /**
   *  @brief Issue an error with positional informations
   */
  void error (const std::string &txt) throw (tl::Exception);

  /**
   *  @brief Issue a warning with positional informations
   */
  void warn (const std::string &txt);

private:
  Stream &m_stream;
  LayerMap m_layer_map;
  tl::Progress m_progress;
  std::string m_cellname;
  double m_dbu;

  modal_variable<Repetition> mm_repetition;
  modal_variable<db::Layout::cell_index_type> mm_placement_cell;
  modal_variable<db::Coord> mm_placement_x;
  modal_variable<db::Coord> mm_placement_y;
  modal_variable<unsigned int> mm_layer;
  modal_variable<unsigned int> mm_datatype;
  modal_variable<unsigned int> mm_textlayer;
  modal_variable<unsigned int> mm_texttype;
  modal_variable<db::Coord> mm_text_x;
  modal_variable<db::Coord> mm_text_y;
  modal_variable<std::string> mm_text_string;
  modal_variable<db::Coord> mm_geometry_x;
  modal_variable<db::Coord> mm_geometry_y;
  modal_variable<db::Coord> mm_geometry_w;
  modal_variable<db::Coord> mm_geometry_h;
  modal_variable< std::vector<db::Point> > mm_polygon_point_list;
  modal_variable<db::Coord> mm_path_halfwidth;
  modal_variable<db::Coord> mm_path_start_extension;
  modal_variable<db::Coord> mm_path_end_extension;
  modal_variable< std::vector<db::Point> > mm_path_point_list;
  modal_variable<unsigned int> mm_ctrapezoid_type;
  modal_variable<db::Coord> mm_circle_radius;
  modal_variable<std::string> mm_last_property_name;
  modal_variable<bool> mm_last_property_is_sprop;
  modal_variable<property_value_list> mm_last_value_list;

  std::map <unsigned long, std::string> m_cellnames;
  std::map <unsigned long, std::string> m_textstrings;
  std::map <unsigned long, std::string> m_propstrings;
  std::map <unsigned long, std::string> m_propnames;
  tl::interval_map <db::ld_type, tl::interval_map <db::ld_type, std::string> > m_layernames;

  std::map <unsigned long, db::Layout::cell_index_type> m_cells_by_id;
  std::map <std::string, db::Layout::cell_index_type> m_cells_by_name;

  std::set <unsigned long> m_defined_cells_by_id;
  std::set <std::string> m_defined_cells_by_name;

  void do_read (db::Layout &layout, bool create_layers) throw (tl::Exception);
  void do_read_cell (db::Layout::cell_index_type cell_index, 
                     db::Layout &layout, bool create_layers,
                     tl::vector<db::CellInstArray> &instances) throw (tl::Exception);

  void read_repetition ();
  void read_pointlist (modal_variable <std::vector <db::Point> > &pointlist, bool for_polygon);
  void read_properties ();
  std::pair <bool, db::PropertiesRepository::properties_id_type> read_element_properties (db::PropertiesRepository &rep);

  unsigned char get_byte () throw (tl::Exception)
  {
    unsigned char *b = (unsigned char *) m_stream.get (1);
    if (! b) {
      error (tl::translate ("Unexpected end-of-file"));
    }
    return *b;
  }

  long get_long () throw (tl::Exception);
  unsigned long get_ulong () throw (tl::Exception);
  int get_int () throw (tl::Exception);
  unsigned int get_uint () throw (tl::Exception);

  void get (long &l) throw (tl::Exception)
  {
    l = get_long ();
  }

  void get (unsigned long &l) throw (tl::Exception)
  {
    l = get_ulong ();
  }

  void get (int &l) throw (tl::Exception)
  {
    l = get_int ();
  }

  void get (unsigned int &l) throw (tl::Exception)
  {
    l = get_uint ();
  }

  void get (double &d) throw (tl::Exception)
  {
    d = get_real ();
  }

  std::string get_str () throw (tl::Exception);
  void get_str (std::string &s) throw (tl::Exception);
  double get_real () throw (tl::Exception);
  db::Point get_gdelta (long grid = 1) throw (tl::Exception);
  db::Point get_3delta (long grid = 1) throw (tl::Exception);
  db::Point get_2delta (long grid = 1) throw (tl::Exception);
  db::Coord get_coord (long grid = 1) throw (tl::Exception);
  db::Coord get_ucoord (unsigned long grid = 1) throw (tl::Exception);

  std::pair <bool, unsigned int> open_dl (db::Layout &layout, const LDPair &dl, bool create) throw (tl::Exception);
};


template <class T>
inline const T &
modal_variable<T>::get () const
{
  if (! m_initialized) {
    mp_reader->error (tl::translate ("Modal variable accessed before being defined: ") + m_name);
  }
  return m_t;
}

}

#endif

