
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <dbPropertiesRepository.h>
#include <tlException.h>
#include <tlString.h>

namespace db
{

// ----------------------------------------------------------------------------------
//  PropertyValue implementation

PropertyValue::PropertyValue ()
  : m_type (t_void), m_string_current (true)
{
  //  .. nothing yet ..
}

PropertyValue::PropertyValue (long l)
  : m_type (t_long), m_string_current (false)
{
  m_numeric.v_long = l;
}

PropertyValue::PropertyValue (double d)
  : m_type (t_double), m_string_current (false)
{
  m_numeric.v_double = d;
}

PropertyValue::PropertyValue (const std::string &s)
  : m_type (t_string), m_string_current (true), m_string (s)
{
  // .. nothing yet ..
}

PropertyValue::PropertyValue (const PropertyValue &d)
  : m_type (t_void), m_string ()
{
  operator= (d);
}

PropertyValue &
PropertyValue::operator= (long v)
{
  m_type = t_long;
  m_numeric.v_long = v;
  m_string_current = false;
  return *this;
}

PropertyValue &
PropertyValue::operator= (double v)
{
  m_type = t_double;
  m_numeric.v_double = v;
  m_string_current = false;
  return *this;
}

PropertyValue &
PropertyValue::operator= (const std::string &v)
{
  m_type = t_string;
  m_string = v;
  m_string_current = true;
  return *this;
}

PropertyValue &
PropertyValue::operator= (const std::vector<PropertyValue> &l)
{
  m_type = t_list;
  m_list = l;
  m_string_current = false;
  return *this;
}

PropertyValue &
PropertyValue::operator= (const PropertyValue &d)
{
  if (&d != this) {

    m_type = d.m_type;
    m_string_current = false;

    if (m_type == t_double) {
      m_numeric.v_double = d.m_numeric.v_double;
    } else if (m_type == t_long) {
      m_numeric.v_long = d.m_numeric.v_long;
    } else if (m_type == t_string) {
      m_string = d.m_string;
      m_string_current = true;
    } else if (m_type == t_list) {
      m_list = d.m_list;
    }

  }

  return *this;
}

bool 
PropertyValue::operator== (const PropertyValue &d) const
{
  if (m_type != d.m_type) {
    return false;
  }

  if (m_type == t_void) {
    return true;
  } else if (m_type == t_long) {
    return m_numeric.v_long == d.m_numeric.v_long;
  } else if (m_type == t_double) {
    return m_numeric.v_double == d.m_numeric.v_double;
  } else if (m_type == t_string) {
    return m_string == d.m_string;
  } else if (m_type == t_list) {
    return m_list == d.m_list;
  }

  tl_assert (false);
}

bool 
PropertyValue::operator!= (const PropertyValue &d) const
{
  return ! operator== (d);
}

bool 
PropertyValue::operator< (const PropertyValue &d) const
{
  if (m_type != d.m_type) {
    return int (m_type) < int (d.m_type);
  }

  if (m_type == t_void) {
    return false;
  } else if (m_type == t_long) {
    return m_numeric.v_long < d.m_numeric.v_long;
  } else if (m_type == t_double) {
    return m_numeric.v_double < d.m_numeric.v_double;
  } else if (m_type == t_string) {
    return m_string < d.m_string;
  } else if (m_type == t_list) {
    return m_list < d.m_list;
  }

  tl_assert (false);
}

bool 
PropertyValue::is_string () const
{
  return true;
}

bool 
PropertyValue::is_double () const
{
  return m_type == t_double || m_type == t_long;
}

bool 
PropertyValue::is_long () const
{
  return m_type == t_long;
}

bool 
PropertyValue::is_list () const
{
  return m_type == t_list;
}

long 
PropertyValue::get_long () const
{
  if (m_type == t_long) {
    return m_numeric.v_long;
  } 
  tl_assert (false);
}

double 
PropertyValue::get_double () const
{
  if (m_type == t_long) {
    return m_numeric.v_long;
  } else if (m_type == t_double) {
    return m_numeric.v_double;
  } 
  tl_assert (false);
}

std::string
PropertyValue::to_string () const
{
  if (m_type == t_long) {
    return "#" + tl::to_string (m_numeric.v_long);
  } else if (m_type == t_double) {
    return "##" + tl::to_string (m_numeric.v_double);
  } else if (m_type == t_string) {
    return tl::to_word_or_quoted_string (m_string);
  } else if (m_type == t_list) {
    std::string r = "(";
    for (iterator l = begin (); l != end (); ++l) {
      if (l != begin ()) {
        r += ",";
      }
      r += l->to_string ();
    }
    r += ")";
    return r;
  } else {
    return "()";
  }
}

const std::string &
PropertyValue::get_string () const
{
  if (! m_string_current) {
    if (m_type == t_void) {
      m_string = "";
    } else if (m_type == t_long) {
      m_string = tl::to_string (m_numeric.v_long);
    } else if (m_type == t_double) {
      m_string = tl::to_string (m_numeric.v_double);
    } else if (m_type == t_list) {
      m_string = "";
      for (iterator l = begin (); l != end (); ++l) {
        if (l != begin ()) {
          m_string += ",";
        }
        m_string += l->get_string ();
      }
    } 
    m_string_current = true;
  }
  return m_string;
}

PropertyValue::iterator
PropertyValue::begin () const
{
  return m_list.begin ();
}

PropertyValue::iterator
PropertyValue::end () const
{
  return m_list.end ();
}

// ----------------------------------------------------------------------------------
//  PropertiesRepository implementation

PropertiesRepository::PropertiesRepository ()
{
  //  install empty property set
  properties_set empty_set;
  properties_id_type id = properties_id (empty_set);
  tl_assert (id == 0);
}

PropertiesRepository::PropertiesRepository (const PropertiesRepository &d)
  : m_propnames_by_id (d.m_propnames_by_id), 
    m_propname_ids_by_name (d.m_propname_ids_by_name),
    m_properties_by_id (d.m_properties_by_id),
    m_properties_ids_by_set (d.m_properties_ids_by_set),
    m_properties_component_table (d.m_properties_component_table)
{
  //  .. nothing yet ..
}

PropertiesRepository &
PropertiesRepository::operator= (const PropertiesRepository &d)
{
  if (&d != this) {
    m_propnames_by_id            = d.m_propnames_by_id;
    m_propname_ids_by_name       = d.m_propname_ids_by_name;
    m_properties_by_id           = d.m_properties_by_id;
    m_properties_ids_by_set      = d.m_properties_ids_by_set;
    m_properties_component_table = d.m_properties_component_table;
  }
  return *this;
}

std::pair<bool, PropertiesRepository::property_names_id_type>
PropertiesRepository::get_id_of_name (const PropertyValue &name) const
{
  std::map <PropertyValue, property_names_id_type>::const_iterator pi = m_propname_ids_by_name.find (name);
  if (pi == m_propname_ids_by_name.end ()) {
    return std::make_pair (false, property_names_id_type (0));
  } else {
    return std::make_pair (true, pi->second);
  }
}

PropertiesRepository::property_names_id_type 
PropertiesRepository::prop_name_id (const PropertyValue &name)
{
  std::map <PropertyValue, property_names_id_type>::const_iterator pi = m_propname_ids_by_name.find (name);
  if (pi == m_propname_ids_by_name.end ()) {
    property_names_id_type id = m_propnames_by_id.size ();
    m_propnames_by_id.insert (std::make_pair (id, name));
    m_propname_ids_by_name.insert (std::make_pair (name, id));
    return id;
  } else {
    return pi->second;
  }
}

const PropertyValue &
PropertiesRepository::prop_name (property_names_id_type id) const
{
  return m_propnames_by_id.find (id)->second;
}

PropertiesRepository::properties_id_type 
PropertiesRepository::properties_id (const properties_set &props)
{
  std::map <properties_set, properties_id_type>::const_iterator pi = m_properties_ids_by_set.find (props);
  if (pi == m_properties_ids_by_set.end ()) {
    properties_id_type id = m_properties_ids_by_set.size ();
    m_properties_ids_by_set.insert (std::make_pair (props, id));
    m_properties_by_id.insert (std::make_pair (id, props));
    for (properties_set::const_iterator nv = props.begin (); nv != props.end (); ++nv) {
      m_properties_component_table.insert (std::make_pair (*nv, properties_id_vector ())).first->second.push_back (id);
    }
    return id;
  } else {
    return pi->second;
  }
}

const PropertiesRepository::properties_set &
PropertiesRepository::properties (properties_id_type id) const
{
  return m_properties_by_id.find (id)->second;
}

const PropertiesRepository::properties_id_vector &
PropertiesRepository::properties_ids_by_name_value (const name_value_pair &nv) const
{
  std::map <name_value_pair, properties_id_vector>::const_iterator idv = m_properties_component_table.find (nv);
  if (idv == m_properties_component_table.end ()) {
    static properties_id_vector empty;
    return empty;
  } else {
    return idv->second;
  }
}

} // namespace db

// ----------------------------------------------------------------------------------
//  Extractor implementation

namespace tl
{

template <>
void extractor_impl (tl::Extractor &ex, db::PropertyValue &v)
{
  if (! test_extractor_impl (ex, v)) {
    ex.error (tl::translate ("Expected a property value specification"));
  }
}

template <>
bool test_extractor_impl (tl::Extractor &ex, db::PropertyValue &v)
{
  std::string s;

  if (ex.test ("##")) {

    double x = 0;
    ex.read (x);
    v = x;
    return true;

  } else if (ex.test ("#")) {

    long x = 0;
    ex.read (x);
    v = x;
    return true;

  } else if (ex.test ("(")) {

    std::vector<db::PropertyValue> values;
    if (! ex.test (")")) {
      while (true) {
        values.push_back (db::PropertyValue ());
        ex.read (values.back ());
        if (ex.test (",")) {
          //  .. continue
        } else {
          ex.expect (")");
          break;
        }
      }
    }
    if (values.size () == 0) {
      v = db::PropertyValue ();
    } else if (values.size () == 1){
      v = values.back ();
    } else {
      v = db::PropertyValue (values.begin (), values.end ());
    }
    return true;

  } else if (ex.try_read_word_or_quoted (s)) {

    v = db::PropertyValue (s);
    return true;

  } else {

    return false;

  }
}


} // namespace tl

