
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbPropertiesRepository
#define HDR_dbPropertiesRepository

#include <vector>
#include <string>
#include <map>

namespace db
{

/**
 *  @brief The property value
 *
 *  This is basically a variant specialized for use with property values.
 *  It can be a long integer, a double or a string value.
 *  Implict conversion is provided: a double value is converted to a string
 *  if required. 
 *  A property value can be "void" which renders an empty string.
 */

class PropertyValue
{
public:
  typedef std::vector<PropertyValue>::const_iterator iterator;

  /**  
   *  @brief Default constructor: create a "void" property
   */
  PropertyValue ();

  /**  
   *  @brief Construction from a long value
   */
  PropertyValue (long l);

  /**  
   *  @brief Construction from a double value
   */
  PropertyValue (double d);

  /**  
   *  @brief Construction from a string value
   */
  PropertyValue (const std::string &s);

  /**  
   *  @brief Copy constructor
   */
  PropertyValue (const PropertyValue &d);

  /**  
   *  @brief Construction from a list 
   */
  template <class PI>
  PropertyValue (const PI &from, const PI &to)
    : m_type (t_list), m_list (from, to), m_string_current (false)
  {
    // .. nothing yet ..
  }


  /**  
   *  @brief Assignment from a long value
   */
  PropertyValue &operator= (long v);

  /**  
   *  @brief Assignment from a double value
   */
  PropertyValue &operator= (double v);

  /**  
   *  @brief Assignment from a string value
   */
  PropertyValue &operator= (const std::string &v);

  /**  
   *  @brief Assignment operator
   */
  PropertyValue &operator= (const PropertyValue &d);

  /**  
   *  @brief Assignment operator
   */
  PropertyValue &operator= (const std::vector<PropertyValue> &d);

  /**  
   *  @brief Equality
   */
  bool operator== (const PropertyValue &d) const;

  /**  
   *  @brief Inequality
   */
  bool operator!= (const PropertyValue &d) const;

  /**  
   *  @brief Comparison
   */
  bool operator< (const PropertyValue &d) const;

  /**  
   *  @brief "is_string" predicate
   *
   *  This method returns true, if the value is a string or can be 
   *  converted into one. Basically this is the case for all content.
   */
  bool is_string () const;

  /**  
   *  @brief "is_double" predicate
   *
   *  This method returns true, if the value is a double or can be 
   *  converted into one. This is true for double or long.
   */
  bool is_double () const;

  /**  
   *  @brief "is_long" predicate
   *
   *  This method returns true, if the value is a long value.
   *  It returns false for all others.
   */
  bool is_long () const;

  /**  
   *  @brief "is_list" predicate
   *
   *  This method returns true, if the value is a list.
   */
  bool is_list () const;

  /**
   *  @brief Return the long value
   *
   *  This method will assert if "is_long" is false.
   */
  long get_long () const;

  /**
   *  @brief Return the double value
   *
   *  This method will assert if "is_double" is false.
   */
  double get_double () const;

  /**
   *  @brief Return the begin iterator for the list, if the property is one
   *
   *  This method will assert if "is_list" is false.
   */
  iterator begin () const;

  /**
   *  @brief Return the end iterator for the list, if the property is one
   *
   *  This method will assert if "is_list" is false.
   */
  iterator end () const;

  /**
   *  @brief Return the string value
   *
   *  This method will assert if "is_string" is false.
   */
  const std::string &get_string () const;

  /**
   *  @brief Convert to a string that can be parsed back
   */
  std::string to_string () const;

private:
  enum { t_void, t_long, t_double, t_string, t_list } m_type;

  union {
    long v_long;
    double v_double;
  } m_numeric;

  std::vector<PropertyValue> m_list;

  mutable bool m_string_current;
  mutable std::string m_string;
};


/**
 *  @brief The properties repository
 *
 *  This repository associates a set of property name/value pairs with 
 *  an unique Id which can be stored with a shape_with_properties element.
 *  For performance reasons property names (which are strings) are not
 *  stored as such but as integers.
 */

class PropertiesRepository
{
public:
  typedef size_t properties_id_type;
  typedef size_t property_names_id_type;
  typedef std::map <property_names_id_type, PropertyValue> properties_set;
  typedef std::map <properties_id_type, properties_set> properties_map;
  typedef std::map <properties_id_type, properties_set>::const_iterator iterator;
  typedef std::pair <property_names_id_type, PropertyValue> name_value_pair;
  typedef std::vector <properties_id_type> properties_id_vector;

  /**
   *  @brief Default constructor
   */
  PropertiesRepository ();

  /**
   *  @brief Copy constructor
   */
  PropertiesRepository (const PropertiesRepository &d);

  /**
   *  @brief Assignment
   */
  PropertiesRepository &operator= (const PropertiesRepository &d);

  /**
   *  @brief Associate a name with a name Id
   * 
   *  A name of a property can be either a integer or a string as specified
   *  with the PropertyValue variant. In principle it could even be a list or void.
   *  This method will assign a new Id to the given name if required and
   *  return the Id associated with it.
   */
  property_names_id_type prop_name_id (const PropertyValue &name);
  
  /**
   *  @brief Get the id for a name 
   * 
   *  This method checks wheter the given name is present as a name and returns the 
   *  id in the second member of the pair. The first member is true, if the name is
   *  present.
   */
  std::pair<bool, property_names_id_type> get_id_of_name (const PropertyValue &name) const;
  
  /**
   *  @brief Associate a name with a name Id
   * 
   *  This method will return the name associated with the given Id.
   *  It will assert if the Id is not a valid one.
   */
  const PropertyValue &prop_name (property_names_id_type id) const;
  
  /**
   *  @brief Associate a properties set with a properties Id
   * 
   *  This method will assign a new Id to the given set if required and
   *  return the Id associated with it.
   *  An empty property set is associated with property Id 0.
   */
  properties_id_type properties_id (const properties_set &props);
  
  /**
   *  @brief Associate a properties set with a properties Id
   * 
   *  This method will return the properties set associated with the given Id.
   *  It will assert if the Id is not a valid one.
   *  Id 0 always delivers an empty property set.
   */
  const properties_set &properties (properties_id_type id) const;

  /**
   *  @brief Iterate over Id/Properties sets
   */
  iterator begin () const
  {
    return m_properties_by_id.begin ();
  }

  /**
   *  @brief Iterate over Id/Properties sets: end iterator
   */
  iterator end () const
  {
    return m_properties_by_id.end ();
  }

  /**
   *  @brief Obtain the first properties id in the repository
   */
  properties_id_type begin_id () 
  {
    return 0;
  }
  
  /**
   *  @brief Obtain the last properties id in the repository plus 1
   */
  properties_id_type end_id () 
  {
    return m_properties_by_id.size ();
  }
  
  /**
   *  @brief Lookup a table of properties id's by a name value pair
   *
   *  For a given name/value pair, this method returns a vector of ids
   *  of property sets that contain the given name/value pair. This method
   *  is intended for use with the properties_id resolution algorithm.
   */
  const properties_id_vector &properties_ids_by_name_value (const name_value_pair &nv) const;

private:
  std::map <property_names_id_type, PropertyValue> m_propnames_by_id;
  std::map <PropertyValue, property_names_id_type> m_propname_ids_by_name;

  std::map <properties_id_type, properties_set> m_properties_by_id;
  std::map <properties_set, properties_id_type> m_properties_ids_by_set;
  std::map <name_value_pair, properties_id_vector> m_properties_component_table;
};

} // namespace db

/**
 *  @brief Special extractors for the property values
 */

namespace tl 
{

  class Extractor;
  template <class T> void extractor_impl (tl::Extractor &ex, T &);
  template <class T> bool test_extractor_impl (tl::Extractor &ex, T &);

  template <> void extractor_impl (tl::Extractor &ex, db::PropertyValue &p);
  template <> bool test_extractor_impl (tl::Extractor &ex, db::PropertyValue &p);

} // namespace tl

#endif

