
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbShape
#define HDR_dbShape

#include <tlException.h>
#include <dbManager.h>
#include <dbPolygon.h>
#include <dbPath.h>
#include <dbEdge.h>
#include <dbText.h>
#include <dbBox.h>
#include <dbShapes.h>

#include <set>

namespace db 
{

/**
 *  @brief A generic edge iterator 
 *
 *  This edge iterator is used to provide edge iterators for simple and
 *  complex polygons as well.
 *  The implementation is based on a union scheme. It assumes that the
 *  iterators represented by the generic instance do not require a special
 *  copy semantics but can be copied by bytewise copy simply.
 */
template <class C>
class generic_polygon_edge_iterator
{
public:
  typedef C coord_type;
  typedef db::edge<C> edge_type;
  typedef db::polygon<C> polygon_type;
  typedef typename polygon_type::polygon_edge_iterator polygon_edge_iterator_type;
  typedef db::simple_polygon<C> simple_polygon_type;
  typedef typename simple_polygon_type::polygon_edge_iterator simple_polygon_edge_iterator_type;
  typedef db::polygon_ref<polygon_type> polygon_ref_type;
  typedef typename polygon_ref_type::polygon_edge_iterator polygon_ref_edge_iterator_type;
  typedef db::polygon_ref<simple_polygon_type> simple_polygon_ref_type;
  typedef typename simple_polygon_ref_type::polygon_edge_iterator simple_polygon_ref_edge_iterator_type;

  enum iterator_type
  {
    SimplePolygon, SimplePolygonRef, Polygon, PolygonRef
  };

  /**
   *  @brief This constructor creates an edge iterator at "begin" for polygons
   * 
   *  The iterator stores begin and end and provides a "at_end" method to
   *  test if the sequence is at the end.
   *
   *  @param begin The start of the sequence
   */
  generic_polygon_edge_iterator (const polygon_edge_iterator_type &begin)
  {
    m_type = Polygon;
    new (m_d.iter) polygon_edge_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for polygon references
   * 
   *  See former ctor.
   */
  generic_polygon_edge_iterator (const polygon_ref_edge_iterator_type &begin)
  {
    m_type = PolygonRef;
    new (m_d.iter) polygon_ref_edge_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for simple polygons
   * 
   *  See former ctor.
   */
  generic_polygon_edge_iterator (const simple_polygon_edge_iterator_type &begin)
  {
    m_type = SimplePolygon;
    new (m_d.iter) simple_polygon_edge_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for simple polygon references
   * 
   *  See former ctor.
   */
  generic_polygon_edge_iterator (const simple_polygon_ref_edge_iterator_type &begin)
  {
    m_type = SimplePolygonRef;
    new (m_d.iter) simple_polygon_ref_edge_iterator_type (begin);
  }
    
  /**
   *  @brief at_end predicate
   */
  bool at_end () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief at_begin predicate
   */
  bool at_begin () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief Increment operator
   */
  generic_polygon_edge_iterator &operator++ () 
  {
    generic_f<bool, inc_f> ();
    return *this;
  }

  /**
   *  @brief Decrement operator
   */
  generic_polygon_edge_iterator &operator-- () 
  {
    generic_f<bool, dec_f> ();
    return *this;
  }

  /**
   *  @brief access operator
   */
  edge_type operator* () const
  {
    return generic_const_f<edge_type, deref_f> ();
  }
  
private:
  //  This union is there to compute the maximum size required to hold all
  //  kind of iterators 
  union iter_size {
    char sz1 [sizeof (polygon_edge_iterator_type)];
    char sz2 [sizeof (polygon_ref_edge_iterator_type)];
    char sz3 [sizeof (simple_polygon_edge_iterator_type)];
    char sz4 [sizeof (simple_polygon_ref_edge_iterator_type)];
  };

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (iter_size)];
  } m_d;

  iterator_type m_type;

  struct at_end_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_end ();
    } 
  };

  struct inc_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator++ ();
      return false;
    } 
  };

  struct dec_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator-- ();
      return false;
    } 
  };

  struct deref_f {
    template <class Iter>
    edge_type operator() (const Iter &iter) const
    {
      return iter.operator* ();
    } 
  };

  struct at_begin_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_begin ();
    } 
  };

  template <class Ret, class Func>
  Ret generic_const_f () const
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((const polygon_edge_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((const polygon_ref_edge_iterator_type *) m_d.iter));
    } else if (m_type == SimplePolygon) {
      return f (*((const simple_polygon_edge_iterator_type *) m_d.iter));
    } else {
      return f (*((const simple_polygon_ref_edge_iterator_type *) m_d.iter));
    }
  }

  template <class Ret, class Func>
  Ret generic_f () 
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((polygon_edge_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((polygon_ref_edge_iterator_type *) m_d.iter));
    } else if (m_type == SimplePolygon) {
      return f (*((simple_polygon_edge_iterator_type *) m_d.iter));
    } else {
      return f (*((simple_polygon_ref_edge_iterator_type *) m_d.iter));
    }
  }
};

/**
 *  @brief A generic point iterator 
 *
 *  This point iterator is used to provide point iterators for simple and
 *  complex polygons as well as paths.
 *  The implementation is based on a union scheme. It assumes that the
 *  iterators represented by the generic instance do not require a special
 *  copy semantics but can be copied by bytewise copy simply.
 */
template <class C>
class generic_point_iterator
{
public:
  typedef C coord_type;
  typedef db::point<C> point_type;
  typedef db::polygon<C> polygon_type;
  typedef typename polygon_type::polygon_contour_iterator polygon_point_iterator_type;
  typedef db::polygon_ref<polygon_type> polygon_ref_type;
  typedef typename polygon_ref_type::polygon_contour_iterator polygon_ref_point_iterator_type;
  typedef db::path<C> path_type;
  typedef typename path_type::iterator path_point_iterator_type;
  typedef db::path_ref<path_type> path_ref_type;
  typedef typename path_ref_type::iterator path_ref_point_iterator_type;

  enum iterator_type
  {
    Polygon, PolygonRef, Path, PathRef
  };

  /**
   *  @brief This constructor creates an point iterator at "begin" for polygons
   * 
   *  The iterator stores begin and end and provides a "at_end" method to
   *  test if the sequence is at the end.
   *  This method also applies to simple polygon contours.
   *
   *  @param begin The start of the sequence
   */
  generic_point_iterator (const polygon_point_iterator_type &begin)
  {
    m_type = Polygon;
    new (m_d.iter) polygon_point_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an point iterator at "begin" for polygon references
   * 
   *  This method also applies to simple polygon references.
   *  See former ctor.
   */
  generic_point_iterator (const polygon_ref_point_iterator_type &begin)
  {
    m_type = PolygonRef;
    new (m_d.iter) polygon_ref_point_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an point iterator at "begin" for paths
   * 
   *  See former ctor.
   */
  generic_point_iterator (const path_point_iterator_type &begin)
  {
    m_type = Path;
    new (m_d.iter) path_point_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an point iterator at "begin" for path references
   * 
   *  See former ctor.
   */
  generic_point_iterator (const path_ref_point_iterator_type &begin)
  {
    m_type = PathRef;
    new (m_d.iter) path_ref_point_iterator_type (begin);
  }
    
  /**
   *  @brief at_end predicate
   */
  bool at_end () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief at_begin predicate
   */
  bool at_begin () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief Increment operator
   */
  generic_point_iterator &operator++ () 
  {
    generic_f<bool, inc_f> ();
    return *this;
  }

  /**
   *  @brief Decrement operator
   */
  generic_point_iterator &operator-- () 
  {
    generic_f<bool, dec_f> ();
    return *this;
  }

  /**
   *  @brief access operator
   */
  point_type operator* () const
  {
    return generic_const_f<point_type, deref_f> ();
  }

  /**
   *  @brief Equality 
   */
  bool operator== (const generic_point_iterator &d) const
  {
    tl_assert (m_type == d.m_type);

    if (m_type == Polygon) {
      return *((const polygon_point_iterator_type *) m_d.iter) == *((const polygon_point_iterator_type *) d.m_d.iter);
    } else if (m_type == PolygonRef) {
      return *((const polygon_ref_point_iterator_type *) m_d.iter) == *((const polygon_ref_point_iterator_type *) d.m_d.iter);
    } else if (m_type == Path) {
      return *((const path_point_iterator_type *) m_d.iter) == *((const path_point_iterator_type *) d.m_d.iter);
    } else {
      return *((const path_ref_point_iterator_type *) m_d.iter) == *((const path_ref_point_iterator_type *) d.m_d.iter);
    }
  }

  /**
   *  @brief Inequality
   */
  bool operator!= (const generic_point_iterator &d) const
  {
    return !operator== (d);
  }
  
private:
  //  This union is there to compute the maximum size required to hold all
  //  kind of iterators 
  union iter_size {
    char sz1 [sizeof (polygon_point_iterator_type)];
    char sz2 [sizeof (polygon_ref_point_iterator_type)];
    char sz3 [sizeof (path_point_iterator_type)];
    char sz4 [sizeof (path_ref_point_iterator_type)];
  };

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (iter_size)];
  } m_d;

  iterator_type m_type;

  struct at_end_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_end ();
    } 
  };

  struct inc_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator++ ();
      return false;
    } 
  };

  struct dec_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator-- ();
      return false;
    } 
  };

  struct deref_f {
    template <class Iter>
    point_type operator() (const Iter &iter) const
    {
      return iter.operator* ();
    } 
  };

  struct at_begin_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_begin ();
    } 
  };

  template <class Ret, class Func>
  Ret generic_const_f () const
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((const polygon_point_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((const polygon_ref_point_iterator_type *) m_d.iter));
    } else if (m_type == Path) {
      return f (*((const path_point_iterator_type *) m_d.iter));
    } else {
      return f (*((const path_ref_point_iterator_type *) m_d.iter));
    }
  }

  template <class Ret, class Func>
  Ret generic_f () 
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((polygon_point_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((polygon_ref_point_iterator_type *) m_d.iter));
    } else if (m_type == Path) {
      return f (*((path_point_iterator_type *) m_d.iter));
    } else {
      return f (*((path_ref_point_iterator_type *) m_d.iter));
    }
  }
};

/**
 *  @brief A shape proxy 
 *
 *  The shape proxy is basically a pointer to a shape of different kinds 
 *  No copy of the shape is created: if the shape proxy is copied the copy still
 *  points to the original shape. If the original shape is modified or deleted,
 *  the shape proxy will also point to a modified or invalid shape.
 *  The proxy can be "null" which means a invalid reference.
 */ 
template <class C>
class shape
{
public:
  typedef C coord_type;
  typedef db::polygon<C> polygon_type;
  typedef db::simple_polygon<C> simple_polygon_type;
  typedef db::polygon_ref<polygon_type> polygon_ref_type;
  typedef db::polygon_ref<simple_polygon_type> simple_polygon_ref_type;
  typedef db::path<C> path_type;
  typedef db::path_ref<path_type> path_ref_type;
  typedef db::edge<C> edge_type;
  typedef db::text<C> text_type;
  typedef db::text_ref<text_type> text_ref_type;
  typedef db::user_object<C> user_object_type;
  typedef db::box<C> box_type;
  typedef db::point<C> point_type;
  typedef db::generic_polygon_edge_iterator<C> polygon_edge_iterator;
  typedef db::generic_point_iterator<C> point_iterator;

  enum object_type {
    Null,
    Polygon,
    PolygonRef,
    SimplePolygon,
    SimplePolygonRef,
    Edge,
    Path,
    PathRef,
    Box,
    Text,
    TextRef,
    UserObject
  };
    
  /** 
   *  @brief Construct a shape proxy as a null object
   */
  shape ()
    : m_props (0), m_type (Null)
  {
    m_generic.any = 0;
  }

  /**
   *  @brief Construct as a proxy to a certain object
   */
  template <class Obj>
  shape (const Obj &obj)
  {
    init (obj);
  }

  /** 
   *  @brief Construct a shape proxy as a shape with properties
   */
  template <class Sh>
  void init (const db::shape_with_properties<Sh> &swp)
  {
    init ((const Sh &) swp);
    m_props = swp.properties_id ();
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon
   */
  void init (const polygon_type &polygon)
  {
    m_props = 0;
    m_type = Polygon;
    m_generic.polygon = &polygon;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon reference
   */
  void init (const polygon_ref_type &polygon_ref)
  {
    m_props = 0;
    m_type = PolygonRef;
    m_generic.polygon_ref = &polygon_ref;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a simple olygon
   */
  void init (const simple_polygon_type &polygon)
  {
    m_props = 0;
    m_type = SimplePolygon;
    m_generic.simple_polygon = &polygon;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon reference
   */
  void init (const simple_polygon_ref_type &polygon_ref)
  {
    m_props = 0;
    m_type = SimplePolygonRef;
    m_generic.simple_polygon_ref = &polygon_ref;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path
   */
  void init (const path_type &path)
  {
    m_props = 0;
    m_type = Path;
    m_generic.path = &path;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path reference
   */
  void init (const path_ref_type &path)
  {
    m_props = 0;
    m_type = PathRef;
    m_generic.path_ref = &path;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a text
   */
  void init (const text_type &text)
  {
    m_props = 0;
    m_type = Text;
    m_generic.text = &text;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a text reference
   */
  void init (const text_ref_type &text_ref)
  {
    m_props = 0;
    m_type = TextRef;
    m_generic.text_ref = &text_ref;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a edge
   */
  void init (const edge_type &edge)
  {
    m_props = 0;
    m_type = Edge;
    m_generic.edge = &edge;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a box
   */
  void init (const box_type &box)
  {
    m_props = 0;
    m_type = Box;
    m_generic.box = &box;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a user object
   */
  void init (const user_object_type &obj)
  {
    m_props = 0;
    m_type = UserObject;
    m_generic.user_object = &obj;
  }

  /**
   *  @brief Associate a properties Id with the shape
   */
  void prop_id (db::PropertiesRepository::properties_id_type id)
  {
    m_props = id;
  }

  /**
   *  @brief Reset the properties associated with the shape
   */
  void prop_id ()
  {
    m_props = 0;
  }

  /**
   *  @brief Get the properties Id associated with the shape
   */
  db::PropertiesRepository::properties_id_type prop_id () const
  {
    return m_props;
  }

  /**
   *  @brief Check, if the shape is associated with a properties Id
   */
  bool has_prop_id () const
  {
    return m_props != 0;
  }

  /**
   *  @brief Create the "begin" point iterator
   *
   *  This method applies to paths.
   */
  point_iterator begin_point () const
  {
    if (m_type == Path) {
      return point_iterator (path ().begin ());
    } else if (m_type == PathRef) {
      return point_iterator (path_ref ().begin ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "end" point iterator
   *
   *  This method applies to paths.
   */
  point_iterator end_point () const
  {
    if (m_type == Path) {
      return point_iterator (path ().end ());
    } else if (m_type == PathRef) {
      return point_iterator (path_ref ().end ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "begin" hull point iterator
   *
   *  This method applies to polygons.
   */
  point_iterator begin_hull () const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().begin_hull ());
    } else if (m_type == SimplePolygonRef) {
      return point_iterator (simple_polygon_ref ().begin_hull ());
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().begin_hull ());
    } else if (m_type == PolygonRef) {
      return point_iterator (polygon_ref ().begin_hull ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "end" hull point iterator
   *
   *  This method applies to polygons.
   */
  point_iterator end_hull () const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().end_hull ());
    } else if (m_type == SimplePolygonRef) {
      return point_iterator (simple_polygon_ref ().end_hull ());
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().end_hull ());
    } else if (m_type == PolygonRef) {
      return point_iterator (polygon_ref ().end_hull ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "begin" hole point iterator
   *
   *  This method applies to polygons.
   *  Simple polygons deliver an empty sequence.
   *
   *  @param hole The hole index (see holes () method)
   */
  point_iterator begin_hole (unsigned int hole) const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().begin_hole (hole));
    } else if (m_type == SimplePolygonRef) {
      return point_iterator (simple_polygon_ref ().begin_hole (hole));
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().begin_hole (hole));
    } else if (m_type == PolygonRef) {
      return point_iterator (polygon_ref ().begin_hole (hole));
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "end" hole point iterator
   *
   *  This method applies to polygons.
   *  Simple polygons deliver an empty sequence.
   *
   *  @param hole The hole index (see holes () method)
   */
  point_iterator end_hole (unsigned int hole) const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().end_hole (hole));
    } else if (m_type == SimplePolygonRef) {
      return point_iterator (simple_polygon_ref ().end_hole (hole));
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().end_hole (hole));
    } else if (m_type == PolygonRef) {
      return point_iterator (polygon_ref ().end_hole (hole));
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Return the number of holes
   *
   *  This method applies to polygons.
   *  Simple polygons deliver a zero value.
   */
  unsigned int holes () const
  {
    if (m_type == SimplePolygon) {
      return simple_polygon ().holes ();
    } else if (m_type == SimplePolygonRef) {
      return simple_polygon_ref ().obj ().holes ();
    } else if (m_type == Polygon) {
      return polygon ().holes ();
    } else if (m_type == PolygonRef) {
      return polygon_ref ().obj ().holes ();
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create and return an edge iterator
   */
  polygon_edge_iterator begin_edge () const
  {
    if (m_type == SimplePolygon) {
      return polygon_edge_iterator (simple_polygon ().begin_edge ());
    } else if (m_type == SimplePolygonRef) {
      return polygon_edge_iterator (simple_polygon_ref ().begin_edge ());
    } else if (m_type == Polygon) {
      return polygon_edge_iterator (polygon ().begin_edge ());
    } else if (m_type == PolygonRef) {
      return polygon_edge_iterator (polygon_ref ().begin_edge ());
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Return the type of the shape reference
   */
  object_type type () const
  {
    return m_type;
  }

  /** 
   *  @brief Test if the shape proxy is a null object
   */
  bool is_null () const
  {
    return m_type == Null;
  }

  /** 
   *  @brief Return a reference to the polygon if one is referenced
   */
  const polygon_type &polygon () const
  {
    tl_assert (m_type == Polygon);
    return *m_generic.polygon;
  }

  /** 
   *  @brief Return a reference to the polygon reference if one is referenced
   */
  const polygon_ref_type &polygon_ref () const
  {
    tl_assert (m_type == PolygonRef);
    return *m_generic.polygon_ref;
  }

  /** 
   *  @brief Return a reference to the simple polygon if one is referenced
   */
  const simple_polygon_type &simple_polygon () const
  {
    tl_assert (m_type == SimplePolygon);
    return *m_generic.simple_polygon;
  }

  /** 
   *  @brief Return a reference to the simple polygon reference if one is referenced
   */
  const simple_polygon_ref_type &simple_polygon_ref () const
  {
    tl_assert (m_type == SimplePolygonRef);
    return *m_generic.simple_polygon_ref;
  }

  /** 
   *  @brief Test if the shape proxy points to a polygon
   */
  bool is_polygon () const
  {
    return (m_type == Polygon || m_type == PolygonRef || m_type == SimplePolygon || m_type == SimplePolygonRef);
  }

  /** 
   *  @brief Instantiate the polygon object
   * 
   *  If a polygon is referenced, this object is instantiated
   *  by this method.
   */
  void polygon (polygon_type &p) const
  {
    if (m_type == Polygon) {
      p = *m_generic.polygon;
    } else if (m_type == PolygonRef) {
      m_generic.polygon_ref->instantiate (p);
    } else if (m_type == SimplePolygon) {
      p.clear ();
      p.assign_hull (m_generic.simple_polygon->begin_hull (), m_generic.simple_polygon->end_hull ());
    } else if (m_type == SimplePolygonRef) {
      p.clear ();
      p.assign_hull (m_generic.simple_polygon_ref->obj ().begin_hull (), m_generic.simple_polygon_ref->obj ().end_hull (), m_generic.simple_polygon_ref->trans ());
    } else if (m_type == Path) {
      p = m_generic.path->polygon ();
    } else if (m_type == PathRef) {
      p = m_generic.path_ref->obj ().polygon ();
      p.transform (m_generic.path_ref->trans ());
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Return a reference to the path if one is referenced
   */
  const path_type &path () const
  {
    tl_assert (m_type == Path);
    return *m_generic.path;
  }

  /** 
   *  @brief Return a reference to the path reference if one is referenced
   */
  const path_ref_type &path_ref () const
  {
    tl_assert (m_type == PathRef);
    return *m_generic.path_ref;
  }

  /** 
   *  @brief Test if the shape proxy points to a path
   */
  bool is_path () const
  {
    return (m_type == Path || m_type == PathRef);
  }

  /**
   *  @brief Obtain the path width
   *
   *  Applies to paths only. Will assert if not a text.
   */
  db::Path::coord_type path_width () const
  {
    tl_assert (m_type == Path || m_type == PathRef);
    if (m_type == Path) {
      return m_generic.path->width ();
    } else {
      return m_generic.path_ref->obj ().width ();
    }
  }
  
  /**
   *  @brief Obtain the path extensions
   *
   *  Applies to paths only. Will assert if not a text.
   *
   *  @return A pair consisting of the begin and end extensions.
   */
  const std::pair<db::Path::coord_type, db::Path::coord_type> &path_extensions () const
  {
    tl_assert (m_type == Path || m_type == PathRef);
    if (m_type == Path) {
      return m_generic.path->extensions ();
    } else {
      return m_generic.path_ref->obj ().extensions ();
    }
  }
  
  /** 
   *  @brief Instantiate the path object
   * 
   *  If a path is referenced, this object is instantiated
   *  by this method.
   */
  void path (path_type &p) const
  {
    if (m_type == Path) {
      p = *m_generic.path;
    } else if (m_type == PathRef) {
      m_generic.path_ref->instantiate (p);
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Return a reference to the edge if one is referenced
   */
  const edge_type &edge () const
  {
    tl_assert (m_type == Edge);
    return *m_generic.edge;
  }

  /** 
   *  @brief Test if the shape proxy points to a edge
   */
  bool is_edge () const
  {
    return (m_type == Edge);
  }

  /** 
   *  @brief Instantiate the edge object
   * 
   *  If a edge is referenced, this object is instantiated
   *  by this method.
   */
  void edge (edge_type &e) const
  {
    tl_assert (m_type == Edge);
    e = *m_generic.edge;
  }

  /** 
   *  @brief Return a reference to the text if one is referenced
   */
  const text_type &text () const
  {
    tl_assert (m_type == Text);
    return *m_generic.text;
  }

  /** 
   *  @brief Return a reference to the text if one is referenced
   */
  const text_ref_type &text_ref () const
  {
    tl_assert (m_type == TextRef);
    return *m_generic.text_ref;
  }

  /** 
   *  @brief Test if the shape proxy points to a text
   */
  bool is_text () const
  {
    return (m_type == Text || m_type == TextRef);
  }

  /** 
   *  @brief Instantiate the text object
   * 
   *  If a text is referenced, this object is instantiated
   *  by this method.
   */
  void text (text_type &t) const
  {
    tl_assert (m_type == Text || m_type == TextRef);
    if (m_type == Text) {
      t = *m_generic.text;
    } else {
      m_generic.text_ref->instantiate (t);
    }
  }

  /**
   *  @brief Obtain the text string
   *
   *  Applies to texts only. Will assert if not a text.
   */
  const std::string &text_string () const
  {
    tl_assert (m_type == Text || m_type == TextRef);
    if (m_type == Text) {
      return m_generic.text->string ();
    } else {
      return m_generic.text_ref->obj ().string ();
    }
  }
  
  /**
   *  @brief Obtain the text transformation
   *
   *  Applies to texts only. Will assert if not a text.
   */
  db::Text::trans_type text_trans () const
  {
    tl_assert (m_type == Text || m_type == TextRef);
    if (m_type == Text) {
      return m_generic.text->trans ();
    } else {
      return m_generic.text_ref->trans () * m_generic.text_ref->obj ().trans ();
    }
  }

  /**
   *  @brief Obtain the text size
   *
   *  Applies to texts only. Will assert if not a text.
   */
  db::Text::coord_type text_size () const
  {
    tl_assert (m_type == Text || m_type == TextRef);
    if (m_type == Text) {
      return m_generic.text->size ();
    } else {
      return m_generic.text_ref->obj ().size ();
    }
  }

  /**
   *  @brief Obtain the text's font id
   *
   *  Applies to texts only. Will assert if not a text.
   */
  db::Text::font_id_type text_font_id () const
  {
    tl_assert (m_type == Text || m_type == TextRef);
    if (m_type == Text) {
      return m_generic.text->font_id ();
    } else {
      return m_generic.text_ref->obj ().font_id ();
    }
  }

  /** 
   *  @brief Return a reference to the box if one is referenced
   */
  const box_type &box () const
  {
    tl_assert (m_type == Box);
    return *m_generic.box;
  }

  /** 
   *  @brief Test if the shape proxy points to a box
   */
  bool is_box () const
  {
    return (m_type == Box);
  }

  /** 
   *  @brief Instantiate the box object
   * 
   *  If a box is referenced, this object is instantiated
   *  by this method.
   */
  void box (box_type &b) const
  {
    tl_assert (m_type == Box);
    b = *m_generic.box;
  }

  /** 
   *  @brief Return a reference to the user object if one is referenced
   */
  const user_object_type &user_object () const
  {
    tl_assert (m_type == UserObject);
    return *m_generic.user_object;
  }

  /** 
   *  @brief Test if the shape proxy points to a user object
   */
  bool is_user_object () const
  {
    return (m_type == UserObject);
  }

  /** 
   *  @brief Instantiate the user object
   * 
   *  If a user object is referenced, this object is instantiated
   *  by this method.
   */
  void user_object (user_object_type &u) const
  {
    tl_assert (m_type == UserObject);
    u = *m_generic.user_object;
  }

  /** 
   *  @brief Compute the bounding box of the shape the is referenced
   */
  box_type bbox () const
  {
    switch (m_type) {
    case Null:
      return box_type ();
    case Polygon:
      return m_generic.polygon->box ();
    case PolygonRef:
      return m_generic.polygon_ref->box ();
    case SimplePolygon:
      return m_generic.simple_polygon->box ();
    case SimplePolygonRef:
      return m_generic.simple_polygon_ref->box ();
    case Text:
      return m_generic.text->box ();
    case TextRef:
      return m_generic.text_ref->box ();
    case Edge:
      return box_type (m_generic.edge->p1 (), m_generic.edge->p2 ());
    case Path:
      return m_generic.path->box ();
    case PathRef:
      return m_generic.path_ref->box ();
    case Box:
      return *m_generic.box;
    case UserObject:
      return m_generic.user_object->box ();
    default:
      return box_type ();
    }
  }

  /**
   *  @brief inequality 
   */
  bool operator!= (const shape &d) const
  {
    return ! operator== (d);
  }

  /**
   *  @brief equality 
   *
   *  Equality of shapes is not specified by the identity of the objects but by the
   *  identity of the pointers - both shapes must reference the same object.
   */
  bool operator== (const shape &d) const
  {
    if (m_type != d.m_type) {
      return false;
    }
    return m_generic.any == d.m_generic.any;
  }

  /**
   *  @brief less-than operator 
   *
   *  Ordering of shapes is not specified by the identity of the objects but by the
   *  order of the pointers.
   */
  bool operator< (const shape &d) const
  {
    if (m_type != d.m_type) {
      return m_type < d.m_type;
    }
    return m_generic.any < d.m_generic.any;
  }

  size_t mem_used () const
  {
    size_t s = sizeof (*this);
    switch (m_type) {
    case Polygon:
      return s + mem_used (*polygon ());
    case PolygonRef:
      return s + mem_used (*polygon_ref ());
    case SimplePolygon:
      return s + mem_used (*simple_polygon ());
    case SimplePolygonRef:
      return s + mem_used (*simple_polygon_ref ());
    case Text:
      return s + mem_used (*text ());
    case TextRef:
      return s + mem_used (*text_ref ());
    case Edge:
      return s + mem_used (*edge ());
    case Path:
      return s + mem_used (*path ());
    case PathRef:
      return s + mem_used (*path_ref ());
    case Box:
      return s + mem_used (*box ());
    case UserObject:
      return s + mem_used (*user_object ());
    default:
      return s;
    }
  }

  size_t mem_reqd () const
  {
    size_t s = sizeof (*this);
    switch (m_type) {
    case Polygon:
      return s + mem_reqd (*polygon ());
    case PolygonRef:
      return s + mem_reqd (*polygon_ref ());
    case SimplePolygon:
      return s + mem_reqd (*simple_polygon ());
    case SimplePolygonRef:
      return s + mem_reqd (*simple_polygon_ref ());
    case Text:
      return s + mem_reqd (*text ());
    case TextRef:
      return s + mem_reqd (*text_ref ());
    case Edge:
      return s + mem_reqd (*edge ());
    case Path:
      return s + mem_reqd (*path ());
    case PathRef:
      return s + mem_reqd (*path_ref ());
    case Box:
      return s + mem_reqd (*box ());
    case UserObject:
      return s + mem_reqd (*user_object ());
    default:
      return s;
    }
  }

public:
  union generic {
    const polygon_type *polygon;
    const polygon_ref_type *polygon_ref;
    const simple_polygon_type *simple_polygon;
    const simple_polygon_ref_type *simple_polygon_ref;
    const text_type *text;
    const text_ref_type *text_ref;
    const edge_type *edge;
    const path_type *path;
    const path_ref_type *path_ref;
    const box_type *box;
    const user_object_type *user_object;
    void *any;
  };

  db::PropertiesRepository::properties_id_type m_props;
  object_type m_type;
  generic m_generic;
};

template <class C>
size_t mem_used (const shape<C> &s)
{
  return s.mem_used (); 
}

template <class C>
size_t mem_reqd (const shape<C> &s)
{
  return s.mem_reqd (); 
}


/** 
 *  @brief A typedef for a standard shape
 */
typedef db::shape<db::Coord> Shape;


/**
 *  @brief A generic shape iterator
 *
 *  This iterator can iterate any kind of shape from a "shapes" container.
 *  It allows to select certain kind of shapes. The dereferecing operator
 *  returns a shape proxy object that can be used to access the actual shape.
 *  It can iterator over all shapes and over a region selected.
 *  If the iterator is constructed, it always points to the beginning of the 
 *  list of shapes in the container or the first shape selected by the region.
 *  The iterator is incremented as usual with the ++ operator. The end of the
 *  sequence is tested with the at_end method.
 */
template <class C, class Graph>
class shape_iterator
{
public:
  typedef C coord_type;
  typedef db::polygon<C> polygon_type;
  typedef db::simple_polygon<C> simple_polygon_type;
  typedef db::polygon_ref<polygon_type> polygon_ref_type;
  typedef db::polygon_ref<simple_polygon_type> simple_polygon_ref_type;
  typedef db::path<C> path_type;
  typedef db::path_ref<path_type> path_ref_type;
  typedef db::edge<C> edge_type;
  typedef db::text<C> text_type;
  typedef db::text_ref<text_type> text_ref_type;
  typedef db::box<C> box_type;
  typedef db::point<C> point_type;
  typedef db::user_object<C> user_object_type;
  typedef db::shape<C> shape_type;
  typedef db::shapes<C, Graph> shapes_type;
  typedef db::PropertiesRepository::properties_id_type properties_id_type;
  typedef std::set<properties_id_type> property_selector;

  enum region_mode
  {
    None,
    Overlapping,
    Touching
  };

  enum object_type 
  {
    Polygon = 0,
    PolygonRef = 1,
    SimplePolygon = 2,
    SimplePolygonRef = 3,
    Edge = 4,
    Path = 5,
    PathRef = 6,
    Box = 7,
    Text = 8,
    TextRef = 9,
    UserObject = 10,
    Null = 11 //  must be last!
  };

  enum flags_type 
  {
    //  These flags are a combination of all bits representing the
    //  object type (Bit 0 for Polygon and so on)
    Polygons          = 0xf,    
    Edges             = 0x10,
    Paths             = 0x60,
    Boxes             = 0x80,
    Texts             = 0x300,
    UserObjects       = 0x400,
    Properties        = 0x800,
    All               = 0x7ff,
    AllWithProperties = 0xfff 
  };
    
  /**
   *  @brief The default constructor of the iterator
   */
  shape_iterator ()
    : m_region_mode (None), m_type (Null), m_box (), m_flags (0), mp_shapes (0), mp_prop_sel (0), m_inv_prop_sel (false)
  {
    m_valid = false;
    m_with_props = false;
  }

  /**
   *  @brief The normal iterator
   *
   *  This constructor creates an iterator that iterates over all shapes of
   *  the selected kind. The kind of shape is selected by the "flags" argument
   *  which is an or-ed sum of flags_type constants.
   *
   *  @param shapes The shapes container to iterate over
   *  @param flags The kind of shapes to iterate over (or-ed constants of flags_type)
   */
  shape_iterator (const shapes_type &shapes, unsigned int flags = All, const property_selector *prop_sel = 0, bool inv_prop_sel = false)
    : m_region_mode (None), m_type (object_type (0)), m_box (), m_flags (flags), mp_shapes (&shapes), mp_prop_sel (prop_sel), m_inv_prop_sel (inv_prop_sel)
  {
    //  optimize: empty property selection plus inverse = no property selection at all
    //            any property selection and not inverse = only shapes with properties
    if (mp_prop_sel) {
      if (mp_prop_sel->empty () && m_inv_prop_sel) {
        mp_prop_sel = 0;
        m_inv_prop_sel = false;
      } else if (! m_inv_prop_sel) {
        m_flags |= Properties;
      }
    }
    m_valid = false;
    m_with_props = false;
    inc ();
  }

  /**
   *  @brief The region iterator
   *
   *  This constructor creates an iterator that iterates over all shapes of
   *  the selected kind that interact with the given region with their bounding box. 
   *  The kind of shape is selected by the "flags" argument which is an or-ed sum 
   *  of flags_type constants. The region_mode argument can be either Overlapping
   *  or Touching. In the former case all shapes whose bounding box overlaps with
   *  the given region are selected. In the latter case all shapes whose bounding
   *  box touches the given region are selected.
   *
   *  @param shapes The shapes container to iterate over
   *  @param box The region to select
   *  @param mode See above
   *  @param flags The kind of shapes to iterate over (or-ed constants of flags_type)
   */
  shape_iterator (const shapes_type &shapes, const box_type &box, region_mode mode, unsigned int flags = All, const property_selector *prop_sel = 0, bool inv_prop_sel = false)
    : m_region_mode (mode), m_type (object_type (0)), m_box (box), m_flags (flags), mp_shapes (&shapes), mp_prop_sel (prop_sel), m_inv_prop_sel (inv_prop_sel)
  {
    //  optimize: empty property selection plus inverse = no property selection at all
    //            any property selection and not inverse = only shapes with properties
    if (mp_prop_sel) {
      if (mp_prop_sel->empty () && m_inv_prop_sel) {
        mp_prop_sel = 0;
        m_inv_prop_sel = false;
      } else if (! m_inv_prop_sel) {
        m_flags |= Properties;
      }
    }
    m_valid = false;
    m_with_props = false;
    inc ();
  }

  /**
   *  @brief The destructor
   */
  ~shape_iterator ()
  {
    //  this trick destroys all iterators that have been allocated in the generic union
    if (m_type != Null) {
      m_flags = 0;
      inc ();
      tl_assert (m_type == Null);
    }
  }

  /** 
   *  @brief Access the object 
   * 
   *  This method delivers a proxy object that can be copied but still points
   *  to the original objects. This must be taken under consideration if the
   *  container that is iterated over is destroyed: the shape object will then
   *  become invalid instantly.
   */
  const shape_type &operator* () const
  {
    return m_shape;
  }

  /** 
   *  @brief Access the object (see operator*)
   */
  const shape_type *operator-> () const
  {
    return &m_shape;
  }

  /**
   *  @brief Increment the iterator
   */
  shape_iterator &operator++ () 
  {
    inc ();
    return *this;
  }

  /**
   *  @brief Test if the iterator is at the end
   *
   *  @return true if there are no more elements
   */
  bool at_end () const
  {
    return m_type == Null;
  }

private:
  //  this union is simply there to determine the maximum size required for all
  //  possible iterators.
  union iter_size {

    char sz1_n [sizeof (typename db::layer<polygon_type>::iterator)];
    char sz2_n [sizeof (typename db::layer<polygon_ref_type>::iterator)];
    char sz3_n [sizeof (typename db::layer<simple_polygon_type>::iterator)];
    char sz4_n [sizeof (typename db::layer<simple_polygon_ref_type>::iterator)];
    char sz5_n [sizeof (typename db::layer<path_type>::iterator)];
    char sz6_n [sizeof (typename db::layer<path_ref_type>::iterator)];
    char sz7_n [sizeof (typename db::layer<edge_type>::iterator)];
    char sz8_n [sizeof (typename db::layer<box_type>::iterator)];
    char sz9_n [sizeof (typename db::layer<text_type>::iterator)];
    char sz10_n [sizeof (typename db::layer<text_ref_type>::iterator)];
    char sz11_n [sizeof (typename db::layer<user_object_type>::iterator)];

    char sz1_np [sizeof (typename db::layer< db::shape_with_properties<polygon_type> >::iterator)];
    char sz2_np [sizeof (typename db::layer< db::shape_with_properties<polygon_ref_type> >::iterator)];
    char sz3_np [sizeof (typename db::layer< db::shape_with_properties<simple_polygon_type> >::iterator)];
    char sz4_np [sizeof (typename db::layer< db::shape_with_properties<simple_polygon_ref_type> >::iterator)];
    char sz5_np [sizeof (typename db::layer< db::shape_with_properties<path_type> >::iterator)];
    char sz6_np [sizeof (typename db::layer< db::shape_with_properties<path_ref_type> >::iterator)];
    char sz7_np [sizeof (typename db::layer< db::shape_with_properties<edge_type> >::iterator)];
    char sz8_np [sizeof (typename db::layer< db::shape_with_properties<box_type> >::iterator)];
    char sz9_np [sizeof (typename db::layer< db::shape_with_properties<text_type> >::iterator)];
    char sz10_np [sizeof (typename db::layer< db::shape_with_properties<text_ref_type> >::iterator)];
    char sz11_np [sizeof (typename db::layer< db::shape_with_properties<user_object_type> >::iterator)];

    char sz1_t [sizeof (typename db::layer<polygon_type>::touching_iterator)];
    char sz2_t [sizeof (typename db::layer<polygon_ref_type>::touching_iterator)];
    char sz3_t [sizeof (typename db::layer<simple_polygon_type>::touching_iterator)];
    char sz4_t [sizeof (typename db::layer<simple_polygon_ref_type>::touching_iterator)];
    char sz5_t [sizeof (typename db::layer<path_type>::touching_iterator)];
    char sz6_t [sizeof (typename db::layer<path_ref_type>::touching_iterator)];
    char sz7_t [sizeof (typename db::layer<edge_type>::touching_iterator)];
    char sz8_t [sizeof (typename db::layer<box_type>::touching_iterator)];
    char sz9_t [sizeof (typename db::layer<text_type>::touching_iterator)];
    char sz10_t [sizeof (typename db::layer<text_ref_type>::touching_iterator)];
    char sz11_t [sizeof (typename db::layer<user_object_type>::touching_iterator)];

    char sz1_tp [sizeof (typename db::layer< db::shape_with_properties<polygon_type> >::touching_iterator)];
    char sz2_tp [sizeof (typename db::layer< db::shape_with_properties<polygon_ref_type> >::touching_iterator)];
    char sz3_tp [sizeof (typename db::layer< db::shape_with_properties<simple_polygon_type> >::touching_iterator)];
    char sz4_tp [sizeof (typename db::layer< db::shape_with_properties<simple_polygon_ref_type> >::touching_iterator)];
    char sz5_tp [sizeof (typename db::layer< db::shape_with_properties<path_type> >::touching_iterator)];
    char sz6_tp [sizeof (typename db::layer< db::shape_with_properties<path_ref_type> >::touching_iterator)];
    char sz7_tp [sizeof (typename db::layer< db::shape_with_properties<edge_type> >::touching_iterator)];
    char sz8_tp [sizeof (typename db::layer< db::shape_with_properties<box_type> >::touching_iterator)];
    char sz9_tp [sizeof (typename db::layer< db::shape_with_properties<text_type> >::touching_iterator)];
    char sz10_tp [sizeof (typename db::layer< db::shape_with_properties<text_ref_type> >::touching_iterator)];
    char sz11_tp [sizeof (typename db::layer< db::shape_with_properties<user_object_type> >::touching_iterator)];

    char sz1_o [sizeof (typename db::layer<polygon_type>::overlapping_iterator)];
    char sz2_o [sizeof (typename db::layer<polygon_ref_type>::overlapping_iterator)];
    char sz3_o [sizeof (typename db::layer<simple_polygon_type>::overlapping_iterator)];
    char sz4_o [sizeof (typename db::layer<simple_polygon_ref_type>::overlapping_iterator)];
    char sz5_o [sizeof (typename db::layer<path_type>::overlapping_iterator)];
    char sz6_o [sizeof (typename db::layer<path_ref_type>::overlapping_iterator)];
    char sz7_o [sizeof (typename db::layer<edge_type>::overlapping_iterator)];
    char sz8_o [sizeof (typename db::layer<box_type>::overlapping_iterator)];
    char sz9_o [sizeof (typename db::layer<text_type>::overlapping_iterator)];
    char sz10_o [sizeof (typename db::layer<text_ref_type>::overlapping_iterator)];
    char sz11_o [sizeof (typename db::layer<user_object_type>::overlapping_iterator)];
   
    char sz1_op [sizeof (typename db::layer< db::shape_with_properties<polygon_type> >::overlapping_iterator)];
    char sz2_op [sizeof (typename db::layer< db::shape_with_properties<polygon_ref_type> >::overlapping_iterator)];
    char sz3_op [sizeof (typename db::layer< db::shape_with_properties<simple_polygon_type> >::overlapping_iterator)];
    char sz4_op [sizeof (typename db::layer< db::shape_with_properties<simple_polygon_ref_type> >::overlapping_iterator)];
    char sz5_op [sizeof (typename db::layer< db::shape_with_properties<path_type> >::overlapping_iterator)];
    char sz6_op [sizeof (typename db::layer< db::shape_with_properties<path_ref_type> >::overlapping_iterator)];
    char sz7_op [sizeof (typename db::layer< db::shape_with_properties<edge_type> >::overlapping_iterator)];
    char sz8_op [sizeof (typename db::layer< db::shape_with_properties<box_type> >::overlapping_iterator)];
    char sz9_op [sizeof (typename db::layer< db::shape_with_properties<text_type> >::overlapping_iterator)];
    char sz10_op [sizeof (typename db::layer< db::shape_with_properties<text_ref_type> >::overlapping_iterator)];
    char sz11_op [sizeof (typename db::layer< db::shape_with_properties<user_object_type> >::overlapping_iterator)];
  };

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (iter_size)];
  } m_d;
  
  bool m_valid, m_with_props;
  region_mode m_region_mode;
  object_type m_type;
  box_type m_box;
  shape_type m_shape;
  unsigned int m_flags;
  const shapes_type *mp_shapes;
  const property_selector *mp_prop_sel;
  bool m_inv_prop_sel;

  template <class Sh>
  bool advance ()
  {
    bool sel = (m_flags & (1 << (unsigned int) m_type)) != 0;
    bool props_only = (m_flags & Properties) != 0;

    if (! m_with_props && ! props_only) {
    
      typename db::layer<Sh>::iterator *iter = (typename db::layer<Sh>::iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer<Sh>::iterator i (mp_shapes->begin (typename Sh::tag ()));
        if (i != mp_shapes->end (typename Sh::tag ())) {
          new (iter) typename db::layer<Sh>::iterator (i);
          m_valid = true;
        }
      } 

      if (m_valid) {
        if (!sel || *iter == mp_shapes->end (typename Sh::tag ())) {
          m_valid = false;
        } else {
          m_shape = **iter;
          ++*iter;
          return true;
        }
      }

      m_with_props = true;

    }

    if (m_with_props || props_only) {
    
      typename db::layer< db::shape_with_properties<Sh> >::iterator *iter = (typename db::layer< db::shape_with_properties<Sh> >::iterator *) m_d.iter;
      typename db::layer< db::shape_with_properties<Sh> >::iterator iend (mp_shapes->end (typename db::shape_with_properties <Sh>::tag ()));

      if (!m_valid && sel) {
        typename db::layer< db::shape_with_properties<Sh> >::iterator i (mp_shapes->begin (typename db::shape_with_properties <Sh>::tag ()));
        if (mp_prop_sel) {
          while (i != iend && (mp_prop_sel->find (i->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel) {
            ++i;
          }
        }
        if (i != iend) {
          new (iter) typename db::layer< db::shape_with_properties<Sh> >::iterator (i);
          m_valid = true;
        }
      } 

      if (m_valid) {
        if (!sel || *iter == iend) {
          m_valid = false;
        } else {
          m_shape = **iter;
          do {
            ++*iter;
          } while (mp_prop_sel && *iter != iend && (mp_prop_sel->find ((*iter)->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel);
          return true;
        }
      }

    }

    m_with_props = false;
    return false;
  }

  template <class Sh>
  bool advance_touching ()
  {
    bool sel = (m_flags & (1 << (unsigned int) m_type)) != 0;
    bool props_only = (m_flags & Properties) != 0;

    if (! m_with_props && ! props_only) {

      typename db::layer<Sh>::touching_iterator *iter = (typename db::layer<Sh>::touching_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer<Sh>::touching_iterator i (mp_shapes->begin_touching (m_box, typename Sh::tag ()));
        if (!i.at_end ()) {
          new (iter) typename db::layer<Sh>::touching_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
        } else {
          m_shape = **iter;
          ++*iter;
          return true;
        }
      }

      m_with_props = true;

    }

    if (m_with_props || props_only) {

      typename db::layer< db::shape_with_properties<Sh> >::touching_iterator *iter = (typename db::layer< db::shape_with_properties<Sh> >::touching_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer< db::shape_with_properties<Sh> >::touching_iterator i (mp_shapes->begin_touching (m_box, typename  db::shape_with_properties<Sh> ::tag ()));
        if (mp_prop_sel) {
          while (!i.at_end () && (mp_prop_sel->find (i->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel) {
            ++i;
          }
        }
        if (!i.at_end ()) {
          new (iter) typename db::layer< db::shape_with_properties<Sh> >::touching_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
          m_with_props = false;
        } else {
          m_shape = **iter;
          do {
            ++*iter;
          } while (mp_prop_sel && !iter->at_end () && (mp_prop_sel->find ((*iter)->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel);
          return true;
        }
      }

    }

    m_with_props = false;
    return false;
  }

  template <class Sh>
  bool advance_overlapping ()
  {
    bool sel = (m_flags & (1 << (unsigned int) m_type)) != 0;
    bool props_only = (m_flags & Properties) != 0;

    if (! m_with_props && ! props_only) {

      typename db::layer<Sh>::overlapping_iterator *iter = (typename db::layer<Sh>::overlapping_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer<Sh>::overlapping_iterator i (mp_shapes->begin_overlapping (m_box, typename Sh::tag ()));
        if (!i.at_end ()) {
          new (iter) typename db::layer<Sh>::overlapping_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
        } else {
          m_shape = **iter;
          ++*iter;
          return true;
        }
      }

      m_with_props = true;

    }

    if (m_with_props || props_only) {

      typename db::layer< db::shape_with_properties<Sh> >::overlapping_iterator *iter = (typename db::layer< db::shape_with_properties<Sh> >::overlapping_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer< db::shape_with_properties<Sh> >::overlapping_iterator i (mp_shapes->begin_overlapping (m_box, typename  db::shape_with_properties<Sh> ::tag ()));
        if (mp_prop_sel) {
          while (!i.at_end () && (mp_prop_sel->find (i->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel) {
            ++i;
          }
        }
        if (!i.at_end ()) {
          new (iter) typename db::layer< db::shape_with_properties<Sh> >::overlapping_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
        } else {
          m_shape = **iter;
          do {
            ++*iter;
          } while (mp_prop_sel && !iter->at_end () && (mp_prop_sel->find ((*iter)->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel);
          return true;
        }
      }

    }

    m_with_props = false;
    return false;
  }

  void inc ()
  {
    if (m_region_mode == None) {
      
      switch (m_type) {
      //  The order of the switch statements must be identical to the order of
      //  enum constants
      case Polygon:
        if (advance<polygon_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case PolygonRef:
        if (advance<polygon_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case SimplePolygon:
        if (advance<simple_polygon_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case SimplePolygonRef:
        if (advance<simple_polygon_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Edge:
        if (advance<edge_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Path:
        if (advance<path_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case PathRef:
        if (advance<path_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Box:
        if (advance<box_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Text:
        if (advance<text_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case TextRef:
        if (advance<text_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case UserObject:
        if (advance<user_object_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Null:
        break;
      }

    } else if (m_region_mode == Touching) {
      
      switch (m_type) {
      //  The order of the switch statements must be identical to the order of
      //  enum constants
      case Polygon:
        if (advance_touching<polygon_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case PolygonRef:
        if (advance_touching<polygon_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case SimplePolygon:
        if (advance_touching<simple_polygon_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case SimplePolygonRef:
        if (advance_touching<simple_polygon_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Edge:
        if (advance_touching<edge_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Path:
        if (advance_touching<path_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case PathRef:
        if (advance_touching<path_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Box:
        if (advance_touching<box_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Text:
        if (advance_touching<text_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case TextRef:
        if (advance_touching<text_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case UserObject:
        if (advance_touching<user_object_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Null:
        break;
      }

    } else if (m_region_mode == Overlapping) {
      
      switch (m_type) {
      //  The order of the switch statements must be identical to the order of
      //  enum constants
      case Polygon:
        if (advance_overlapping<polygon_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case PolygonRef:
        if (advance_overlapping<polygon_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case SimplePolygon:
        if (advance_overlapping<simple_polygon_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case SimplePolygonRef:
        if (advance_overlapping<simple_polygon_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Edge:
        if (advance_overlapping<edge_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Path:
        if (advance_overlapping<path_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case PathRef:
        if (advance_overlapping<path_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Box:
        if (advance_overlapping<box_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Text:
        if (advance_overlapping<text_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case TextRef:
        if (advance_overlapping<text_ref_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case UserObject:
        if (advance_overlapping<user_object_type> ()) return;
        m_type = object_type ((unsigned int) m_type + 1);
      case Null:
        break;
      }

    }

  }

};

/** 
 *  @brief A typedef for a standard shape iterator
 */
typedef db::shape_iterator<db::Coord, db::CellGraph> ShapeIterator;

}  // namespace db
  
#endif

