
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbShapeWithProperties
#define HDR_dbShapeWithProperties

#include <tlException.h>
#include <dbPolygon.h>
#include <dbPath.h>
#include <dbEdge.h>
#include <dbText.h>
#include <dbBox.h>
#include <dbPropertiesRepository.h>

namespace db
{

/**
 *  @brief A shape with properties template
 *
 *  This template provides some kind of "enhanced shape". A shape can be supplied
 *  with additional properties. For performance reasons, the properties are stored
 *  as an index within a lookup table. Each index refers to a set of properties.
 *  The "PropertiesRepository" class manages the properties available and associates
 *  a property set with an index. 
 *  The shape-with-properties template adds the properties repository index to a 
 *  shape and inherits all of the shape's methods.
 */

template <class Sh>
class shape_with_properties
  : public Sh
{
public:
  typedef Sh shape_type;
  typedef PropertiesRepository::properties_id_type properties_id_type;

  typedef typename Sh::box_type box_type;
  typedef typename Sh::coord_type coord_type;
  typedef typename Sh::point_type point_type;

  typedef db::shape_tag< shape_with_properties<Sh> > tag;

  /**
   *  @brief The default constructor
   */
  shape_with_properties ()
    : Sh (), m_id (0)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Create myself from a shape and an id
   */
  shape_with_properties (const Sh &sh, properties_id_type id)
    : Sh (sh), m_id (id)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief The copy constructor
   */
  shape_with_properties (const shape_with_properties<Sh> &d)
    : Sh (d), m_id (d.m_id)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Assignment
   */
  shape_with_properties &operator= (const shape_with_properties<Sh> &d)
  {
    if (this != &d) {
      Sh::operator= (d);
      m_id = d.m_id;
    }
    return *this;
  }

  /**
   *  @brief Equality
   */
  bool operator== (const shape_with_properties<Sh> &d) const
  {
    return Sh::operator== (d) && m_id == d.m_id;
  }

  /**
   *  @brief Inequality
   */
  bool operator!= (const shape_with_properties<Sh> &d) const
  {
    return ! operator== (d);
  }

  /**
   *  @brief Comparison
   */
  bool operator< (const shape_with_properties<Sh> &d) const
  {
    if (! Sh::operator== (d)) {
      return Sh::operator< (d);
    } 
    return m_id < d.m_id;
  }

  /**
   *  @brief Properties Id read accessor
   */
  properties_id_type properties_id () const
  {
    return m_id;
  }

  /**
   *  @brief Properties Id write accessor
   */
  void properties_id (properties_id_type id) 
  {
    m_id = id;
  }

private:
  properties_id_type m_id;
};

typedef shape_with_properties<Polygon> PolygonWithProperties;
typedef shape_with_properties<DPolygon> DPolygonWithProperties;
typedef shape_with_properties<SimplePolygon> SimplePolygonWithProperties;
typedef shape_with_properties<DSimplePolygon> DSimplePolygonWithProperties;
typedef shape_with_properties<PolygonRef> PolygonRefWithProperties;
typedef shape_with_properties<DPolygonRef> DPolygonRefWithProperties;
typedef shape_with_properties<SimplePolygonRef> SimplePolygonRefWithProperties;
typedef shape_with_properties<DSimplePolygonRef> DSimplePolygonRefWithProperties;

typedef shape_with_properties<Path> PathWithProperties;
typedef shape_with_properties<DPath> DPathWithProperties;
typedef shape_with_properties<PathRef> PathRefWithProperties;
typedef shape_with_properties<DPathRef> DPathRefWithProperties;

typedef shape_with_properties<Edge> EdgeWithProperties;
typedef shape_with_properties<DEdge> DEdgeWithProperties;

typedef shape_with_properties<Text> TextWithProperties;
typedef shape_with_properties<DText> DTextWithProperties;
typedef shape_with_properties<TextRef> TextRefWithProperties;
typedef shape_with_properties<DTextRef> DTextRefWithProperties;

typedef shape_with_properties<Box> BoxWithProperties;
typedef shape_with_properties<DBox> DBoxWithProperties;

} // namespace db

#endif

