
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/




#ifndef HDR_dbShapes
#define HDR_dbShapes

#include <dbObject.h>
#include <dbManager.h>
#include <dbLayer.h>
#include <tlVector.h>

namespace db 
{

class CellGraph;

template <class Shapes, class Sh> class layer_op;
template <class Box, class Graph> class shapes;
template <class Coord, class Graph> class shape_iterator;
template <class Coord> class generic_polygon_edge_iterator;
template <class Coord> class generic_repository;

/**
 *  @brief A base class for layer operations 
 *
 *  This class is used for the Op classes for the undo/redo queuing mechanism.
 */
template <class Shapes>
class layer_op_base
  : public db::Op
{
public:
  layer_op_base () : db::Op () { }

  virtual void undo (Shapes *shapes) = 0;
  virtual void redo (Shapes *shapes) = 0;
};

/**
 *  @brief A helper class for shape generalisation
 *
 *  This class serves first as a RTTI token for the 
 *  various shape-specific layer implementations and 
 *  provides some common methods though it's interface
 */

template <class Shapes>
class layer_base 
{
public:
  typedef typename Shapes::box_type box_type;
  typedef typename Shapes::coord_type coord_type;

  virtual ~layer_base () { }

  virtual box_type bbox () const = 0;
  virtual void update_bbox () = 0;
  virtual bool is_bbox_dirty () const = 0;
  virtual void sort () = 0;
  virtual void clear (Shapes *target, db::Manager *manager) = 0;
  virtual layer_base<Shapes> *clone (Shapes *target, db::Manager *manager) const = 0;
  virtual layer_base<Shapes> *translated (Shapes *target, generic_repository<coord_type> &rep) const = 0;

  virtual void collect_mem_stat (db::MemStatistics &m) const = 0;
};

/**
 *  @brief Actual implementation of the layer_base class
 *  
 *  For a certain shape type this template implements the 
 *  shape specific layer methods.
 */

template <class Shapes, class Sh>
class layer_class 
  : public layer_base<Shapes>
{
public:
  typedef db::layer<Sh> layer_type;
  typedef typename Shapes::box_type box_type;
  typedef typename Shapes::coord_type coord_type;

  layer_class ()
    : layer_base<Shapes> (), m_layer () 
  { }

  layer_type &layer () 
  {
    return m_layer;
  }

  const layer_type &layer () const
  {
    return m_layer;
  }

  virtual box_type bbox () const 
  {
    return m_layer.bbox ();
  }

  virtual void update_bbox () 
  {
    m_layer.update_bbox ();
  }

  virtual bool is_bbox_dirty () const
  {
    return m_layer.is_bbox_dirty ();
  }

  virtual void sort () 
  {
    m_layer.sort ();
  }

  virtual void clear (Shapes *target, db::Manager *manager)
  {
    if (manager->transacting ()) {
      manager->queue (target, new db::layer_op<Shapes, Sh> (false /*not insert*/, m_layer.begin (), m_layer.end ()));
    }
    m_layer.clear ();
  }

  virtual layer_base<Shapes> *clone (Shapes *target, db::Manager *manager) const 
  {
    layer_class<Shapes, Sh> *r = new layer_class<Shapes, Sh> ();
    if (manager->transacting ()) {
      manager->queue (target, new db::layer_op<Shapes, Sh> (true /*insert*/, m_layer.begin (), m_layer.end ()));
    }
    r->m_layer = m_layer;
    return r;
  }
  
  virtual layer_base<Shapes> *translated (Shapes *target, generic_repository<coord_type> &rep) const 
  {
    layer_class<Shapes, Sh> *r = new layer_class<Shapes, Sh> ();
    r->m_layer.translate (m_layer, rep);
    return r;
  }
  
  virtual void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.shapes_info ((const layer_base<Shapes> &)(*this)); // base class contribution
    m_layer.collect_mem_stat (m);
  }

private:
  layer_type m_layer;
};

/**
 *  @brief A "shapes" collection
 *  
 *  A shapes collection is a collection of geometrical objects.
 *  The implementation is based on a set of layers of different
 *  shape types. 
 *  The general idea is that is rarely required to operate with
 *  different shape types at once. Therefore it is practical to
 *  request only a single shape type at once. This is done through
 *  the various begin.. and end.. methods that are specialized on
 *  a certain shape type.
 */

template <class Coord, class Graph>
class shapes 
  : public db::Object
{
public:
  typedef Coord coord_type;
  typedef Graph graph_type;
  typedef db::box<Coord> box_type;
  typedef db::shape_iterator<Coord, Graph> shape_iterator;
  typedef db::generic_polygon_edge_iterator<Coord> polygon_edge_iterator;

  /**
   *  @brief Default ctor: create an empty collection not referencing a graph
   */
  shapes (db::Manager *manager = 0)
    : db::Object (manager), mp_graph (0)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Standard ctor: create an empty collection referencing a graph
   *
   *  The graph reference is used to invalid the bbox flag of the graph
   *  whenever something changes on the shapes list.
   */
  shapes (graph_type &g) 
    : db::Object (g.manager ()), mp_graph (&g)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Dtor: clear all ..
   */
  ~shapes () 
  {
    clear ();
    mp_graph = 0;
  }

  /**
   *  @brief Copy ctor
   */
  shapes (const shapes &d) 
    : db::Object (d), mp_graph (d.mp_graph)
  {
    operator= (d);
  }

  /**
   *  @brief Assignment operator
   */
  shapes &
  operator= (const shapes &d)
  {
    if (&d != this) {

      clear ();

      m_layers.reserve (d.m_layers.size ());
      for (typename tl::vector<layer_base<shapes> *>::const_iterator l = d.m_layers.begin (); l != d.m_layers.end (); ++l) {
        m_layers.push_back ((*l)->clone (this, manager ()));
      }

    }
    return *this;
  }

  /**
   *  @brief The translation operator
   *
   *  This operator is used to copy one shapes layer to another repository space.
   *  The current shapes layer will be overwritten. 
   *
   *  @param src The source shapes collection
   *  @param rep The repository that is associated with *this and into which the
   *             shapes will be copied
   */
  void translate (const shapes<Coord, Graph> &d, generic_repository<Coord> &rep)
  {
    tl_assert (&d != this);
    //  no undo support for this currently
    tl_assert (! manager () || ! manager ()->transacting ());

    clear ();

    m_layers.reserve (d.m_layers.size ());
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = d.m_layers.begin (); l != d.m_layers.end (); ++l) {
      m_layers.push_back ((*l)->translated (this, rep));
    }
  }

  /**
   *  @brief Swap the contents of this shapes collection with another one
   */
  void swap (shapes &d)
  {
    tl_assert (! manager ()->transacting ()); // TODO: implement undo/redo support?
    m_layers.swap (d.m_layers);
    d.invalidate_state ();
    invalidate_state ();
  }

  /**
   *  @brief Insert a shape of the given type
   *
   *  This inserts a shape of the given kind into the collection.
   *  Basically a new layer is created if no shape of this kind
   *  already existed.
   *  Inserting a shape will invalidate the bbox and the sorting
   *  state.
   *
   *  @param sh The shape to insert (copy)
   *  
   *  @return A reference to the object created
   */
  template <class Sh>
  const Sh &insert (const Sh &sh)
  {
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, Sh> (true /*insert*/, sh));
    }
    invalidate_state ();
    return layer<Sh> ().insert (sh);
  }

  /**
   *  @brief Insert a shape sequence
   *
   *  Inserts a sequence of shapes [from,to)
   */

  template <class Iter>
  void insert (Iter from, Iter to)
  {
    typedef typename std::iterator_traits <Iter>::value_type value_type;
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, value_type> (true /*insert*/, from, to));
    }
    layer<value_type> ().insert (from, to);
    invalidate_state ();
  }

  /**
   *  @brief Reserve the number of elements for a shape type
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param n The number of elements to reserve
   */
  template <class Tag> 
  void reserve (Tag tag, size_t n)
  {
    layer<typename Tag::shape_type> ().reserve (n);
  }

  /**
   *  @brief Erase an element 
   *
   *  Erases a shape at the given position
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param pos The position of the shape to erase
   */
  template <class Tag> 
  void erase (Tag tag, typename db::layer<typename Tag::shape_type>::iterator pos)
  {
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, typename Tag::shape_type> (false /*not insert*/, *pos));
    }
    layer<typename Tag::shape_type> ().erase (pos);
    invalidate_state ();
  }

  /**
   *  @brief Erase an element 
   *
   *  Erases a shapes at the given positions [from,to)
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param from The first position of the shapes to erase
   *  @param to The past-end position of the shapes to erase
   */
  template <class Tag> 
  void erase (Tag tag, typename db::layer<typename Tag::shape_type>::iterator from,
                       typename db::layer<typename Tag::shape_type>::iterator to)
  {
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, typename Tag::shape_type> (false /*not insert*/, from, to));
    }
    layer<typename Tag::shape_type> ().erase (from, to);
    invalidate_state ();
  }

  /**
   *  @brief Erasing of multiple elements
   *
   *  Erase a set of positions given by an iterator I: *(from,to).
   *  *I must render an "iterator" object.
   *  The iterators in the sequence from, to must be sorted in
   *  "later" order.
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  @param first The start of the sequence of iterators
   *  @param last The end of the sequence of iterators
   */
  template <class Tag, class I>
  void erase_positions (Tag tag, I first, I last)
  {
    if (manager ()->transacting ()) {
      for (I p = first; p != last; ++p) {
        manager ()->queue (this, new db::layer_op<shapes, typename Tag::shape_type> (false /*not insert*/, **p));
      }
    }
    layer<typename Tag::shape_type> ().erase_positions (first, last);
    invalidate_state ();
  }

  /**
   *  @brief Replace an element at the given position 
   *
   *  Replaces the element at the position pos with the
   *  new element.
   *  Replacing a shape will invalidate the bbox and the sorting
   *  state.
   *
   *  @param pos The position at which to replace the shape
   *  @param sh The shape to replace 
   *  
   *  @return A reference to the object created
   */
  template <class Sh>
  const Sh &replace (typename db::layer<Sh>::iterator pos, const Sh &sh)
  {
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::layer_op<shapes, Sh> (false /*not insert*/, *pos));
      manager ()->queue (this, new db::layer_op<shapes, Sh> (true /*insert*/, sh));
    }
    invalidate_state ();
    return layer<Sh> ().replace (pos, sh);
  }

  /**
   *  @brief updates the bbox 
   *
   *  Updating the bbox is required after insert operations
   *  and is performed only as far as necessary.
   */
  void update_bbox ()
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->update_bbox ();
    }
  }

  /**
   *  @brief check if the bounding box needs update
   *
   *  Returns true if the bounding box of the shapes has changed and 
   *  requires an update.
   */
  bool is_bbox_dirty () const
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      if ((*l)->is_bbox_dirty ()) {
        return true;
      }
    }
    return false;
  }

  /**
   *  @brief Retrieve the bbox 
   *
   *  Retrieving the bbox might required an update_bbox
   *  before the bbox is valid. It will assert if the bbox
   *  was not valid.
   *  Doing a update_bbox before does not imply a performance penalty
   *  since the state is cached.
   */
  box_type bbox () const
  {
    box_type box;
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      box += (*l)->bbox ();
    }
    return box;
  }

  /**
   *  @brief sorts the trees
   *
   *  Sorting the trees is required after insert operations
   *  and is performed only as far as necessary.
   */
  void sort () 
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->sort ();
    }
  }

  /**
   *  @brief Clears the collection
   */
  void clear ()
  {
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->clear (this, manager ());
      delete *l;
    }
    m_layers.clear ();
    invalidate_state ();
  }

  /**
   *  @brief Do a region search for a certain shape type in "touching" mode
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The region iterator 
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::touching_iterator begin_touching (const box_type &b, Tag type) const 
  {
    return layer<typename Tag::shape_type> ().begin_touching (b);
  }

  /**
   *  @brief The region iterator end token
   */
  template <class Tag> 
  size_t end_touching (Tag type) const 
  {
    return 0; // KLUDGE: some shortcut but basically an implementation detail ..
  }

  /**
   *  @brief Do a region search for a certain shape type in "overlapping" mode
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The region iterator 
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::overlapping_iterator begin_overlapping (const box_type &b, Tag type) const 
  {
    return layer<typename Tag::shape_type> ().begin_overlapping (b);
  }

  /**
   *  @brief The region iterator end token
   */
  template <class Tag> 
  size_t end_overlapping (Tag type) const 
  {
    return 0; // KLUDGE: some shortcut but basically an implementation detail ..
  }

  /**
   *  @brief begin iterator of all elements of a certain shape type
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The first position of the shapes of the requested type
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::iterator begin (Tag type) const
  {
    return layer<typename Tag::shape_type> ().begin ();
  }

  /**
   *  @brief end iterator of all elements of a certain shape type
   *
   *  @param tag The shape type's tag (i.e. db::Polygon::tag)
   *  
   *  @return The post-end position of the shapes of the requested type
   */
  template <class Tag> 
  typename db::layer<typename Tag::shape_type>::iterator end (Tag type) const
  {
    return layer<typename Tag::shape_type> ().end ();
  }

  /**
   *  @brief begin iterator of all elements 
   *
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The generic iterator
   */
  shape_iterator begin (unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    return shape_iterator (*this, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief begin iterator of all elements with a touching mode region query
   *
   *  @param box The region to query the shapes from
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The generic iterator
   */
  shape_iterator begin_touching (const box_type &box, unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    return shape_iterator (*this, box, shape_iterator::Touching, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief begin iterator of all elements with a overlapping mode region query
   *
   *  @param box The region to query the shapes from
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The generic iterator
   */
  shape_iterator begin_overlapping (const box_type &box, unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    return shape_iterator (*this, box, shape_iterator::Overlapping, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief find a given shape (exactly)
   *
   *  @param s The shape to find
   *  
   *  @return end(Sh::tag) if the shape was not found, the position otherwise
   */
  template <class Sh> 
  typename db::layer<Sh>::iterator find (const Sh &s) const
  {
    return layer<Sh> ().find (s);
  }

  /** 
   *  @brief Implementation of the redo method
   */
  void redo (db::Op *op) throw ()
  {
    db::layer_op_base<shapes> *layop = dynamic_cast< db::layer_op_base<shapes> *> (op);
    if (layop) {
      layop->redo (this);
    } 
  }

  /** 
   *  @brief Implementation of the undo method
   */
  void undo (db::Op *op) throw ()
  {
    db::layer_op_base<shapes> *layop = dynamic_cast< db::layer_op_base<shapes> *> (op);
    if (layop) {
      layop->undo (this);
    } 
  }

  void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.cell_info (m_layers);
    m.cell_info (mp_graph);

    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      (*l)->collect_mem_stat (m);
    }
  }

private:
  template <class Sh>
  const db::layer<Sh> &layer () const
  {
    typedef layer_class<shapes, Sh> lay_cls;
    
    for (typename tl::vector<layer_base<shapes> *>::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      const lay_cls *lc = dynamic_cast <const lay_cls *> (*l);
      if (lc) {
        return lc->layer ();
      }
    }

    //  there seems to be a bug in gcc that disables
    //  the use of a simple static object here:
    static const db::layer<Sh> *empty_layer = 0;
    if (! empty_layer) {
      empty_layer = new db::layer<Sh> ();
    }
    return *empty_layer;
  }

  template <class Sh>
  db::layer<Sh> &layer () 
  {
    typedef layer_class<shapes, Sh> lay_cls;
    lay_cls *lc;
    
    for (typename tl::vector<layer_base<shapes> *>::iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
      lc = dynamic_cast <lay_cls *> (*l);
      if (lc) {
        //  this is what optimizes access times for another access
        //  with this type 
        std::swap (m_layers.front (), *l);
        return lc->layer ();
      }
    }

    //  create 
    lc = new lay_cls ();
    m_layers.push_back (lc);
    std::swap (m_layers.front (), m_layers.back ());
    return lc->layer ();
  }

  void invalidate_state ()
  {
    if (mp_graph) {
      mp_graph->invalidate_bboxes ();
    }
  }

  tl::vector<layer_base<shapes> *> m_layers;
  graph_type *mp_graph;
};

/**
 *  @brief A undo/redo queue object for the layer
 *
 *  This class is used internally to queue an insert or erase operation
 *  into the db::Object manager's undo/redo queue.
 */
template <class Shapes, class Sh>
class layer_op
  : public layer_op_base<Shapes> 
{
public:
  layer_op (bool insert, const Sh &sh)
    : m_insert (insert)
  {
    m_shapes.push_back (sh);
  }
  
  template <class Iter>
  layer_op (bool insert, Iter from, Iter to)
    : m_insert (insert)
  {
    m_shapes.insert (m_shapes.end (), from, to);
  }

  virtual void undo (Shapes *shapes)
  {
    if (m_insert) {
      erase (shapes);
    } else {
      insert (shapes);
    }
  }

  virtual void redo (Shapes *shapes)
  {
    if (m_insert) {
      insert (shapes);
    } else {
      erase (shapes);
    }
  }

private:
  bool m_insert;
  std::vector<Sh> m_shapes;

  void insert (Shapes *shapes)
  {
    shapes->insert (m_shapes.begin (), m_shapes.end ());
  }

  void erase (Shapes *shapes)
  {
    if (size_t (std::distance (shapes->begin (typename Sh::tag ()), shapes->end (typename Sh::tag ()))) <= m_shapes.size ()) {
      //  If all shapes are to be removed, just clear the shapes
      shapes->erase (typename Sh::tag (), shapes->begin (typename Sh::tag ()), shapes->end (typename Sh::tag ()));
    } else {
      //  This is not quite effective but seems to be the simpliest way
      //  of implementing this: search for each element and erase exactly
      //  the one:
      for (typename std::vector<Sh>::const_iterator sh = m_shapes.begin (); sh != m_shapes.end (); ++sh) {
        for (typename db::layer<Sh>::iterator lsh = shapes->begin (typename Sh::tag ()); lsh != shapes->end (typename Sh::tag ()); ++lsh) {
          if (*lsh == *sh) {
            shapes->erase (typename Sh::tag (), lsh);
            break; 
          }
        }
      }
    }
  }
};

/** 
 *  @brief A typedef for a standard shapes container 
 */
typedef db::shapes<db::Coord, db::CellGraph> Shapes;

}  // namespace db
  
#endif

