
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include <errno.h> 
#include <string.h> 

#include "dbStream.h"

#include "tlException.h"
#include "tlString.h"

namespace db
{

// ---------------------------------------------------------------
//  Stream implementation

Stream::Stream ()
  : m_pos (0), mp_bptr (0)
{ 
  m_bcap = 4096; // initial buffer capacity
  m_blen = 0;
  mp_buffer = new char [m_bcap];
}

const char * 
Stream::get (size_t n)
{
  if (m_blen < n) {

    //  to keep move activity low, allocate twice as much as required
    if (m_bcap < n * 2) {

      while (m_bcap < n) {
        m_bcap *= 2;
      }

      char *buffer = new char [m_bcap];
      memcpy (buffer, mp_bptr, m_blen);
      delete [] mp_buffer;
      mp_buffer = buffer;

    } else {
      memmove (mp_buffer, mp_bptr, m_blen);
    }

    m_blen += read (mp_buffer + m_blen, m_bcap - m_blen); 
    mp_bptr = mp_buffer;

  }

  if (m_blen >= n) {
    const char *r = mp_bptr;
    mp_bptr += n;
    m_blen -= n;
    m_pos += n;
    return r;
  } else {
    return 0;
  }

}

void
Stream::put (const char *b, size_t n)
{
  write (b, n);
  m_pos += n;
}

void
Stream::unget (size_t n)
{
  mp_bptr -= n;
  m_blen += n;
  m_pos -= n;
}

// ---------------------------------------------------------------
//  FileStream implementation

FileStream::FileStream ()
  : m_file (NULL), m_reading (false)
{
  // .. nothing yet ..
}

FileStream::~FileStream ()
{
  if (m_file != NULL) {
    close ();
  }  
}

void 
FileStream::open (const std::string &path, bool read) throw (tl::Exception)
{
  tl_assert (m_file == NULL);
  reset_pos ();

  m_source = path;
  m_reading = read;
  m_file = fopen (path.c_str (), read ? "rb" : "wb");
  if (m_file == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

void 
FileStream::close ()
{
  tl_assert (m_file != NULL);
  fclose (m_file);
  m_file = NULL;
}

size_t 
FileStream::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_reading);
  tl_assert (m_file != NULL);
  size_t ret = fread (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FileReadErrorException (m_source, pos (), ferror (m_file));
    }
  }

  return ret;
}

void 
FileStream::write (const char *b, size_t n) throw (tl::Exception)
{
  tl_assert (! m_reading);
  tl_assert (m_file != NULL);
  size_t ret = fwrite (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FileWriteErrorException (m_source, pos (), ferror (m_file));
    }
  }
}

// ---------------------------------------------------------------
//  ZLibStream implementation

ZLibStream::ZLibStream ()
  : m_zs (NULL), m_reading (false)
{
  // .. nothing yet ..
}

ZLibStream::~ZLibStream ()
{
  if (m_zs != NULL) {
    close ();
  }  
}

void 
ZLibStream::open (const std::string &path, bool read) throw (tl::Exception)
{
  tl_assert (m_zs == NULL);
  reset_pos ();

  m_source = path;
  m_reading = read;
  m_zs = gzopen (path.c_str (), read ? "rb" : "wb");
  if (m_zs == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

void 
ZLibStream::close ()
{
  tl_assert (m_zs != NULL);
  gzclose (m_zs);
  m_zs = NULL;
}

size_t 
ZLibStream::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_reading);
  tl_assert (m_zs != NULL);
  int ret = gzread (m_zs, b, n);
  if (ret < 0) {
    int gz_err = 0;
    const char *em = gzerror (m_zs, &gz_err);
    if (gz_err == Z_ERRNO) {
      throw FileReadErrorException (m_source, pos (), errno);
    } else {
      throw ZLibReadErrorException (m_source, pos (), em);
    }
  }

  return ret;
}

void 
ZLibStream::write (const char *b, size_t n) throw (tl::Exception)
{
  tl_assert (! m_reading);
  tl_assert (m_zs != NULL);
  int ret = gzwrite (m_zs, (char *) b, n);
  if (ret < 0) {
    int gz_err = 0;
    const char *em = gzerror (m_zs, &gz_err);
    if (gz_err == Z_ERRNO) {
      throw FileWriteErrorException (m_source, pos (), errno);
    } else {
      throw ZLibWriteErrorException (m_source, pos (), em);
    }
  }
}

// ---------------------------------------------------------------
//  FileStream implementation

FilePipe::FilePipe ()
  : m_file (NULL), m_reading (false)
{
  // .. nothing yet ..
}

FilePipe::~FilePipe ()
{
  if (m_file != NULL) {
    close ();
  }  
}

void 
FilePipe::open (const std::string &path, bool read) throw (tl::Exception)
{
  tl_assert (m_file == NULL);
  reset_pos ();

  m_source = path;
  m_reading = read;
  m_file = popen (path.c_str (), read ? "r" : "w");
  if (m_file == NULL) {
    throw FilePOpenErrorException (m_source, errno);
  }
}

void 
FilePipe::close ()
{
  tl_assert (m_file != NULL);
  fclose (m_file);
  m_file = NULL;
}

size_t 
FilePipe::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_reading);
  tl_assert (m_file != NULL);
  size_t ret = fread (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FilePReadErrorException (m_source, pos (), ferror (m_file));
    }
  }

  return ret;
}

void 
FilePipe::write (const char *b, size_t n) throw (tl::Exception)
{
  tl_assert (! m_reading);
  tl_assert (m_file != NULL);
  size_t ret = fwrite (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FilePWriteErrorException (m_source, pos (), ferror (m_file));
    }
  }
}

}

