
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbStream
#define HDR_dbStream

#include "tlException.h"

#include <string>
#include <stdio.h>
#include <zlib.h>


namespace db
{

/**
 *  @brief A stream abstraction class
 *
 *  The basic objective of this interface is to provide
 *  the capability to read a block of n bytes into a buffer.
 */

class Stream
{
public:
  /**
   *  @brief Default constructor
   */
  Stream ();

  /**
   *  @brief Default (empty) destructor
   */
  virtual ~Stream () { }

  /** 
   *  @brief This is the outer write method to call
   *  
   *  This implementation writes data through the 
   *  protected write call.
   */
  void put (const char *b, size_t n);

  /** 
   *  @brief This is the outer read method to call
   *  
   *  This implementation obtains data through the 
   *  protected read call and buffers the data accordingly so
   *  a contigous memory block can be returned
   *
   *  @return 0 if not enough data can be obtained
   */
  const char *get (size_t n);

  /** 
   *  @brief Undo a previous get call
   *  
   *  This call puts back the bytes read by a previous get call.
   *  Only one call can be made undone.
   */
  void unget (size_t n);

  /**
   *  @brief Obtain the current file position
   */
  size_t pos () const 
  {
    return m_pos;
  }
    
  /** 
   *  @brief Read a block of n bytes
   *
   *  Read the requested number of bytes or less.
   *  May throw an exception if a read error occures.
   *  
   *  @param b The buffer where to write to
   *  @param n The number of bytes to read (or less)
   *
   *  @return The number of bytes read. Should report 0 on EOF
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception) = 0;

  /**
   *  @brief Write a block a n bytes
   *
   *  May throw an exception if a write error occures.
   *
   *  @param b What to write
   *  @param n The number of bytes to write 
   */
  virtual void write (const char *b, size_t n) throw (tl::Exception) = 0;

protected:
  void reset_pos ()
  {
    m_pos = 0;
  }

private:
  size_t m_pos;
  char *mp_buffer;
  size_t m_bcap;
  size_t m_blen;
  char *mp_bptr;
};

class FileOpenErrorException
  : public tl::Exception
{
public:
  FileOpenErrorException (const std::string &f, int en)
    : tl::Exception (tl::translate ("Unable to open file: %s (errno=%d)").c_str (), f.c_str (), en)
  { }
};

class FileWriteErrorException
  : public tl::Exception
{
public:
  FileWriteErrorException (const std::string &f, size_t p, int en)
    : tl::Exception (tl::translate ("Read error on file: %s (position=%d,errno=%d)").c_str (), f.c_str (), long (p), en)
  { }
};

class FileReadErrorException
  : public tl::Exception
{
public:
  FileReadErrorException (const std::string &f, size_t p, int en)
    : tl::Exception (tl::translate ("Read error on file: %s (position=%d,errno=%d)").c_str (), f.c_str (), long (p), en)
  { }
};

class ZLibWriteErrorException
  : public tl::Exception
{
public:
  ZLibWriteErrorException (const std::string &f, size_t p, const char *em)
    : tl::Exception (tl::translate ("Write error on file in decompression library: %s (position=%d,message=%s)").c_str (), f.c_str (), long (p), em)
  { }
};

class ZLibReadErrorException
  : public tl::Exception
{
public:
  ZLibReadErrorException (const std::string &f, size_t p, const char *em)
    : tl::Exception (tl::translate ("Read error on file in decompression library: %s (position=%d,message=%s)").c_str (), f.c_str (), long (p), em)
  { }
};

/**
 *  @brief A zlib file adapter class
 *
 *  Implements the stream adapter for a zlib stream
 */
class ZLibStream
  : public Stream
{
public:
  ZLibStream ();
  virtual ~ZLibStream ();

  /**
   *  @brief Open a file with the given path
   *
   *  Opening a file is a prerequisite for reading from the 
   *  object. open() will throw a FileOpenErrorException if 
   *  an error occurs.
   *
   *  @param path The (relative) path of the file to open
   *  @param read True, if the file should be read, false on write.
   */
  void open (const std::string &path, bool read = true) throw (tl::Exception);

  /**
   *  @brief Close the file
   *
   *  The destructor will automatically close the file.
   */
  void close ();

  /**
   *  @brief Read from a file 
   *
   *  Implements the basic read method. 
   *  Will throw a ZLibReadErrorException if an error occurs.
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception);

  /**
   *  @brief Write to a file 
   *
   *  Implements the basic write method. 
   *  Will throw a ZLibWriteErrorException if an error occurs.
   */
  virtual void write (const char *b, size_t n) throw (tl::Exception);

private:
  std::string m_source;
  gzFile m_zs;
  bool m_reading;
};

/**
 *  @brief A simple file adapter class
 *
 *  Implements the stream adapter for a simple file
 */
class FileStream
  : public Stream
{
public:
  FileStream ();
  virtual ~FileStream ();

  /**
   *  @brief Open a file with the given path
   *
   *  Opening a file is a prerequisite for reading from the 
   *  object. open() will throw a FileOpenErrorException if 
   *  an error occurs.
   *
   *  @param path The (relative) path of the file to open
   *  @param read True, if the file should be read, false on write.
   */
  void open (const std::string &path, bool read = true) throw (tl::Exception);

  /**
   *  @brief Close the file
   *
   *  The destructor will automatically close the file.
   */
  void close ();

  /**
   *  @brief Read from a file 
   *
   *  Implements the basic read method. 
   *  Will throw a FileReadErrorException if an error occurs.
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception);

  /**
   *  @brief Write to a file 
   *
   *  Implements the basic write method. 
   *  Will throw a FileWriteErrorException if an error occurs.
   */
  virtual void write (const char *b, size_t n) throw (tl::Exception);

private:
  std::string m_source;
  FILE *m_file;
  bool m_reading;
};

class FilePOpenErrorException
  : public tl::Exception
{
public:
  FilePOpenErrorException (const std::string &f, int en)
    : tl::Exception (tl::translate ("Unable to get input from command through pipe: %s (errno=%d)").c_str (), f.c_str (), en)
  { }
};

class FilePReadErrorException
  : public tl::Exception
{
public:
  FilePReadErrorException (const std::string &f, size_t p, int en)
    : tl::Exception (tl::translate ("Read error on pipe from command: %s (position=%d,errno=%d)").c_str (), f.c_str (), long (p), en)
  { }
};

class FilePWriteErrorException
  : public tl::Exception
{
public:
  FilePWriteErrorException (const std::string &f, size_t p, int en)
    : tl::Exception (tl::translate ("Write error on pipe from command: %s (position=%d,errno=%d)").c_str (), f.c_str (), long (p), en)
  { }
};

/**
 *  @brief A simple file adapter class
 *
 *  Implements the pipe adapter for a simple file
 */
class FilePipe
  : public Stream
{
public:
  FilePipe ();
  virtual ~FilePipe ();

  /**
   *  @brief Open a stream by connecting with the stdout of a given command
   *
   *  Opening a pipe is a prerequisite for reading from the 
   *  object. open() will throw a FilePOpenErrorException if 
   *  an error occurs - commonly if the command cannot be executed.
   *  This implementation is based on popen ().
   *
   *  @param cmd The command to execute
   *  @param read True, if the file should be read, false on write.
   */
  void open (const std::string &path, bool read = true) throw (tl::Exception);

  /**
   *  @brief Close the pipe
   *
   *  The destructor will automatically close the pipe.
   */
  void close ();

  /**
   *  @brief Read from the pipe 
   *
   *  Implements the basic read method. 
   *  Will throw a FilePReadErrorException if an error occurs.
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception);

  /**
   *  @brief Write to a file 
   *
   *  Implements the basic write method. 
   *  Will throw a FilePWriteErrorException if an error occurs.
   */
  virtual void write (const char *b, size_t n) throw (tl::Exception);

private:
  FILE *m_file;
  std::string m_source;
  bool m_reading;
};

}

#endif

