
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbText
#define HDR_dbText

#include <dbTypes.h>
#include <dbMemStatistics.h>
#include <dbPoint.h>
#include <dbTrans.h>
#include <dbShapeRepository.h>
#include <dbShapeTag.h>

#include <string>

namespace db {

template <class Coord> class generic_repository;

/**
 *  @brief A text object
 *
 *  A text object has a point (location), a text, a text transformation,
 *  a text size and a font id. Text size and font id are provided to be
 *  be able to render the text correctly.
 * 
 */

template <class C>
struct text
{
  typedef C coord_type;
  typedef db::coord_traits<C> coord_traits;
  typedef db::point<C> point_type;
  typedef db::box<C> box_type;
  typedef db::simple_trans<C> trans_type;
  typedef db::shape_tag< text<C> > tag;
  typedef int font_id_type;

  /**
   *  @brief The standard constructor 
   *
   *  The standard constructor is taking a transformation (with point),
   *  a text size value (which can be zero) and a font id which can 0 
   *  also.
   * 
   *  @param s The text 
   *  @param t The transformation of the text
   *  @param h The size
   *  @param f The font id
   */
  text (const std::string &s, const trans_type &t, coord_type h = 0, font_id_type f = 0)
    : m_string (s), m_trans (t), m_size (h), m_font_id (f)
  {
    // .. nothing yet ..
  }

  /** 
   *  @brief Default constructor
   *
   *  Creates an empty text object at (0,0) with empty text.
   */
  text ()
    : m_string (""), m_trans (), m_size (0), m_font_id (0)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief The (dummy) translation operator
   */
  void translate (const text<C> &d, db::generic_repository<C> &)
  {
    *this = d;
  }

  /**
   *  @brief Ordering operator
   */
  bool operator< (const text<C> &b) const
  {
    return m_trans < b.m_trans || (m_trans == b.m_trans && m_string < b.m_string);
  }

  /**
   *  @brief Equality test
   */
  bool operator== (const text<C> &b) const
  {
    return m_trans == b.m_trans && m_string == b.m_string;
  }

  /** 
   *  @brief Inequality test
   */
  bool operator!= (const text<C> &b) const
  {
    return !operator== (b);
  }

  /**
   *  @brief The text string accessor
   */
  const std::string &string () const 
  { 
    return m_string;
  }
   
  /**
   *  @brief The transformation accessor
   */
  const trans_type &trans () const 
  { 
    return m_trans;
  }
   
  /**
   *  @brief The size accessor
   */
  const coord_type &size () const 
  { 
    return m_size;
  }
   
  /**
   *  @brief The font id accessor
   */
  const font_id_type &font_id () const 
  { 
    return m_font_id;
  }
   
  /**
   *  @brief Transform the text.
   *
   *  Transforms the text with the given transformation.
   *  Modifies the text with the transformed text.
   *  
   *  @param t The transformation to apply.
   *
   *  @return The transformed text.
   */
  template <class Tr>
  text<C> &transform (const Tr &t)
  {
    m_trans = t * m_trans;
    return *this;
  }

  /**
   *  @brief Transform the text.
   *
   *  Transforms the text with the given transformation.
   *  Does not modify the text but returns the transformed text.
   *  
   *  @param t The transformation to apply.
   *
   *  @return The transformed text.
   */
  template <class Tr>
  text<typename Tr::target_coord_type> transformed (const Tr &t) const
  {
    typedef typename Tr::target_coord_type target_coord_type;
    return text<target_coord_type> (m_string, simple_trans<target_coord_type> (t (fixpoint_trans (m_trans.rot ())).rot (), t (m_trans.disp ())), t (m_size), m_font_id);
  }

  /**
   *  @brief Return the bounding box
   *
   *  The bounding box is a box consisting of a single point
   */
  box_type box () const
  {
    return box_type (m_trans.disp (), m_trans.disp ());
  }

  /**
   *  @brief String conversion
   */
  std::string to_string () const
  {
    return std::string ("('") + m_string + "'," + m_trans.to_string () + ")";
  }

  /**
   *  @brief Reduce the text
   *
   *  Reduction of a text normalizes the text by extracting
   *  a suitable transformation and placing the text in a unique
   *  way.
   *
   *  @return The transformation that must be applied to render the original text
   */
  trans_type reduce ()
  {
    trans_type tr = m_trans;
    m_trans = trans_type ();
    return tr;
  }

  size_t mem_used () const
  {
    return db::mem_used (m_string) + db::mem_used (m_trans) + db::mem_used (m_size) + db::mem_used (m_font_id);
  }

  size_t mem_reqd () const
  {
    return db::mem_reqd (m_string) + db::mem_reqd (m_trans) + db::mem_reqd (m_size) + db::mem_reqd (m_font_id);
  }

private:
  std::string m_string;
  trans_type m_trans;
  coord_type m_size;
  font_id_type m_font_id;
};


/**
 *  @brief Binary * operator (transformation)
 *
 *  Transforms the text with the given transformation and 
 *  returns the result.
 *
 *  @param t The transformation to apply
 *  @param s The text to transform
 *  @return t * s
 */
template <class C, class Tr>
inline text<typename Tr::target_coord_type> 
operator* (const Tr &t, const text<C> &s)
{
  return s.transformed (t);
}

/**
 *  @brief Output stream insertion operator
 */
template <class C>
inline std::ostream &
operator<< (std::ostream &os, const text<C> &s)
{
  return (os << s.to_string ());
}

/**
 *  @brief The standard text typedef
 */
typedef text<db::Coord> Text;

/**
 *  @brief The double coordinate text typedef
 */
typedef text<db::DCoord> DText;

/** 
 *  @brief A text reference
 *
 *  A text reference is basically a proxy to a text and
 *  is used to implement text references with a repository.
 */

template <class Text>
struct text_ref
  : public shape_ref<Text>
{
  typedef typename Text::coord_type coord_type;
  typedef typename Text::box_type box_type;
  typedef typename Text::point_type point_type;
  typedef db::simple_trans<coord_type> trans_type;
  typedef Text text_type;
  typedef db::generic_repository<coord_type> repository_type;
  typedef db::shape_tag< text_ref<Text> > tag;

  /**
   *  @brief The default constructor.
   *  
   *  The default constructor creates a invalid text reference
   */
  text_ref ()
    : shape_ref<Text> ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief The constructor creating a reference from an actual text
   */
  text_ref (const text_type &p, repository_type &rep)
    : shape_ref<Text> (p, rep)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief The translation constructor.
   *  
   *  This constructor allows to copy a text reference from one
   *  repository to another
   */
  text_ref (const text_ref &ref, repository_type &rep)
    : shape_ref<Text> (ref, rep)
  {
    // .. nothing yet ..
  }

  /** 
   *  @brief Return the transformed object
   * 
   *  This version does not change the object and is const.
   */
  text_ref<Text> transformed (const trans_type &t) const
  {
    text_ref<Text> tref (*this);
    tref.transform (t);
    return tref;
  }
};

/**
 *  @brief Binary * operator (transformation)
 *
 *  Transforms the text reference with the given transformation and 
 *  returns the result.
 *
 *  @param t The transformation to apply
 *  @param p The text reference to transform
 *  @return t * p
 */
template <class Text, class Tr>
inline text_ref<Text>
operator* (const Tr &t, const text_ref<Text> &p)
{
  return p.transformed (t);
}

/**
 *  @brief The text reference typedef
 */
typedef text_ref<Text> TextRef;

/**
 *  @brief The text reference typedef for double coordinates
 */
typedef text_ref<DText> DTextRef;


template <class X>
size_t mem_used (const text<X> &x)
{
  return x.mem_used ();
}

template <class X>
size_t mem_reqd (const text<X> &x)
{
  return x.mem_reqd ();
}

} // namespace db

#endif

