
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbTrans
#define HDR_dbTrans

#include <dbPoint.h>
#include <tlString.h>
#include <string>
#include <iostream>

namespace tl {
  class Extractor;
}

namespace db {

/**
 *  @brief A dummy unit transformation
 *
 *  This transformation is supplied in order to allow generic transformation
 *  parameters being passed a "dummy" transformation.
 */

struct unit_trans
{
  typedef unit_trans inverse_trans;

  /**
   *  @brief The default constructor (unity transformation)
   */
  unit_trans ()
  {
    // .. nothing else ..
  }

  /**
   *  @brief Mirror predicate
   */
  bool is_mirror () const
  {
    return false;
  }

  /**
   *  @brief Inversion
   */
  unit_trans inverted () const
  {
    return *this;
  }

  /** 
   *  @brief In-place inversion
   */
  unit_trans invert ()
  {
    return *this;
  }

  /**
   *  @brief The transformation of a point 
   */
  template <class C>
  point<C> operator() (const point<C> &p) const
  {
    return p;
  }

  /**
   *  @brief Transform a distance
   */
  template <class C>
  C ctrans (C d) const
  {
    return d;
  }

  /**
   *  @brief String conversion
   */
  std::string to_string () const;
};

/**
 *  @brief A fixpoint transformation
 *
 *  The fixpoint transformation applies a rotation and/or mirroring operation.
 */

struct fixpoint_trans
{
  typedef fixpoint_trans inverse_trans;

  /**
   *  @brief The default constructor (unity transformation)
   */
  fixpoint_trans ()
    : m_f (0)
  {
    // .. nothing else ..
  }

  /**
   *  @brief The standard constructor using angle and mirror flag
   *
   *  The sequence of operations is: mirroring at x axis,
   *  rotation, application of displacement.
   *  
   *  @param rot The rotation in units of 90 degree
   *  @param mirrx True, if mirrored at x axis
   */
  fixpoint_trans (int rot, bool mirrx)
    : m_f ((rot & 3) + (mirrx ? 4 : 0))
  {
    // .. nothing else ..
  }
  
  /**
   *  @brief The copy constructor 
   *
   *  @param d The source from which to copy
   */
  fixpoint_trans (const fixpoint_trans &d) 
    : m_f (d.rot ())
  { }

  /**
   *  @brief The standard constructor using a code rather than angle and mirror and no displacement
   *  
   *  @param f The rotation/mirror code (r0 .. m135 constants)
   */
  fixpoint_trans (int f)
    : m_f (f)
  {
    // .. nothing else ..
  }

  /** 
   *  @brief The rotation/mirror codes
   */
  static const int r0   = 0;  //  No rotation
  static const int r90  = 1;  //  Rotation by 90 degree counterclockwise
  static const int r180 = 2;  //  Rotation by 180 degree counterclockwise
  static const int r270 = 3;  //  Rotation by 270 degree counterclockwise
  static const int m0   = 4;  //  Mirroring at x-axis
  static const int m45  = 5;  //  Mirroring at 45-degree axis
  static const int m90  = 6;  //  Mirroring at y-axis
  static const int m135 = 7;  //  Mirroring at 135-degree axis

  /**
   *  @brief Inversion
   *
   *  Returns the inverted transformation
   *
   *  @return The inverted transformation
   */
  fixpoint_trans inverted () const
  {
    fixpoint_trans t (*this);
    t.invert ();
    return t;
  }

  /** 
   *  @brief In-place inversion
   *
   *  Inverts the transformation and replaces *this by the
   *  inverted one.
   *
   *  @return The inverted transformation
   */
  fixpoint_trans invert ()
  {
    if (m_f < 4) {
      m_f = (4 - m_f) & 3;
    }
    return *this;
  }

  /**
   *  @brief The transformation of a point 
   *
   *  The operator() method transforms the given point.
   *  q = t(p)
   *  
   *  @param p The point to transform
   *  @return The transformed point
   */
  template <class C>
  point<C> operator() (const point<C> &p) const
  {
    switch (m_f) {
    default:
      return point<C> (p.x (), p.y ());
    case 1:
      return point<C> (-p.y (), p.x ());
    case 2:
      return point<C> (-p.x (), -p.y ());
    case 3:
      return point<C> (p.y (), -p.x ());
    case 4:
      return point<C> (p.x (), -p.y ());
    case 5:
      return point<C> (p.y (), p.x ());
    case 6:
      return point<C> (-p.x (), p.y ());
    case 7:
      return point<C> (-p.y (), -p.x ());
    }
  }

  /**
   *  @brief Transform a fixpoint transformation
   */
  fixpoint_trans ftrans (fixpoint_trans t) const
  {
    fixpoint_trans tt (*this);
    tt *= t;
    return tt;
  }

  /**
   *  @brief Transform a distance
   */
  template <class C>
  C ctrans (C d) const
  {
    return d;
  }

  /**
   *  @brief Multiplication (concatenation) of transformations
   *
   *  The *= operator modifies the transformation by 
   *  replacing *this with *this * t (t is applied before *this).
   *
   *  @param t The transformation to apply before
   *  @return The modified transformation
   */
  fixpoint_trans &operator*= (const fixpoint_trans &t);

  /**
   *  @brief A sorting criterion
   */
  bool operator< (const fixpoint_trans &t) const
  {
    return m_f < t.m_f;
  }

  /**
   *  @brief Equality test
   */
  bool operator== (const fixpoint_trans &t) const
  {
    return m_f == t.m_f;
  }

  /**
   *  @brief Inequality test
   */
  bool operator!= (const fixpoint_trans &t) const
  {
    return !operator== (t);
  }

  /**
   *  @brief String conversion
   */
  std::string to_string () const;

  /**
   *  @brief Accessor to the rotation/mirror code
   */
  int rot () const 
  {
    return m_f;
  }

  /**
   *  @brief Mirror flag.
   *
   *  The result of this operation is true, if the transformation is
   *  mirroring, i.e. det(M) == -1.
   */
  bool is_mirror () const
  {
    return m_f >= 4;
  }

private:
  int m_f;
};

/**
 *  @brief Multiplication (concatenation) of transformations
 *
 *  t = t1 * t2 is the resulting transformation that is effectively
 *  applied if first t2 and then t1 is applied.
 *
 *  @param t1 The transformation to apply last
 *  @param t2 The transformation to apply first
 *  @return t1 * t2
 */
inline fixpoint_trans 
operator* (const fixpoint_trans &t1, const fixpoint_trans &t2)
{
  fixpoint_trans t (t1);
  t *= t2;
  return t;
}

/**
 *  @brief Output stream insertion operator
 */
inline std::ostream &
operator<< (std::ostream &os, const fixpoint_trans &t)
{
  return (os << t.to_string ());
}

/**
 *  @brief A simple transformation
 *
 *  The simple transformation applies a displacement vector
 *  and a simple fixpoint transformation. 
 */

template <class C>
struct simple_trans
  : private fixpoint_trans 
{
  typedef C coord_type;
  typedef C target_coord_type;
  typedef simple_trans<C> inverse_trans;

  /**
   *  @brief The default constructor (unity transformation)
   */
  simple_trans ()
    : fixpoint_trans (0), m_u ()
  {
    // .. nothing else ..
  }

  /**
   *  @brief The standard constructor using angle and mirror flag
   *
   *  The sequence of operations is: mirroring at x axis,
   *  rotation, application of displacement.
   *  
   *  @param rot The rotation in units of 90 degree
   *  @param mirrx True, if mirrored at x axis
   *  @param u The displacement
   */
  template <class D>
  simple_trans (int rot, bool mirrx, const point<D> &u)
    : fixpoint_trans (rot, mirrx), m_u (u)
  {
    // .. nothing else ..
  }
  
  /**
   *  @brief The copy constructor that converts also
   *
   *  The copy constructor allows to convert between different
   *  coordinate types, if possible.
   *
   *  @param d The source from which to copy
   */
  template <class D>
  simple_trans (const simple_trans<D> &d) 
    : fixpoint_trans (d.rot ()), m_u (d.disp ())
  { }

  /**
   *  @brief The standard constructor using a code rather than angle and mirror
   *  
   *  @param f The rotation/mirror code (r0 .. m135 constants)
   *  @param u The displacement
   */
  simple_trans (int f, const point<C> &u)
    : fixpoint_trans (f), m_u (u)
  {
    // .. nothing else ..
  }

  /**
   *  @brief The standard constructor using a displacement only
   *  
   *  @param u The displacement
   */
  simple_trans (const point<C> &u)
    : fixpoint_trans (0), m_u (u)
  {
    // .. nothing else ..
  }

  /**
   *  @brief The standard constructor using a code rather than angle and mirror and no displacement
   *  
   *  @param f The rotation/mirror code (r0 .. m135 constants)
   */
  simple_trans (int f)
    : fixpoint_trans (f), m_u (0, 0)
  {
    // .. nothing else ..
  }

  /**
   *  @brief Inversion
   *
   *  Returns the inverted transformation
   *
   *  @return The inverted transformation
   */
  simple_trans<C> inverted () const
  {
    simple_trans<C> t (*this);
    t.invert ();
    return t;
  }

  /** 
   *  @brief In-place inversion
   *
   *  Inverts the transformation and replaces *this by the
   *  inverted one.
   *
   *  @return The inverted transformation
   */
  simple_trans<C> invert ()
  {
    fixpoint_trans::invert ();
    point<C> u = m_u;
    m_u = point<C> ();
    m_u = -operator() (u);
    return *this;
  }

  /**
   *  @brief The transformation of a distance
   *
   *  The ctrans method transforms the given distance.
   *  e = t(d). For the simple transformations, there
   *  is no magnification and no modification of the distance
   *  therefore.
   *  
   *  @param d The distance to transform
   *  @return The transformed distance
   */
  template <class D>
  C ctrans (D d) const
  {
    return d;
  }
  
  /**
   *  @brief The transformation of a point 
   *
   *  The operator() method transforms the given point.
   *  q = t(p)
   *  
   *  @param p The point to transform
   *  @return The transformed point
   */
  template <class D>
  point<C> operator() (const point<D> &p) const
  {
    return point<C> (fixpoint_trans::operator() (p)) + m_u;
  }

  /**
   *  @brief Transform a fixpoint transformation
   */
  fixpoint_trans ftrans (fixpoint_trans t) const
  {
    return fixpoint_trans::ftrans (t);
  }

  /**
   *  @brief Multiplication (concatenation) of transformations
   *
   *  The *= operator modifies the transformation by 
   *  replacing *this with *this * t (t is applied before *this).
   *
   *  @param t The transformation to apply before
   *  @return The modified transformation
   */
  simple_trans<C> &operator*= (const simple_trans<C> &t)
  {
    m_u = operator() (t.m_u); 
    fixpoint_trans::operator*= (t);
    return *this;
  }

  /**
   *  @brief A sorting criterion
   */
  bool operator< (const simple_trans<C> &t) const
  {
    return fixpoint_trans::operator< (t) || (fixpoint_trans::operator== (t) && m_u < t.m_u);
  }

  /**
   *  @brief Equality test
   */
  bool operator== (const simple_trans<C> &t) const
  {
    return fixpoint_trans::operator== (t) && m_u == t.m_u;
  }

  /**
   *  @brief Inequality test
   */
  bool operator!= (const simple_trans<C> &t) const
  {
    return !operator== (t);
  }

  /**
   *  @brief String conversion
   */
  std::string to_string () const
  {
    std::string s1 = fixpoint_trans::to_string ();
    std::string s2 = m_u.to_string ();
    if (! s1.empty () && ! s2.empty ()) {
      return s1 + " " + s2;
    } else {
      return s1 + s2;
    }
  }

  /** 
   *  @brief Accessor to the point
   */
  const point<C> &disp () const
  {
    return m_u;
  }

  /** 
   *  @brief Accessor to the fp_trans
   */
  const fixpoint_trans &fp_trans () const
  {
    return *this;
  }

  using fixpoint_trans::rot;
  using fixpoint_trans::is_mirror;

  //  make the constants usable
  using fixpoint_trans::r0;
  using fixpoint_trans::r90;
  using fixpoint_trans::r180;
  using fixpoint_trans::r270;
  using fixpoint_trans::m0;
  using fixpoint_trans::m45;
  using fixpoint_trans::m90;
  using fixpoint_trans::m135;

private:
  point<C> m_u;
};

/**
 *  @brief Multiplication (concatenation) of transformations
 *
 *  t = t1 * t2 is the resulting transformation that is effectively
 *  applied if first t2 and then t1 is applied.
 *
 *  @param t1 The transformation to apply last
 *  @param t2 The transformation to apply first
 *  @return t1 * t2
 */
template <class C>
inline simple_trans<C> 
operator* (const simple_trans<C> &t1, const simple_trans<C> &t2)
{
  simple_trans<C> t (t1);
  t *= t2;
  return t;
}

/**
 *  @brief Output stream insertion operator
 */
template <class C>
inline std::ostream &
operator<< (std::ostream &os, const simple_trans<C> &t)
{
  return (os << t.to_string ());
}

/**
 *  @brief A magnifying transformation
 *
 *  This transformation does a isotropic magnification, but no
 *  rotation or displacement.
 *  F is the type of the outgoing coordinate, I the type of the
 *  incoming coordinate. Since the actual transformation methods
 *  are template members, the incoming type is only used for 
 *  determining the inverse of the transformation.
 */

template <class I, class F>
struct mag_trans
{
  typedef I coord_type;
  typedef F target_coord_type;
  typedef mag_trans<F, I> inverse_trans;

  /**
   *  @brief The default constructor (unity transformation)
   */
  mag_trans ()
    : m_f (1)
  {
    // .. nothing else ..
  }

  /**
   *  @brief The standard constructor with a fixed magnification
   *
   *  @param f The magnification factor
   */
  mag_trans (double f)
    : m_f (f)
  {
    // .. nothing else ..
  }

  /**
   *  @brief Inversion
   *
   *  Returns the inverted transformation
   *
   *  @return The inverted transformation
   */
  mag_trans<F, I> inverted () const
  {
    return mag_trans<F, I> (1.0 / m_f);
  }

  /**
   *  @brief The transformation of a distance
   *
   *  The ctrans method transforms the given distance.
   *  e = t(d). For the simple transformations, there
   *  is no magnification and no modification of the distance
   *  therefore.
   *  
   *  @param d The distance to transform
   *  @return The transformed distance
   */
  template <typename C>
  F ctrans (C d) const
  {
    return db::coord_traits<F>::rounded (d * m_f);
  }
  
  /**
   *  @brief The transformation of a point 
   *
   *  The operator() method transforms the given point.
   *  q = t(p)
   *  
   *  @param p The point to transform
   *  @return The transformed point
   */
  template <class C>
  point<F> operator() (const point<C> &p) const
  {
    return point<F> (db::coord_traits<F>::rounded (p.x () * m_f), db::coord_traits<F>::rounded (p.y () * m_f));
  }

  /**
   *  @brief Transform a fixpoint transformation
   */
  fixpoint_trans ftrans (fixpoint_trans t) const
  {
    return t;
  }

  /**
   *  @brief Multiplication (concatenation) of transformations
   *
   *  The *= operator modifies the transformation by 
   *  replacing *this with *this * t (t is applied before *this).
   *
   *  @param t The transformation to apply before
   *  @return The modified transformation
   */
  mag_trans<I, F> &operator*= (const mag_trans<I, F> &t)
  {
    m_f *= t.m_f;
    return *this;
  }

  /**
   *  @brief A sorting criterion
   */
  bool operator< (const mag_trans<I, F> &t) const
  {
    return m_f < t.m_f;
  }

  /**
   *  @brief Equality test
   */
  bool operator== (const mag_trans<I, F> &t) const
  {
    return fabs (m_f - t.m_f) < 1e-10;
  }

  /**
   *  @brief Inequality test
   */
  bool operator!= (const mag_trans<I, F> &t) const
  {
    return !operator== (t);
  }

  /**
   *  @brief String conversion
   */
  std::string to_string () const
  {
    return "*" + tl::to_string (m_f);
  }

  /**
   *  @brief Factor accessor function
   */
  double factor () const 
  { 
    return m_f;
  }

private:
  double m_f;
};

/**
 *  @brief Multiplication (concatenation) of transformations
 *
 *  t = t1 * t2 is the resulting transformation that is effectively
 *  applied if first t2 and then t1 is applied.
 *
 *  @param t1 The transformation to apply last
 *  @param t2 The transformation to apply first
 *  @return t1 * t2
 */
template <class I, class F>
inline mag_trans<I, F> 
operator* (const mag_trans<I, F> &t1, const mag_trans<I, F> &t2)
{
  mag_trans<I, F> t (t1);
  t *= t2;
  return t;
}

/**
 *  @brief Output stream insertion operator
 */
template <class I, class F>
inline std::ostream &
operator<< (std::ostream &os, const mag_trans<I, F> &t)
{
  return (os << t.to_string ());
}

/**
 *  @brief A combined transformation
 *  
 *  A combined transformation is the combination of two 
 *  transformations T1 and T2 (T=T1*T2). Although the multiplication
 *  of two transformations may render the same result, but
 *  usually is more efficient. To combine two different
 *  transformations however, the combined_trans template
 *  is better suited.
 */

template <class T1, class T2>
struct combined_trans
{
  typedef typename T2::coord_type coord_type;
  typedef typename T1::target_coord_type target_coord_type;
  typedef combined_trans<typename T2::inverse_trans, typename T1::inverse_trans> inverse_trans;

  /**
   *  @brief Default constructor
   *
   *  Creates a unity transformation
   */
  combined_trans ()
    : t1 (), t2 ()
  {
    // .. nothing else ..
  }

  /**
   *  @brief Standard constructor
   *
   *  Takes two transformations and combines both.
   */
  combined_trans (const T1 &_t1, const T2 &_t2)
    : t1 (_t1), t2 (_t2)
  {
    // .. nothing else ..
  }

  /** 
   *  @brief Inversion
   *
   *  Inverts the transformation and returns the inverted
   *  transformation which swaps T1 and T2 in the type definition.
   *
   *  @return The inverted transformation
   */
  inverse_trans inverted () const
  {
    return inverse_trans (t2.inverted (), t1.inverted ());
  }

  /**
   *  @brief The transformation of a point 
   *
   *  The operator() method transforms the given point.
   *  q = t(p)
   *  
   *  @param p The point to transform
   *  @return The transformed point
   */
  template <class C>
  point<target_coord_type> operator() (const point<C> &p) const
  {
    typedef typename T2::target_coord_type intern_coord_type;
    point<intern_coord_type> q = t2.operator() (p);
    return t1.operator() (q);
  }

  /**
   *  @brief The transformation of a distance 
   *
   *  The ctrans method transforms the given distance.
   *  s = t(d)
   *  
   *  @param d The distance to transform
   *  @return The transformed distance
   */
  template <class C>
  target_coord_type ctrans (C p) const
  {
    typedef typename T2::target_coord_type intern_coord_type;
    intern_coord_type q = t2.ctrans (p);
    return t1.ctrans (q);
  }

  /**
   *  @brief The transformation of a fixpoint transformation 
   *
   *  The ftrans method transforms the given fixpoint transformation.
   *  f = t(f)
   *  
   *  @param f The fixpoint transformation to transform
   *  @return The transformed fixpoint transformation
   */
  fixpoint_trans ftrans (fixpoint_trans f) const
  {
    return t1 (t2 (f));
  }

  /**
   *  @brief A sorting criterion
   */
  bool operator< (const combined_trans<T1, T2> &t) const
  {
    return t1 < t.t1 || (t1 == t.t1 && t2 < t.t2);
  }

  /**
   *  @brief Equality test
   */
  bool operator== (const combined_trans<T1, T2> &t) const
  {
    return t1 == t.t1 && t2 == t.t2;
  }

  /**
   *  @brief Inequality test
   */
  bool operator!= (const combined_trans<T1, T2> &t) const
  {
    return !operator== (t);
  }

  /**
   *  @brief String conversion
   */
  std::string to_string () const
  {
    std::string s1 = t1.to_string ();
    std::string s2 = t2.to_string ();
    if (! s1.empty () && ! s2.empty ()) {
      return s1 + " " + s2;
    } else {
      return s1 + s2;
    }
  }

  T1 t1;
  T2 t2;
};

/**
 *  @brief Output stream insertion operator
 */
template <class T1, class T2>
inline std::ostream &
operator<< (std::ostream &os, const combined_trans<T1, T2> &t)
{
  return (os << t.to_string ());
}

/**
 *  @brief The standard fixpoint transformation
 */
typedef fixpoint_trans FTrans;

/**
 *  @brief The standard transformation
 */
typedef simple_trans<db::Coord> Trans;

/**
 *  @brief The double coordinate transformation
 */
typedef simple_trans<db::DCoord> DTrans;

/**
 *  @brief The standard magnifying transformation
 */
typedef mag_trans<db::Coord, db::DCoord>  Mag;

/**
 *  @brief The double coordinate magnifying transformation
 */
typedef mag_trans<db::DCoord, db::DCoord> DMag;

/**
 *  @brief The combined magnifying transformation
 */
typedef combined_trans<DTrans, Mag>  MagTrans;

/**
 *  @brief The double coordinate combined magnifying transformation
 */
typedef combined_trans<DTrans, DMag> DMagTrans;

/**
 *  @brief A special implementation for operator* on MagTrans 
 */
inline MagTrans 
operator* (const MagTrans &a, const MagTrans &b)
{
  return MagTrans (a.t1 * db::DTrans (b.t1.rot (), b.t1.disp () * a.t2.factor ()), Mag (a.t2.factor () * b.t2.factor ()));
}
 
/**
 *  @brief A special implementation for operator* on MagTrans 
 */
inline MagTrans 
operator* (const MagTrans &a, const Trans &b)
{
  return MagTrans (a.t1 * db::DTrans (b.rot (), b.disp () * a.t2.factor ()), Mag (a.t2.factor ()));
}
 
/**
 *  @brief A special implementation for operator* on MagTrans 
 */
inline MagTrans 
operator* (const MagTrans &a, const Mag &b)
{
  return MagTrans (a.t1, a.t2 * b);
}
 
/**
 *  @brief A special implementation for operator* on DMagTrans 
 */
inline DMagTrans 
operator* (const DMagTrans &a, const DMagTrans &b)
{
  return DMagTrans (a.t1 * db::DTrans (b.t1.rot (), b.t1.disp () * a.t2.factor ()), DMag (a.t2.factor () * b.t2.factor ()));
}
 
/**
 *  @brief A special implementation for operator* on DMagTrans 
 */
inline DMagTrans 
operator* (const DMagTrans &a, const DTrans &b)
{
  return DMagTrans (a.t1 * db::DTrans (b.rot (), b.disp () * a.t2.factor ()), DMag (a.t2.factor ()));
}
 
/**
 *  @brief A special implementation for operator* on MagTrans 
 */
inline DMagTrans 
operator* (const DMagTrans &a, const DMag &b)
{
  return DMagTrans (a.t1, a.t2 * b);
}
 
/**
 *  @brief A special implementation for operator* on MagTrans 
 */
inline MagTrans 
operator* (const DMagTrans &a, const MagTrans &b)
{
  return MagTrans (a.t1 * db::DTrans (b.t1.rot (), b.t1.disp () * a.t2.factor ()), Mag (a.t2.factor () * b.t2.factor ()));
}
 
/**
 *  @brief Transformation operator
 *
 *  Transforms the point with the given transformation
 *  (q = T * p).
 *
 *  @param t The transformation to apply
 *  @param p The point to transform
 *  @return The transformed point
 */
template <class C, class Tr> 
inline point<typename Tr::target_coord_type> 
operator* (const Tr &t, const point<C> &p)
{
  return t.operator() (p);
}

/**
 *  @brief Transformation operator with fixpoint transformation
 *
 *  Transforms the point with the given transformation
 *  (q = T * p).
 *
 *  @param t The transformation to apply
 *  @param p The point to transform
 *  @return The transformed point
 */
template <class C> 
inline point<C> 
operator* (const fixpoint_trans &t, const point<C> &p)
{
  return t.operator() (p);
}

/**
 *  @brief A special implementation for inverted MagTrans 
 */
inline db::MagTrans 
inverted (const db::MagTrans &t)
{
  db::MagTrans::inverse_trans tinv (t.inverted ());
  return db::MagTrans (db::DTrans (tinv.t2.rot (), tinv.t2.disp () * tinv.t1.factor ()), db::Mag (tinv.t1.factor ()));
}

/**
 *  @brief A special implementation for inverted DMagTrans 
 */
inline db::DMagTrans 
inverted (const db::DMagTrans &t)
{
  db::DMagTrans::inverse_trans tinv (t.inverted ());
  return db::DMagTrans (db::DTrans (tinv.t2.rot (), tinv.t2.disp () * tinv.t1.factor ()), db::DMag (tinv.t1.factor ()));
}

/**
 *  @brief Fuzzy compare function for transformation objects
 *
 *  This compare function implements a fuzzy "less" operator.
 *  Fuzzyness refers to the fact, that the magnification part
 *  is compared allowing a certain small difference, as defined
 *  by the coordinate trait's "equal" operator. 
 */
template <class T>
class TransFuzzyLessF 
{
public:
  bool operator() (const T &a, const T &b) const
  {
    typedef typename T::target_coord_type target_coord_type;
    typedef typename T::coord_type coord_type;
    typedef db::coord_traits <target_coord_type> coord_traits;

    //  the base principle is to check two test points for equality. 
    db::point<coord_type> in1 (0, 0);
    db::point<coord_type> in2 (1, 2);

    db::point<target_coord_type> out_a, out_b;
    
    out_a = a (in1); out_b = b (in1);
    if (! coord_traits::equals (out_a.x (), out_b.x ())) {
      return out_a.x () < out_b.x ();
    }
    if (! coord_traits::equals (out_a.y (), out_b.y ())) {
      return out_a.y () < out_b.y ();
    }

    out_a = a (in2); out_b = b (in2);
    if (! coord_traits::equals (out_a.x (), out_b.x ())) {
      return out_a.x () < out_b.x ();
    }
    if (! coord_traits::equals (out_a.y (), out_b.y ())) {
      return out_a.y () < out_b.y ();
    }

    return false;
  }
};

/**
 *  @brief Fuzzy compare function for transformation objects
 *
 *  This compare function implements a fuzzy "equal" operator.
 *  Fuzzyness refers to the fact, that the magnification part
 *  is compared allowing a certain small difference, as defined
 *  by the coordinate trait's "equal" operator. 
 */
template <class T>
class TransFuzzyEqualF 
{
public:
  bool operator() (const T &a, const T &b) const
  {
    typedef typename T::target_coord_type target_coord_type;
    typedef typename T::coord_type coord_type;
    typedef db::coord_traits <coord_type> coord_traits;

    //  the base principle is to check two test points for equality. 
    db::point<coord_type> in1 (0, 0);
    db::point<coord_type> in2 (1, 2);

    db::point<target_coord_type> out_a, out_b;
    
    out_a = a (in1); out_b = b (in1);
    if (! coord_traits::equals (out_a.x (), out_b.x ())) {
      return false;
    }
    if (! coord_traits::equals (out_a.y (), out_b.y ())) {
      return false;
    }

    out_a = a (in2); out_b = b (in2);
    if (! coord_traits::equals (out_a.x (), out_b.x ())) {
      return false;
    }
    if (! coord_traits::equals (out_a.y (), out_b.y ())) {
      return false;
    }

    return true;
  }
};

} // namespace db

/**
 *  @brief Special extractors for the transformations
 */

namespace tl 
{

  template<> void extractor_impl (tl::Extractor &ex, db::FTrans &t);
  template<> void extractor_impl (tl::Extractor &ex, db::Trans &t);
  template<> void extractor_impl (tl::Extractor &ex, db::DTrans &t);
  template<> void extractor_impl (tl::Extractor &ex, db::Mag &t);
  template<> void extractor_impl (tl::Extractor &ex, db::DMag &t);
  template<> void extractor_impl (tl::Extractor &ex, db::MagTrans &t);
  template<> void extractor_impl (tl::Extractor &ex, db::DMagTrans &t);

  template<> bool test_extractor_impl (tl::Extractor &ex, db::FTrans &t);
  template<> bool test_extractor_impl (tl::Extractor &ex, db::Trans &t);
  template<> bool test_extractor_impl (tl::Extractor &ex, db::DTrans &t);
  template<> bool test_extractor_impl (tl::Extractor &ex, db::Mag &t);
  template<> bool test_extractor_impl (tl::Extractor &ex, db::DMag &t);
  template<> bool test_extractor_impl (tl::Extractor &ex, db::MagTrans &t);
  template<> bool test_extractor_impl (tl::Extractor &ex, db::DMagTrans &t);

} // namespace tl

#endif

