
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtInstService.h"
#include "edtUtils.h"
#include "edtInstPropertiesPage.h"
#include "layLayoutView.h"
#include "tlProgress.h"

namespace edt
{

// -------------------------------------------------------------

//  search range for select in pixels
unsigned int inst_search_range = 5;   // @@@ make variable?

//  max. number of tries in single-click selection before giving up
int inst_point_sel_tests = 10000;

// -------------------------------------------------------------

class InstFinder
  : public edt::Finder
{
public:
  struct StopException { };

  typedef std::vector<edt::InstService::InstInstPath> founds_vector_type;
  typedef founds_vector_type::const_iterator iterator;

  InstFinder (bool point_mode)
    : Finder (point_mode), m_progress (tl::translate ("Selecting ..."))
  {
    m_progress.set_unit (1000);
    m_progress.set_format ("");
  }

  bool find (lay::LayoutView *view, unsigned int cv_index, const db::MagTrans &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, int min_level, int max_level)
  {
    m_cv_index = cv_index;
    m_tries = inst_point_sel_tests;
    mp_view = view;

    //  actually find
    try {
      start (view, cv_index, trans, layout, region, cell, min_level, max_level);
    } catch (StopException) {
      return false;
    }

    //  return true if anything was found
    return ! m_founds.empty ();
  }

  iterator begin () const
  {
    return m_founds.begin ();
  }

  iterator end () const
  {
    return m_founds.end ();
  }

private:
  virtual void 
  visit_cell (const db::Cell &cell, const db::Box &search_box, const db::MagTrans &t, int level)
  {
    if (! point_mode ()) {

      ++m_progress;

      //  shortcut for cases where nothing can be found
      if (level != max_level () - 1 && (mp_view->hidden_cells ().size () <= m_cv_index || mp_view->hidden_cells () [m_cv_index].empty ())) {
        return;
      }

      //  look for instances to check here ..
      db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
      while (! inst.at_end ()) {

        const db::CellInstArray &cell_inst = *inst;

        ++m_progress;

        //  just consider the instances exactly at the last level of 
        //  hierarchy (this is where the boxes are drawn) or if of cells that
        //  are hidden.
        if (level == max_level () - 1 || mp_view->is_cell_hidden (cell_inst.object ().cell_index (), m_cv_index)) {

          db::box_convert <db::CellInst> bc (layout ());
          for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
          
            ++m_progress;

            const db::DBox box = db::MagTrans (*p, cell_inst.mag ()) * layout ().cell (cell_inst.object ().cell_index ()).bbox ();

            //  in box mode, just test the boxes
            if (box.inside (db::DBox (search_box))) {

              //  in point mode just store that found that has the least "distance"
              m_founds.push_back (edt::InstService::InstInstPath ());
              m_founds.back ().cv_index = m_cv_index;
              m_founds.back ().cell = &cell;
              m_founds.back ().path = path ();
              m_founds.back ().trans = t;
              m_founds.back ().cell_inst = &cell_inst;
              m_founds.back ().array_inst = p;

            }

          }

        }

        ++inst;

      }

    } else {

      if (--m_tries < 0) {
        throw StopException ();
      }

      //  look for instances to check here ..
      db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
      while (! inst.at_end ()) {

        if (--m_tries < 0) {
          throw StopException ();
        }

        const db::CellInstArray &cell_inst = *inst;

        //  just consider the instances exactly at the last level of 
        //  hierarchy (this is where the boxes are drawn) or if of cells that
        //  are hidden.
        if (level == max_level () - 1 || mp_view->is_cell_hidden (cell_inst.object ().cell_index (), m_cv_index)) {

          db::box_convert <db::CellInst> bc (layout ());
          for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
          
            if (--m_tries < 0) {
              throw StopException ();
            }

            bool match = false;
            double d = std::numeric_limits<double>::max ();

            const db::DBox box = db::MagTrans (*p, cell_inst.mag ()) * layout ().cell (cell_inst.object ().cell_index ()).bbox ();

            //  convert to polygon and test those edges
            db::DPolygon poly (box);
            for (db::DPolygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
              test_edge (t * *e, d, match);
            }

            if (! match && box.contains (search_box.center ())) {
              d = t.ctrans (poly_dist (poly.begin_edge (), db::DPoint (search_box.center ()))); 
              match = true;
            }

            if (match && closer (d)) {

              //  in point mode just store that found that has the least "distance"
              if (m_founds.empty ()) {
                m_founds.push_back (edt::InstService::InstInstPath ());
              }

              m_founds.back ().cv_index = m_cv_index;
              m_founds.back ().cell = &cell;
              m_founds.back ().path = path ();
              m_founds.back ().trans = t;
              m_founds.back ().cell_inst = &cell_inst;
              m_founds.back ().array_inst = p;

            }

          }

        }

        ++inst;

      }

    }

  }

  unsigned int m_cv_index;
  std::vector<edt::InstService::InstInstPath> m_founds;
  int m_tries;
  tl::Progress m_progress;
  lay::LayoutView *mp_view;
};


// -------------------------------------------------------------

InstService::InstService (db::Manager *manager, lay::LayoutView *view)
  : lay::ViewService (view->canvas ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    m_min_level (0), m_max_level (0)
{ 
  //  .. nothing yet ..
}

InstService::~InstService ()
{
  for (std::vector<lay::Marker *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.clear ();
}

void 
InstService::clear_highlights ()
{
  for (std::vector<lay::Marker *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (false);
  }
}

void 
InstService::restore_highlights ()
{
  for (std::vector<lay::Marker *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (true);
  }
}

void 
InstService::highlight (unsigned int n)
{
  for (std::vector<lay::Marker *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (n-- == 0);
  }
}

bool 
InstService::has_selection ()
{
  return ! m_selection.empty ();
}

double 
InstService::click_proximity (const db::DPoint &pos)
{
  InstFinder finder (true);

  //  go through all cell views
  std::set< std::pair<db::MagTrans, int> > variants = mp_view->cv_transform_variants ();
  for (std::set< std::pair<db::MagTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {

    const lay::CellView &cv = mp_view->cellview (v->second);
    if (cv.is_valid ()) {

      double dbu = cv->layout ().dbu ();

      //  compute search box
      double l = double (inst_search_range) / (canvas ()->trans () * v->first).t2.factor ();
      db::Box search_box = db::Box::from_double (db::DMag (1.0 / dbu) * db::DBox (pos, pos).enlarged (db::DPoint (l, l)));

      finder.find (mp_view, v->second, v->first, cv->layout (), search_box, *cv.cell (), int (m_min_level), int (m_max_level));

    }

  }

  //  Return the finder's proximity value
  if (finder.begin () != finder.end ()) {
    return finder.proximity ();
  } else {
    return lay::Editable::click_proximity (pos); 
  }
}

bool 
InstService::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selection.empty ()) {
      m_selection.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selection.empty ()) {
        m_selection.clear ();
        needs_update = true;
      }
    } else {

      //  extract all insts

      //  @@@ not implemented yet @@@

    }

  } else {

    InstFinder finder (box.is_point ());

    //  go through all cell views
    std::set< std::pair<db::MagTrans, int> > variants = mp_view->cv_transform_variants ();
    for (std::set< std::pair<db::MagTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {

      const lay::CellView &cv = mp_view->cellview (v->second);
      if (cv.is_valid ()) {

        double dbu = cv->layout ().dbu ();

        //  compute search box
        double l = double (inst_search_range) / (canvas ()->trans () * v->first).t2.factor ();
        db::Box search_box = db::Box::from_double (db::DMag (1.0 / dbu) * box.enlarged (db::DPoint (l, l)));

        finder.find (mp_view, v->second, v->first, cv->layout (), search_box, *cv.cell (), int (m_min_level), int (m_max_level));

      }

    }

    //  collect the founds from the finder
    for (InstFinder::iterator f = finder.begin (); f != finder.end (); ++f) {
      select (*f, mode);
      needs_update = true;
      any_selected = true;
    }
      
  }

  //  if required, update the list of ruler objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  return any_selected;
}

lay::PropertiesPage *
InstService::properties_page (QWidget *parent)
{
  return new edt::InstPropertiesPage (this, parent);
}

void 
InstService::get_selection (std::vector <InstInstPath> &sel) const
{
  sel.clear ();
  sel.reserve (m_selection.size ());

  //  positions will hold a set of iterators that are to be erased
  for (std::set<InstInstPath>::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    sel.push_back (*r);
  }
}

bool 
InstService::configure (const std::string &name, const std::string &value)
{

  // @@@
  return false;

}

bool 
InstService::select (const InstInstPath &obj, lay::Editable::SelectionMode mode)
{
  if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
    //  select
    if (m_selection.find (obj) == m_selection.end ()) {
      m_selection.insert (obj);
      return true;
    }
  } else if (mode == lay::Editable::Reset) {
    //  unselect
    if (m_selection.find (obj) != m_selection.end ()) {
      m_selection.erase (obj);
      return true;
    }
  } else {
    //  invert selection
    if (m_selection.find (obj) != m_selection.end ()) {
      m_selection.erase (obj);
    } else {
      m_selection.insert (obj);
    }
    return true;
  }
  return false;
}

void 
InstService::clear_selection ()
{
  select (db::DBox (), lay::Editable::Reset);
}

void
InstService::hier_levels (std::pair<unsigned int, unsigned int> l)
{
  m_max_level = l.second;
  m_min_level = l.first;
}

void 
InstService::selection_to_view ()
{
  //  the selection objects need to be recreated since we destroyed the old rulers
  for (std::vector<lay::Marker *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.clear ();
  m_markers.reserve (m_selection.size ());
  for (std::set<InstInstPath>::iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    m_markers.push_back (new lay::Marker (mp_view, r->cv_index));
    const lay::CellView &cv = mp_view->cellview (r->cv_index);
    db::box_convert<db::CellInst> bc (cv->layout ());
    m_markers.back ()->set (bc (r->cell_inst->object ()), r->trans * db::MagTrans (*r->array_inst, db::Mag (r->cell_inst->mag ())));
  }
}

// -------------------------------------------------------------

} // namespace edt


