
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_edtInstService
#define HDR_edtInstService

#include "layEditable.h"
#include "layPlugin.h"
#include "layViewObject.h"
#include "layMarker.h"
#include "dbLayout.h"
#include "edtUtils.h"

#include <set>
#include <vector>

namespace edt {

class LayoutView;
class LayoutCanvas;
class Service;
class CellView;

// -------------------------------------------------------------

class InstService
  : public lay::ViewService,
    public lay::Editable,
    public lay::Plugin,
    public db::Object
{
public: 
  /**
   *  @brief A helper class that describes a shape instance completely
   *
   *  It holds these informations: cv_index (the cellview index where the shape
   *  was found), cell (the cell where the shape is located in), path (the instantiation
   *  path of this cell from the cellview's cell), trans (the transformation of the cell 
   *  relative to the cellview's cell), layer (the layer where the shape is found as an
   *  index in the layer table) and shape (the shape reference itself).
   *  
   *  trans and cell are redundant - in principle this information can be
   *  derived from the path. It is kept for performance reasons however.
   */
  struct InstInstPath
  {
    unsigned int cv_index;
    const db::Cell *cell;
    tl::vector<edt::InstElement> path;
    db::MagTrans trans;
    const db::CellInstArray *cell_inst;
    db::CellInstArray::iterator array_inst;


    bool operator< (const InstInstPath &d) const
    {
      if (cv_index != d.cv_index) {
        return cv_index < d.cv_index;
      }
      if (cell != d.cell) {
        return cell < d.cell;
      }
      if (cell_inst != d.cell_inst) {
        return cell_inst < d.cell_inst;
      }
      if (*array_inst != *d.array_inst) {
        return *array_inst < *d.array_inst;
      }
      //  at last, compare the path
      return path < d.path;
    }

    bool operator== (const InstInstPath &d) const
    {
      return cv_index == d.cv_index &&
             cell == d.cell &&
             cell_inst == d.cell_inst &&
             *array_inst == *d.array_inst && 
             path == d.path;
    }

    bool operator!= (const InstInstPath &d) const
    {
      return !operator== (d);
    }
  };

  typedef std::set<InstInstPath> objects;
  typedef objects::const_iterator obj_iterator;

  /**
   *  @brief The constructor
   */
  InstService (db::Manager *manager, lay::LayoutView *view);

  /**
   *  @brief The destructor
   */
  ~InstService ();

  /** 
   *  @brief Clear all highlights (for current object highlighting)
   */
  void clear_highlights ();

  /** 
   *  @brief Restore all highlights (for current object highlighting)
   */
  void restore_highlights ();

  /** 
   *  @brief Highlight a certain object
   */
  void highlight (unsigned int n);

  /**
   *  @brief Tell if anything is selected
   */
  virtual bool has_selection ();

  /** 
   *  @brief "select" operation
   */
  virtual double click_proximity (const db::DPoint &pos);

  /** 
   *  @brief "select" operation
   */
  virtual bool select (const db::DBox &box, lay::Editable::SelectionMode mode);

  /**
   *  @brief Create the properties page
   */
  virtual lay::PropertiesPage *properties_page (QWidget * /*parent*/);

  /**
   *  @brief Get the selection for the properties page
   */
  void get_selection (std::vector <InstInstPath> &selection) const;

  /**
   *  @brief Implementation of "Plugin" interface
   */
  bool configure (const std::string &name, const std::string &value);

  /**
   *  @brief Receive notifications about hierarchy levels
   */
  void hier_levels (std::pair<unsigned int, unsigned int> l);
  
  /**
   *  @brief Color accessor
   */
  QColor color () const
  {
    return m_color;
  }

  /**
   *  @brief Obtain the lay::ViewService interface
   */
  lay::ViewService *view_service_interface ()
  {
    return this;
  }

  /**
   *  @brief Obtain the lay::Editable interface
   */
  lay::Editable *editable_interface ()
  {
    return this;
  }

  /**
   *  @brief Get the selection containter
   */
  const objects &selection () const
  {
    return m_selection;
  }

  /**
   *  @brief Access to the view object
   */
  lay::LayoutView *view () const
  {
    return mp_view;
  }

private:
  //  Ruler display and snapping configuration
  QColor m_color;

  //  The layout view that the editor service is attached to
  lay::LayoutView *mp_view;

  //  The hierarchy levels shown
  unsigned int m_min_level, m_max_level;

  //  The marker objects representing the selection
  std::vector<lay::Marker *> m_markers;

  //  The selection
  objects m_selection;

  /**
   *  @brief Select a certain object
   *
   *  @return true, if the selection has changed
   */
  bool select (const InstInstPath &obj, lay::Editable::SelectionMode mode);

  /**
   *  @brief Clear the selection
   */
  void clear_selection ();

  /**
   *  @brief Update m_markers to reflect the selection
   */
  void selection_to_view ();
};

}

#endif

