
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtPropertiesPages.h"
#include "edtUtils.h"
#include "layLayoutView.h"
#include "tlString.h"

#include <string>

namespace edt
{

// -------------------------------------------------------------------------
//  ShapePropertiesPage implementation

ShapePropertiesPage::ShapePropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), 
    mp_service (service), m_enable_cb_callback (true)
{
  m_pos = service->selection ().begin ();
  m_index = 0;
  mp_service->clear_highlights ();
}

ShapePropertiesPage::~ShapePropertiesPage ()
{
  mp_service->restore_highlights ();
}

void 
ShapePropertiesPage::setup ()
{
  connect (dbu_checkbox (), SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_checkbox (), SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
}

void 
ShapePropertiesPage::back ()
{
  m_pos = mp_service->selection ().end ();
  m_index = mp_service->selection ().size ();
}

void 
ShapePropertiesPage::front ()
{
  m_pos = mp_service->selection ().begin ();
  m_index = 0;
}

bool 
ShapePropertiesPage::at_begin () const
{
  return (m_pos == mp_service->selection ().begin ());
}

bool 
ShapePropertiesPage::at_end () const
{
  return (m_pos == mp_service->selection ().end ());
}

void 
ShapePropertiesPage::operator-- ()
{
  --m_pos;
  --m_index;
}

void 
ShapePropertiesPage::operator++ ()
{
  ++m_pos;
  ++m_index;
}

void 
ShapePropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

bool 
ShapePropertiesPage::dbu_units () const
{
  return dbu_checkbox ()->isChecked ();
}

bool 
ShapePropertiesPage::abs_trans () const
{
  return abs_checkbox ()->isChecked ();
}

db::MagTrans
ShapePropertiesPage::trans () const
{
  db::MagTrans t;
  if (abs_trans ()) {
    for (tl::vector<edt::InstElement>::const_iterator p = m_pos->path.begin (); p != m_pos->path.end (); ++p) {
      t = t * db::MagTrans (*(p->array_inst), db::Mag (p->cell_inst->mag ()));
    }
  }
  return t;
}

void 
ShapePropertiesPage::display_mode_changed (bool)
{
  if (! m_enable_cb_callback) {
    return;
  }

  const lay::CellView &cv = mp_service->view ()->cellview (m_pos->cv_index);
  double dbu = cv->layout ().dbu ();

  do_update (m_pos->shape, dbu);
}

void 
ShapePropertiesPage::update ()
{
  mp_service->highlight (m_index);

  m_enable_cb_callback = false;
  dbu_checkbox ()->setChecked (mp_service->view ()->dbu_coordinates ());
  abs_checkbox ()->setChecked (mp_service->view ()->absolute_coordinates ());
  m_enable_cb_callback = true;

  const lay::CellView &cv = mp_service->view ()->cellview (m_pos->cv_index);
  double dbu = cv->layout ().dbu ();

  do_update (m_pos->shape, dbu);
}

void
ShapePropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), m_pos->cv_index, m_pos->path);
}

void
ShapePropertiesPage::show_props ()
{
  UserPropertiesForm props_form (this);
  props_form.show (mp_service->view (), m_pos->cv_index, m_pos->shape);
}

// -------------------------------------------------------------------------
//  helper functions to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

static std::string
coords_to_string (const db::DPoint &dp, double dbu, bool du, const char *sep = "\t")
{
  return coord_to_string (dp.x (), dbu, du) + sep + coord_to_string (dp.y (), dbu, du);
}

// -------------------------------------------------------------------------
//  PolygonPropertiesPage implementation

PolygonPropertiesPage::PolygonPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
PolygonPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Polygon poly;
  shape.polygon (poly);

  std::string ptlist;
  ptlist.reserve (4096);

  db::MagTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Polygon::polygon_contour_iterator pt = poly.begin_hull (); pt != poly.end_hull (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }

  pointListEdit->setText (ptlist.c_str ());
}

bool 
PolygonPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
PolygonPropertiesPage::apply ()
{

  //  @@@

}

// -------------------------------------------------------------------------
//  BoxPropertiesPage implementation

BoxPropertiesPage::BoxPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
BoxPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Box box;
  shape.box (box);

  db::MagTrans t = trans ();
  bool du = dbu_units ();

  db::DBox dbox (t * box);
  x1_le->setText (coord_to_string (dbox.left (), dbu, du).c_str ());
  y1_le->setText (coord_to_string (dbox.bottom (), dbu, du).c_str ());
  x2_le->setText (coord_to_string (dbox.right (), dbu, du).c_str ());
  y2_le->setText (coord_to_string (dbox.top (), dbu, du).c_str ());
  w_le->setText (coord_to_string (dbox.width (), dbu, du).c_str ());
  h_le->setText (coord_to_string (dbox.height (), dbu, du).c_str ());
}

bool 
BoxPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
BoxPropertiesPage::apply ()
{

  //  @@@

}

// -------------------------------------------------------------------------
//  TextPropertiesPage implementation

TextPropertiesPage::TextPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
TextPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Text text;
  shape.text (text);

  db::MagTrans t = trans ();
  bool du = dbu_units ();
  db::DPoint dp = t * text.trans ().disp ();

  text_le->setText (text.string ().c_str ());
  x_le->setText (coord_to_string (dp.x (), dbu, du).c_str ());
  y_le->setText (coord_to_string (dp.y (), dbu, du).c_str ());
}

bool 
TextPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
TextPropertiesPage::apply ()
{

  //  @@@

}

// -------------------------------------------------------------------------
//  PathPropertiesPage implementation

PathPropertiesPage::PathPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
PathPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Path path;
  shape.path (path);

  std::string ptlist;
  ptlist.reserve (4096);

  db::MagTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Path::iterator pt = path.begin (); pt != path.end (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }
  ptlist_le->setText (ptlist.c_str ());

  width_le->setText (coord_to_string (t.ctrans (path.width ()), dbu, du).c_str ());
  start_ext_le->setText (coord_to_string (t.ctrans (path.extensions ().first), dbu, du).c_str ());
  end_ext_le->setText (coord_to_string (t.ctrans (path.extensions ().second), dbu, du).c_str ());
}

bool 
PathPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
PathPropertiesPage::apply ()
{

  //  @@@

}

}

