
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtUtils.h"
#include "layCellView.h"
#include "layLayoutView.h"
#include "layMarker.h"

#include <dbLayout.h>

namespace edt {

// ----------------------------------------------------------------------
//  edt::Finder implementation

Finder::Finder (bool point_mode)
  : m_min_level (0), m_max_level (0),
    mp_layout (0), m_point_mode (point_mode)
{
  m_distance = std::numeric_limits<double>::max ();
}

Finder::~Finder ()
{
  //  .. nothing yet ..
}

bool 
Finder::closer (double d)
{
  if (d <= m_distance) {
    m_distance = d;
    return true;
  } else {
    return false;
  }
}

void 
Finder::start (lay::LayoutView *view, unsigned int cv_index, const db::MagTrans &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, int min_level, int max_level, int layer)
{
  m_layer = (unsigned int) layer;
  m_region = region; 
  mp_layout = &layout;
  mp_view = view;
  m_cv_index = cv_index;
  m_min_level = min_level;
  m_max_level = max_level;

  if (layer >= 0) {
    m_box_convert = db::box_convert <db::CellInst> (layout, (unsigned int) layer);
    m_cell_box_convert = db::box_convert <db::Cell> ((unsigned int) layer);
  } else {
    m_box_convert = db::box_convert <db::CellInst> (layout);
    m_cell_box_convert = db::box_convert <db::Cell> ();
  }

  m_path.erase (m_path.begin (), m_path.end ());

  do_find (cell, 0, trans);
}

void
Finder::test_edge (const db::DEdge &edg, double &distance, bool &match)
{
  //  we hit the region with the edge end points - take the closest vertex
  if (m_region.contains (edg.p1 ()) || m_region.contains (edg.p2 ())) {
    double d = std::min (edg.p1 ().distance (m_region.center ()), 
                         edg.p2 ().distance (m_region.center ()));
    if (! match || d < distance) {
      distance = d;
    }
    match = true;
  }
  
  //  if the edge cuts through the active region: test the
  //  edge as a whole
  if (edg.clipped (m_region).first) {
    double d = edg.distance_abs (m_region.center ());
    if (! match || d < distance) {
      distance = d;
    }
    match = true;
  }
}

void
Finder::do_find (const db::Cell &cell, int level, const db::MagTrans &t)
{
  if (level >= m_max_level || ! (t * m_cell_box_convert (cell)).touches (m_region) || mp_view->is_cell_hidden (cell.cell_index (), m_cv_index)) {
    return;
  }

  db::Box touch_box (t.inverted () * m_region); 

  if (level >= m_min_level) {
    visit_cell (cell, touch_box, t, level);
  }

  db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
  while (! inst.at_end ()) {

    const db::CellInstArray &cell_inst = *inst;
    for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, m_box_convert); ! p.at_end (); ++p) {

      m_path.push_back (edt::InstElement (cell_inst, p));

      do_find (mp_layout->cell (cell_inst.object ().cell_index ()), 
               level + 1,
               t * db::MagTrans (*p, cell_inst.mag ()));

      m_path.pop_back ();

    }

    ++inst;

  }

}

// ----------------------------------------------------------------------
//  edt::InstantiationForm implementation

InstantiationForm::InstantiationForm (QWidget *parent)
  : QDialog (parent), mp_view (0), m_cv_index (-1), mp_path (0), mp_marker (0), m_enable_cb_callbacks (false)
{
  Ui::InstantiationForm::setupUi (this);

  connect (list, SIGNAL (itemDoubleClicked (QListWidgetItem *)), this, SLOT (double_clicked (QListWidgetItem *)));
  connect (dbu_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
}

InstantiationForm::~InstantiationForm ()
{
  if (mp_marker) {
    delete mp_marker;
    mp_marker = 0;
  }
}

void 
InstantiationForm::display_mode_changed (bool)
{
  if (! m_enable_cb_callbacks) {
    return;
  }

  mp_view->dbu_coordinates (dbu_cb->isChecked ());
  mp_view->absolute_coordinates (abs_cb->isChecked ());

  update ();
}

void 
InstantiationForm::double_clicked (QListWidgetItem *item)
{
  int row = list->row (item);
  if (row < 0) {
    return;
  }

  lay::CellView::cell_path_type path = mp_view->cellview (m_cv_index).path ();
  for (tl::vector<edt::InstElement>::const_iterator p = mp_path->begin (); p != mp_path->begin () + row; ++p) {
    path.push_back (p->cell_inst->object ().cell_index ());
  }

  mp_view->set_current_cell_path (m_cv_index, path);

  if (! mp_marker) {
    mp_marker = new lay::Marker (mp_view, m_cv_index);
  }

  const db::Layout &layout = mp_view->cellview (m_cv_index)->layout ();
  db::Box box;
  if (row == 0) {
    box = mp_view->cellview (m_cv_index).cell ()->bbox ();
  } else {
    box = layout.cell (mp_path->begin ()[row - 1].cell_inst->object ().cell_index ()).bbox ();
  }

  //  TODO: this does not consider global transformation and variants of this
  db::MagTrans abs_trans;
  for (tl::vector<edt::InstElement>::const_iterator p = mp_path->begin (); p != mp_path->begin () + row; ++p) {
    db::MagTrans trans (*(p->array_inst), db::Mag (p->cell_inst->mag ()));
    abs_trans = abs_trans * trans;
  }

  mp_marker->set (box, abs_trans, mp_view->cv_transform_variants (m_cv_index));
}

void 
InstantiationForm::show (lay::LayoutView *view, unsigned int cv_index, const tl::vector<edt::InstElement> &path)
{
  mp_view = view;
  m_cv_index = cv_index;
  mp_path = &path;

  m_enable_cb_callbacks = false;
  dbu_cb->setChecked (mp_view->dbu_coordinates ());
  abs_cb->setChecked (mp_view->absolute_coordinates ());
  m_enable_cb_callbacks = true;

  update ();
  exec ();

  mp_view = 0;
  mp_path = 0;
  m_cv_index = -1;
}

void 
InstantiationForm::update ()
{
  bool dbu_coord = dbu_cb->isChecked ();
  bool abs_coord = abs_cb->isChecked ();

  const lay::CellView &cv = mp_view->cellview (m_cv_index);
  double dbu = cv->layout ().dbu ();

  layout_le->setText (cv->name ().c_str ());

  list->clear ();

  list->addItem (cv->layout ().cell_name (cv.cell_index ()));
  db::MagTrans abs_trans;

  for (tl::vector<edt::InstElement>::const_iterator p = mp_path->begin (); p != mp_path->end (); ++p) {

    //  build the instance information from the path

    db::MagTrans trans (*(p->array_inst), db::Mag (p->cell_inst->mag ()));
    abs_trans = abs_trans * trans;

    if (abs_coord) {
      trans = abs_trans;
    }

    std::string line;
    line += cv->layout ().cell_name (p->cell_inst->object ().cell_index ());
    line += "\tat ";
    if (! dbu_coord) {
      line += tl::micron_to_string (trans.t1.disp ().x () * dbu);
      line += ",";
      line += tl::micron_to_string (trans.t1.disp ().y () * dbu);
    } else {
      line += tl::db_to_string (trans.t1.disp ().x ());
      line += ",";
      line += tl::db_to_string (trans.t1.disp ().y ());
    }

    line += " ";
    line += trans.t1.fp_trans ().to_string ();

    if (fabs (trans.t2.factor () - 1.0) > 1e-9) {
      line += tl::sprintf (" *%.9g", trans.t2.factor ());
    }

    list->addItem (line.c_str ());
  }
}

// ----------------------------------------------------------------------
//  edt::UserPropertiesForm implementation

UserPropertiesForm::UserPropertiesForm (QWidget *parent)
  : QDialog (parent)
{
  Ui::UserPropertiesForm::setupUi (this);
}

void 
UserPropertiesForm::show (lay::LayoutView *view, unsigned int cv_index, const db::Shape &shape)
{
  const lay::CellView &cv = view->cellview (cv_index);
  const db::PropertiesRepository &prep = cv->layout ().properties_repository ();

  prop_list->clear ();

  if (shape.has_prop_id ()) {

    const db::PropertiesRepository::properties_set &props = prep.properties (shape.prop_id ());
    for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
      QTreeWidgetItem *entry = new QTreeWidgetItem (prop_list);
      entry->setText (0, prep.prop_name (p->first).get_string ().c_str ());
      entry->setText (1, p->second.get_string ().c_str ());
    }

  }

  exec ();
}

}

