
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layCellView.h"
#include "dbLayout.h"
#include "tlLog.h"

#include <algorithm>

namespace lay
{

// -------------------------------------------------------------
//  LayoutHandle implementation


LayoutHandle::LayoutHandle (db::Layout *layout, const std::string &filename)
  : mp_layout (layout),
    m_ref_count (0),
    m_filename (filename)
{
  static int nn = 0;

  std::string n;
  do {
    n = tl::sprintf ("L%d", ++nn);
  } while (find (n) != 0);

  m_name = n;
  ms_dict.insert (std::make_pair (n, this));

  if (tl::verbosity () >= 30) {
    tl::info << "Created layout " << name ();
  }
}

LayoutHandle::~LayoutHandle ()
{
  if (tl::verbosity () >= 30) {
    tl::info << "Deleted layout " << name ();
  }

  delete mp_layout;
  mp_layout = 0;

  ms_dict.erase (m_name);
}

void 
LayoutHandle::rename (const std::string &name) throw (tl::Exception)
{
  std::string n (name);
  int nn = 0;

  if (n != m_name) {
    do {
      if (find (n) == 0) {
        ms_dict.erase (m_name);
        if (tl::verbosity () >= 40) {
          tl::info << "Renamed layout from " << m_name << " to " << n;
        }
        m_name = n;
        ms_dict.insert (std::make_pair (n, this));
        return;
      }
      n = name + tl::sprintf ("[%d]", ++nn);
    } while (1);
  }
}

db::Layout &
LayoutHandle::layout () const
{
  return *mp_layout;
}

const std::string &
LayoutHandle::filename () const
{
  return m_filename;
}

const std::string &
LayoutHandle::name () const
{
  return m_name;
}

void 
LayoutHandle::add_ref ()
{
  if (tl::verbosity () >= 40) {
    tl::info << "Add reference to " << m_name;
  }
  ++m_ref_count;
}

void 
LayoutHandle::remove_ref ()
{
  if (tl::verbosity () >= 40) {
    tl::info << "Remove reference from " << m_name;
  }
  if (--m_ref_count <= 0) {
    //  not nice, but hopefully we can do so:
    delete this;
  }
}

std::map <std::string, LayoutHandle *> LayoutHandle::ms_dict;

LayoutHandle *
LayoutHandle::find (const std::string &name)
{
  std::map <std::string, LayoutHandle *>::const_iterator h = ms_dict.find (name);
  if (h == ms_dict.end ()) {
    return 0;
  } else {
    return h->second;
  }
}

void 
LayoutHandle::get_names (std::vector <std::string> &names)
{
  names.clear ();
  names.reserve (ms_dict.size ());
  for (std::map <std::string, LayoutHandle *>::const_iterator h = ms_dict.begin (); h != ms_dict.end (); ++h) {
    names.push_back (h->first);
  }
}

// -------------------------------------------------------------
//  LayoutHandleRef implementation

LayoutHandleRef::LayoutHandleRef ()
  : mp_handle (0)
{
  // .. nothing yet ..
}

LayoutHandleRef::LayoutHandleRef (LayoutHandle *h)
  : mp_handle (0)
{
  set (h);
}

LayoutHandleRef::LayoutHandleRef (const LayoutHandleRef &r)
  : mp_handle (0)
{
  set (r.mp_handle);
}

LayoutHandleRef::~LayoutHandleRef ()
{
  set (0);
}

bool 
LayoutHandleRef::operator== (const LayoutHandleRef &r) const
{
  return mp_handle == r.mp_handle;
}

LayoutHandleRef &
LayoutHandleRef::operator= (const LayoutHandleRef &r)
{
  if (&r != this) {
    set (r.mp_handle);
  }
  return *this;
}

void
LayoutHandleRef::set (LayoutHandle *h)
{
  if (mp_handle) {
    mp_handle->remove_ref ();
    mp_handle = 0;
  } 
  mp_handle = h;
  if (mp_handle) {
    mp_handle->add_ref ();
  }
}

LayoutHandle *
LayoutHandleRef::operator-> () const
{
  return mp_handle;
}

LayoutHandle *
LayoutHandleRef::get () const
{
  return mp_handle;
}

// -------------------------------------------------------------
//  LayoutHandleRef implementation

CellView::CellView () 
  : mp_cell (0), m_cell_index (0), m_deco_layer (-1)
{ }

bool 
CellView::operator== (const CellView &cv) const
{
  return m_layout_href == cv.m_layout_href && mp_cell == cv.mp_cell 
        && m_cell_index == cv.m_cell_index && m_path == cv.m_path;
}

bool 
CellView::is_valid () const
{
  return m_layout_href.get () != 0 && mp_cell != 0;
}

void 
CellView::set_path (const cell_path_type &p)
{
  mp_cell = 0;
  m_cell_index = 0;
  m_path = p;

  if (p.size () > 0 && m_layout_href.get () && p.back () < m_layout_href->layout ().cells ()) {
    m_cell_index = p.back ();
    mp_cell = &m_layout_href->layout ().cell (p.back ());
  }
}

void
CellView::set_cell (cell_index_type index)
{
  db::Layout &layout = m_layout_href->layout ();

  m_cell_index = index;
  mp_cell = &layout.cell (m_cell_index);

  m_path.clear ();
  m_path.push_back (index);

  while (! layout.cell (index).is_top ()) {
    index = *layout.cell (index).begin_parent_cells ();
    m_path.push_back (index);
  }

  std::reverse (m_path.begin (), m_path.end ());
}

void 
CellView::set_cell (const std::string &name)
{
  std::pair<bool, db::Layout::cell_index_type> cp = m_layout_href->layout ().cell_by_name (name.c_str ());
  if (cp.first) {
    set_cell (cp.second);
  } else {
    reset_cell ();
  }
}

void 
CellView::reset_cell ()
{
  mp_cell = 0;
  m_cell_index = 0;
  m_path.clear ();
}

void 
CellView::set (lay::LayoutHandle *handle)
{
  reset_cell ();
  m_layout_href.set (handle);
}


}

