
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <inttypes.h>

#include <string>
#include <algorithm>

#include "EditStipplesFormImpl.h"
#include "SelectStippleFormImpl.h"
#include "layLayerControlPanel.h"
#include "layLayerTreeModel.h"
#include "layColorPalette.h"
#include "layViewOp.h"
#include "layLayoutCanvas.h"
#include "tlInternational.h"

#include <QtGui/QVBoxLayout>
#include <QtGui/QHBoxLayout>
#include <QtGui/QImage>
#include <QtGui/QInputDialog>
#include <QtGui/QColorDialog>
#include <QtGui/QLineEdit>
#include <QtGui/QCheckBox>
#include <QtGui/QBitmap>
#include <QtGui/QPixmap>
#include <QtGui/QTreeView>
#include <QtGui/QHeaderView>
#include <QtGui/QMenu>
#include <QtGui/QToolButton>
#include <QtGui/QMessageBox>
#include <QtGui/QItemDelegate>

namespace lay
{

// --------------------------------------------------------------------
//  LCPRemitter implementation

LCPRemitter::LCPRemitter (int index, QObject *parent, const char *name)
  : QObject (parent, name),
    m_index (index)
{
  // .. 
}

void 
LCPRemitter::the_slot ()
{
  emit the_signal (m_index);
}

// --------------------------------------------------------------------
//  LCPActiveLabel implementation

LCPActiveLabel::LCPActiveLabel (int index, QWidget *parent)
  : LabelWithBackground (parent), m_index (index), m_grabbed (false)
{
  setFrameStyle (QFrame::Panel | QFrame::Raised);
  setBackgroundRole (QPalette::Window);
}

void 
LCPActiveLabel::mousePressEvent (QMouseEvent *e)
{
  if (! m_grabbed && e->button () == Qt::LeftButton) {
    setFrameShadow (QFrame::Sunken);
    m_grabbed = true;
    grabMouse ();
  }
}

void 
LCPActiveLabel::mouseReleaseEvent (QMouseEvent *e)
{
  if (m_grabbed) {
   
    setFrameShadow (QFrame::Raised);
    releaseMouse ();
    m_grabbed = false;

    if (e->button () == Qt::LeftButton && rect ().contains (e->pos ())) {
      emit clicked (m_index);
    }

  }
}

// --------------------------------------------------------------------
//  LCPDitherPalette implementation

LCPDitherPalette::LCPDitherPalette (QWidget *parent, lay::LayoutView *view, const char *name)
  : QFrame (parent, name), mp_view (view)
{
  const lay::DitherPattern &pattern = view->dither_pattern ();

  int n = 0;
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  for (unsigned int i = 0; i < 4; ++i) {

    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    for (unsigned int j = 0; j < 4; ++j) {
      
      LCPActiveLabel *b = new LCPActiveLabel (n, f);

      b->setMinimumSize (28, 28);
      b->setAlignment (Qt::AlignCenter);
      b->setLineWidth (1);
      b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));

      color_t color0 = b->paletteBackgroundColor ().rgb ();
      color_t color1 = b->paletteForegroundColor ().rgb ();

      const unsigned int h = 24;
      const unsigned int w = 24;

      QImage image (w, h, 32);
      const uint32_t *dp = pattern.pattern (n);
      for (unsigned int l = 0; l < h; ++l, ++dp) {
        uint32_t m = *dp;
        if (l == 0 || l == h - 1) { 
          m |= ((1 << w) - 1);
        } else {
          m |= ((1 << (w - 1)) | 1);
        }
        color_t *pt = (color_t *) image.scanLine (h - 1 - l);
        for (unsigned int b = 0; b < w; ++b) {
          *pt++ = (m & 1) ? color1 : color0;
          m >>= 1;
        }
      }
      b->setPixmap (QPixmap (image));

      connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));

      ll->addWidget (b);

      ++n;

    }

  }

  {
    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    LCPActiveLabel *b;

    //  No pattern
    b = new LCPActiveLabel (-3, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("None").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    //  More pattern 
    b = new LCPActiveLabel (-2, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("More ..").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);
  }

  {
    //  Edit pattern 
    LCPActiveLabel *b;
    b = new LCPActiveLabel (-1, this);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("Custom Pattern ..").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    l->addWidget (b);
  }

}

void 
LCPDitherPalette::button_clicked (int index)
{
  if (index == -1) {
    
    //  edit pattern
    lay::DitherPattern pattern (mp_view->dither_pattern ());
    EditStipplesFormImpl stipples_form (0, pattern);
    if (stipples_form.exec () && stipples_form.pattern () != pattern) {
      emit pattern_changed (stipples_form.pattern ());
    }

  } else if (index == -2) {
    
    //  edit pattern
    SelectStippleFormImpl stipples_form (0, mp_view->dither_pattern ());
    if (stipples_form.exec () && stipples_form.selected () >= 0) {
      emit dither_selected (int (stipples_form.selected ()));
    }

  } else if (index == -3) {
    
    emit dither_selected (-1);

  } else {
    emit dither_selected (index);
  }
}

// --------------------------------------------------------------------
//  LCPVisibilityPalette implementation

LCPVisibilityPalette::LCPVisibilityPalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  QFrame *f;
  LCPActiveLabel *b;
  QHBoxLayout *ll;

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (0, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Show").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (1, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Hide").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (2, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Transp.").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (3, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Opaque").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);
}

void 
LCPVisibilityPalette::button_clicked (int index)
{
  if (index == 0) {
    emit visibility_change (true);
  } else if (index == 1) {
    emit visibility_change (false);
  } else if (index == 2) {
    emit transparency_change (true);
  } else if (index == 3) {
    emit transparency_change (false);
  }
}

// --------------------------------------------------------------------
//  LCPAnimationPalette implementation

LCPAnimationPalette::LCPAnimationPalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  QFrame *f;
  LCPActiveLabel *b;
  QHBoxLayout *ll;

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (0, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("None").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (1, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Scroll").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (2, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Blink").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (3, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("/Blink").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);
}

void 
LCPAnimationPalette::button_clicked (int index)
{
  emit animation_selected (index);
}

// --------------------------------------------------------------------
//  LCPStylePalette implementation

LCPStylePalette::LCPStylePalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  QFrame *f;
  LCPActiveLabel *b;
  QHBoxLayout *ll;

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (0, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("0px").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (1, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("1px").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (2, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText ("2px");
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (3, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("3px").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (100, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Simple").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (101, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Marked").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);
}

void 
LCPStylePalette::button_clicked (int index)
{
  if (index >= 0 && index < 16) {
    emit width_selected (index);
  } else if (index == 100) {
    emit marked_selected (false);
  } else if (index == 101) {
    emit marked_selected (true);
  }
}

// --------------------------------------------------------------------
//  LCPColorPalette implementation

LCPColorPalette::LCPColorPalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);
  for (unsigned int i = 0; i < 6; ++i) {

    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    for (unsigned int j = 0; j < 7; ++j) {
      
      unsigned int n = j * 6 + i;

      LCPActiveLabel *b = new LCPActiveLabel (n, f);
      while (m_color_buttons.size () <= n) {
        m_color_buttons.push_back (0);
      }
      m_color_buttons [n] = b;
      b->setMinimumSize (16, 16);
      b->setLineWidth (1);
      b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
      b->setText ("");
      connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));

      ll->addWidget (b);

    }

  }

  {
    //  darker and brighter colors
    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    LCPActiveLabel *b;

    //  No color
    b = new LCPActiveLabel (-1, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("None").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    //  More colors ..
    b = new LCPActiveLabel (-2, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("More ..").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);
  }

  {
    //  darker and brighter colors
    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    LCPActiveLabel *b;

    QLabel *lbl = new QLabel (tl::translate ("S/V").c_str (), f);
    ll->addWidget (lbl);

    b = new LCPActiveLabel (-10, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setPixmap (QPixmap (":dark.png"));
    b->setBackgroundMode (Qt::PaletteButton);
    b->setAlignment (Qt::AlignHCenter);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    b = new LCPActiveLabel (-11, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setPixmap (QPixmap (":bright.png"));
    b->setBackgroundMode (Qt::PaletteButton);
    b->setAlignment (Qt::AlignHCenter);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    b = new LCPActiveLabel (-12, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setPixmap (QPixmap (":neutral.png"));
    b->setBackgroundMode (Qt::PaletteButton);
    b->setAlignment (Qt::AlignHCenter);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

  }

}

void
LCPColorPalette::set_palette (const lay::ColorPalette &palette)
{
  if (palette != m_palette) {

    m_palette = palette;

    for (unsigned int i = 0; i < m_color_buttons.size (); ++i) {
      QColor color;
      if (i < m_palette.colors ()) {
        color = QColor (m_palette.color_by_index (i));
      }
      if (m_color_buttons [i]) {
        QPalette pl;
        pl.setColor (QColorGroup::Window, color);
        m_color_buttons [i]->setPalette (pl);
      }
    }

  }
}

void 
LCPColorPalette::button_clicked (int index)
{
  if (index >= 0) {
    QColor color;
    if (index < int (m_palette.colors ())) {
      color = QColor (m_palette.color_by_index (index));
    }
    emit color_selected (color);
  } else if (index == -1) {
    //  no color
    emit color_selected (QColor ());
  } else if (index == -2) {
    QColor c = QColorDialog::getColor ();
    if (c.isValid ()) {
      emit color_selected (c);
    }
  } else if (index == -10) {
    emit color_brightness_selected (-16); // darker
  } else if (index == -11) {
    emit color_brightness_selected (16); // brighter
  } else if (index == -12) {
    emit color_brightness_selected (0); // brighter
  }
}

// --------------------------------------------------------------------
//  LCPTreeItemDelegate declaration & implementation

/**
 *  @brief A layer tree widget helper class
 *
 *  A specialization of the ItemDelegate that bypasses the computation
 *  of sizeHint and returns the pixmap's size directly for higher
 *  performance.
 */

class LCPItemDelegate : public QItemDelegate
{
public:
  LCPItemDelegate (QWidget *parent) 
    : QItemDelegate (parent)
  { }

private:
  virtual QSize 
  sizeHint (const QStyleOptionViewItem &style, const QModelIndex &index) const
  {
    if (index.column () == 0) {
      return QSize (40, 16);
    } else {
      return QItemDelegate::sizeHint (style, index);
    }
  }
};

// --------------------------------------------------------------------
//  LCPTreeWidget declaration & implementation

/**
 *  @brief A layer tree widget helper class
 *
 *  A specialization of the TreeView that allows us to control
 *  sizeHint which otherwise is pretty large (around 100 pixel per column)
 *  and attaches the layer tree model to the view.
 */

class LCPTreeWidget : public QTreeView
{
public:
  LCPTreeWidget (QWidget *parent, lay::LayerTreeModel *model)
    : QTreeView (parent), mp_model (model)
  {
    setModel (model);
    setItemDelegate (new LCPItemDelegate (this));
  }

  ~LCPTreeWidget ()
  {
    delete mp_model;
    mp_model = 0;
  }

  virtual QSize sizeHint () const
  {
    return QSize (0, 0);
  }

  void set_selection (const std::vector<lay::LayerPropertiesConstIterator> &sel) 
  {
    clearSelection ();
    for (std::vector<lay::LayerPropertiesConstIterator>::const_iterator s = sel.begin (); s != sel.end (); ++s) {
      selectionModel ()->select (mp_model->index (*s), QItemSelectionModel::Select | QItemSelectionModel::Rows);
    }
    selectionModel ()->setCurrentIndex (QModelIndex (), QItemSelectionModel::Current | QItemSelectionModel::Rows);
  }

  void set_current (const lay::LayerPropertiesConstIterator &sel) 
  {
    QModelIndex index = mp_model->index (sel);
    selectionModel ()->select (index, QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);
    selectionModel ()->setCurrentIndex (index, QItemSelectionModel::Current | QItemSelectionModel::Rows);
  }

  //  overload the double click event, because the default behaviour is
  //  expanding/collapsing the items
  void 
  mouseDoubleClickEvent (QMouseEvent *event)
  {
    QModelIndex index (indexAt (event->pos ()));
    if (index.isValid ()) {
      emit doubleClicked (index);
    }
  }

  void collapse_all ()
  {
#if QT_VERSION >= 0x040200
    collapseAll ();
#else
    //  Qt <4.1 does not have an "collapseAll". So we emulate it.
    std::vector<QModelIndex> indices;
    QModelIndex index (mp_model->upperLeft ());
    while (index.isValid ()) {
      if (isExpanded (index)) {
        indices.push_back (index);
      }
      index = indexBelow (index);
    }
    for (std::vector<QModelIndex>::const_iterator i = indices.begin (); i != indices.end (); ++i) {
      collapse (*i);
    }
#endif
  }

  void expand_all ()
  {
#if QT_VERSION >= 0x040200
    expandAll ();
#else
    //  Qt <4.1 does not have an "expandAll". So we emulate it.
    std::vector<QModelIndex> indices;
    QModelIndex index (mp_model->upperLeft ());
    while (index.isValid ()) {
      if (! isExpanded (index)) {
        indices.push_back (index);
      }
      index = indexBelow (index);
    }
    for (std::vector<QModelIndex>::const_iterator i = indices.begin (); i != indices.end (); ++i) {
      expand (*i);
    }
#endif
  }

private:
  lay::LayerTreeModel *mp_model;
};

// --------------------------------------------------------------------
//  LayerControlPanel implementation

LayerControlPanel::LayerControlPanel (lay::LayoutView *view, db::Manager *manager, QWidget *parent, const char *name)
  : QFrame (parent, name), db::Object (manager),
    mp_view (view), m_needs_update (true), m_in_update (false)
{
  setSizePolicy (QSizePolicy (QSizePolicy::Minimum, QSizePolicy::Preferred, 0, 0));

  mp_layer_frame = new QFrame (this);
  QVBoxLayout *l = new QVBoxLayout (mp_layer_frame);
  l->setMargin (0);
  l->setSpacing (0);

  mp_model = new lay::LayerTreeModel (this, view);
  mp_layer_list = new LCPTreeWidget (mp_layer_frame, mp_model);
  mp_model->set_font (mp_layer_list->font ());
  mp_layer_list->setUniformRowHeights (true);
  l->addWidget (mp_layer_list);
  connect (mp_layer_list, SIGNAL(doubleClicked (const QModelIndex &)), this, SLOT (double_clicked (const QModelIndex &)));
  mp_layer_list->setContextMenuPolicy (Qt::CustomContextMenu);
  connect (mp_layer_list, SIGNAL(customContextMenuRequested (const QPoint &)), this, SLOT (context_menu (const QPoint &)));
  mp_layer_list->header ()->hide ();
  mp_layer_list->setSelectionMode (QTreeView::ExtendedSelection);
  mp_layer_list->setRootIsDecorated (false);
  mp_layer_list->setIconSize (QSize (32, 16));
  //  Custom resize mode makes the columns as narrow as possible
#if QT_VERSION >= 0x040200
  mp_layer_list->header ()->setResizeMode (QHeaderView::ResizeToContents);
#else
  mp_layer_list->header ()->setResizeMode (QHeaderView::Custom);
#endif
  QFrame *button_frame = new QFrame (mp_layer_frame);
  l->addWidget (button_frame);
  QHBoxLayout *ll = new QHBoxLayout (button_frame);
  ll->setMargin (0);
  ll->setSpacing (0);

  QToolButton *b;

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":downdown.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (downdown_clicked ()));
  ll->addWidget (b);

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":down.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (down_clicked ()));
  ll->addWidget (b);

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":up.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (up_clicked ()));
  ll->addWidget (b);

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":upup.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (upup_clicked ()));
  ll->addWidget (b);
  ll->addStretch (0);

  LCPVisibilityPalette *vp = new LCPVisibilityPalette (this);
  add_panel (vp, tl::translate ("Visibility").c_str ());
  connect (vp, SIGNAL(visibility_change (bool)), this, SLOT (visibility_changed (bool)));
  connect (vp, SIGNAL(transparency_change (bool)), this, SLOT (transparency_changed (bool)));

  LCPStylePalette *sp = new LCPStylePalette (this);
  add_panel (sp, tl::translate ("Style").c_str ());
  connect (sp, SIGNAL(width_selected (int)), this, SLOT (width_changed (int)));
  connect (sp, SIGNAL(marked_selected (bool)), this, SLOT (marked_changed (bool)));

  LCPAnimationPalette *ap = new LCPAnimationPalette (this);
  add_panel (ap, tl::translate ("Animation").c_str ());
  connect (ap, SIGNAL(animation_selected (int)), this, SLOT (animation_changed (int)));

  LCPDitherPalette *dp = new LCPDitherPalette (this, mp_view);
  mp_dither_palette = dp;
  add_panel (dp, tl::translate ("Stipple").c_str ());
  connect (dp, SIGNAL(dither_selected (int)), this, SLOT (dither_changed (int)));
  connect (dp, SIGNAL(pattern_changed (const lay::DitherPattern &)), this, SLOT (dither_pattern_changed (const lay::DitherPattern &)));

  LCPColorPalette *pf = new LCPColorPalette (this);
  mp_frame_palette = pf;
  add_panel (pf, tl::translate ("Frame color").c_str ());
  connect (pf, SIGNAL(color_selected (QColor)), this, SLOT (frame_color_changed (QColor)));
  connect (pf, SIGNAL(color_brightness_selected (int)), this, SLOT (frame_color_brightness (int)));

  LCPColorPalette *p = new LCPColorPalette (this);
  mp_palette = p;
  add_panel (p, tl::translate ("Color").c_str ());
  connect (p, SIGNAL(color_selected (QColor)), this, SLOT (fill_color_changed (QColor)));
  connect (p, SIGNAL(color_brightness_selected (int)), this, SLOT (fill_color_brightness (int)));

  mp_context_menu = new QMenu (this);
  mp_context_menu->addAction (tl::translate ("Select All").c_str (), this, SLOT (cm_select_all ()));
  //  It is not sure, whether "expandAll" destabilizes the tree widget:
  //  mp_context_menu->addAction (tl::translate ("Expand All").c_str (), this, SLOT (cm_expand_all ()));
  mp_context_menu->addSeparator ();
  mp_context_menu->addAction (tl::translate ("Hide").c_str (), this, SLOT (cm_hide ()));
  mp_context_menu->addAction (tl::translate ("Show").c_str (), this, SLOT (cm_show ()));
  mp_context_menu->addAction (tl::translate ("Rename").c_str (), this, SLOT (cm_rename ()));
  mp_context_menu->addSeparator ();
  mp_context_menu->addAction (tl::translate ("Select Source").c_str (), this, SLOT (cm_source ()));
  mp_context_menu->addSeparator ();
  QMenu *sort_menu = mp_context_menu->addMenu (tl::translate ("Sort By").c_str ());
  sort_menu->addAction (tl::translate ("Layout Index, Layer And Datatype").c_str (), this, SLOT (cm_sort_by_ild ()));
  sort_menu->addAction (tl::translate ("Layout Index, Datatype And Layer").c_str (), this, SLOT (cm_sort_by_idl ()));
  sort_menu->addAction (tl::translate ("Layer, Datatype And Layout Index").c_str (), this, SLOT (cm_sort_by_ldi ()));
  sort_menu->addAction (tl::translate ("Datatype, Layer And Layout Index").c_str (), this, SLOT (cm_sort_by_dli ()));
  sort_menu->addAction (tl::translate ("Name").c_str (), this, SLOT (cm_sort_by_name ()));
  mp_context_menu->addSeparator ();
  mp_context_menu->addAction (tl::translate ("Delete View").c_str (), this, SLOT (cm_delete ()));
  mp_context_menu->addAction (tl::translate ("Insert View").c_str (), this, SLOT (cm_insert ()));
  mp_context_menu->addAction (tl::translate ("Add Other Views").c_str (), this, SLOT (cm_add_missing ()));
  mp_context_menu->addAction (tl::translate ("Clean Up Views").c_str (), this, SLOT (cm_remove_unused ()));
  mp_context_menu->addSeparator ();
  mp_context_menu->addAction (tl::translate ("Group").c_str (), this, SLOT (cm_group ()));
  mp_context_menu->addAction (tl::translate ("Ungroup").c_str (), this, SLOT (cm_ungroup ()));
  QMenu *regroup_menu = mp_context_menu->addMenu (tl::translate ("Regroup Views").c_str ());
  regroup_menu->addAction (tl::translate ("By Layout Index").c_str (), this, SLOT (cm_regroup_by_index ()));
  regroup_menu->addAction (tl::translate ("By Datatype").c_str (), this, SLOT (cm_regroup_by_datatype ()));
  regroup_menu->addAction (tl::translate ("By Layer").c_str (), this, SLOT (cm_regroup_by_layer ()));
  regroup_menu->addAction (tl::translate ("Flatten").c_str (), this, SLOT (cm_regroup_flatten ()));

  mp_view->add_layer_list_observer (*this); 
}

LayerControlPanel::~LayerControlPanel ()
{
  //  .. nothing yet ..
}

void 
LayerControlPanel::add_panel (QWidget *panel_widget, const char *text)
{
  panel_widget->hide ();

  QFrame *f = new FrameWithBackground (this);
  QHBoxLayout *l = new QHBoxLayout (f, 3, 0);
  l->setMargin (0);
  l->setSpacing (0);

  f->setFrameStyle (QFrame::Panel | QFrame::Raised);
  f->setLineWidth (1);
  f->setBackgroundRole (QPalette::Highlight);

  QCheckBox *b = new QCheckBox (f);
  l->addWidget (b);

  b->setBackgroundRole (QPalette::Highlight);
  QPalette pl (b->palette ());
  pl.setColor (QColorGroup::Foreground, pl.color (QPalette::Active, QColorGroup::HighlightedText));
  b->setPalette (pl);

  b->setText (text);

  LCPRemitter *e = new LCPRemitter (int (m_tool_panels.size ()), this);
  connect (b, SIGNAL(clicked ()), e, SLOT (the_slot ()));
  connect (e, SIGNAL(the_signal (int)), this, SLOT (panel_button_clicked (int)));

  m_tool_panels.push_back (std::make_pair (f, panel_widget));
}

QSize
LayerControlPanel::sizeHint () const
{
  int w = 0;
  for (std::vector <std::pair <QWidget *, QWidget *> >::const_iterator i = m_tool_panels.begin (); i != m_tool_panels.end (); ++i) {
    w = std::max (std::max (i->first->sizeHint ().width (), i->second->sizeHint ().width ()), w);
  }
  w = std::max (mp_layer_list->sizeHint ().width (), w);
  return QSize (w, 0);
}

void 
LayerControlPanel::resizeEvent (QResizeEvent *re)
{
  rearrange (re->size ().width (), re->size ().height ());
}

void 
LayerControlPanel::resize (int w, int h)
{
  QFrame::resize (w, h);
  rearrange (w, h);
}

void 
LayerControlPanel::setGeometry (int x, int y, int w, int h)
{
  QFrame::setGeometry (x, y, w, h);
  rearrange (w, h);
}

void 
LayerControlPanel::rearrange (int w, int h)
{
  for (std::vector <std::pair <QWidget *, QWidget *> >::iterator i = m_tool_panels.begin (); i != m_tool_panels.end (); ++i) {

    int hh;

    if (i->second->isVisible ()) {
      hh = i->second->sizeHint ().height ();
      h -= hh;
      i->second->setGeometry (0, h, w, hh);
    }

    hh = i->first->sizeHint ().height () - 6;
    h -= hh;
    i->first->setGeometry (0, h, w, hh);

  }

  if (h < 0) {
    mp_layer_frame->hide ();
  } else {
    mp_layer_frame->setGeometry (0, 0, w, h);
    mp_layer_frame->show ();
  }
}

template <class Op>
void 
LayerControlPanel::foreach_selected (const Op &op)
{
  std::vector<lay::LayerPropertiesConstIterator> sel = selected_layers ();

  for (std::vector<lay::LayerPropertiesConstIterator>::const_iterator l = sel.begin (); l != sel.end (); ++l) {
    lay::LayerProperties props (**l);
    op (props);
    mp_view->set_properties (*l, props);
  }
}

struct SetColor
{
  /** 
   *  @brief set (some) colors of the properties
   *
   *  @param flags Bitmask that defines which colors to change: #0=fill,#1=frame,#2=vertex,#3=text
   *  @param c The color to apply
   */
  SetColor (QColor c, unsigned int flags)
    : m_color (c), m_flags (flags)
  {
    // .. nothing yet ..
  }

  void operator() (lay::LayerProperties &props) const
  {
    if (m_flags & 2) {
      if (! m_color.isValid ()) {
        props.clear_fill_color ();
      } else {
        props.set_fill_color (m_color.rgb ());
        props.set_fill_brightness (0);
      }
    } 
    if (m_flags & 1) {
      if (! m_color.isValid ()) {
        props.clear_frame_color ();
      } else {
        props.set_frame_color (m_color.rgb ());
        props.set_frame_brightness (0);
      }
    }
  }

private:
  QColor m_color;
  unsigned int m_flags;
};

void 
LayerControlPanel::fill_color_changed (QColor c)
{
  manager ()->transaction (tl::translate ("Change fill color")); 
  SetColor op (c, 3 /*fill,frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

void 
LayerControlPanel::frame_color_changed (QColor c)
{
  manager ()->transaction (tl::translate ("Change frame color")); 
  SetColor op (c, 1 /*frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

struct SetBrightness
{
  /** 
   *  @brief set (some) colors of the properties
   *
   *  @param flags Bitmask that defines which colors to change: #0=fill,#1=frame,#2=vertex,#3=text
   *  @param c The color to apply
   */
  SetBrightness (int delta, unsigned int flags)
    : m_delta (delta), m_flags (flags)
  {
    // .. nothing yet ..
  }

  void operator() (lay::LayerProperties &props) const
  {
    if (m_flags & 2) {
      if (m_delta == 0) {
        props.set_fill_brightness (0);
      } else {
        props.set_fill_brightness (props.fill_brightness (false) + m_delta);
      }
    } 
    if (m_flags & 1) {
      if (m_delta == 0) {
        props.set_frame_brightness (0);
      } else {
        props.set_frame_brightness (props.frame_brightness (false) + m_delta);
      }
    }
  }

private:
  int m_delta;
  unsigned int m_flags;
};

void 
LayerControlPanel::fill_color_brightness (int delta)
{
  manager ()->transaction (tl::translate ("Change fill color brightness")); 
  SetBrightness op (delta, 3 /*fill,frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

void 
LayerControlPanel::frame_color_brightness (int delta)
{
  manager ()->transaction (tl::translate ("Change frame color brightness")); 
  SetBrightness op (delta, 1 /*frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

struct SetDither
{
  SetDither (int di)
    : m_di (di)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    if (m_di < 0) {
      props.clear_dither_pattern ();
    } else {
      props.set_dither_pattern ((unsigned int) m_di);
    }
  }

private:
  int m_di;
};

void
LayerControlPanel::dither_pattern_changed (const lay::DitherPattern &pattern)
{
  manager ()->transaction (tl::translate ("Edit stipple pattern")); 
  mp_view->set_dither_pattern (pattern);
  manager ()->commit ();
}

void
LayerControlPanel::dither_changed (int di)
{
  manager ()->transaction (tl::translate ("Set stipple pattern")); 
  SetDither op (di);
  foreach_selected (op);
  manager ()->commit ();
  emit stipple_changed ();
}

struct SetVisible
{
  SetVisible (bool v)
    : m_visible (v)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.set_visible (m_visible);
  }

private:
  bool m_visible;
};

void 
LayerControlPanel::visibility_changed (bool visible)
{
  if (visible) {
    manager ()->transaction (tl::translate ("Show layer")); 
  } else {
    manager ()->transaction (tl::translate ("Hide layer")); 
  }

  SetVisible op (visible);
  foreach_selected (op);
  manager ()->commit ();

  emit visibility_changed ();
}

struct SetTransparency
{
  SetTransparency (bool t)
    : m_transparent (t)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.set_transparent (m_transparent);
  }

private:
  bool m_transparent;
};

void 
LayerControlPanel::transparency_changed (bool transparent)
{
  manager ()->transaction (tl::translate ("Change transparency")); 
  SetTransparency op (transparent);
  foreach_selected (op);
  manager ()->commit ();

  emit transparency_changed ();
}

struct SetAnimation
{
  SetAnimation (int mode)
    : m_mode (mode)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.set_animation (m_mode);
  }

private:
  int m_mode;
};

void 
LayerControlPanel::animation_changed (int mode)
{
  manager ()->transaction (tl::translate ("Change animation mode")); 
  SetAnimation op (mode);
  foreach_selected (op);
  manager ()->commit ();

  emit animation_changed ();
}

struct SetWidth
{
  SetWidth (int w)
    : m_width (w)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.set_width (m_width);
  }

private:
  int m_width;
};

void 
LayerControlPanel::width_changed (int width)
{
  manager ()->transaction (tl::translate ("Change line width")); 
  SetWidth op (width);
  foreach_selected (op);
  manager ()->commit ();

  emit width_changed ();
}

struct SetMarked
{
  SetMarked (bool m)
    : m_marked (m)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.set_marked (m_marked);
  }

private:
  bool m_marked;
};

void 
LayerControlPanel::marked_changed (bool marked)
{
  manager ()->transaction (tl::translate ("Change marked vertices")); 
  SetMarked op (marked);
  foreach_selected (op);
  manager ()->commit ();

  emit marked_changed ();
}

struct LayerSelectionClearOp
  : public db::Op
{
  LayerSelectionClearOp ()
    : db::Op ()
  { }
};

void 
LayerControlPanel::cm_delete ()
{
  try {

    std::vector<lay::LayerPropertiesConstIterator> sel = selected_layers ();
    if (! sel.empty ()) {

      begin_updates ();

      manager ()->transaction (tl::translate ("Delete layer views")); 

      std::sort (sel.begin (), sel.end (), CompareLayerIteratorBottomUp ());
      for (std::vector<lay::LayerPropertiesConstIterator>::iterator s = sel.begin (); s != sel.end (); ++s) {
        mp_view->delete_layer (*s);
      }

      if (manager ()->transacting ()) {
        manager ()->queue (this, new LayerSelectionClearOp ());
      }

      emit order_changed ();
        
      manager ()->commit ();

      update_content ();

    }

  } catch (tl::Exception &ex) {
    cancel_updates ();
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_remove_unused ()
{
  try {

    begin_updates ();

    manager ()->transaction (tl::translate ("Clean up views")); 

    bool any_deleted;
    do {

      std::vector <lay::LayerPropertiesConstIterator> sel;

      lay::LayerPropertiesConstIterator l = mp_view->begin_layers ();
      while (! l.at_end ()) {
        if (! l->has_children () && l->bbox ().empty ()) {
          sel.push_back (l);
        }
        ++l;
      }

      std::sort (sel.begin (), sel.end (), CompareLayerIteratorBottomUp ());
      any_deleted = false;
      for (std::vector<lay::LayerPropertiesConstIterator>::iterator s = sel.begin (); s != sel.end (); ++s) {
        mp_view->delete_layer (*s);
        any_deleted = true;
      }

    } while (any_deleted);

    emit order_changed ();

    manager ()->commit ();

    update_content ();

  } catch (tl::Exception &ex) {
    cancel_updates ();
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_add_missing ()
{
  try {

    begin_updates ();

    manager ()->transaction (tl::translate ("Add other views")); 

    std::set <lay::ParsedLayerSource> present;
    LayerPropertiesConstIterator l = mp_view->begin_layers ();
    while (! l.at_end ()) {
      if (! l->has_children ()) {
        present.insert (l->source (true /*real*/));
      }
      ++l;
    }

    std::vector <lay::ParsedLayerSource> actual;
    for (unsigned int cv = 0; cv < mp_view->cellviews (); ++cv) {
      const db::Layout &layout = mp_view->cellview (cv)->layout ();
      for (unsigned int l = 0; l < layout.layers (); ++l) {
        const db::LayerProperties &props = layout.get_properties (l);
        if (props.layer >= 0 && props.datatype >= 0) {
          actual.push_back (lay::ParsedLayerSource (props, cv));
        }
      }
    }

    std::sort (actual.begin (), actual.end ());

    for (std::vector <lay::ParsedLayerSource>::const_iterator a = actual.begin (); a != actual.end (); ++a) {
      if (present.find (*a) == present.end ()) {
        lay::LayerPropertiesNode node;
        node.attach_view (mp_view); 
        node.set_source (*a);
        if (! node.bbox ().empty ()) {
          mp_view->init_layer_properties (node);
          mp_view->insert_layer (mp_view->end_layers (), node);
        }
      }
    }

    emit order_changed ();

    manager ()->commit ();

    update_content ();

  } catch (tl::Exception &ex) {
    cancel_updates ();
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_insert ()
{
  lay::LayerPropertiesConstIterator sel = current_layer ();
  if (! sel.is_null ()) {

    lay::LayerProperties props;

    bool ok = false;
    QString n = QInputDialog::getText (tl::translate ("Select Source").c_str (),
                                       tl::translate ("Specify source string").c_str (),
                                       QLineEdit::Normal, 
                                       props.source_string (false).c_str (),
                                       &ok);
   
    if (ok) {

      try {

        manager ()->transaction (tl::translate ("Insert views")); 

        props.set_source (n.ascii ());
        mp_view->init_layer_properties (props);
        mp_view->insert_layer (sel, props);

        if (manager ()->transacting ()) {
          manager ()->queue (this, new LayerSelectionClearOp ());
        }
        mp_layer_list->set_current (sel);

        emit order_changed ();

        manager ()->commit ();

      } catch (tl::Exception &ex) {
        manager ()->clear ();
        tl::error << ex.msg ();
        QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
      } 

    }

  }
}

void 
LayerControlPanel::cm_group ()
{
  try {

    std::vector<lay::LayerPropertiesConstIterator> sel = selected_layers ();
    if (! sel.empty ()) {

      begin_updates ();

      manager ()->transaction (tl::translate ("Group layer views")); 

      lay::LayerPropertiesNode node;
      for (std::vector<lay::LayerPropertiesConstIterator>::const_iterator s = sel.begin (); s != sel.end (); ++s) {
        node.add_child (**s);
      }

      //  this establishes a true bottom-up order
      std::sort (sel.begin (), sel.end (), CompareLayerIteratorBottomUp ());
      for (std::vector<lay::LayerPropertiesConstIterator>::iterator s = sel.begin (); s != sel.end (); ++s) {
        mp_view->delete_layer (*s);
      }

      mp_view->insert_layer (*sel.rbegin (), node);

      if (manager ()->transacting ()) {
        manager ()->queue (this, new LayerSelectionClearOp ());
      }

      emit order_changed ();

      manager ()->commit ();

      update_content ();

      mp_layer_list->set_current (*sel.rbegin ());

    }

  } catch (tl::Exception &ex) {
    cancel_updates ();
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_ungroup ()
{
  try {

    lay::LayerPropertiesConstIterator sel = current_layer ();
    if (! sel.is_null () && sel->has_children ()) {

      begin_updates ();

      manager ()->transaction (tl::translate ("Ungroup layer views")); 

      lay::LayerPropertiesNode node = *sel;
      mp_view->delete_layer (sel);

      for (lay::LayerPropertiesNode::const_iterator c = node.end_children (); c != node.begin_children (); ) {
        --c;
        mp_view->insert_layer (sel, c->flat ());
      }

      if (manager ()->transacting ()) {
        manager ()->queue (this, new LayerSelectionClearOp ());
      }
      set_selection (std::vector<lay::LayerPropertiesConstIterator> ()); // clear selection

      emit order_changed ();

      manager ()->commit ();

      update_content ();

    }

  } catch (tl::Exception &ex) {
    cancel_updates ();
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_source ()
{
  lay::LayerPropertiesConstIterator sel = current_layer ();
  if (! sel.is_null ()) {

    lay::LayerProperties props = *sel;

    bool ok = false;
    QString n = QInputDialog::getText (tl::translate ("Select Source").c_str (),
                                       tl::translate ("Specify source string").c_str (),
                                       QLineEdit::Normal, 
                                       props.source_string (false).c_str (),
                                       &ok);
   
    if (ok) {

      try {

        props.set_source (n.ascii ());

        manager ()->transaction (tl::translate ("Select source")); 
        mp_view->set_properties (sel, props);
        manager ()->commit ();

      } catch (tl::Exception &ex) {
        manager ()->clear ();
        tl::error << ex.msg ();
        QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
      } 

    }

  }
}

void 
LayerControlPanel::cm_rename ()
{
  lay::LayerPropertiesConstIterator sel = current_layer ();
  if (! sel.is_null ()) {

    lay::LayerProperties props = *sel;

    bool ok = false;
    QString n = QInputDialog::getText (tl::translate ("Rename layer").c_str (),
                                       tl::translate ("Enter new name of layer").c_str (),
                                       QLineEdit::Normal, 
                                       props.display_string (false).c_str (),
                                       &ok);
   
    if (ok) {

      try {

        props.set_name (n.ascii ());

        manager ()->transaction (tl::translate ("Rename layer")); 
        mp_view->set_properties (sel, props);
        manager ()->commit ();

      } catch (tl::Exception &ex) {
        manager ()->clear ();
        tl::error << ex.msg ();
        QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
      } 

    }

  }
}

void 
LayerControlPanel::cm_show ()
{
  try {

    manager ()->transaction (tl::translate ("Show layer")); 

    SetVisible op (true);
    foreach_selected (op);

    manager ()->commit ();

    emit visibility_changed ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_hide ()
{
  try {

    manager ()->transaction (tl::translate ("Hide layer")); 

    SetVisible op (false);
    foreach_selected (op);

    manager ()->commit ();

    emit visibility_changed ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::cm_select_all ()
{
  try {
    mp_layer_list->selectAll ();
  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::up_clicked ()
{
  try {
    manager ()->transaction (tl::translate ("Move up")); 
    do_move (1 /*up*/);
    manager ()->commit ();
  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::down_clicked ()
{
  try {
    manager ()->transaction (tl::translate ("Move down")); 
    do_move (0 /*down*/);
    manager ()->commit ();
  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::downdown_clicked ()
{
  try {
    manager ()->transaction (tl::translate ("Move fully down")); 
    do_move (2 /*downdown*/);
    manager ()->commit ();
  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::upup_clicked ()
{
  try {
    manager ()->transaction (tl::translate ("Move fully up")); 
    do_move (3 /*upup*/);
    manager ()->commit ();
  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::do_move (int mode)
{
  std::vector<lay::LayerPropertiesConstIterator> sel = selected_layers ();
  std::vector<lay::LayerPropertiesConstIterator> new_sel;
  new_sel.reserve (sel.size ());

  lay::LayerPropertiesList new_props (mp_view->get_properties ());

  move_algo (sel.begin (), sel.end (), lay::LayerPropertiesConstIterator (mp_view->get_properties (), size_t (0)), lay::LayerPropertiesIterator (new_props, size_t (0)), new_sel, mode);

  mp_view->set_properties (new_props);

  set_selection (new_sel);

  emit order_changed ();
}


void
LayerControlPanel::move_algo (std::vector<lay::LayerPropertiesConstIterator>::const_iterator from, 
                              std::vector<lay::LayerPropertiesConstIterator>::const_iterator to, 
                              lay::LayerPropertiesConstIterator parent,
                              lay::LayerPropertiesIterator new_parent,
                              std::vector<lay::LayerPropertiesConstIterator> &new_sel,
                              int mode)
{
  size_t nsel = new_sel.size ();

  std::vector<lay::LayerPropertiesConstIterator> org_sel;

  std::vector<lay::LayerPropertiesConstIterator>::const_iterator s = from;
  lay::LayerPropertiesConstIterator c = parent;
  lay::LayerPropertiesConstIterator nc = new_parent;
  for (c.first_child (), nc.first_child (); ! c.at_end (); c.next_sibling (), nc.next_sibling ()) {
    s = std::lower_bound (s, to, c);
    if (s != to && *s == c) {
      //  a selected child: remember this position
      new_sel.push_back (nc);
      org_sel.push_back (c);
    }
  }

  //  compute new positions from the current ones ..
  if (mode == 0 /*down*/) {

    lay::LayerPropertiesConstIterator l;
    for (std::vector<lay::LayerPropertiesConstIterator>::iterator i = new_sel.end (); i != new_sel.begin () + nsel; ) {
      lay::LayerPropertiesConstIterator ns = *--i;
      ns.next_sibling ();
      if (! ns.at_end () && ns != l) {
        *i = ns;
      } 
      l = *i;
    }

  } else if (mode == 1 /*up*/) {

    lay::LayerPropertiesConstIterator l;
    for (std::vector<lay::LayerPropertiesConstIterator>::iterator i = new_sel.begin () + nsel; i != new_sel.end (); ++i) {
      lay::LayerPropertiesConstIterator ns = *i;
      if (ns.child_index () > 0) {
        ns.next_sibling (-1);
      }
      if (ns != l) {
        *i = ns;
      } 
      l = *i;
    }

  } else if (mode == 2 /*downdown*/) {

    if (new_sel.end () != new_sel.begin () + nsel) {
      size_t n = new_sel.begin ()[nsel].num_siblings ();
      for (std::vector<lay::LayerPropertiesConstIterator>::iterator i = new_sel.end (); i != new_sel.begin () + nsel; ) {
        --i;
        i->to_sibling (--n);
      }
    }

  } else if (mode == 3 /*upup*/) {
    
    size_t n = 0;
    for (std::vector<lay::LayerPropertiesConstIterator>::iterator i = new_sel.begin () + nsel; i != new_sel.end (); ++i) {
      i->to_sibling (n++);
    }

  }

  std::vector<lay::LayerPropertiesConstIterator>::const_iterator inew = new_sel.begin () + nsel;
  std::vector<lay::LayerPropertiesConstIterator>::const_iterator iorg = org_sel.begin ();

  lay::LayerPropertiesIterator ins = new_parent;
  ins.first_child ();

  std::vector< std::pair<lay::LayerPropertiesConstIterator, lay::LayerPropertiesIterator> > rec;

  s = from;
  c = parent;
  for (c.first_child (); ! c.at_end (); c.next_sibling ()) {

    while (inew != new_sel.end () && ins == *inew) {
      *ins = **iorg;
      ins.next_sibling ();
      ++inew;
      ++iorg;
    }

    s = std::lower_bound (s, to, c);
    if (s == to || *s != c) {
      *ins = *c;
      if (c->has_children ()) {
        rec.push_back (std::make_pair (c, ins));
      }
      ins.next_sibling ();
    }

  }

  while (inew != new_sel.end () && ins == *inew) {
    *ins = **iorg;
    ins.next_sibling ();
    ++inew;
    ++iorg;
  }

  //  now treat all nodes with children (it is important to do this at last, because then the
  //  child iterators will be valid finally and can be inserted into "new_sel")
  for (std::vector< std::pair<lay::LayerPropertiesConstIterator, lay::LayerPropertiesIterator> >::iterator r = rec.begin (); r != rec.end (); ++r) {
    move_algo (from, to, r->first, r->second, new_sel, mode);
  }

}

void 
LayerControlPanel::set_selection (const std::vector<lay::LayerPropertiesConstIterator> &new_sel)
{
  mp_layer_list->set_selection (new_sel);

  //  :TODO: save selection for undo? Currently we just clear it.
  if (manager ()->transacting ()) {
    manager ()->queue (this, new LayerSelectionClearOp ());
  }
}

void 
LayerControlPanel::clear_selection ()
{
  std::vector<lay::LayerPropertiesConstIterator> empty_sel;
  set_selection (empty_sel);
}

void
LayerControlPanel::cm_regroup_flatten ()
{
  try {

    manager ()->transaction (tl::translate ("Flatten layers")); 
    regroup_layers (RegroupFlatten);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_regroup_by_index ()
{
  try {

    manager ()->transaction (tl::translate ("Regroup layers")); 
    regroup_layers (RegroupByIndex);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_regroup_by_datatype ()
{
  try {

    manager ()->transaction (tl::translate ("Regroup layers")); 
    regroup_layers (RegroupByDatatype);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_regroup_by_layer ()
{
  try {

    manager ()->transaction (tl::translate ("Regroup layers")); 
    regroup_layers (RegroupByLayer);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_sort_by_name ()
{
  try {

    manager ()->transaction (tl::translate ("Sort layers")); 
    sort_layers (ByName);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_sort_by_ild ()
{
  try {

    manager ()->transaction (tl::translate ("Sort layers")); 
    sort_layers (ByIndexLayerDatatype);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_sort_by_idl ()
{
  try {

    manager ()->transaction (tl::translate ("Sort layers")); 
    sort_layers (ByIndexDatatypeLayer);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_sort_by_ldi ()
{
  try {

    manager ()->transaction (tl::translate ("Sort layers")); 
    sort_layers (ByLayerDatatypeIndex);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
LayerControlPanel::cm_sort_by_dli ()
{
  try {

    manager ()->transaction (tl::translate ("Sort layers")); 
    sort_layers (ByDatatypeLayerIndex);
    emit order_changed ();
    manager ()->commit ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

struct LDSortingProps
  : std::vector <int>
{
  LDSortingProps (int l1, int l2, int l3)
  {
    reserve (3);
    push_back (l1);
    push_back (l2);
    push_back (l3);
  }
};

struct LayerSorter 
{
  LayerSorter (lay::LayerControlPanel::SortOrder order)
    : m_order (order)
  {
    //  .. nothing yet ..
  }

  bool operator() (const lay::LayerPropertiesNode &a, const lay::LayerPropertiesNode &b) const
  {
    if (m_order == lay::LayerControlPanel::ByName) {

      return a.display_string (false) < b.display_string (false);

    } else if (m_order == lay::LayerControlPanel::ByIndexLayerDatatype) {

      if (a.source (false).cv_index () != b.source (false).cv_index ()) {
        return (a.source (false).cv_index () < b.source (false).cv_index ());
      }
      if (a.source (false).layer () != b.source (false).layer ()) {
        return (a.source (false).layer () < b.source (false).layer ());
      }
      if (a.source (false).datatype () != b.source (false).datatype ()) {
        return (a.source (false).datatype () < b.source (false).datatype ());
      }
      return false;

    } else if (m_order == lay::LayerControlPanel::ByIndexDatatypeLayer) {

      if (a.source (false).cv_index () != b.source (false).cv_index ()) {
        return (a.source (false).cv_index () < b.source (false).cv_index ());
      }
      if (a.source (false).datatype () != b.source (false).datatype ()) {
        return (a.source (false).datatype () < b.source (false).datatype ());
      }
      if (a.source (false).layer () != b.source (false).layer ()) {
        return (a.source (false).layer () < b.source (false).layer ());
      }
      return false;

    } else if (m_order == lay::LayerControlPanel::ByLayerDatatypeIndex) {

      if (a.source (false).layer () != b.source (false).layer ()) {
        return (a.source (false).layer () < b.source (false).layer ());
      }
      if (a.source (false).datatype () != b.source (false).datatype ()) {
        return (a.source (false).datatype () < b.source (false).datatype ());
      }
      if (a.source (false).cv_index () != b.source (false).cv_index ()) {
        return (a.source (false).cv_index () < b.source (false).cv_index ());
      }
      return false;

    } else if (m_order == lay::LayerControlPanel::ByDatatypeLayerIndex) {

      if (a.source (false).datatype () != b.source (false).datatype ()) {
        return (a.source (false).datatype () < b.source (false).datatype ());
      }
      if (a.source (false).layer () != b.source (false).layer ()) {
        return (a.source (false).layer () < b.source (false).layer ());
      }
      if (a.source (false).cv_index () != b.source (false).cv_index ()) {
        return (a.source (false).cv_index () < b.source (false).cv_index ());
      }
      return false;

    } else {
      return false;
    }
  }

private:
  lay::LayerControlPanel::SortOrder m_order;
};

void
LayerControlPanel::sort_layers (SortOrder order)
{
  std::vector<lay::LayerPropertiesNode> new_props;

  lay::LayerPropertiesConstIterator p = current_layer ().parent ();
  if (p.is_null ()) {
    new_props.assign (mp_view->get_properties ().begin_const (), mp_view->get_properties ().end_const ());
  } else {
    new_props.assign (p->begin_children (), p->end_children ());
  }

  std::sort (new_props.begin (), new_props.end (), LayerSorter (order));

  lay::LayerPropertiesList prop_list;

  if (p.is_null ()) {

    for (std::vector<lay::LayerPropertiesNode>::const_iterator np = new_props.begin (); np != new_props.end (); ++np) {
      //  :KLUDGE: the list should have an insert with a begin..end iterator pair ..
      prop_list.push_back (*np);
    }

  } else {

    prop_list = mp_view->get_properties ();

    lay::LayerPropertiesIterator pp (prop_list, p.uint ());
    pp->clear_children ();
    for (std::vector<lay::LayerPropertiesNode>::const_iterator np = new_props.begin (); np != new_props.end (); ++np) {
      //  :KLUDGE: the list should have an insert with a begin..end iterator pair ..
      pp->add_child (*np);
    }

  }

  mp_view->set_properties (prop_list);
}

struct LayerRegroupSorter 
{
  LayerRegroupSorter (lay::LayerControlPanel::RegroupMode mode)
    : m_mode (mode)
  {
    //  .. nothing yet ..
  }

  bool operator() (const lay::LayerPropertiesNode &a, const lay::LayerPropertiesNode &b) const
  {
    if (m_mode == lay::LayerControlPanel::RegroupByIndex) {
      return (a.source (false).cv_index () < b.source (false).cv_index ());
    } else if (m_mode == lay::LayerControlPanel::RegroupByDatatype) {
      return (a.source (false).datatype () < b.source (false).datatype ());
    } else if (m_mode == lay::LayerControlPanel::RegroupByLayer) {
      return (a.source (false).layer () < b.source (false).layer ());
    } else {
      return false;
    }
  }

private:
  lay::LayerControlPanel::RegroupMode m_mode;
};

void
LayerControlPanel::regroup_layers (RegroupMode mode)
{
  std::vector<lay::LayerProperties> linear_props;
  LayerPropertiesConstIterator l = mp_view->begin_layers ();
  while (! l.at_end ()) {
    if (! l->has_children ()) {
      linear_props.push_back (l->flat ()); 
    }
    ++l;
  }

  LayerRegroupSorter sorter (mode);
  std::stable_sort (linear_props.begin (), linear_props.end (), sorter);

  lay::LayerPropertiesList prop_list;

  std::vector<lay::LayerProperties>::const_iterator i = linear_props.begin ();
  while (i != linear_props.end ()) {

    std::vector<lay::LayerProperties>::const_iterator f = i;
    do {
      ++f;
    } while (f != linear_props.end () && !sorter (*i, *f));

    //  make a new node from [i..f)
    if (mode == lay::LayerControlPanel::RegroupByIndex) {

      prop_list.push_back (lay::LayerPropertiesNode ());
      lay::ParsedLayerSource source;
      source.cv_index (i->source (true /*real*/).cv_index ());
      prop_list.back ().set_source (source);

    } else if (mode == lay::LayerControlPanel::RegroupByDatatype) {

      prop_list.push_back (lay::LayerPropertiesNode ());
      lay::ParsedLayerSource source;
      source.datatype (i->source (true /*real*/).datatype ());
      prop_list.back ().set_source (source);

    } else if (mode == lay::LayerControlPanel::RegroupByLayer) {

      prop_list.push_back (lay::LayerPropertiesNode ());
      lay::ParsedLayerSource source;
      source.layer (i->source (true /*real*/).layer ());
      prop_list.back ().set_source (source);

    } 

    for (std::vector<lay::LayerProperties>::const_iterator p = i; p != f; ++p) {

      lay::LayerProperties pp (*p);
      lay::ParsedLayerSource source (pp.source (true /*real*/));

      if (mode == lay::LayerControlPanel::RegroupByIndex) {
        source.cv_index (-1);
      } else if (mode == lay::LayerControlPanel::RegroupByDatatype) {
        source.datatype (-1);
      } else if (mode == lay::LayerControlPanel::RegroupByLayer) {
        source.layer (-1);
      }

      pp.set_source (source);
      if (mode != lay::LayerControlPanel::RegroupFlatten) {
        prop_list.back ().add_child (pp);
      } else {
        prop_list.push_back (pp);
      }

    }

    i = f;
  
  }

  mp_view->set_properties (prop_list);
}

void
LayerControlPanel::cm_expand_all ()
{
  mp_layer_list->expand_all ();
}

void
LayerControlPanel::context_menu (const QPoint &p)
{
  mp_context_menu->exec (mp_layer_list->mapToGlobal (p));
}

void 
LayerControlPanel::double_clicked (const QModelIndex &index)
{
  try {

    if (! index.isValid ()) {
      return;
    }

    lay::LayerPropertiesConstIterator item (mp_view->get_properties (), size_t (index.internalId ()));

    lay::LayerProperties props = *item;
    props.set_visible (! props.visible (false));

    if (props.visible (false)) {
      manager ()->transaction (tl::translate ("Show layer")); 
    } else {
      manager ()->transaction (tl::translate ("Hide layer")); 
    }

    mp_view->set_properties (item, props);

    manager ()->commit ();

    emit visibility_changed ();

  } catch (tl::Exception &ex) {
    manager ()->clear ();
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
LayerControlPanel::panel_button_clicked (int index)
{
  if (index < 0 || index >= int (m_tool_panels.size ())) {
    return;
  }

  if (m_tool_panels [index].second->isVisible ()) {
    m_tool_panels [index].second->hide ();
  } else {
    m_tool_panels [index].second->show ();
  }

  rearrange (width (), height ());
}

void
LayerControlPanel::set_palette (const lay::ColorPalette &p)
{
  mp_palette->set_palette (p);
  mp_frame_palette->set_palette (p);
}

void
LayerControlPanel::set_background_color (QColor c)
{
  QPalette pl (mp_layer_list->palette ());
  pl.setColor (QColorGroup::Base, c);
  mp_layer_list->setPalette (pl);
  mp_model->set_background_color (c);
}

void
LayerControlPanel::set_text_color (QColor c)
{
  QPalette pl (mp_layer_list->palette ());
  pl.setColor (QColorGroup::Text, c);
  mp_layer_list->setPalette (pl);
  mp_model->set_text_color (c);
}

void
LayerControlPanel::begin_updates ()
{
  if (m_in_update) {
    tl::warn << "begin_layer_list_change() called recursively";
  } else {

    //  we force a clear_selection in this case, since we cannot make sure the
    //  selecting remains valid
    clear_selection ();
    save_expanded ();

    mp_model->signal_begin_layer_changed ();  //  this makes the view redraw the data

    m_in_update = true;

  }
}

void 
LayerControlPanel::check_updated ()
{
  tl_assert (! m_in_update);
}

void
LayerControlPanel::cancel_updates ()
{
  m_in_update = 0;
  m_needs_update = false;
}

void
LayerControlPanel::update_content (int phase)
{
  if (phase >= 0) {
    mp_model->set_phase (phase);
  }

  if (m_in_update) {

    m_in_update = false;

    mp_model->signal_layer_changed ();  //  this makes the view redraw the data

    bool has_children = false;
    for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); l != mp_view->end_layers () && ! has_children; ++l) {
      if (l->has_children ()) {
        has_children = true;
      }
    }
    mp_layer_list->setRootIsDecorated (has_children);
    mp_layer_list->doItemsLayout ();

    restore_expanded ();

    m_needs_update = false;

  } else {

    if (m_needs_update) {

      m_needs_update = false;

      bool has_children = false;
      for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); l != mp_view->end_layers () && ! has_children; ++l) {
        if (l->has_children ()) {
          has_children = true;
        }
      }
      mp_layer_list->setRootIsDecorated (has_children);
      mp_layer_list->reset ();

    } else {
      mp_model->signal_data_changed ();  //  this makes the view redraw the data
    }

  }
}

lay::LayerPropertiesConstIterator 
LayerControlPanel::current_layer () const
{
  QModelIndex index = mp_layer_list->selectionModel ()->currentIndex ();
  if (index.isValid ()) {
    return lay::LayerPropertiesConstIterator (mp_view->get_properties (), size_t (index.internalId ()));
  } else {
    return lay::LayerPropertiesConstIterator ();
  }
}

std::vector <lay::LayerPropertiesConstIterator>
LayerControlPanel::selected_layers () const
{
  QModelIndexList selected = mp_layer_list->selectionModel ()->selectedIndexes ();

  std::vector <lay::LayerPropertiesConstIterator> llist;
  llist.reserve (selected.size ());
  for (QModelIndexList::const_iterator i = selected.begin (); i != selected.end (); ++i) {
    if (i->column () == 0) {
      llist.push_back (lay::LayerPropertiesConstIterator (mp_view->get_properties (), size_t (i->internalId ())));
    }
  }

  //  filter out the children:
  //  we employ the fact, that the LayerPropertiesConstIterator's are ordered
  //  parents first and children before siblings.
  std::sort (llist.begin (), llist.end ());

  std::vector<lay::LayerPropertiesConstIterator>::iterator write = llist.begin ();
  for (std::vector<lay::LayerPropertiesConstIterator>::iterator read = llist.begin (); read != llist.end (); ) {

    lay::LayerPropertiesConstIterator n = *read;
    *write++ = n;
    n.next_sibling (); 

    read = std::lower_bound (read + 1, llist.end (), n);

  }

  llist.erase (write, llist.end ());
  return llist;
}

void 
LayerControlPanel::undo (db::Op *op) throw ()
{
  LayerSelectionClearOp *clrop = dynamic_cast <LayerSelectionClearOp *> (op);
  if (clrop) {
    set_selection (std::vector<lay::LayerPropertiesConstIterator> ()); // clear selection
    return;
  }
}

void 
LayerControlPanel::redo (db::Op *op) throw ()
{
  LayerSelectionClearOp *clrop = dynamic_cast <LayerSelectionClearOp *> (op);
  if (clrop) {
    set_selection (std::vector<lay::LayerPropertiesConstIterator> ()); // clear selection
    return;
  }
}

void 
LayerControlPanel::signal_observed_int (int f)
{
  if (f & 2) {
    m_needs_update = true;
  }
}

void
LayerControlPanel::save_expanded ()
{
  m_expanded.clear ();

  QModelIndex index (mp_model->upperLeft ());
  while (index.isValid ()) {
    if (mp_layer_list->isExpanded (index)) {
      lay::LayerPropertiesConstIterator iter = mp_model->iterator (index);
      if (! iter.is_null () && ! iter.at_end ()) {
        m_expanded.insert (iter->id ());
      }
    }
    index = mp_layer_list->indexBelow (index);
  }

  mp_layer_list->collapse_all ();
}

void
LayerControlPanel::restore_expanded ()
{
  lay::LayerPropertiesConstIterator l = mp_view->begin_layers ();
  while (! l.at_end ()) {
    if (m_expanded.find (l->id ()) != m_expanded.end ()) {
      QModelIndex index = mp_model->index (l);
      mp_layer_list->expand (index);
    }
    ++l;
  }

  m_expanded.clear ();
}


} // namespace lay 
