
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtCore/QEvent>
#include <QtGui/QPixmap>
#include <QtGui/QBitmap>
#include <QtGui/QPainter>
#include <QtGui/QApplication>

#include "tlTimer.h"
#include "tlLog.h"
#include "layLayoutCanvas.h"
#include "layRedrawThread.h"
#include "layLayoutView.h"
#include "layMarker.h"

#include "layBitmapsToImage.h"

namespace lay
{


LayoutCanvas::LayoutCanvas (QWidget *parent, lay::LayoutView *view, const char *name)
  : lay::ViewObjectCanvas (parent, name), 
    mp_view (view),
    mp_status_bar (0),
    mp_image (0), mp_pixmap (0), 
    m_width (1), m_height (1),
    m_background (0), m_foreground (0), m_active (0),
    m_need_redraw (false),
    m_redraw_clearing (false),
    m_update_ignore (0)
{
  mp_redraw_thread = new lay::RedrawThread (0, view);
  mp_redraw_thread->set_canvas (this);

  setBackgroundMode (Qt::NoBackground);
  set_colors (paletteBackgroundColor (),
              palette ().color (QPalette::Normal, QColorGroup::Text),
              palette ().color (QPalette::Normal, QColorGroup::Mid));
  setAttribute (Qt::WA_PaintOnScreen);
  setAttribute (Qt::WA_NoSystemBackground);
}

LayoutCanvas::~LayoutCanvas ()
{
  if (mp_image) {
    delete mp_image;
    mp_image = 0;
  }
  if (mp_pixmap) {
    delete mp_pixmap;
    mp_pixmap = 0;
  }
  if (mp_redraw_thread) {
    delete mp_redraw_thread;
    mp_redraw_thread = 0;
  }

  lock ();
  clear_bitmaps ();
  clear_dyn_bitmaps ();
  unlock ();
}

void
LayoutCanvas::clear_bitmaps ()
{
  while (! mp_bitmap_buffers.empty ()) {
    delete mp_bitmap_buffers.back ();
    mp_bitmap_buffers.pop_back ();
  }

  while (! mp_drawing_bitmap_buffers.empty ()) {
    while (! mp_drawing_bitmap_buffers.back ().empty ()) {
      delete mp_drawing_bitmap_buffers.back ().back ();
      mp_drawing_bitmap_buffers.back ().pop_back ();
    }
    mp_drawing_bitmap_buffers.pop_back ();
  }
}

void 
LayoutCanvas::set_colors (QColor background, QColor foreground, QColor active)
{
  m_background = background.rgb ();
  m_foreground = foreground.rgb ();
  m_active = active.rgb ();

  update_image ();
}

void
LayoutCanvas::set_view_ops (std::vector <lay::ViewOp> &view_ops)
{
  if (view_ops != m_view_ops) {
    m_view_ops.swap (view_ops);
    update_image ();
  }
}

void
LayoutCanvas::set_dither_pattern (const lay::DitherPattern &p)
{
  if (p != m_dither_pattern) {
    m_dither_pattern = p;
    update_image ();
  }
}

void
LayoutCanvas::clear (unsigned int nlayers)
{
  m_width = width ();
  m_height = height ();

  if (! mp_image ||
      (unsigned int) mp_image->width () != m_width || (unsigned int) mp_image->height () != m_height) {
    if (mp_image) {
      delete mp_image;
    }
    mp_image = new QImage (m_width, m_height, QImage::Format_RGB32);
    if (mp_pixmap) {
      delete mp_pixmap;
      mp_pixmap = 0;
    }
  }

  mp_image->fill (m_background);

  lock ();

  //  delete the buffers and create new ones 
  clear_bitmaps ();

  for (unsigned int i = 0; i < nlayers; ++i) {
    mp_bitmap_buffers.push_back (new lay::Bitmap (m_width, m_height));
  }

  for (lay::Drawings::iterator d = lay::Drawings::begin (); d != lay::Drawings::end (); ++d) {
    mp_drawing_bitmap_buffers.push_back (std::vector <lay::Bitmap *> ());
    for (unsigned int i = 0; i < d->num_bitmaps (); ++i) {
      mp_drawing_bitmap_buffers.back ().push_back (new lay::Bitmap (m_width, m_height));
    }
  }

  unlock ();
}

lay::Bitmap *
LayoutCanvas::bitmap (const lay::ViewOp &style)
{
  std::map <lay::ViewOp, unsigned int>::iterator b = m_dyn_bitmap_table.find (style);
  if (b == m_dyn_bitmap_table.end ()) {

    //  we need to create a new bitmap
    m_dyn_bitmap_table.insert (std::make_pair (style, mp_dyn_bitmaps.size ()));
    mp_dyn_bitmaps.push_back (new lay::Bitmap (m_width, m_height));
    m_dyn_view_ops.push_back (style);
    return mp_dyn_bitmaps.back ();
    
  } else {
    //  we can recycle a current one
    return mp_dyn_bitmaps [b->second];
  }
}

void 
LayoutCanvas::clear_dyn_bitmaps ()
{
  for (unsigned int i = 0; i < mp_dyn_bitmaps.size (); ++i) {
    if (mp_dyn_bitmaps [i]) {
      delete mp_dyn_bitmaps [i];
    }
  }
  mp_dyn_bitmaps.clear ();
  m_dyn_view_ops.clear ();
  m_dyn_bitmap_table.clear ();
}

void
LayoutCanvas::update_image ()
{
  //  touch_bg() will make paintEvent recreate the image and issue an update()
  touch_bg ();
}

void
LayoutCanvas::free_resources ()
{
  if (mp_pixmap) {
    delete mp_pixmap;
    mp_pixmap = 0;
  }
}

static void
bitmap_to_bitmap (const lay::ViewOp &view_op, lay::Bitmap &bitmap,
                  unsigned char *data,
                  unsigned int width, unsigned int height,
                  const lay::DitherPattern &dp)
{
  //  determine endianess ..
  unsigned int x = 0xc0000001;
  unsigned char x0 = ((unsigned char *) &x) [0];

  for (unsigned int y = 0; y < height; ++y) {

    unsigned nbytes = ((width + 7) / 8);

    if (! bitmap.is_scanline_empty (height - 1 - y)) {

      uint32_t dither = dp.pattern (view_op.dither_index ()) [(height - 1 - y) & (lay::wordlen - 1)];

      if (x0 == 0xc0) {

        //  MSB first ..
        const uint32_t *p = bitmap.scanline (height - 1 - y);
        uint32_t d = 0;
        const char *dp = (const char *)&d;
        while (nbytes >= 4) {
          d = *p++ & dither;
          *data++ |= dp[3];
          *data++ |= dp[2];
          *data++ |= dp[1];
          *data++ |= dp[0];
          nbytes -= 4;
        }
        if (nbytes > 0) {
          d = *p++ & dither;
          dp += 4;
          while (nbytes > 0) {
            *data++ |= *--dp;
            --nbytes;
          }
        }

      } else if (x0 == 0x01) {

        //  LSB first ..
        const uint32_t *p = bitmap.scanline (height - 1 - y);
        uint32_t d = 0;
        const char *dp = (const char *)&d;
        while (nbytes >= 4) {
          d = *p++ & dither;
          *data++ |= dp[0];
          *data++ |= dp[1];
          *data++ |= dp[2];
          *data++ |= dp[3];
          nbytes -= 4;
        }
        if (nbytes > 0) {
          d = *p++ & dither;
          while (nbytes > 0) {
            *data++ |= *dp++;
            --nbytes;
          }
        }

      } else {
        //  unable to determine endianess
        tl_assert (false);
      }

    } else {
      data += nbytes;
    }
    
  }

}

#ifdef _WIN32
// Windows build runs well with double-buffering
// #  define SINGLE_BUFFERED_MASK
// #  define SINGLE_BUFFERED
#else
// X11 runs better with the SINGLE_BUFFERED_MASK scheme
// #  define SINGLE_BUFFERED
#  define SINGLE_BUFFERED_MASK
#endif

void
LayoutCanvas::paintEvent (QPaintEvent *p)
{
  //  this is the update image request
  tl::SelfTimer timer_info (tl::verbosity () >= 31, tl::translate ("PaintEvent"));

  //  if required, start the redraw thread ..
  bool redrawn = m_need_redraw;
  if (m_need_redraw) {
    mp_redraw_thread->start (m_redraw_clearing, m_need_redraw_layer);
    m_need_redraw = false;
  }

  if (mp_image) {
    
    //  check, if the background needs to be updated 
    if (redrawn || needs_update_bg ()) {

      tl_assert (mp_image->width () == int (m_width));
      tl_assert (mp_image->height () == int (m_height));

      //  clear the image
      mp_image->fill (m_background);

      //  and paint the background objects
      do_render_bg (*mp_image);

      //  convert the bitmap data to image data
      bitmaps_to_image (m_view_ops, mp_bitmap_buffers, dither_pattern (), mp_image, m_width, m_height, &m_mutex);

      //  convert the bitmaps of the "drawing" objects too:
      std::vector <std::vector <lay::Bitmap *> >::const_iterator bt = mp_drawing_bitmap_buffers.begin ();
      for (lay::Drawings::iterator d = lay::Drawings::begin (); d != lay::Drawings::end () && bt != mp_drawing_bitmap_buffers.end (); ++d, ++bt) {
        bitmaps_to_image (d->get_view_ops (), *bt, dither_pattern (), mp_image, m_width, m_height, &m_mutex);
      }

      if (mp_pixmap) {
        delete mp_pixmap;
        mp_pixmap = 0;
      }

    }

    //  prepare a buffer
    size_t nbytes = (m_width + 7) / 8 * m_height;
    unsigned char *p_data = new unsigned char [nbytes];
    memset (p_data, 0, nbytes);

    //  create a base pixmap consisting of the layout with background
    //  and static foreground objects

    if (! mp_pixmap || needs_update_static () || 
        mp_image->size ().width () != mp_pixmap->size ().width () ||
        mp_image->size ().height () != mp_pixmap->size ().height ()) {

      if (mp_pixmap) {
        delete mp_pixmap;
      } 

      //  render the dynamic parts ..
      clear_dyn_bitmaps ();
      do_render (true);

      if (mp_dyn_bitmaps.size () > 0) {
        QImage full_image (*mp_image);
        bitmaps_to_image (m_dyn_view_ops, mp_dyn_bitmaps, dither_pattern (), &full_image, m_width, m_height, &m_mutex);
        mp_pixmap = new QPixmap (full_image);
      } else {
        mp_pixmap = new QPixmap (*mp_image);
      }

    }

    //  erase any previous data
    clear_dyn_bitmaps ();

    //  render dynamic content
    do_render (false);

#if defined(SINGLE_BUFFERED)

    //  Single-buffering scheme - somewhat more flickering 
    //  but less resource requirements.

    //  produce the pixmap first and then overdraw with dynamic content
    bitBlt (this, 0, 0, mp_pixmap, 0, 0, mp_pixmap->size ().width (), mp_pixmap->size ().height ());

    //  dynamic bitmaps are always drawn in transparent mode ..
    QPainter painter (this);
    painter.setBackgroundMode (Qt::TransparentMode);

    //  create QBitmap objects and paint them for the bitmaps
    //  present
    for (unsigned int n = 0; n < mp_dyn_bitmaps.size (); ++n) {
      if (mp_dyn_bitmaps [n] != 0) {
        memset (p_data, 0, nbytes);
        bitmap_to_bitmap (m_dyn_view_ops [n], *mp_dyn_bitmaps [n], p_data, m_width, m_height, dither_pattern ());
        QBitmap bitmap (m_width, m_height, p_data, true);
        painter.setPen (QRgb (m_dyn_view_ops [n].ormask ()));
        painter.drawPixmap (0, 0, bitmap);
      }
    }

#elif defined(SINGLE_BUFFERED_MASK)

    //  Single-buffering scheme with mask - somewhat more flickering 
    //  but less resource requirements.

    //  combine the bitmaps into a mask
    for (unsigned int n = 0; n < mp_dyn_bitmaps.size (); ++n) {
      if (mp_dyn_bitmaps [n] != 0) {
        bitmap_to_bitmap (m_dyn_view_ops [n], *mp_dyn_bitmaps [n], p_data, m_width, m_height, dither_pattern ());
      }
    }

    //  invert the mask
    unsigned char *d = p_data;
    for (unsigned int i = 0; i < nbytes; ++i) {
      *d++ ^= 0xff;
    }

    //  create the mask
    QBitmap mask (m_width, m_height, p_data, true);
    mp_pixmap->setMask (mask);

    //  produce the pixmap with the mask
    bitBlt (this, 0, 0, mp_pixmap, 0, 0, mp_pixmap->size ().width (), mp_pixmap->size ().height ());

    //  dynamic bitmaps are always drawn in transparent mode ..
    QPainter painter (this);
    painter.setBackgroundMode (Qt::TransparentMode);

    //  create QBitmap objects and paint them for the bitmaps
    //  present
    for (unsigned int n = 0; n < mp_dyn_bitmaps.size (); ++n) {
      if (mp_dyn_bitmaps [n] != 0) {
        memset (p_data, 0, nbytes);
        bitmap_to_bitmap (m_dyn_view_ops [n], *mp_dyn_bitmaps [n], p_data, m_width, m_height, dither_pattern ());
        QBitmap bitmap (m_width, m_height, p_data, true);
        painter.setPen (QRgb (m_dyn_view_ops [n].ormask ()));
        painter.drawPixmap (0, 0, bitmap);
      }
    }

#else

    //  Double-buffering scheme - somewhat less flickering but
    //  only efficient if the QImage has the precise size and
    //  at the cost of twice the X resource requirements.

    //  create an intermediate pixmap
    QPixmap px (*mp_pixmap);

    //  dynamic bitmaps are always drawn in transparent mode ..
    QPainter painter (&px);
    painter.setBackgroundMode (Qt::TransparentMode);

    //  create QBitmap objects and paint them for the bitmaps
    //  present
    for (unsigned int n = 0; n < mp_dyn_bitmaps.size (); ++n) {
      if (mp_dyn_bitmaps [n] != 0) {
        memset (p_data, 0, nbytes);
        bitmap_to_bitmap (m_dyn_view_ops [n], *mp_dyn_bitmaps [n], p_data, m_width, m_height, dither_pattern ());
        QBitmap bitmap (m_width, m_height, p_data, true);
        painter.setPen (QRgb (m_dyn_view_ops [n].ormask ()));
        painter.drawPixmap (0, 0, bitmap);
      }
    }

    //  produce the combined pixmap on the window
    bitBlt (this, 0, 0, &px, 0, 0, mp_pixmap->size ().width (), mp_pixmap->size ().height ());

#endif

    //  delete buffer
    delete [] p_data;

    //  erase dynamic bitmaps 
    clear_dyn_bitmaps ();

    QApplication::syncX ();

  }

}

QImage 
LayoutCanvas::screenshot () 
{
  if (mp_image) {

    QImage full_image (*mp_image);

    //  render the dynamic parts ..

    clear_dyn_bitmaps ();
    do_render (true);
    bitmaps_to_image (m_dyn_view_ops, mp_dyn_bitmaps, dither_pattern (), &full_image, m_width, m_height, &m_mutex);

    clear_dyn_bitmaps ();
    do_render (false);
    bitmaps_to_image (m_dyn_view_ops, mp_dyn_bitmaps, dither_pattern (), &full_image, m_width, m_height, &m_mutex);

    clear_dyn_bitmaps ();

    return full_image;

  } else {

    QImage img (size (), QImage::Format_RGB32);
    img.fill (QColor (m_background).rgb ());
    return img;

  }
}

void 
LayoutCanvas::resizeEvent (QResizeEvent *)
{
  redraw (true /*clear all*/, std::set<int> ());
}

void 
LayoutCanvas::zoom_to (const db::DMagTrans &tr)
{
  if (trans () != tr) {
    trans (tr);
    for (service_iterator svc = begin_services (); svc != end_services (); ++svc) {
      svc->update ();
    }
    redraw (true /*clear all*/, std::set<int> ());
  }
}

void
LayoutCanvas::pan_center (const db::DPoint &p)
{
  double f = trans ().t2.factor ();

  db::DPoint fp = p * f;

  //  as a safety measure we treat the zero factor case somewhat more graceful
  double dx = fp.x () - double (width ()) * 0.5;
  double dy = fp.y () - double (height ()) * 0.5;
  db::DMagTrans t = db::DMagTrans (db::DTrans (0, db::DPoint (-dx, -dy)), db::DMag (f));

  zoom_to (t);
}

void
LayoutCanvas::zoom_box (const db::DBox &box)
{
  //  use double arithmetics to avoid overflows
  double w = double (box.right ()) - double (box.left ());
  double h = double (box.top ()) - double (box.bottom ());
  double fx = w / double (std::max (width (), 1));
  double fy = h / double (std::max (height (), 1));
  double f = std::max (fx, fy);

  db::DMagTrans t;

  //  as a safety measure we treat the zero factor case somewhat more graceful
  if (f < 1e-13) {
    f = 0.001; // default mag. factor
  }

  double mx = double (box.right ()) + double (box.left ());
  double my = double (box.top ()) + double (box.bottom ());
  double dx = (mx / f - double (width ())) * 0.5;
  double dy = (my / f - double (height ())) * 0.5;
  t = db::DMagTrans (db::DTrans (0, db::DPoint (-dx, -dy)), db::DMag (1.0 / f));

  zoom_to (t);
}

void 
LayoutCanvas::connect_status_bar (lay::StatusBar *sb)
{
  if (! sb && mp_status_bar) {
    mp_status_bar->current_pos ();
  }
  mp_status_bar = sb;
}

void 
LayoutCanvas::current_pos (double x, double y, bool dbu_units)
{
  if (mp_status_bar) {
    mp_status_bar->current_pos (x, y, dbu_units);
  }
}

bool
LayoutCanvas::event (QEvent *e) 
{
   if (e->type () == QEvent::User + 2 ||
       (e->type () == QEvent::User + 1 && m_update_ignore <= 0)) {

    //  do the update and measure the time required for it.
    tl::Timer timer;
    timer.start ();
    update_image ();
    timer.stop ();

    //  If the time required is sufficiently long, ignore the next snapshot update events 
    m_update_ignore = 0;
    if (timer.sec_user () > double (update_interval) * 0.3 * 0.001) {
      m_update_ignore = (unsigned int) (timer.sec_user () / (double (update_interval) * 0.001)) + 1;
    }

    return true;

  } else if (e->type () == QEvent::User + 1) {
    --m_update_ignore;
    return true;
  } else {
    return QWidget::event (e);
  }
}

void
LayoutCanvas::redraw (bool clearing, const std::set<int> &layers)
{
  mp_redraw_thread->stop ();

  if (! m_need_redraw) {
    m_redraw_clearing = false;
    m_need_redraw_layer.clear ();
  }

  m_need_redraw = true;
  if (clearing) {
    m_redraw_clearing = true;
  }
  m_need_redraw_layer.insert (layers.begin (), layers.end ());

  update (); // produces a paintEvent()
}

void
LayoutCanvas::stop_redraw ()
{
  mp_redraw_thread->stop ();
}

void
LayoutCanvas::update_drawings ()
{
  update_image ();
}

}  // namespace lay

