
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_layLayoutCanvas
#define HDR_layLayoutCanvas

#include <vector>
#include <map>
#include <set>
#include <utility>

#include <QtCore/QMutex>

#include "dbTrans.h"
#include "dbBox.h"
#include "layViewOp.h"
#include "layViewObject.h"
#include "layBitmap.h"
#include "layDrawing.h"
#include "layDitherPattern.h"


namespace lay
{

class LayoutView;
class RedrawThread;
class ProgressWidget;

/**
 *  @brief A status bar interface 
 *
 *  This is the interface that the main window provides
 *  in order to display status information
 */
class StatusBar
{
public:
  virtual ~StatusBar () { }

  virtual void current_pos (double x, double y, bool dbu_units) = 0;
  virtual void current_pos () = 0;
  virtual ProgressWidget *progress_widget () = 0;
  virtual void show_progress_widget (bool show) = 0;
};

/** 
 *  @brief A "canvas" object to paint layouts on
 *
 *  The purpose of this object is to provide a canvas QWidget
 *  and to manage the auxiliary objects like rulers etc.
 */

class LayoutCanvas 
  : public lay::ViewObjectCanvas,
    public lay::Drawings
{
Q_OBJECT

public:
  LayoutCanvas (QWidget *parent, lay::LayoutView *view, const char *name = "bitmap_view");
  ~LayoutCanvas ();

  void set_colors (QColor background, QColor foreground, QColor active);
  void set_view_ops (std::vector <lay::ViewOp> &view_ops);

  void lock ()
  {
    m_mutex.lock ();
  }

  void unlock ()
  {
    m_mutex.unlock ();
  }

  const std::vector <lay::ViewOp> &get_view_ops () const
  { 
    return m_view_ops; 
  }

  QColor background_color () const
  { 
    return QColor (m_background); 
  }

  QColor foreground_color () const
  { 
    return QColor (m_foreground); 
  }

  QColor active_color () const
  { 
    return QColor (m_active); 
  }

  void clear (unsigned int nlayers);

  void set_bitmap (unsigned int n, const lay::Bitmap &bitmap)
  { 
    lock ();
    if (n < mp_bitmap_buffers.size ()) {
      *(mp_bitmap_buffers [n]) = bitmap; 
    }
    unlock ();
  }

  void set_drawing_bitmap (unsigned int d, unsigned int n, const lay::Bitmap &bitmap)
  { 
    lock ();
    if (d < mp_drawing_bitmap_buffers.size () && n < mp_drawing_bitmap_buffers [d].size ()) {
      *(mp_drawing_bitmap_buffers [d][n]) = bitmap; 
    }
    unlock ();
  }

  unsigned int bitmaps () const
  {
    return mp_bitmap_buffers.size ();
  }

  unsigned int drawing_bitmaps (unsigned int d) const
  {
    return mp_drawing_bitmap_buffers [d].size ();
  }

  QImage screenshot ();

  void update_image ();

  virtual void paintEvent (QPaintEvent *);

  void zoom_to (const db::DMagTrans &trans);
  void zoom_box (const db::DBox &box);
  void pan_center (const db::DPoint &p);

  /**
   *  @brief Free internal resources, in particular pixmaps
   */
  void free_resources (); 

  /**
   *  @brief Obtain a new bitmap object
   */
  virtual lay::Bitmap *bitmap (const lay::ViewOp &viewop);

  /**
   *  @brief Free all bitmaps
   */
  void clear_dyn_bitmaps ();
  
  /**
   *  @brief Stop the redraw thread
   */
  void stop_redraw ();

  /**
   *  @brief Start the redraw or issue an request to do so
   */
  void redraw (bool clearing, const std::set<int> &layers);

  void connect_status_bar (lay::StatusBar *sb);
  void current_pos (double x, double y, bool dbu_units);

  void set_dither_pattern (const lay::DitherPattern &p);

  const lay::DitherPattern &dither_pattern () const
  {
    return m_dither_pattern;
  }

private:
  lay::LayoutView *mp_view;
  StatusBar *mp_status_bar;
  QImage *mp_image;
  QPixmap *mp_pixmap;
  std::vector <lay::Bitmap *> mp_bitmap_buffers;
  std::vector <std::vector <lay::Bitmap *> > mp_drawing_bitmap_buffers;
  unsigned int m_width;
  unsigned int m_height;
  lay::color_t m_background;
  lay::color_t m_foreground;
  lay::color_t m_active;
  std::vector <lay::ViewOp> m_view_ops;
  lay::DitherPattern m_dither_pattern;

  bool m_need_redraw;
  bool m_redraw_clearing;
  std::set<int> m_need_redraw_layer;

  lay::RedrawThread *mp_redraw_thread;
  unsigned int m_update_ignore;

  std::map <lay::ViewOp, unsigned int> m_dyn_bitmap_table;
  std::vector <lay::Bitmap *> mp_dyn_bitmaps;
  std::vector <lay::ViewOp> m_dyn_view_ops;

  QMutex m_mutex;

  virtual void resizeEvent (QResizeEvent *);
  virtual bool event (QEvent *e);

  //  implementation of the lay::Drawings interface
  void update_drawings ();

  //  clear the bitmaps
  void clear_bitmaps ();
};

} //  namespace lay

#endif

