
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QIcon>
#include <QtGui/QApplication>
#include <QtCore/QDir>

#include "layMainWindow.h"
#include "layVersion.h"

#ifdef _WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif

int 
main (int argc, char *argv [])
{
  QApplication app (argc, argv);

  app.setWindowIcon (QIcon (":logo.png"));

  std::vector <std::string> files;
  std::string layer_props_file;
  bool same_view = false;
  bool sync_mode = false;

  //  locate the layviewrc file in the home directory
  std::string config_file = QDir::home ().absoluteFilePath (".layviewrc").ascii ();

  for (int i = 1; i < argc; ++i) {

    std::string a = argv [i];

    if (a == "-d" && (i + 1) < argc) {

      ++i;
      int v = std::max (0, atoi (argv [i]));
      tl::verbosity (v);

    } else if (a == "-l" && (i + 1) < argc) {

      layer_props_file = argv [++i];

    } else if (a == "-c" && (i + 1) < argc) {

      config_file = argv [++i];

    } else if (a == "-p" && (i + 1) < argc) {

#ifdef _WIN32
      //  there is no "dlopen" on mingw, so we need to emulate it.
      HINSTANCE handle = LoadLibrary (argv[++i]);
      if (! handle) {
        std::cerr << tl::translate ("Unable to load plugin: ") << argv[i] << " " << tl::translate ("with error message: ") << GetLastError () << std::endl;
        exit (1);
      }
#else
      void *handle;
      handle = dlopen (argv[++i], RTLD_LAZY);
      if (! handle) {
        std::cerr << tl::translate ("Unable to load plugin: ") << argv[i] << std::endl;
        exit (1);
      }
#endif

      tl::info << "Loaded plugin '" << argv[i] << "'";

    } else if (a == "-s") {

      same_view = true;
      
    } else if (a == "-x") {

      sync_mode = true;
      
    } else if (a == "-v") {

      std::cout << lay::Version::name () << " " << lay::Version::version () << std::endl;
      exit (0);
      
    } else if (a == "-h") {

      std::cout << lay::Version::exe_name () << " [<options>] [<file>] .." << std::endl;
      std::cout << "options:" << std::endl;
      std::cout << "  -d <debug level>   Set debug level" << std::endl;
      std::cout << "  -l <layer file>    Use layer definition file" << std::endl;
      std::cout << "  -s                 Load files into same view" << std::endl;
      std::cout << "  -x                 Synchroneous redraw mode" << std::endl;
      std::cout << "  -c <config file>   Use this configuration file" << std::endl;
      std::cout << "  -v                 Print program version and exit" << std::endl;
      std::cout << "  -p <plugin>        Load the plugin (can be used multiple times)" << std::endl;
      exit (0);

    } else if (a[0] == '-') {

      std::cerr << tl::translate ("Unknown option: ") << a << std::endl;
      exit (1);

    } else {
      files.push_back (a);
    }

  }

  lay::RedrawThread::set_synchroneous (sync_mode);

  lay::MainWindow *mw = new lay::MainWindow (&app);
  app.setMainWidget (mw);
  mw->setCaption ((std::string (lay::Version::name ()) + " " + lay::Version::version ()).c_str ());
  mw->resize (800, 600);
  mw->show ();

  int result = 0;

  try {

    mw->read_config (config_file);

    for (std::vector <std::string>::const_iterator f = files.begin (); f != files.end (); ++f) {
      mw->load_layout (*f, same_view ? 2 /*same view*/ : 1 /*new view*/);
    }

    if (! layer_props_file.empty ()) {
      mw->load_layer_props (layer_props_file, true /*all views*/);
    }

    result = app.exec ();

    mw->write_config (config_file); 

  } catch (std::exception &ex) {
    std::cerr << "*** ERROR: " << ex.what () << std::endl;
    result = 1;
  } catch (tl::Exception &ex) {
    std::cerr << "*** ERROR: " << ex.msg () << std::endl;
    result = 1;
  } catch (...) {
    std::cerr << "*** ERROR: unspecific error" << std::endl;
    result = 1;
  }

  return result;

}


