
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <iostream>
#include <fstream>

#include <QtGui/QWindowsStyle>
#include <QtGui/QCDEStyle>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QToolBar>
#include <QtGui/QStatusBar>
#include <QtGui/QInputDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QPainter>

#include "tlInternational.h"
#include "tlTimer.h"
#include "tlLog.h"
#include "dbMemStatistics.h"
#include "dbManager.h"
#include "dbClipboard.h"
#include "layVersion.h"
#include "layMainWindow.h"
#include "layProgress.h"
#include "layLayerControlPanel.h" // because of LabelWithBackground
#include "layFileDialog.h"
#include "layRedrawThread.h"
#include "SettingsFormImpl.h"
#include "CellSelectionFormImpl.h"
#include "SelectCellViewFormImpl.h"
#include "MainConfigPage.h"
#include "MainConfigPage2.h"
#include "tlXMLParser.h"

namespace lay
{

// ------------------------------------------------------------
//  Declaration of the configuration options
//  The configuration pages are declared via two "dummy" plugins

const std::string cfg_grid ("grid-micron");
const std::string cfg_synchronized_views ("synchronized-views");


class MainPluginDeclaration
  : public lay::PluginDeclaration
{
public:
  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const
  {
    options.push_back (std::pair<std::string, std::string> (cfg_grid, "0.001"));
  }

  virtual lay::ConfigPage *config_page (QWidget *parent, std::string &title) const 
  {
    title = tl::translate ("Grid");
    return new MainConfigPage (parent); 
  }
};

static tl::Registrar<lay::PluginDeclaration>::Class config_decl (new MainPluginDeclaration (), 1000, "MainPlugin(1)");

class MainPluginDeclaration2
  : public lay::PluginDeclaration
{
public:
  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const
  {
    options.push_back (std::pair<std::string, std::string> (cfg_synchronized_views, "false"));
  }

  virtual lay::ConfigPage *config_page (QWidget *parent, std::string &title) const 
  {
    title = tl::translate ("Misc.");
    return new MainConfigPage2 (parent); 
  }
};

static tl::Registrar<lay::PluginDeclaration>::Class config_decl2 (new MainPluginDeclaration2 (), 1001, "MainPlugin(2)");

// -------------------------------------------------------------
//  The "grid" configuration page

MainConfigPage::MainConfigPage (QWidget *parent)
  : lay::ConfigPage (parent)
{
  mp_ui = new Ui::MainConfigPage ();
  mp_ui->setupUi (this);
}

MainConfigPage::~MainConfigPage ()
{
  delete mp_ui;
  mp_ui = 0;
}

void 
MainConfigPage::setup (lay::Plugin *root)
{
  double grid_micron = 0.0;
  root->config_get (cfg_grid, grid_micron);
  mp_ui->grid_edit->setText (tl::to_string (grid_micron).c_str ());
}

void 
MainConfigPage::commit (lay::Plugin *root)
{
  try {
    double g;
    tl::from_string (mp_ui->grid_edit->text ().ascii (), g);
    root->config_set (cfg_grid, g);
  } catch (...) { }
}

// ------------------------------------------------------------
//  The "misc" config page

MainConfigPage2::MainConfigPage2 (QWidget *parent)
  : lay::ConfigPage (parent)
{
  mp_ui = new Ui::MainConfigPage2 ();
  mp_ui->setupUi (this);
}

MainConfigPage2::~MainConfigPage2 ()
{
  delete mp_ui;
  mp_ui = 0;
}

void 
MainConfigPage2::setup (lay::Plugin *root)
{
  bool flag = false;
  root->config_get (cfg_synchronized_views, flag);
  mp_ui->sync_views_cbx->setChecked (flag);
}

void 
MainConfigPage2::commit (lay::Plugin *root)
{
  root->config_set (cfg_synchronized_views, mp_ui->sync_views_cbx->isChecked ());
}

// -------------------------------------------------------------

class ViewWidgetStack 
  : public QWidget
{
public:
  ViewWidgetStack (QWidget *parent = 0, const char *name = 0)
    : QWidget (parent, name)
  {
    mp_bglabel = new LabelWithBackground (this);
    mp_bglabel->setText (tl::translate ("Use File/Open to open a layout").c_str ());
    mp_bglabel->setAlignment (Qt::AlignVCenter | Qt::AlignHCenter);
    mp_bglabel->show ();
  }

  void addWidget (LayoutView *w)
  {
    m_widgets.push_back (w);
    updateGeometry ();
    raiseWidget (w);
  }

  void removeWidget (LayoutView *w)
  {
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child == w) {
        m_widgets.erase (child, child + 1);
        if (m_widgets.size () == 0) {
          mp_bglabel->show ();
        }
        return;
      }
    }
  }

  void raiseWidget (LayoutView *w)
  {
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child != w) {
        (*child)->deactivate ();
      }
    }
    mp_bglabel->hide ();
    w->activate ();
    w->show ();
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child != w) {
        (*child)->hide ();
      }
    }
  }
  
  LayoutView *widget (size_t index)
  {
    if (index < m_widgets.size ()) {
      return m_widgets [index];
    } else {
      return 0;
    }
  }

  QWidget *background_widget ()
  {
    return mp_bglabel;
  }

protected:
  virtual void resizeEvent (QResizeEvent *)
  {
    updateGeometry ();
  }

  void updateGeometry ()
  {
    //  set the geometry of all children 
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      (*child)->setGeometry (0, 0, width (), height ());
    }
    mp_bglabel->setGeometry (0, 0, width (), height ());
  }

  std::vector <LayoutView *> m_widgets;
  QLabel *mp_bglabel;
};

// -------------------------------------------------------------

/**
 *  @brief Create a action from a string 
 *
 *  The format of the string is: <text>["("shortcut")"]["<"icon-resource">"]
 *  ..
 */
QAction *
create_action (const std::string &s, QWidget *parent)
{
  QString title;
  QString shortcut;
  QString res;

  bool has_shortcut = false;
  bool has_icon = false;

  const char *p = s.c_str ();
  while (*p && *p != '(' && *p != '<') {
    title += *p++;
  }

  if (*p == '(') {
    has_shortcut = true;
    ++p;
    while (*p && *p != ')') {
      shortcut += *p++;
    }
    if (*p == ')') {
      ++p;
    }
    while (*p && isspace (*p)) {
      ++p;
    }
  }

  if (*p == '<') {
    has_icon = true;
    ++p;
    while (*p && *p != '>') {
      res += *p++;
    }
  }
  
  if (has_icon) {
    return new QAction (QIcon (res), title, parent);
  } else {
    return new QAction (title, parent);
  }
}


// -------------------------------------------------------------

MainWindow::MainWindow (QApplication *app, const char *name)
    : QMainWindow (0, name), 
      m_mode (lay::LayoutView::default_mode ()),
      mp_setup_form (new SettingsFormImpl (0, this, "setup_form")),
      m_current_view (0),
      m_edits_enabled (true),
      m_disable_tab_selected (false),
      m_grid_micron (0.001), 
      m_synchronized_views (false),
      m_disabled_edits (0),
      m_work_in_progress (false),
      mp_app (app)
{
  mp_pr = new lay::ProgressReporter (this, app);

  QFrame *mf = new QFrame (this);
  setCentralWidget (mf);

  QVBoxLayout *vbl = new QVBoxLayout (mf);
  vbl->setMargin (0);
  vbl->setSpacing (0);

  mp_tab_bar = new QTabBar (mf);
  vbl->addWidget (mp_tab_bar);
  connect (mp_tab_bar, SIGNAL (currentChanged (int)), this, SLOT (view_selected (int)));

  mp_view_stack = new ViewWidgetStack (mf);
  vbl->addWidget (mp_view_stack);

  QToolBar *toolbar = new QToolBar (this);
  toolbar->setMovable (false);
  toolbar->setToolButtonStyle (Qt::ToolButtonTextUnderIcon);
  addToolBar (Qt::TopToolBarArea, toolbar);

  QMenuBar *mbar = menuBar ();

  //  create actions
  mp_open_action = new QAction (tl::translate ("Open").c_str (), this);
  mp_open_same_panel_action = new QAction (tl::translate ("Open In Same Panel").c_str (), this);
  mp_open_new_panel_action = new QAction (tl::translate ("Open In New Panel").c_str (), this);
  mp_clone_action = new QAction (tl::translate ("Clone Panel").c_str (), this);
  mp_close_action = new QAction (tl::translate ("Close").c_str (), this);
  mp_reload_action = new QAction (tl::translate ("Reload").c_str (), this);
  mp_setup_action = new QAction (tl::translate ("Setup").c_str (), this);
  mp_screenshot_action = new QAction (tl::translate ("Screenshot").c_str (), this);
  mp_load_layer_props_action = new QAction (tl::translate ("Load Layer Properties").c_str (), this);
  mp_save_layer_props_action = new QAction (tl::translate ("Save Layer Properties").c_str (), this);
  mp_exit_action = new QAction (tl::translate ("Exit").c_str (), this);

  mp_screenshot_action->setShortcut (QKeySequence ("Print"));
  mp_open_new_panel_action->setShortcut (QKeySequence ("Ctrl+O"));
  mp_open_same_panel_action->setShortcut (QKeySequence ("Shift+Ctrl+O"));
  mp_close_action->setShortcut (QKeySequence ("Ctrl+W"));
  mp_reload_action->setShortcut (QKeySequence ("Ctrl+R"));
  mp_exit_action->setShortcut (QKeySequence ("Ctrl+Q"));

  connect (mp_open_action, SIGNAL (triggered ()), this, SLOT (open_current_view ()));
  connect (mp_open_same_panel_action, SIGNAL (triggered ()), this, SLOT (open_too ()));
  connect (mp_open_new_panel_action, SIGNAL (triggered ()), this, SLOT (open_new_view ()));
  connect (mp_clone_action, SIGNAL (triggered ()), this, SLOT (clone ()));
  connect (mp_close_action, SIGNAL (triggered ()), this, SLOT (close ()));
  connect (mp_reload_action, SIGNAL (triggered ()), this, SLOT (reload_current_view ()));
  connect (mp_setup_action, SIGNAL (triggered ()), this, SLOT (setup ()));
  connect (mp_screenshot_action, SIGNAL (triggered ()), this, SLOT (screen_shot ()));
  connect (mp_load_layer_props_action, SIGNAL (triggered ()), this, SLOT (load_layer_props ()));
  connect (mp_save_layer_props_action, SIGNAL (triggered ()), this, SLOT (save_layer_props ()));
  connect (mp_exit_action, SIGNAL (triggered ()), this, SLOT (exit ()));

  mp_file_menu = new QMenu (this);
  mp_file_menu->addAction (mp_open_action);
  mp_file_menu->addAction (mp_open_same_panel_action);
  mp_file_menu->addAction (mp_open_new_panel_action);
  mp_file_menu->addAction (mp_close_action);
  mp_file_menu->addAction (mp_clone_action);
  mp_file_menu->addAction (mp_reload_action);
  mp_file_menu->insertSeparator ();
  mp_file_menu->addAction (mp_setup_action);
  mp_file_menu->insertSeparator ();
  mp_file_menu->addAction (mp_screenshot_action);
  mp_file_menu->insertSeparator ();
  mp_file_menu->addAction (mp_load_layer_props_action);
  mp_file_menu->addAction (mp_save_layer_props_action);
  mp_file_menu->insertSeparator ();
  mp_file_menu->addAction (mp_exit_action);
  mp_file_menu->setTitle (tl::translate ("&File").c_str ());
  mbar->addMenu (mp_file_menu);

  mp_mode_menu = new QMenu (this);
  mp_mode_menu->setTitle (tl::translate ("Mode").c_str ());

  //  Fill the mode menu file items from the intrinsic mouse modes followed by the 
  //  modes obtained from the plugins

  int mode_index = 0;

  std::vector <std::string> mode_titles;
  lay::LayoutView::intrinsic_mouse_modes (&mode_titles);
  for (std::vector <std::string>::const_iterator t = mode_titles.begin (); t != mode_titles.end (); ++t, ++mode_index) {

    QAction *action = create_action (*t, this);
    mp_mode_menu->addAction (action);
    m_mode_menu_actions.push_back (action);

    action->setCheckable (true);
    action->setData (QVariant (mode_index));
    action->setChecked (mode_index == int (m_mode));
    toolbar->addAction (action);

    connect (action, SIGNAL (triggered ()), this, SLOT (mode ()));

  }

  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls, ++mode_index) {
    std::string title;
    if (cls->implements_mouse_mode (title)) {

      QAction *action = create_action (title, this);
      mp_mode_menu->addAction (action);
      m_mode_menu_actions.push_back (action);

      action->setCheckable (true);
      action->setData (QVariant (mode_index));
      action->setChecked (mode_index == int (m_mode));
      toolbar->addAction (action);

      connect (action, SIGNAL (triggered ()), this, SLOT (mode ()));

    }
  }

  mp_plugin_enabled_menu = new QMenu (this);
  mp_plugin_enabled_menu->setTitle (tl::translate ("Select").c_str ());

  //  make all plugins that return true on "implements_editable" into menu entries and
  //  set up the actions accordingly
  int n_plugin = 0;
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls, ++n_plugin) {
    std::string title;
    QAction *action = 0;
    if (cls->implements_editable (title)) {
      action = mp_plugin_enabled_menu->addAction (title.c_str ());
      action->setData (QVariant (n_plugin));
      m_plugin_enabled.push_back (true);
    } else {
      m_plugin_enabled.push_back (false);
    }
    m_plugin_enable_actions.push_back (action);
  }

  connect (mp_plugin_enabled_menu, SIGNAL (triggered (QAction *)), this, SLOT (enable (QAction *)));

  mp_undo_action = new QAction (tl::translate ("Undo").c_str (), this);
  mp_redo_action = new QAction (tl::translate ("Redo").c_str (), this);
  mp_del_action = new QAction (tl::translate ("Delete").c_str (), this);
  mp_show_properties_action = new QAction (tl::translate ("Properties").c_str (), this);
  mp_copy_action = new QAction (tl::translate ("Copy").c_str (), this);
  mp_cut_action = new QAction (tl::translate ("Cut").c_str (), this);
  mp_paste_action = new QAction (tl::translate ("Paste").c_str (), this);
  mp_cancel_action = new QAction (tl::translate ("Cancel").c_str (), this);

  mp_undo_action->setShortcut (QKeySequence ("Ctrl+Z"));
  mp_redo_action->setShortcut (QKeySequence ("Ctrl+Y"));
  mp_del_action->setShortcut (QKeySequence ("Del"));
  mp_copy_action->setShortcut (QKeySequence ("Ctrl+C"));
  mp_paste_action->setShortcut (QKeySequence ("Ctrl+V"));
  mp_cut_action->setShortcut (QKeySequence ("Ctrl+X"));
  mp_cancel_action->setShortcut (QKeySequence ("Esc"));

  connect (mp_undo_action, SIGNAL (triggered ()), this, SLOT (undo ()));
  connect (mp_redo_action, SIGNAL (triggered ()), this, SLOT (redo ()));
  connect (mp_del_action, SIGNAL (triggered ()), this, SLOT (del ()));
  connect (mp_show_properties_action, SIGNAL (triggered ()), this, SLOT (show_properties ()));
  connect (mp_copy_action, SIGNAL (triggered ()), this, SLOT (copy ()));
  connect (mp_paste_action, SIGNAL (triggered ()), this, SLOT (paste ()));
  connect (mp_cut_action, SIGNAL (triggered ()), this, SLOT (cut ()));
  connect (mp_cancel_action, SIGNAL (triggered ()), this, SLOT (cancel ()));

  mp_edit_menu = new QMenu (this);
  mp_edit_menu->addAction (mp_undo_action);
  mp_edit_menu->addAction (mp_redo_action);
  mp_edit_menu->insertSeparator ();
  mp_edit_menu->addAction (mp_del_action);
  mp_edit_menu->addAction (mp_show_properties_action);
  mp_edit_menu->insertSeparator ();
  mp_edit_menu->addAction (mp_copy_action);
  mp_edit_menu->addAction (mp_paste_action);
  mp_edit_menu->addAction (mp_cut_action);
  mp_edit_menu->insertSeparator ();
  mp_edit_menu->addMenu (mp_mode_menu);
  mp_edit_menu->addMenu (mp_plugin_enabled_menu);
  mp_edit_menu->insertSeparator ();
  //  @@@ good menu layout?:
  mp_edit_menu->addAction (mp_cancel_action);
  mp_edit_menu->insertSeparator ();
  mp_edit_menu->setTitle (tl::translate ("&Edit").c_str ());
  mbar->addMenu (mp_edit_menu);

  mp_goto_bookmark_menu = new QMenu (this);
  connect (mp_goto_bookmark_menu, SIGNAL (triggered (QAction *)), this, SLOT (goto_bookmark (QAction *)));
  mp_goto_bookmark_menu->setTitle (tl::translate ("Goto Bookmark").c_str ());

  mp_bookmark_view_action = new QAction (tl::translate ("Bookmark This View").c_str (), this);
  mp_manage_bookmarks_action = new QAction (tl::translate ("Manage Bookmarks").c_str (), this);
  mp_load_bookmarks_action = new QAction (tl::translate ("Load Bookmarks").c_str (), this);
  mp_save_bookmarks_action = new QAction (tl::translate ("Save Bookmarks").c_str (), this);

  mp_bookmark_view_action->setShortcut (QKeySequence ("Ctrl+D"));

  connect (mp_bookmark_view_action, SIGNAL (triggered ()), this, SLOT (bookmark_view ()));
  connect (mp_manage_bookmarks_action, SIGNAL (triggered ()), this, SLOT (manage_bookmarks ()));
  connect (mp_load_bookmarks_action, SIGNAL (triggered ()), this, SLOT (load_bookmarks ()));
  connect (mp_save_bookmarks_action, SIGNAL (triggered ()), this, SLOT (save_bookmarks ()));

  mp_bookmark_menu = new QMenu (this);
  mp_bookmark_menu->setTitle (tl::translate ("&Bookmarks").c_str ());
  mp_bookmark_menu->addAction (mp_bookmark_view_action);
  mp_goto_bookmark_action = mp_bookmark_menu->addMenu (mp_goto_bookmark_menu);
  mp_bookmark_menu->insertSeparator ();
  mp_bookmark_menu->addAction (mp_load_bookmarks_action);
  mp_bookmark_menu->addAction (mp_save_bookmarks_action);
  mp_bookmark_menu->addAction (mp_manage_bookmarks_action);
  mbar->addMenu (mp_bookmark_menu);

  mp_max_hier_action = new QAction (tl::translate ("Full Hierarchy").c_str (), this);
  mp_max_hier_0_action = new QAction (tl::translate ("Box Only").c_str (), this);
  mp_max_hier_1_action = new QAction (tl::translate ("Top Level Only").c_str (), this);
  mp_inc_max_hier_action = new QAction (tl::translate ("Increment Hierarchy").c_str (), this);
  mp_dec_max_hier_action = new QAction (tl::translate ("Decrement Hierarchy").c_str (), this);
  mp_zoom_fit_action = new QAction (tl::translate ("Zoom Fit").c_str (), this);
  mp_zoom_in_action = new QAction (tl::translate ("Zoom In").c_str (), this);
  mp_zoom_out_action = new QAction (tl::translate ("Zoom Out").c_str (), this);
  mp_pan_up_action = new QAction (tl::translate ("Pan Up").c_str (), this);
  mp_pan_down_action = new QAction (tl::translate ("Pan Down").c_str (), this);
  mp_pan_left_action = new QAction (tl::translate ("Pan Left").c_str (), this);
  mp_pan_right_action = new QAction (tl::translate ("Pan Right").c_str (), this);
  mp_redraw_action = new QAction (tl::translate ("Redraw").c_str (), this);
  mp_last_display_state_action = new QAction (tl::translate ("Last State").c_str (), this);
  mp_next_display_state_action = new QAction (tl::translate ("Next State").c_str (), this);
  mp_select_cell_action = new QAction (tl::translate ("Select Cell").c_str (), this);
  mp_select_current_cell_action = new QAction (tl::translate ("Select Current Cell").c_str (), this);
  mp_goto_position_action = new QAction (tl::translate ("Goto Position").c_str (), this);

  mp_max_hier_action->setShortcut (QKeySequence ("*"));
  mp_max_hier_0_action->setShortcut (QKeySequence ("0"));
  mp_max_hier_1_action->setShortcut (QKeySequence ("1"));
  mp_inc_max_hier_action->setShortcut (QKeySequence ("+"));
  mp_dec_max_hier_action->setShortcut (QKeySequence ("-"));
  mp_zoom_fit_action->setShortcut (QKeySequence ("F2"));
  mp_zoom_in_action->setShortcut (QKeySequence ("Return"));
  mp_zoom_out_action->setShortcut (QKeySequence ("Shift+Return"));
  mp_pan_up_action->setShortcut (QKeySequence ("Up"));
  mp_pan_down_action->setShortcut (QKeySequence ("Down"));
  mp_pan_left_action->setShortcut (QKeySequence ("Left"));
  mp_pan_right_action->setShortcut (QKeySequence ("Right"));
  mp_last_display_state_action->setShortcut (QKeySequence ("Shift+Tab"));
  mp_next_display_state_action->setShortcut (QKeySequence ("Tab"));
  mp_select_current_cell_action->setShortcut (QKeySequence ("Ctrl+S")); 
  mp_goto_position_action->setShortcut (QKeySequence ("Ctrl+G"));

  connect (mp_max_hier_action, SIGNAL (triggered ()), this, SLOT (max_hier ()));
  connect (mp_max_hier_0_action, SIGNAL (triggered ()), this, SLOT (max_hier_0 ()));
  connect (mp_max_hier_1_action, SIGNAL (triggered ()), this, SLOT (max_hier_1 ()));
  connect (mp_inc_max_hier_action, SIGNAL (triggered ()), this, SLOT (inc_max_hier ()));
  connect (mp_dec_max_hier_action, SIGNAL (triggered ()), this, SLOT (dec_max_hier ()));
  connect (mp_zoom_fit_action, SIGNAL (triggered ()), this, SLOT (zoom_fit ()));
  connect (mp_zoom_in_action, SIGNAL (triggered ()), this, SLOT (zoom_in ()));
  connect (mp_zoom_out_action, SIGNAL (triggered ()), this, SLOT (zoom_out ()));
  connect (mp_pan_up_action, SIGNAL (triggered ()), this, SLOT (pan_up ()));
  connect (mp_pan_down_action, SIGNAL (triggered ()), this, SLOT (pan_down ()));
  connect (mp_pan_left_action, SIGNAL (triggered ()), this, SLOT (pan_left ()));
  connect (mp_pan_right_action, SIGNAL (triggered ()), this, SLOT (pan_right ()));
  connect (mp_redraw_action, SIGNAL (triggered ()), this, SLOT (redraw ()));
  connect (mp_last_display_state_action, SIGNAL (triggered ()), this, SLOT (last_display_state ()));
  connect (mp_next_display_state_action, SIGNAL (triggered ()), this, SLOT (next_display_state ()));
  connect (mp_select_cell_action, SIGNAL (triggered ()), this, SLOT (select_cell ()));
  connect (mp_select_current_cell_action, SIGNAL (triggered ()), this, SLOT (select_current_cell ()));
  connect (mp_goto_position_action, SIGNAL (triggered ()), this, SLOT (goto_position ()));

  mp_zoom_menu = new QMenu (this);
  mp_zoom_menu->setTitle (tl::translate ("&Display").c_str ());
  mp_zoom_menu->addAction (mp_max_hier_action);
  mp_zoom_menu->addAction (mp_max_hier_0_action);
  mp_zoom_menu->addAction (mp_max_hier_1_action);
  mp_zoom_menu->addAction (mp_inc_max_hier_action);
  mp_zoom_menu->addAction (mp_dec_max_hier_action);
  mp_zoom_menu->insertSeparator ();
  mp_zoom_menu->addAction (mp_zoom_fit_action);
  mp_zoom_menu->addAction (mp_zoom_in_action);
  mp_zoom_menu->addAction (mp_zoom_out_action);
  mp_zoom_menu->insertSeparator ();
  mp_zoom_menu->addAction (mp_pan_up_action);
  mp_zoom_menu->addAction (mp_pan_down_action);
  mp_zoom_menu->addAction (mp_pan_left_action);
  mp_zoom_menu->addAction (mp_pan_right_action);
  mp_zoom_menu->insertSeparator ();
  mp_zoom_menu->addAction (mp_redraw_action);
  mp_zoom_menu->insertSeparator ();
  mp_zoom_menu->addAction (mp_last_display_state_action);
  mp_zoom_menu->addAction (mp_next_display_state_action);
  mp_zoom_menu->insertSeparator ();
  mp_zoom_menu->addAction (mp_select_cell_action);
  mp_zoom_menu->addAction (mp_select_current_cell_action);
  mp_zoom_menu->addAction (mp_goto_position_action);
  mbar->addMenu (mp_zoom_menu);

  //  @@@ currently this is where all the custom menus go ..
  QMenu *tools_menu = new QMenu (this);
  tools_menu->setTitle (tl::translate ("&Tools").c_str ());
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    std::vector<lay::MenuEntry> menu_entries;
    cls->get_menu_entries (menu_entries);
    for (std::vector<lay::MenuEntry>::const_iterator m = menu_entries.begin (); m != menu_entries.end (); ++m) {
      QAction *action = tools_menu->addAction (QString (m->title.c_str ()));
      action->setData (QVariant (QString (m->symbol.c_str ())));
      //  make these actions "edit" actions that are disabled or enabled of edits are disabled or enabled
      m_edit_actions.push_back (action);
    }
  }
  connect (tools_menu, SIGNAL (triggered (QAction *)), this, SLOT (generic_menu (QAction *)));
  mbar->addMenu (tools_menu);
  //  @@@
  
  mbar->insertSeparator ();

  QAction *about_action = new QAction (tl::translate ("About").c_str (), this);
  connect (about_action, SIGNAL (triggered ()), this, SLOT (help_about ()));
  QMenu *help_menu = new QMenu (this);
  help_menu->setTitle (tl::translate ("&Help").c_str ());
  help_menu->addAction (about_action);
  mbar->addMenu (help_menu);

  mp_status_bar = statusBar ();

  mp_msg_label = new QLabel (mp_status_bar, "msg_label");
  mp_status_bar->addWidget (mp_msg_label, 1);

  mp_right_stack = new QStackedWidget (mp_status_bar);

  mp_cp_frame = new QFrame (mp_right_stack);
  mp_right_stack->addWidget (mp_cp_frame);

  QHBoxLayout *cp_frame_ly = new QHBoxLayout (mp_cp_frame);
  cp_frame_ly->setMargin (0);
  cp_frame_ly->setSpacing (0);
  QLabel *xl = new QLabel (" x:", mp_cp_frame, "x");
  cp_frame_ly->addWidget (xl);
  mp_cpx_label = new QLabel (mp_cp_frame, "cpx_label");
  mp_cpx_label->setAlignment (Qt::AlignVCenter | Qt::AlignRight);
  mp_cpx_label->setMinimumSize (100, 0); // @@@
  cp_frame_ly->addWidget (mp_cpx_label);
  QLabel *yl = new QLabel (" y:", mp_cp_frame, "y");
  cp_frame_ly->addWidget (yl);
  mp_cpy_label = new QLabel (mp_cp_frame, "cpy_label");
  mp_cpy_label->setAlignment (Qt::AlignVCenter | Qt::AlignRight);
  mp_cpy_label->setMinimumSize (100, 0); // @@@
  cp_frame_ly->addWidget (mp_cpy_label);

  mp_progress_widget = new ProgressWidget (mp_pr, mp_right_stack);
  mp_right_stack->addWidget (mp_progress_widget);

  mp_right_stack->setCurrentIndex (0);
  mp_status_bar->addWidget (mp_right_stack);

  connect (mp_edit_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));
  connect (mp_zoom_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));
  connect (mp_bookmark_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));
  connect (mp_goto_bookmark_menu, SIGNAL (aboutToShow ()), this, SLOT (goto_bookmark_menu_show ()));

  m_edit_actions.push_back (mp_open_action);
  m_edit_actions.push_back (mp_open_same_panel_action);
  m_edit_actions.push_back (mp_open_new_panel_action);
  m_edit_actions.push_back (mp_close_action);
  m_edit_actions.push_back (mp_reload_action);
  m_edit_actions.push_back (mp_setup_action);
  m_edit_actions.push_back (mp_screenshot_action);
  m_edit_actions.push_back (mp_load_layer_props_action);
  m_edit_actions.push_back (mp_save_layer_props_action);
  m_edit_actions.push_back (mp_undo_action);
  m_edit_actions.push_back (mp_redo_action);
  m_edit_actions.push_back (mp_del_action);
  m_edit_actions.push_back (mp_show_properties_action);
  m_edit_actions.push_back (mp_copy_action);
  m_edit_actions.push_back (mp_cut_action);
  m_edit_actions.push_back (mp_paste_action);
  m_edit_actions.push_back (mp_select_cell_action);

  //  create file dialogs:

  //  bookmarks file dialog
  mp_bookmarks_fdia = new lay::FileDialog (this, 
                            tl::translate ("Bookmarks File"), 
                            tl::translate ("All files (*);;Bookmark files (*.lyb)"), 
                            "lyb");
  //  layer properties
  mp_lprops_fdia = new lay::FileDialog (this, 
                            tl::translate ("Layer Properties File"), 
                            tl::translate ("All files (*);;Layer properties files (*.lyp)"), 
                            "lyp");
  //  screenshots
  mp_screenshot_fdia = new lay::FileDialog (this, 
                            tl::translate ("Screenshot"), 
                            tl::translate ("All files (*);;PNG files (*.png)"), 
                            "png");

  //  layout file dialog
  mp_layout_fdia = new lay::FileDialog (this,
                            tl::translate ("Layout File"), 
                            tl::translate ("All files (*);;GDS files (*.GDS *.gds *.gds.gz *.GDS.gz);;OASIS files (*.OAS *.oas *.oas.gz *.OAS.gz)"));

}

MainWindow::~MainWindow ()
{
  delete mp_pr;
  mp_pr = 0;

  //  explicitly delete the views here. Otherwise they
  //  are deleted by ~QWidget, which is too late since then
  //  the manager does not exist any longer.
  //  First empty the mp_views vector and then delete. This way,
  //  any callbacks issued during the deleting of the views do
  //  not find any invalid view pointers but rather nothing.

  std::vector <lay::LayoutView *> views;
  views.swap (mp_views);

  while (views.size () > 0) {
    delete views.back ();
    views.pop_back ();
  }
}

void
MainWindow::config_finalize ()
{
  //  make the changes visible in the setup form if the form is visible 
  mp_setup_form->setup ();
}

bool 
MainWindow::configure (const std::string &name, const std::string &value)
{
  if (name == cfg_grid) {

    double g;
    tl::from_string (value, g);
    m_grid_micron = g;
    return false; // not taken - let others set it too.

  } else if (name == cfg_synchronized_views) {

    bool flag;
    tl::from_string (value, flag);
    m_synchronized_views = flag;
    return true;

  } else {
    return false;
  }

}

void 
MainWindow::enable_edits (bool enable)
{
  if (enable) {
    if (m_disabled_edits > 0) {
      --m_disabled_edits;
    }
    enable = (m_disabled_edits == 0);
  } else {
    ++m_disabled_edits;
  }

  for (std::vector<QAction *>::iterator a = m_edit_actions.begin (); a != m_edit_actions.end (); ++a) {
    (*a)->setEnabled (enable);
  }

  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->enable_edits (enable);
  }
  
  m_edits_enabled = enable;

  //  call this to establish a consistent set of options on the menu entries
  popup_menu_show ();
}

void
MainWindow::exit ()
{
  QMainWindow::close ();
}

lay::LayoutView *
MainWindow::current_view ()
{
  if (m_current_view >= mp_views.size ()) {
    return 0;
  } else {
    return mp_views [m_current_view];
  }
}

const lay::LayoutView *
MainWindow::current_view () const
{
  if (m_current_view >= mp_views.size ()) {
    return 0;
  } else {
    return mp_views [m_current_view];
  }
}

void 
MainWindow::show_properties ()
{
  if (current_view ()) {
    current_view ()->show_properties (this);
  }
}

void 
MainWindow::del ()
{
  if (current_view ()) {
    current_view ()->del ();
    //  because a "delete" might involve objects currently edited, we cancel the edit after we have deleted the object
    current_view ()->cancel ();
    current_view ()->clear_selection ();
  }
}

void 
MainWindow::copy ()
{
  if (current_view () && current_view ()->has_selection ()) {
    current_view ()->copy ();
    current_view ()->clear_selection ();
  }
}

void 
MainWindow::paste ()
{
  if (current_view () && ! db::Clipboard::instance ().empty ()) {
    current_view ()->cancel ();
    current_view ()->clear_selection ();
    current_view ()->paste ();
  }
}

void 
MainWindow::cut ()
{
  if (current_view () && current_view ()->has_selection ()) {
    current_view ()->cut ();
    current_view ()->cancel (); //  see del() for reason why cancel is after cut
    current_view ()->clear_selection ();
  }
}

void 
MainWindow::mode ()
{
  QAction *action = dynamic_cast<QAction *> (sender ());
  if (action) {

    unsigned int m = (unsigned int) action->data ().toInt ();
    m_mode = m;
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->mode (m);
    }

    //  update mode menu
    for (std::vector<QAction *>::const_iterator i = m_mode_menu_actions.begin (); i != m_mode_menu_actions.end (); ++i) {
      (*i)->setEnabled (m_edits_enabled);
      (*i)->setChecked ((*i)->data ().toInt () == int (m_mode));
    }

  }
}

void 
MainWindow::enable (QAction *action)
{
  int m = action->data ().toInt ();
  m_plugin_enabled [m] = ! m_plugin_enabled [m];
  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->enable_editable (m, m_plugin_enabled [m]);
  }
}

void 
MainWindow::undo ()
{
  if (current_view () && current_view ()->manager ()->available_undo ().first) {
    current_view ()->clear_selection ();
    current_view ()->cancel ();
    current_view ()->undo ();
    current_view ()->update_content ();
  }
}

void
MainWindow::redo ()
{
  if (current_view () && current_view ()->manager ()->available_redo ().first) {
    current_view ()->clear_selection ();
    current_view ()->cancel ();
    current_view ()->redo ();
    current_view ()->update_content ();
  }
}

void 
MainWindow::goto_bookmark_menu_show ()
{
  mp_goto_bookmark_menu->clear ();

  if (current_view ()) {
    const lay::BookmarkList &bookmarks = current_view ()->bookmarks ();
    for (size_t i = 0; i < bookmarks.size (); ++i) {
      QAction *action = mp_goto_bookmark_menu->addAction (bookmarks.name (i).c_str ());
      action->setData (QVariant (int (i))); 
    }
  }
}

void
MainWindow::goto_bookmark (QAction *action)
{
  size_t id = size_t (action->data ().toInt ());
  if (current_view () && current_view ()->bookmarks ().size () > id) {
    current_view ()->goto_view (current_view ()->bookmarks ().state (id));
  }
}

void
MainWindow::goto_position ()
{
  if (current_view ()) {

    while (true) {

      bool ok = false;

      db::DBox box (current_view ()->box ());
      std::string pos;
      pos += tl::micron_to_string (box.center ().x ()) + "," + tl::micron_to_string (box.center ().y ());
      pos += ",";
      pos += tl::micron_to_string (std::min (box.width (), box.height ()));

      QString text = QInputDialog::getText (tl::translate ("Enter Position").c_str (), tl::translate ("Enter position either as pair (x,y) or with window size (x,y,s)\n(x,y) will be the new window center position, s (if given) the new window size").c_str (),
                                            QLineEdit::Normal, QString (pos.c_str ()), &ok, this);

      if (! ok) {

        break;

      } else if (text.isEmpty ()) {

        QMessageBox::critical (this, tl::translate ("Error").c_str (), tl::translate ("Enter the position").c_str ());

      } else {

        double x, y, s;
        if (sscanf (text.ascii (), "%lg,%lg,%lg", &x, &y, &s) == 3) {
          current_view ()->goto_window (db::DPoint (x, y), s);
        } else if (sscanf (text.ascii (), "%lg,%lg", &x, &y) == 2) {
          current_view ()->goto_window (db::DPoint (x, y));
        } else {
          QMessageBox::critical (this, tl::translate ("Error").c_str (), tl::translate ("Not a valid position").c_str ());
        }
        break;

      }

    }

  }
}

void
MainWindow::manage_bookmarks ()
{
  try {
    if (current_view ()) {
      current_view ()->manage_bookmarks ();
    }
  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::bookmark_view ()
{
  try {
    if (current_view ()) {
      while (true) {
        bool ok = false;
        QString text = QInputDialog::getText (tl::translate ("Enter Bookmark Name").c_str (), tl::translate ("Bookmark name").c_str (),
                                              QLineEdit::Normal, QString::null, &ok, this);
        if (! ok) {
          break;
        } else if (text.isEmpty ()) {
          QMessageBox::critical (this, tl::translate ("Error").c_str (), tl::translate ("Enter a name for the bookmark").c_str ());
        } else {
          current_view ()->bookmark_view (text.ascii ());
          break;
        }
      }
    }
  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::update_content ()
{
  mp_setup_form->setup ();
  if (current_view ()) {
    current_view ()->update_content ();
  }
}

void
MainWindow::zoom_fit ()
{
  if (current_view ()) {
    current_view ()->zoom_fit ();
  }
}

void
MainWindow::pan_left ()
{
  if (current_view ()) {
    current_view ()->pan_left ();
  }
}

void
MainWindow::pan_right ()
{
  if (current_view ()) {
    current_view ()->pan_right ();
  }
}

void
MainWindow::pan_up ()
{
  if (current_view ()) {
    current_view ()->pan_up ();
  }
}

void
MainWindow::pan_down ()
{
  if (current_view ()) {
    current_view ()->pan_down ();
  }
}

void
MainWindow::zoom_in ()
{
  if (current_view ()) {
    current_view ()->zoom_in ();
  }
}

void
MainWindow::zoom_out ()
{
  if (current_view ()) {
    current_view ()->zoom_out ();
  }
}

void 
MainWindow::popup_menu_show ()
{
  std::string undo_txt (tl::translate ("&Undo"));
  bool undo_enable = false;
  if (current_view () && current_view ()->manager ()->available_undo ().first) {
    undo_txt += " - " + current_view ()->manager ()->available_undo ().second;
    undo_enable = true;
  }
  mp_undo_action->setText (undo_txt.c_str ()); 
  mp_undo_action->setEnabled (undo_enable && m_edits_enabled);

  std::string redo_txt (tl::translate ("&Redo"));
  bool redo_enable = false;
  if (current_view () && current_view ()->manager ()->available_redo ().first) {
    undo_txt += " - " + current_view ()->manager ()->available_redo ().second;
    redo_enable = true;
  }
  mp_redo_action->setText (redo_txt.c_str ()); 
  mp_redo_action->setEnabled (redo_enable && m_edits_enabled);

  mp_copy_action->setEnabled (current_view () && current_view ()->has_selection () && m_edits_enabled);
  mp_cut_action->setEnabled (current_view () && current_view ()->has_selection () && m_edits_enabled);
  mp_paste_action->setEnabled (! db::Clipboard::instance ().empty () && m_edits_enabled);

  mp_next_display_state_action->setEnabled (has_next_display_state ());
  mp_last_display_state_action->setEnabled (has_last_display_state ());

  bool has_bookmarks = current_view () && current_view ()->bookmarks ().size () > 0;
  mp_goto_bookmark_action->setEnabled (has_bookmarks && m_edits_enabled);

  //  update mode menu
  int n = 0;
  for (std::vector<QAction *>::const_iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i, ++n) {
    if (*i) {
      (*i)->setCheckable (true);
      (*i)->setEnabled (m_edits_enabled);
      (*i)->setChecked (m_plugin_enabled [n]);
    }
  }
}

void
MainWindow::redraw ()
{
  if (current_view ()) {
    current_view ()->redraw ();
  }
}

void
MainWindow::cancel ()
{
  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->cancel ();
  }
}

void
MainWindow::save_layer_props ()
{
  try {

    if (current_view ()) {
      std::string fn;
      if (mp_lprops_fdia->get_save (fn)) {
        current_view ()->save_layer_props (fn);
      }
    } else {
      throw tl::Exception (tl::translate ("No view open to save the layer properties from"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::load_layer_props (const std::string &fn, bool all_views)
{
  if (all_views) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->load_layer_props (fn);
    }
  } else if (current_view ()) {
    current_view ()->load_layer_props (fn);
  }
}

void
MainWindow::load_layer_props ()
{
  try {

    if (current_view ()) {
      std::string fn;
      if (mp_lprops_fdia->get_open (fn)) {
        load_layer_props (fn, false /*current view only*/);
      }
    } else {
      throw tl::Exception (tl::translate ("No view open to load the layer properties for"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::save_bookmarks ()
{
  try {

    if (current_view ()) {
      std::string fn;
      if (mp_bookmarks_fdia->get_save (fn)) {
        current_view ()->bookmarks ().save (fn);
      }
    } else {
      throw tl::Exception (tl::translate ("No view open to save the bookmarks from"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::load_bookmarks ()
{
  try {

    if (current_view ()) {
      std::string fn;
      if (mp_bookmarks_fdia->get_open (fn)) {
        BookmarkList bookmarks;
        bookmarks.load (fn);
        current_view ()->bookmarks (bookmarks);
      }
    } else {
      throw tl::Exception (tl::translate ("No view open to load the bookmarks for"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::select_current_cell ()
{
  try {

    lay::LayoutView *view = current_view ();
    if (view) {
      lay::LayoutView::cell_path_type path;
      int cvi = view->active_cellview_index ();
      view->current_cell_path (path);
      view->select_cell_fit (path, cvi);
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  }
}

void
MainWindow::select_cell ()
{
  try {

    if (current_view ()) {

      CellSelectionFormImpl form (0, current_view (), "cell_selection_form");

      if (form.exec () == QDialog::Accepted && 
          form.selected_cellview_index () >= 0) {
        current_view ()->select_cell (form.selected_cellview ().path (), form.selected_cellview_index ()); 
        current_view ()->zoom_fit ();
      }

    } else {
      throw tl::Exception (tl::translate ("No view open to select a cell for"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 

}

void
MainWindow::screen_shot ()
{
  try {

    if (current_view ()) {
      std::string fn;
      if (mp_screenshot_fdia->get_save (fn)) {
        current_view ()->save_screenshot (fn);
      }
    } else {
      throw tl::Exception (tl::translate ("No view open to create a screenshot from"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::setup ()
{
  mp_setup_form->show ();
  mp_setup_form->setup ();
}

void
MainWindow::view_selected (int index)
{
  if (! m_disable_tab_selected) {

    bool trans_set = (m_synchronized_views && current_view () != 0);
    db::DMagTrans trans;
    if (trans_set) {
      trans = current_view ()->canvas ()->trans ();
    }

    m_current_view = index;

    if (current_view ()) {
      if (trans_set) {
        current_view ()->zoom_to (trans);
      }
      mp_view_stack->raiseWidget (current_view ());
      mp_setup_form->setup ();
    }

  }
}

void 
MainWindow::clone ()
{
  try {

    lay::LayoutView *view = 0;
    lay::LayoutView *curr = current_view ();
    if (! curr) {
      throw tl::Exception (tl::translate ("No view open to clone"));
    }

    //  create a new view
    view = new lay::LayoutView (current_view (), this, mp_view_stack);
    mp_views.push_back (view);

    //  we must resize the widget here to set the geometry properly. 
    //  This is required to make zoom_fit work.
    view->setGeometry (0, 0, mp_view_stack->width (), mp_view_stack->height ());
    view->show ();

    //  set initial attributes
    view->hier_levels (hier_levels ());

    //  select the current mode and select the enabled editables
    view->mode (m_mode);
    for (unsigned int m = 0; m < m_plugin_enabled.size (); ++m) {
      if (m_plugin_enable_actions [m]) {
        view->enable_editable (m, m_plugin_enabled [m]);
      }
    }

    lay::DisplayState state;
    current_view ()->save_view (state);
    view->goto_view (state);

    view->update_content ();

    m_current_view = mp_views.size () - 1;
    mp_view_stack->addWidget (view);
    int index = mp_tab_bar->insertTab (-1, view->title ().c_str ());
    mp_tab_bar->setCurrentIndex (index);

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
MainWindow::close ()
{
  if (current_view () && current_view ()->cellviews () > 1) {

    SelectCellViewFormImpl form (0, "select_cv");

    form.set_title (tl::translate ("Select Layout Views To Close"));

    for (unsigned int i = 0; i < current_view ()->cellviews (); ++i) {
      form.tell_cellview (current_view ()->cellview (i));
    }

    if (form.exec () == QDialog::Accepted) {

      if (form.all_selected ()) {
        close_current_view ();
      } else {

        try {

          //  erase the selected cellviews
          std::vector <int> selected = form.selected_cellviews ();
          std::sort (selected.begin (), selected.end ());
          int offset = 0;
          for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
            current_view ()->erase_cellview ((unsigned int)(*i - offset));
            ++offset;
          }

          //  set the tab's text again
          mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());

        } catch (tl::Exception &ex) {
          tl::error << ex.msg ();
          QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
        } 

      }

    }

  } else {
    //  close view completely
    close_current_view ();
  }

}

void 
MainWindow::close_current_view ()
{
  if (current_view ()) {

    try {

      //  this suppresses view_selected events that would otherwise be created
      m_disable_tab_selected = true;

      bool trans_set = false;
      db::DMagTrans trans;
      if (m_synchronized_views) {
        trans = current_view ()->canvas ()->trans ();
        trans_set = true;
      }

      mp_tab_bar->removeTab (m_current_view);
      mp_view_stack->removeWidget (current_view ());
      delete current_view ();
      mp_views.erase (mp_views.begin () + m_current_view, mp_views.begin () + m_current_view + 1);

      if (m_current_view >= mp_views.size () && m_current_view > 0) {
        --m_current_view;
      }

      if (current_view ()) {
        if (trans_set) {
          current_view ()->zoom_to (trans);
        }
        mp_tab_bar->setCurrentTab (m_current_view);
        mp_view_stack->raiseWidget (current_view ());
      }

      m_disable_tab_selected = false;

    } catch (tl::Exception &ex) {
      tl::error << ex.msg ();
      QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
    } 

  }
}

void
MainWindow::reload_current_view ()
{
  try {

    if (current_view ()) {

      if (current_view ()->cellviews () > 1) {

        SelectCellViewFormImpl form (0, "select_cv");

        form.set_title (tl::translate ("Select Layouts To Reload"));

        for (unsigned int i = 0; i < current_view ()->cellviews (); ++i) {
          form.tell_cellview (current_view ()->cellview (i));
        }

        form.select_all ();
      
        if (form.exec () == QDialog::Accepted) {
          std::vector <int> selected = form.selected_cellviews ();
          for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
            reload_layout (*i);
          }
        }
          
      } else if (current_view ()->cellviews () > 0) {
        reload_layout (0);
      }

    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void 
MainWindow::open (int mode)
{
  try {
    static std::string fn;
    if (mp_layout_fdia->get_open (fn)) {
      load_layout (fn, mode);
    }
  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

void
MainWindow::reload_layout (unsigned int cv_index)
{
  lay::LayoutView *view = current_view ();

  if (view && view->cellviews () > cv_index) {
    view->reload_layout (cv_index);
  }
}

void 
MainWindow::load_layout (const std::string &filename, int mode)
{
  lay::LayoutView *view = 0;

  if (! current_view ()) {
    mode = 1;
  }

  if (mode == 1) {

    //  create a new view
    view = new lay::LayoutView (this, mp_view_stack);

    mp_views.push_back (view);

    //  we must resize the widget here to set the geometry properly. 
    //  This is required to make zoom_fit work.
    view->setGeometry (0, 0, mp_view_stack->width (), mp_view_stack->height ());
    view->show ();

    //  set initial attributes
    view->hier_levels (hier_levels ());

    //  select the current mode and select the enabled editables
    view->mode (m_mode);
    for (unsigned int m = 0; m < m_plugin_enabled.size (); ++m) {
      if (m_plugin_enable_actions [m]) {
        view->enable_editable (m, m_plugin_enabled [m]);
      }
    }

  } else {
    view = current_view ();
  }

  try {

    //  load the layout
    view->load_layout (filename, mode == 2);

    //  make the new view the current one
    if (mode == 1) {
      m_current_view = mp_views.size () - 1;
      mp_view_stack->addWidget (mp_views.back ());
      int index = mp_tab_bar->insertTab (-1, current_view ()->title ().c_str ());
      mp_tab_bar->setCurrentIndex (index);
    } else if (mode == 0) {
      mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());
    } else if (mode == 2) {
      mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());
    }

  } catch (...) {

    //  clean up in case of an error ..
    if (mode == 1) {
      delete mp_views.back ();
      mp_views.pop_back ();
    }

    throw;

  }

}

double
MainWindow::grid_micron () const
{
  return m_grid_micron;
}

void 
MainWindow::inc_max_hier ()
{
  unsigned int new_to = max_hier_levels () + 1;
  hier_levels (std::make_pair (min_hier_levels (), new_to));
}

void 
MainWindow::dec_max_hier ()
{
  unsigned int new_to = max_hier_levels () > 0 ? max_hier_levels () - 1 : 0;
  hier_levels (std::make_pair (std::min (min_hier_levels (), new_to), new_to));
}

void 
MainWindow::max_hier ()
{
  if (current_view ()) {
    current_view ()->max_hier ();
  }
}

void 
MainWindow::max_hier_0 ()
{
  hier_levels (std::make_pair (0, 0));
}

void 
MainWindow::max_hier_1 ()
{
  hier_levels (std::make_pair (0, 1));
}

void 
MainWindow::hier_levels (std::pair<unsigned int, unsigned int> l)
{
  if (l != hier_levels ()) {
    if (current_view ()) {
      current_view ()->hier_levels (l);
    }
  } 
}

std::pair<unsigned int, unsigned int> 
MainWindow::hier_levels () const
{
  if (current_view ()) {
    return current_view ()->hier_levels ();
  } else {
    return std::make_pair ((unsigned int) 0, (unsigned int) 0);
  }
}

void 
MainWindow::last_display_state ()
{
  if (has_last_display_state ()) {
    current_view ()->last_display_state ();
  }
}

bool 
MainWindow::has_last_display_state ()
{
  if (current_view ()) {
    return current_view ()->has_last_display_state ();
  } else {
    return false;
  }
}

void 
MainWindow::next_display_state ()
{
  if (has_next_display_state ()) {
    current_view ()->next_display_state ();
  }
}

bool 
MainWindow::has_next_display_state ()
{
  if (current_view ()) {
    return current_view ()->has_next_display_state ();
  } else {
    return false;
  }
}

void
MainWindow::current_pos (double x, double y, bool dbu_units) 
{
  mp_cpx_label->setText (dbu_units ? tl::db_to_string (x).c_str () : tl::micron_to_string (x).c_str ());
  mp_cpy_label->setText (dbu_units ? tl::db_to_string (y).c_str () : tl::micron_to_string (y).c_str ());
}

void
MainWindow::current_pos ()
{
  mp_cpx_label->setText ("");
  mp_cpy_label->setText ("");
}

ProgressWidget *
MainWindow::progress_widget ()
{
  return mp_progress_widget;
}

bool 
MainWindow::eventFilter (QObject *watched, QEvent *event)
{
  if (event->type () == QEvent::MouseButtonDblClick ||
      event->type () == QEvent::MouseButtonPress ||
      event->type () == QEvent::MouseButtonRelease ||
      event->type () == QEvent::MouseMove ||
      event->type () == QEvent::KeyPress ||
      event->type () == QEvent::KeyRelease) {

    QObject *o = watched;
    while (o) {
      //  if the watched object is a child of the progress widget, pass the event on to this ..
      if (o == progress_widget ()) {
        return QObject::eventFilter(watched, event);
      }
      o = o->parent ();
    }
    return true;

  } else {
    // standard event processing
    return QObject::eventFilter(watched, event);
  }
}

void 
MainWindow::show_progress_widget (bool show)
{
  m_work_in_progress = show;
  if (show) {
    //  to avoid recursions of any kind, disallow any user interaction except 
    //  cancelling the operation
    mp_app->installEventFilter (this);
    QApplication::setOverrideCursor (QCursor (Qt::WaitCursor));
  } else {
    mp_app->removeEventFilter (this);
    QApplication::restoreOverrideCursor ();
  }
  mp_right_stack->setCurrentIndex (show ? 1 : 0);
}

//  Writing and Reading of configuration 

struct ConfigGetAdaptor
{
  ConfigGetAdaptor (const std::string &name)
    : mp_owner (0), m_name (name)
  {
    // .. nothing yet ..
  }
  
  std::string operator () () const
  {
    std::string s;
    mp_owner->config_get (m_name, s);
    return s;
  }

  bool at_end () const 
  {
    return m_done;
  }

  void start (const lay::MainWindow &owner) 
  {
    mp_owner = &owner;
    m_done = false;
  }

  void next () 
  {
    m_done = true;
  }

private:
  const lay::MainWindow *mp_owner;
  bool m_done;
  std::string m_name;
};
 
struct ConfigSetAdaptor
{
  ConfigSetAdaptor (const std::string &name)
    : m_name (name)
  {
    // .. nothing yet ..
  }
  
  void operator () (lay::MainWindow &w, const std::string &v) const
  {
    w.config_set (m_name, v);
  }

private:
  std::string m_name;
};
 
//  the configuration file's XML structure is built dynamically
static tl::XMLStruct<lay::MainWindow> 
config_structure ()
{
  tl::XMLElementList body;

  std::vector <std::string> names; 
  lay::Plugin::get_config_names (names);
  for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {
    body.append (tl::XMLMember<std::string, lay::MainWindow, ConfigGetAdaptor, ConfigSetAdaptor, tl::XMLStdConverter <std::string> > ( 
                       ConfigGetAdaptor (*n), ConfigSetAdaptor (*n), *n));
  }

  //  add a wildcard member to ignore all others
  body.append (tl::make_member<lay::MainWindow> ("*"));

  return tl::XMLStruct<lay::MainWindow> ("config", body);
}


bool 
MainWindow::write_config (const std::string &config_file)
{
  std::ofstream out (config_file.c_str ());
  if (! out.good ()) {
    return false;
  }
  config_structure ().write (out, *this); 
  return true;
}

bool
MainWindow::read_config (const std::string &config_file)
{
  try {
    tl::XMLFileSource file (config_file);
    try {
      //  clear the custom dither pattern
      config_structure ().parse (file, *this); 
      config_end ();
      return true;
    } catch (tl::XMLException &ex) {
      std::cerr << tl::translate ("Problem reading config file:") << std::endl;
      std::cerr << "  " << ex.msg () << std::endl;
      return false;
    } 
  } catch (tl::Exception &) {
    return false;
  }
}

void
MainWindow::help_about ()
{
  std::string s = std::string (lay::Version::name ()) + " " + lay::Version::version () +
                  "\n\n" + lay::Version::about_text ();
  std::string t = tl::translate ("About ") + lay::Version::name ();
  QMessageBox::information (this, t.c_str (), s.c_str ());
}

void 
MainWindow::generic_menu (QAction *action)
{
  try {

    if (current_view ()) {
      current_view ()->menu_activated (std::string (action->data ().toString ().toAscii ()));
    } else {
      throw tl::Exception (tl::translate ("No view to browse the instances in"));
    }

  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    QMessageBox::critical (this, tl::translate ("Error").c_str (), ex.msg ().c_str ());
  } 
}

} // namespace lay

