
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_layMainWindow
#define HDR_layMainWindow

#include <QtGui/QStackedWidget>
#include <QtGui/QMainWindow>

#include <vector>

#include "dbManager.h"
#include "layLayoutView.h"
#include "layRedrawThread.h"
#include "layPlugin.h"
#include "tlException.h"

class QTabBar;
class QLabel;
class QAction;
class QMenu;
class QStackedWidget;
class QFileDialog;

class SettingsFormImpl;

namespace Ui {
  class MainConfigPage;
  class MainConfigPage2;
}

namespace lay {

class ViewWidgetStack;
class ProgressReporter;
class FileDialog;

class MainWindow
  : public QMainWindow,
    public lay::StatusBar,
    public lay::PluginRoot
{
Q_OBJECT

private:
  //  Main menu
  QAction *mp_open_action;
  QAction *mp_open_same_panel_action;
  QAction *mp_open_new_panel_action;
  QAction *mp_clone_action;
  QAction *mp_close_action;
  QAction *mp_reload_action;
  QAction *mp_setup_action;
  QAction *mp_screenshot_action;
  QAction *mp_load_layer_props_action;
  QAction *mp_save_layer_props_action;
  QAction *mp_exit_action;
  QAction *mp_undo_action;
  QAction *mp_redo_action;
  QAction *mp_del_action;
  QAction *mp_show_properties_action;
  QAction *mp_copy_action;
  QAction *mp_cut_action;
  QAction *mp_paste_action;
  QAction *mp_cancel_action;
  QAction *mp_bookmark_view_action;
  QAction *mp_manage_bookmarks_action;
  QAction *mp_goto_position_action;
  QAction *mp_load_bookmarks_action;
  QAction *mp_save_bookmarks_action;
  QAction *mp_goto_bookmark_action;
  QAction *mp_max_hier_action;
  QAction *mp_max_hier_0_action;
  QAction *mp_max_hier_1_action;
  QAction *mp_inc_max_hier_action;
  QAction *mp_dec_max_hier_action;
  QAction *mp_zoom_fit_action;
  QAction *mp_zoom_in_action;
  QAction *mp_zoom_out_action;
  QAction *mp_pan_up_action;
  QAction *mp_pan_down_action;
  QAction *mp_pan_left_action;
  QAction *mp_pan_right_action;
  QAction *mp_redraw_action;
  QAction *mp_last_display_state_action;
  QAction *mp_next_display_state_action;
  QAction *mp_select_cell_action;
  QAction *mp_select_current_cell_action;
  QMenu *mp_file_menu;
  QMenu *mp_edit_menu;
  QMenu *mp_zoom_menu;
  QMenu *mp_bookmark_menu;
  QMenu *mp_goto_bookmark_menu;
  QMenu *mp_mode_menu;
  QMenu *mp_plugin_enabled_menu;
  QTabBar *mp_tab_bar;
  ViewWidgetStack *mp_view_stack;
  lay::FileDialog *mp_bookmarks_fdia;
  lay::FileDialog *mp_lprops_fdia;
  lay::FileDialog *mp_screenshot_fdia;
  lay::FileDialog *mp_layout_fdia;
  std::vector<QAction *> m_mode_menu_actions;
  unsigned int m_mode;
  std::vector<QAction *> m_edit_actions;
  std::vector<QAction *> m_plugin_enable_actions;
  std::vector<bool> m_plugin_enabled;
  SettingsFormImpl *mp_setup_form;
  std::vector <lay::LayoutView *> mp_views;
  size_t m_current_view;
  QStatusBar *mp_status_bar;
  QStackedWidget *mp_right_stack;
  ProgressWidget *mp_progress_widget;
  QFrame *mp_cp_frame;
  QLabel *mp_cpx_label, *mp_cpy_label;
  QLabel *mp_msg_label;
  bool m_edits_enabled;
  bool m_disable_tab_selected;

  double m_grid_micron;
  bool m_synchronized_views;
  int m_disabled_edits;
  bool m_work_in_progress;
  QApplication *mp_app;

  lay::ProgressReporter *mp_pr;

  lay::LayoutView *current_view ();
  const lay::LayoutView *current_view () const;

  /** 
   *  @brief Event filter
   *  
   *  In order to be able to filter out mouse events in "work_in_progress" mode,
   *  we implement an event filter here.
   */
  bool eventFilter (QObject *watched, QEvent *event);

protected:
  void update_content ();

public:

  /**
   *  @brief Constructor
   */
  MainWindow (QApplication *app, const char *name = "main_window");

  /** 
   *  @brief Destructor
   */
  ~MainWindow ();

  /** 
   *  @brief Reload the given cellview into the current view
   *
   *  The cellview is given by index in the current view's cellview list.
   */
  void reload_layout (unsigned int cv_index);

  /** 
   *  @brief Load a (new) file into the layout
   *
   *  The mode param controls how to open the file:
   *    0 = replace current view, 1 = create new view, 2 = add to current view
   */
  void load_layout (const std::string &filename, int mode = 0);

  /**
   *  @brief Load a layer definition file
   *
   *  If "all_views" is true, the layer file is loaded on all views. If false,
   *  it is loaded only on the current view.
   */
  void load_layer_props (const std::string &fn, bool all_views);

  /** 
   *  @brief Grid accessor: get the grid value
   *
   *  This is the version delivering a double value in micron.
   */
  double grid_micron () const;

  /**
   *  @brief Hierarchy level selection setter
   */
  void hier_levels (std::pair<unsigned int, unsigned int> l);

  /**
   *  @brief Hierarchy level selection setter
   */
  void min_hier_levels (unsigned int l)
  {
    hier_levels (std::pair<unsigned int, unsigned int> (l, max_hier_levels ()));
  }

  /**
   *  @brief Hierarchy level selection setter
   */
  void max_hier_levels (unsigned int l)
  {
    hier_levels (std::pair<unsigned int, unsigned int> (min_hier_levels (), l));
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  std::pair<unsigned int, unsigned int> hier_levels () const;

  /**
   *  @brief Hierarchy level selection getter
   */
  unsigned int min_hier_levels () const
  {
    return hier_levels ().first;
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  unsigned int max_hier_levels () const
  {
    return hier_levels ().second;
  }

  /**
   *  @brief Return true, if there is a "last" display state
   */
  bool has_last_display_state ();

  /**
   *  @brief Return true, if there is a "next" display state
   */
  bool has_next_display_state ();

  /**
   *  @brief Display the current position
   */
  void current_pos (double x, double y, bool dbu_units);

  /**
   *  @brief Clear the current position
   */
  void current_pos ();

  /**
   *  @brief Return the widget to display progress informations
   */
  ProgressWidget *progress_widget ();

  /**
   *  @brief Make the progress widget visible or invisible
   */
  void show_progress_widget (bool show);

  /**
   *  @brief Write configuration to a file
   *
   *  If the configuration file cannot be written, false
   *  is returned but no exception is thrown.
   *
   *  @return false, if an error occured.
   */
  bool write_config (const std::string &config_file);

  /**
   *  @brief Read the configuration from a file
   *
   *  This method siletly does nothing, if the config file does not
   *  exist. If it does and an error occured, the error message is printed
   *  on stderr. In both cases, false is returned.
   *
   *  @return false, if an error occured.
   */
  bool read_config (const std::string &config_file);

  /**
   *  @brief Enable or disable the actions for edit functions 
   * 
   *  This method is used by non-modal dialogs that want to suppress any editing
   *  activities (like browsers) while they are open.
   */
  void enable_edits (bool enable);

  /** 
   *  @brief Implementation of the PluginRoot interface
   */
  bool configure (const std::string &name, const std::string &value);

  /** 
   *  @brief Implementation of the PluginRoot interface
   */
  void config_finalize ();

public slots:
  void undo ();
  void redo ();
  void del ();
  void show_properties ();
  void copy ();
  void paste ();
  void cut ();
  void mode ();
  void enable (QAction *);
  void popup_menu_show ();
  void zoom_fit ();
  void zoom_in ();
  void zoom_out ();
  void pan_up ();
  void pan_down ();
  void pan_left ();
  void pan_right ();
  void setup ();
  void reload_current_view ();
  void open (int mode);
  void close_current_view ();
  void close ();
  void clone ();
  void inc_max_hier ();
  void dec_max_hier ();
  void max_hier ();
  void max_hier_0 ();
  void max_hier_1 ();
  void last_display_state ();
  void next_display_state ();
  void cancel ();
  void redraw ();
  void screen_shot ();
  void save_layer_props ();
  void load_layer_props ();
  void save_bookmarks ();
  void load_bookmarks ();
  void select_cell ();
  void select_current_cell ();
  void exit ();
  void goto_bookmark_menu_show ();
  void goto_bookmark (QAction *);
  void bookmark_view ();
  void manage_bookmarks ();
  void goto_position ();
  void help_about ();
  void generic_menu (QAction *);
  
  void open_too ()
  {
    open (2);
  }

  void open_new_view ()
  {
    open (1);
  }

  void open_current_view ()
  {
    open (0);
  }

  void view_selected (int index);
  
};

class MainConfigPage 
  : public lay::ConfigPage
{
Q_OBJECT

public:
  MainConfigPage (QWidget *parent);
  ~MainConfigPage ();

  virtual void setup (lay::Plugin *root);
  virtual void commit (lay::Plugin *root);

private:
  Ui::MainConfigPage *mp_ui;
  lay::ColorButton *mp_background_color_cbtn;
};

class MainConfigPage2 
  : public lay::ConfigPage
{
Q_OBJECT

public:
  MainConfigPage2 (QWidget *parent);
  ~MainConfigPage2 ();

  virtual void setup (lay::Plugin *root);
  virtual void commit (lay::Plugin *root);

private:
  Ui::MainConfigPage2 *mp_ui;
};



}

#endif
