
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layMarker.h"
#include "laySnap.h"
#include "layBitmap.h"
#include "layViewOp.h"
#include "layRenderer.h"
#include "layLayoutView.h"

#include "dbCellGraph.h"

namespace lay
{

// ------------------------------------------------------------------------

ShapeMarker::ShapeMarker (lay::LayoutView *view, unsigned int cv_index)
  : lay::ViewObject (view->canvas ()), m_shape (), mp_trans_vector (0), mp_view (view), m_cv_index (cv_index)
{ 
  // .. nothing yet ..
}

ShapeMarker::~ShapeMarker ()
{
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }
}

void 
ShapeMarker::set (const db::Shape &shape, const db::MagTrans &trans)
{
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }
  m_trans = trans;
  m_shape = shape;
  redraw ();
}

void 
ShapeMarker::set (const db::Shape &shape, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);
  m_trans = db::MagTrans ();
  m_shape = shape;
  redraw ();
}

void 
ShapeMarker::render ()
{
  //  obtain bitmaps
  QColor color = mp_view->foreground_color ();
  
  lay::Bitmap *contour = mp_view->bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1));
  lay::Bitmap *vertex = mp_view->bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 3 /*fat vertices*/));
  if (contour == 0 && vertex == 0) {
    return;
  }

  lay::Renderer r;
  r.reserve_edges (100);

  const lay::CellView &cv = mp_view->cellview (m_cv_index);
  tl_assert (cv.is_valid ());
  double dbu = cv->layout ().dbu ();

  r.set_font (db::Font (mp_view->text_font ()));
  r.apply_text_trans (mp_view->apply_text_trans ());
  r.default_text_size (db::Coord (mp_view->default_text_size () / dbu));

  if (mp_trans_vector) {
    for (std::vector<db::MagTrans>::const_iterator tr = mp_trans_vector->begin (); tr != mp_trans_vector->end (); ++tr) {
      db::MagTrans t = (canvas ()->trans () * db::DMag (dbu)) * (*tr * m_trans);
      r.draw (m_shape, &cv->layout ().properties_repository (), t, 0, contour, vertex, contour);
    }
  } else {
    db::MagTrans t = (canvas ()->trans () * db::DMag (dbu)) * m_trans;
    r.draw (m_shape, &cv->layout ().properties_repository (), t, 0, contour, vertex, contour);
  }
}

// ------------------------------------------------------------------------

Marker::Marker (lay::LayoutView *view, unsigned int cv_index)
  : lay::ViewObject (view->canvas ()), mp_trans_vector (0), mp_view (view), m_cv_index (cv_index)
{ 
  m_type = None;
  m_object.any = 0;
}

Marker::~Marker ()
{
  remove_object ();
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }
}

void 
Marker::set (const db::Box &box, const db::MagTrans &trans)
{
  remove_object ();
  m_trans = trans;
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }

  m_type = Box;
  m_object.box = new db::Box (box);

  redraw ();
}

void 
Marker::set (const db::Box &box, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  remove_object ();
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);

  m_type = Box;
  m_object.box = new db::Box (box);

  redraw ();
}

void 
Marker::set (const db::Polygon &poly, const db::MagTrans &trans)
{
  remove_object ();
  m_trans = trans;
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }

  m_type = Polygon;
  m_object.polygon = new db::Polygon (poly);

  redraw ();
}

void 
Marker::set (const db::Polygon &poly, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  remove_object ();
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);

  m_type = Polygon;
  m_object.polygon = new db::Polygon (poly);

  redraw ();
}

void 
Marker::set (const db::Edge &edge, const db::MagTrans &trans)
{
  remove_object ();
  m_trans = trans;
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }

  m_type = Edge;
  m_object.edge = new db::Edge (edge);

  redraw ();
}

void 
Marker::set (const db::Edge &edge, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  remove_object ();
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);

  m_type = Edge;
  m_object.edge = new db::Edge (edge);

  redraw ();
}

void 
Marker::set (const db::Path &path, const db::MagTrans &trans)
{
  remove_object ();
  m_trans = trans;
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }

  m_type = Path;
  m_object.path = new db::Path (path);

  redraw ();
}

void 
Marker::set (const db::Path &path, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  remove_object ();
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);

  m_type = Path;
  m_object.path = new db::Path (path);

  redraw ();
}

void 
Marker::set (const db::Text &text, const db::MagTrans &trans)
{
  remove_object ();
  m_trans = trans;
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }

  m_type = Text;
  m_object.text = new db::Text (text);

  redraw ();
}

void 
Marker::set (const db::Text &text, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  remove_object ();
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);

  m_type = Text;
  m_object.text = new db::Text (text);

  redraw ();
}

void 
Marker::set (const db::CellInstArray &instance, const db::MagTrans &trans)
{
  remove_object ();
  m_trans = trans;
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }

  m_type = Instance;
  m_object.inst = new db::CellInstArray (instance);

  redraw ();
}

void 
Marker::set (const db::CellInstArray &instance, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans)
{
  remove_object ();
  m_trans = t1;
  if (mp_trans_vector) {
    delete mp_trans_vector;
  }
  mp_trans_vector = new std::vector<db::MagTrans> (trans);

  m_type = Instance;
  m_object.inst = new db::CellInstArray (instance);

  redraw ();
}

void
Marker::remove_object ()
{
  if (m_type == Box) {
    delete m_object.box;
  } else if (m_type == Polygon) {
    delete m_object.polygon;
  } else if (m_type == Edge) {
    delete m_object.edge;
  } else if (m_type == Path) {
    delete m_object.path;
  } else if (m_type == Text) {
    delete m_object.text;
  } else if (m_type == Instance) {
    delete m_object.inst;
  } 

  m_type = None;
  m_object.any = 0;
}

void 
Marker::render ()
{ 
  //  obtain bitmaps
  QColor color = mp_view->canvas ()->foreground_color ();
  
  lay::Bitmap *contour = canvas ()->bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1));
  lay::Bitmap *vertex = canvas ()->bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 3 /*fat vertices*/));
  if (contour == 0 && vertex == 0) {
    return;
  }

  lay::Renderer r;
  r.reserve_edges (100);

  const lay::CellView &cv = mp_view->cellview (m_cv_index);
  tl_assert (cv.is_valid ());
  double dbu = cv->layout ().dbu ();

  r.set_font (db::Font (mp_view->text_font ()));
  r.apply_text_trans (mp_view->apply_text_trans ());
  r.default_text_size (db::Coord (mp_view->default_text_size () / dbu));

  if (! mp_trans_vector) {

    db::MagTrans t = (canvas ()->trans () * db::DMag (dbu)) * m_trans;

    if (m_type == Box) {
      r.draw (*m_object.box, t, 0, contour, vertex, contour);
    } else if (m_type == Polygon) {
      r.draw (*m_object.polygon, t, 0, contour, vertex, contour);
    } else if (m_type == Path) {
      r.draw (*m_object.path, t, 0, contour, vertex, contour);
    } else if (m_type == Text) {
      r.draw (*m_object.text, t, 0, contour, vertex, contour);
    } else if (m_type == Edge) {
      r.draw (*m_object.edge, t, 0, contour, vertex, contour);
    } else if (m_type == Instance) {

      // .. 

    }

  } else {

    for (std::vector<db::MagTrans>::const_iterator tr = mp_trans_vector->begin (); tr != mp_trans_vector->end (); ++tr) {

      db::MagTrans t = (canvas ()->trans () * db::DMag (dbu)) * (*tr * m_trans);

      if (m_type == Box) {
        r.draw (*m_object.box, t, 0, contour, vertex, contour);
      } else if (m_type == Polygon) {
        r.draw (*m_object.polygon, t, 0, contour, vertex, contour);
      } else if (m_type == Path) {
        r.draw (*m_object.path, t, 0, contour, vertex, contour);
      } else if (m_type == Text) {
        r.draw (*m_object.text, t, 0, contour, vertex, contour);
      } else if (m_type == Edge) {
        r.draw (*m_object.edge, t, 0, contour, vertex, contour);
      } else if (m_type == Instance) {

        // .. 

      }

    }

  }

}

}

