
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_layMarker
#define HDR_layMarker

#include "layViewObject.h"
#include "dbShape.h"
#include "dbBox.h"
#include "dbPath.h"
#include "dbPolygon.h"
#include "dbText.h"
#include "dbEdge.h"
#include "dbArray.h"

namespace lay
{

class LayoutView;

/**
 *  @brief The shape marker object
 *
 *  The marker is a visual object that "marks" (highlights) a 
 *  certain shape object in the layout.
 */

class ShapeMarker
  : public lay::ViewObject
{
public: 
  /** 
   *  @brief The constructor 
   */ 
  ShapeMarker (lay::LayoutView *view, unsigned int cv_index);

  /**
   *  @brief The destructor
   */
  ~ShapeMarker ();

  /**
   *  @brief Set the shape the marker is to display
   *
   *  The marker just stores the shape which is a proxy. The actual object
   *  must be stored somewhere else.
   */
  void set (const db::Shape &shape, const db::MagTrans &trans);

  /**
   *  @brief Set the shape the marker is to display
   *
   *  The marker just stores the shape which is a proxy. The actual object
   *  must be stored somewhere else.
   */
  void set (const db::Shape &shape, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

private:
  virtual void render ();

  db::Shape m_shape; 
  db::MagTrans m_trans;
  std::vector<db::MagTrans> *mp_trans_vector;
  lay::LayoutView *mp_view;
  unsigned int m_cv_index;
};

/**
 *  @brief The marker object
 *
 *  The marker is a visual object that "marks" (highlights) a 
 *  certain area of the layout.
 */

class Marker
  : public lay::ViewObject
{
public: 
  /** 
   *  @brief The constructor 
   */ 
  Marker (lay::LayoutView *view, unsigned int cv_index);

  /**
   *  @brief The destructor
   */
  ~Marker ();

  /**
   *  @brief Set the box the marker is to display
   *
   *  If the box is empty, no marker is drawn.
   */
  void set (const db::Box &box, const db::MagTrans &trans);

  /**
   *  @brief Set the box the marker is to display
   *
   *  If the box is empty, no marker is drawn.
   *  This variant draws the box multiple times at the given transformations.
   */
  void set (const db::Box &box, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

  /**
   *  @brief Set the polygon the marker is to display
   */
  void set (const db::Polygon &poly, const db::MagTrans &trans);

  /**
   *  @brief Set the polygon the marker is to display
   *  
   *  This variant draws the polygon multiple times at the given transformations.
   */
  void set (const db::Polygon &poly, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

  /**
   *  @brief Set the edge the marker is to display
   */
  void set (const db::Edge &edge, const db::MagTrans &trans);

  /**
   *  @brief Set the edge the marker is to display
   *  
   *  This variant draws the edge multiple times at the given transformations.
   */
  void set (const db::Edge &edge, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

  /**
   *  @brief Set the path the marker is to display
   */
  void set (const db::Path &path, const db::MagTrans &trans);

  /**
   *  @brief Set the path the marker is to display
   *  
   *  This variant draws the path multiple times at the given transformations.
   */
  void set (const db::Path &path, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

  /**
   *  @brief Set the text the marker is to display
   */
  void set (const db::Text &text, const db::MagTrans &trans);

  /**
   *  @brief Set the text the marker is to display
   *  
   *  This variant draws the text multiple times at the given transformations.
   */
  void set (const db::Text &text, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

  /**
   *  @brief Set the instance the marker is to display
   */
  void set (const db::CellInstArray &instance, const db::MagTrans &trans);

  /**
   *  @brief Set the instance the marker is to display
   *  
   *  This variant draws the instance multiple times at the given transformations.
   */
  void set (const db::CellInstArray &instance, const db::MagTrans &t1, const std::vector<db::MagTrans> &trans);

private:
  virtual void render ();

  void remove_object ();

  enum { 
    None, Box, Polygon, Edge, Path, Text, Instance
  } m_type;

  union {
    db::Box *box;
    db::Polygon *polygon;
    db::Edge *edge;
    db::Path *path;
    db::Text *text;
    db::CellInstArray *inst;
    void *any;
  } m_object;

  db::MagTrans m_trans;
  std::vector<db::MagTrans> *mp_trans_vector;
  lay::LayoutView *mp_view;
  unsigned int m_cv_index;
};

}

#endif


