
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtCore/QEvent>
#include <QtGui/QApplication>

#include "layRedrawThread.h"
#include "layLayoutCanvas.h"
#include "tlLog.h"
#include "dbHershey.h"
#include "dbShape.h"

namespace lay 
{

//  time delay until the first snapshot is taken
const int first_snapshot_delay = 20;

// -------------------------------------------------------------
//  some helpful constants

const int special_bitmaps_before = 1;
const int special_bitmaps_after = 1;
const int special_queue_entries = 2;
const int draw_boxes_queue_entry = -1;
const int draw_custom_queue_entry = -2;

// -------------------------------------------------------------

bool RedrawThread::m_synchroneous = false;

bool 
RedrawThread::synchroneous () 
{
  return m_synchroneous;
}

void
RedrawThread::set_synchroneous (bool sync)
{
  m_synchroneous = sync;
}

// -------------------------------------------------------------
//  Implementation of RedrawLayerInfo

RedrawLayerInfo::RedrawLayerInfo (const lay::LayerProperties &lp)
{
  visible          = lp.visible (true /*real*/);
  layer_index      = lp.layer_index ();
  cellview_index   = lp.cellview_index ();
  trans            = lp.trans ();
  prop_sel         = lp.prop_sel ();
  inverse_prop_sel = lp.inverse_prop_sel ();
}

// -------------------------------------------------------------

RedrawThread::RedrawThread (lay::LayoutCanvas *canvas, lay::LayoutView *view)
  : QThread (),
    m_initial_update (false),
    mp_layout (0),
    mp_cell_var_cache (0),
    m_cache_hits (0),
    m_cache_misses (0),
    m_cv_index (-1),
    mp_canvas (canvas),
    mp_view (view),
    m_start_recursion_sentinel (false),
    m_request_stop (false),
    m_request_terminate (false),
    m_redraw (false),
    m_test_count (0),
    m_from_level (0), m_to_level (0),
    m_min_size_for_label (1),
    m_clock (tl::Clock::current ())
{
  //  .. nothing yet ..
}

RedrawThread::~RedrawThread ()
{
  //  terminate thread if already running
  if (running ()) {
    //  wake up the threads, so the they can process the m_request_terminate 
    //  signal
    m_request_stop = true;
    m_request_terminate = true;
    m_restart_cond.wakeAll ();
    //  no terminate ();: do a smooth shutdown.
    wait ();
  }
}

void RedrawThread::set_canvas (lay::LayoutCanvas *canvas)
{
  mp_canvas = canvas;
}

void RedrawThread::stop ()
{
  m_request_stop = true;

  //  wait until the drawing thread has finished and unlocks the
  //  main lock.
  m_main_lock.lock ();
  m_main_lock.unlock ();

  //  release all cellview references
  //  do this in the main thread since this way there is no problem 
  //  we thread-safetiness of reference counting in the cellview references.
  m_cellviews.clear ();
}

void 
RedrawThread::start (bool clear, const std::set<int> &layers)
{
  //  We need this recursion sentinel because start may be called recursively
  //  via CellGraph::update which may open a progress widget and do processEvents.
  //  This may result in a repaint for example issuing a "start" call.
  //  Recursion however is forbidden because of the deadlocking locks used originally
  //  to lock out the redraw thread.
  if (m_start_recursion_sentinel) {
    return;
  }
  m_start_recursion_sentinel = true;

  //  stop anything that runs already
  m_request_stop = true;

  //  wait until the drawing thread has finished and unlocks the
  //  main lock.
  m_main_lock.lock ();
 
  //  reset stop request flag
  m_request_stop = false;

  //  the redraw flag tells the redraw thread to immediately 
  //  redraw.
  m_redraw = true;

  if (! running ()) {
    //  start the thread if not done already.
    QThread::start ();
  }


  {
    if (tl::verbosity () >= 30) {
      tl::info << tl::translate ("Preparing to draw");
    }
    tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Preparing to draw"));

    //  Update all relevant layout objects.
    for (unsigned int i = 0; i < mp_view->cellviews (); ++i) {
      const lay::CellView &cv = mp_view->cellview (i);
      if (cv.is_valid () && ! cv->layout ().under_construction ()) {
        cv->layout ().update ();
      }
    }

    //  copy everything that we need so there is no need to access 
    //  members of lay::LayoutView in the drawing thread.
    //  Note: copying the cellviews will create new references to the
    //  layout objects. These are not automatically freed when the 
    //  drawing ends but rather on "stop". The advantage of this is
    //  that, since "stop" is called from the main thread like "start",
    //  we don't challenge lay::CellView's MT compliance.
    m_from_level = mp_view->hier_levels ().first;
    m_to_level = mp_view->hier_levels ().second;
    m_min_size_for_label = mp_view->min_inst_label_size ();
    m_box_text_transform = mp_view->cell_box_text_transform ();
    m_box_font = mp_view->cell_box_text_font ();
    m_text_font = mp_view->text_font ();
    m_text_visible = mp_view->text_visible ();
    m_text_lazy_rendering = mp_view->text_lazy_rendering ();
    m_show_properties = mp_view->show_properties_as_text ();
    m_apply_text_trans = mp_view->apply_text_trans ();
    m_default_text_size = mp_view->default_text_size ();
    m_test_count = 0;
    m_initial_update = true;

    m_trans = db::MagTrans ();
    mp_prop_sel = 0;
    m_inv_prop_sel = false;
    m_width = mp_canvas->width ();
    m_height = mp_canvas->height ();

    m_hidden_cells = mp_view->hidden_cells ();

    m_cellviews.clear ();
    m_cellviews.reserve (mp_view->cellviews ());
    for (unsigned int i = 0; i < mp_view->cellviews (); ++i) {
      m_cellviews.push_back (mp_view->cellview (int (i)));
    }

    m_properties.clear ();

    size_t nlayers = 0;
    for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); !l.at_end (); ++l) {
      if (! l->has_children ()) {
        ++nlayers;
      }
    }

    m_properties.reserve (nlayers);

    //  count the layers and collect the cellview/transformation variants for the 
    //  drawing of bounding boxes.
    for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); !l.at_end (); ++l) {
      if (! l->has_children ()) {
        m_properties.push_back (RedrawLayerInfo (*l));
      }
    }
    tl_assert (m_properties.size () == nlayers);

    m_box_variants = mp_view->cv_transform_variants ();

    m_ci = 0;

    m_draw_queue.clear ();

    if (mp_view->cellviews () > 0) {

      if (clear) {
        mp_canvas->clear (nlayers * 4 + special_bitmaps_before + special_bitmaps_after);
        m_already_drawn.clear ();
      } else {
        for (std::set<int>::const_iterator l = layers.begin (); l != layers.end (); ++l) {
          m_already_drawn.erase (*l);
        }
      }

      //  set up the drawing queue: controls the sequence of things to 
      //  draw. First there are the visible layers, then there are the
      //  cell boundaries. Then come the invisible layers.

      m_draw_queue.reserve (nlayers + special_queue_entries);

      for (unsigned int i = 0; i < nlayers; ++i) {
        if (m_properties [i].visible) {
          if (m_already_drawn.find (int (i)) == m_already_drawn.end ()) {
            m_draw_queue.push_back (int (i));
          }
        }
      }

      //  cell box drawing
      if (m_already_drawn.find (draw_boxes_queue_entry) == m_already_drawn.end ()) {
        m_draw_queue.push_back (draw_boxes_queue_entry);
      }

      //  decoration drawing
      if (m_already_drawn.find (draw_custom_queue_entry) == m_already_drawn.end ()) {
        m_draw_queue.push_back (draw_custom_queue_entry);
      }

    } else {
      mp_canvas->clear (1);
    }

  }

  //  setup the wait condition for the initial transfer
  m_initial_wait_lock.lock ();

  //  provide a signal for the waiting thread to restart
  m_restart_cond.wakeAll ();
  //  release the locks and let the thread do it's work
  m_main_lock.unlock ();

  //  wait for the first transfer 
  m_initial_wait_cond.wait (&m_initial_wait_lock);
  m_initial_update = false;
  m_initial_wait_lock.unlock ();

  m_start_recursion_sentinel = false;
}

void 
RedrawThread::iterate_variants (void (RedrawThread::*what) ())
{
  if (m_from_level < 0) {

    //  if we start from above the hierarchy, we need to establish a 
    //  cell variant cache to at least avoid painting the current cell
    //  multiple times.
    std::set <std::pair <db::MagTrans, db::Layout::cell_index_type>, lay::CellVariantCacheCompare> cell_var_cache;
    mp_cell_var_cache = &cell_var_cache;
    m_cache_hits = m_cache_misses = 0;
    iterate_variants_rec (what);
    mp_cell_var_cache = 0;

    if (tl::verbosity () >= 40) {
      tl::info << tl::translate ("Cell variant cache hits/misses: ") << m_cache_hits << "/" << m_cache_misses;
    }

  } else {
    mp_cell_var_cache = 0;
    iterate_variants_rec (what);
  }
}

void 
RedrawThread::iterate_variants_rec (void (RedrawThread::*what) ())
{
  if (m_level > m_from_level) {

    db::Cell::parent_inst_iterator p = mp_layout->cell (m_ci).begin_parent_insts ();
    if (! p.at_end ()) {

      //  one level up ..
      --m_level;

      db::MagTrans trans = m_trans;
      db::Box vp = m_vp;
      db::CellGraph::cell_index_type ci = m_ci;

      while (! p.at_end ()) {

        db::Cell::cell_inst_array_type pi = (*p).inst ();

        m_ci = pi.object ().cell_index ();

        for (db::Cell::cell_inst_array_type::iterator pp = pi.begin (); ! pp.at_end (); ++pp) {

          db::MagTrans t (*pp, db::Mag (pi.mag ()));
          m_trans = trans * t;
          m_vp = t.inverted () * db::DBox (vp);
          iterate_variants_rec (what);

        }

        ++p;

      }

      m_trans = trans;
      m_vp = vp;
      m_ci = ci;

      ++m_level;

      return;

    }

  }

  db::Coord lim = std::numeric_limits<db::Coord>::max ();
  db::DBox world (m_trans * db::Box (db::Point (-lim, -lim), db::Point (lim, lim)));
  m_vp = m_trans.inverted () * (world & db::DBox (db::DPoint (0.0, 0.0), db::DPoint (m_width, m_height)));
  m_vp &= mp_layout->cell (m_ci).bbox (); // this avoids problems with accessing designs through very large viewports

  (this->*what) ();

}

void 
RedrawThread::run ()
{
  while (! m_request_terminate) {

    //  lock against setup code
    m_main_lock.lock ();

    m_redraw = false;

    try {

      tl::SelfTimer timer (tl::verbosity () >= 21, "Redrawing");

      //  allow a very short time to pass before we issue the
      //  first update event.
      m_clock = tl::Clock::current () - tl::Clock ((update_interval - first_snapshot_delay) * 0.001);

      //  initialize the bitmap buffers
      m_frame  = lay::Bitmap (m_width, m_height);
      m_fill   = lay::Bitmap (m_width, m_height);
      m_vertex = lay::Bitmap (m_width, m_height);
      m_text   = lay::Bitmap (m_width, m_height);

      //  give the renderer some initial capacity.
      m_r.reserve_edges (100);

      //  work the drawing queue ..
      for (std::vector <int>::const_iterator q = m_draw_queue.begin (); q != m_draw_queue.end (); ++q) {

        m_level = 0;

        if (*q >= 0) {

          //  draw a layer

          m_frame.clear ();
          m_fill.clear ();
          m_vertex.clear ();
          m_text.clear ();

          m_buffers.clear ();
          m_buffers.push_back (std::make_pair (*q * 4 + special_bitmaps_before,     &m_fill));
          m_buffers.push_back (std::make_pair (*q * 4 + special_bitmaps_before + 1, &m_frame));
          m_buffers.push_back (std::make_pair (*q * 4 + special_bitmaps_before + 2, &m_text));
          m_buffers.push_back (std::make_pair (*q * 4 + special_bitmaps_before + 3, &m_vertex));

          const RedrawLayerInfo &li = m_properties [*q];

          if (li.layer_index >= 0 && li.cellview_index >= 0) {

            m_layer = li.layer_index;
           
            //  determine layout and cell associated with this layer ..
            const lay::CellView &cv = m_cellviews [li.cellview_index];
            if (cv.is_valid () && ! cv->layout ().under_construction ()) {

              mp_layout = &cv->layout ();
              m_cv_index = li.cellview_index;

              if (tl::verbosity () >= 30) {
                tl::info << tl::translate ("Drawing layer: ") << mp_layout->get_properties (m_layer).name;
              }
              tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing layer"));

              m_ci = cv.cell_index ();

              m_trans = db::MagTrans (mp_canvas->trans ().t1, db::Mag (mp_canvas->trans ().t2.factor () * mp_layout->dbu ())) * li.trans;
              mp_prop_sel = &li.prop_sel;
              m_inv_prop_sel = li.inverse_prop_sel;

              m_mi_cache.clear ();

              iterate_variants (&RedrawThread::draw_layer);

              mp_prop_sel = 0;
              m_inv_prop_sel = false;

            }

          }

        } else if (*q == draw_boxes_queue_entry) {

          //  draw the bounding boxes 
          if (tl::verbosity () >= 30) {
            tl::info << tl::translate ("Drawing frames");
          }
          tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing frames"));

          m_frame.clear ();

          m_buffers.clear ();
          m_buffers.push_back (std::make_pair (0, &m_frame));

          for (std::set< std::pair<db::MagTrans, int> >::const_iterator b = m_box_variants.begin (); b != m_box_variants.end (); ++b) {

            const lay::CellView &cv = m_cellviews [b->second];
            if (cv.is_valid () && ! cv->layout ().under_construction ()) {

              mp_layout = &cv->layout ();
              m_ci = cv.cell_index ();
              m_cv_index = b->second;

              m_trans = db::MagTrans (mp_canvas->trans ().t1, db::Mag (mp_canvas->trans ().t2.factor () * mp_layout->dbu ())) * b->first;

              iterate_variants (&RedrawThread::draw_boxes);

            }

          }

        } else if (*q == draw_custom_queue_entry) {

          //  draw the decorations
          if (tl::verbosity () >= 30) {
            tl::info << tl::translate ("Drawing decorations");
          }
          tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing decorations"));

          m_frame.clear ();

          m_buffers.clear ();
          m_buffers.push_back (std::make_pair (m_properties.size () * 4 + special_bitmaps_before, &m_frame));

          unsigned int nd = 0;
          for (lay::Drawings::iterator d = mp_canvas->begin (); d != mp_canvas->end (); ++d, ++nd) {

            //  temporarily create bitmap objects, paint on them,
            //  transfer them to the canvas and clear them again.
            //  This operation is not interrupted by any "test_snapshot".

            std::vector <lay::Bitmap *> bitmaps;
            bitmaps.reserve (d->num_bitmaps ());
            for (unsigned int i = 0; i < d->num_bitmaps (); ++i) {
              bitmaps.push_back (new lay::Bitmap (m_width, m_height));
            }

            //  currently, all cellviews are painted over each other ..
            for (unsigned int i = 0; i < m_cellviews.size (); ++i) {
              test_snapshot (); 
              const lay::CellView &cv = m_cellviews [i];
              if (cv.is_valid () && ! cv->layout ().under_construction ()) {
                db::MagTrans trans (mp_canvas->trans ().t1, db::Mag (mp_canvas->trans ().t2.factor () * cv->layout ().dbu ()));
                d->paint_on_bitmaps (cv, trans, bitmaps);
              }
            }

            for (unsigned int i = 0; i < d->num_bitmaps (); ++i) {
              mp_canvas->set_drawing_bitmap (nd, i, *bitmaps [i]); 
            }

            while (! bitmaps.empty ()) {
              delete bitmaps.back ();
              bitmaps.pop_back ();
            }

          }

        }

        transfer ();
        m_buffers.clear ();

        //  mark this entry as already drawn
        m_already_drawn.insert (*q);

      }

      //  post an event to actually draw
      if (m_initial_update) {
        m_initial_wait_lock.lock ();
        m_initial_wait_cond.wakeAll ();
        m_initial_wait_lock.unlock ();
        m_initial_update = false;
      } else {
        QApplication::postEvent (mp_canvas, new QEvent (QEvent::Type (QEvent::User + 2)));
      }

    } catch (...) {

      //  simply catch all exceptions - they are used to abort.

      //  wake up any waiting start() method
      if (m_initial_update) {
        m_initial_wait_lock.lock ();
        m_initial_wait_cond.wakeAll ();
        m_initial_wait_lock.unlock ();
        m_initial_update = false;
      } 

    }

    m_main_lock.unlock ();

    //  if the redraw flag is set now, a start () has happened
    //  between the last unlock and we do not need to do the 
    //  next lock/unlock pair to wait.
    if (! m_redraw && ! m_request_terminate) {
      //  this will block until the redraw is restarted ..
      m_restart_lock.lock ();
      m_restart_cond.wait (& m_restart_lock);
      m_restart_lock.unlock ();
    }

  } 
}

static bool
cells_in (const db::Layout *layout, const db::Cell &cell, 
          const std::set <db::Cell::cell_index_type> &selected,
          int levels,
          std::set <std::pair <int, db::Cell::cell_index_type> > &cache)
{
  if (selected.find (cell.cell_index ()) != selected.end ()) {
    return true;
  }
  if (levels > 0) {
    for (db::Cell::child_cell_iterator c = cell.begin_child_cells (); c != cell.end_child_cells (); ++c) {
      if (cache.find (std::make_pair (levels, *c)) == cache.end ()) {
        if (cells_in (layout, layout->cell (*c), selected, levels - 1, cache)) {
          return true;
        }
        cache.insert (std::make_pair (levels, *c));
      }
    }
  } 
  return false;
}

static bool 
need_draw_box (const db::Layout *layout, const db::Cell &cell, 
               int level, int to_level, 
               const std::vector <std::set <db::Cell::cell_index_type> > &hidden_cells, unsigned int cv_index) 
{
  if (level > to_level) {
    return false;
  }
  if (hidden_cells.size () > cv_index && ! hidden_cells [cv_index].empty ()) {
    std::set <std::pair <int, db::Cell::cell_index_type> > cache;
    if (cells_in (layout, cell, hidden_cells [cv_index], to_level - level, cache)) {
      return true;
    }
  }
  return int (cell.hierarchy_levels ()) + level >= to_level;
}

void 
RedrawThread::draw_boxes ()
{
  //  do not draw, if there is nothing to draw
  if (mp_layout->cells () <= m_ci || m_vp.empty ()) {
    return;
  }

  const db::Cell &cell = mp_layout->cell (m_ci);

  //  we will never come to a valid level ..
  if (! need_draw_box (mp_layout, cell, m_level, m_to_level, m_hidden_cells, m_cv_index)) {
    return;
  }
  if (mp_cell_var_cache) {
    if (mp_cell_var_cache->find (std::make_pair (m_trans, m_ci)) != mp_cell_var_cache->end ()) {
      ++m_cache_hits;
      return;
    }
    ++m_cache_misses;
    mp_cell_var_cache->insert (std::make_pair (m_trans, m_ci));
  }

  //  For small bboxes, the cell outline can be reduced ..
  db::Box bbox = cell.bbox ();

  if (bbox.empty ()) {

    //  no shapes there and below ..

  } else if (m_level == m_to_level || (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ())) {

    //  paint the box on this level
    draw_cell (bbox, mp_layout->cell_name (m_ci));

  } else if (m_level < m_to_level) {

    db::DBox dbbox = m_trans * bbox;
    if (dbbox.width () < 1.5 && dbbox.height () < 1.5) {

      //  the cell is a very small box and we know there must be 
      //  some level at which to draw the boundary: just draw it 
      //  here and stop looking further down ..
      draw_cell (bbox, "");

    } else {

      db::box_convert <db::CellInst> bc (*mp_layout);
     
      db::MagTrans trans = m_trans;
      db::Box vp = m_vp;
      db::CellGraph::cell_index_type ci = m_ci;
      ++m_level;

      //  dive down into the hierarchy ..
      db::Cell::touching_iterator inst = cell.begin_touching (m_vp); 
      while (! inst.at_end ()) {

        const db::CellInstArray &cell_inst = *inst;

        db::Point a, b;
        unsigned long amax, bmax; 
        bool simplify = false;
        if (cell_inst.is_regular_array (a, b, amax, bmax)) {

          db::DBox inst_box = trans * (mp_layout->cell (cell_inst.object ().cell_index ()).bbox () * cell_inst.mag ());
          if (((a.x () == 0 && b.y () == 0) || (a.y () == 0 && b.x () == 0)) && 
              inst_box.width () < 1.5 && inst_box.height () < 1.5 && 
              (amax <= 1 || trans.ctrans (a.distance ()) < 1.5) && 
              (bmax <= 1 || trans.ctrans (b.distance ()) < 1.5)) {
            simplify = true;
          }

        }

        if (simplify) {

          //  The array can be simplified if there are levels below to draw
          if (need_draw_box (mp_layout, mp_layout->cell (cell_inst.object ().cell_index ()), m_level, m_to_level, m_hidden_cells, m_cv_index)) {
            
            m_r.clear ();

            db::box_convert <db::CellInst> bc (*mp_layout);
            db::DBox array_box = trans * cell_inst.bbox (bc);
            m_r.insert (array_box);

            m_r.render_fill (m_frame);

          }

        } else {

          //  The array (or single instance) must be iterated instance
          //  by instance
          for (db::CellInstArray::iterator p = cell_inst.begin_touching (vp, bc); ! p.at_end (); ++p) {
        
            test_snapshot (); 
            db::MagTrans t (*p, db::Mag (cell_inst.mag ()));
            m_trans = trans * t;
            m_vp = t.inverted () * db::DBox (vp);
            m_ci = cell_inst.object ().cell_index ();
            draw_boxes ();

          }

        }

        ++inst;
      
      }

      --m_level;
      m_trans = trans;
      m_vp = vp;
      m_ci = ci;

    }

  }

}

/**
 *  @brief A helper function to determine if there are any shapes on the cell below to a certain hierarchy level
 *
 *  @return 0, if there are no shapes
 *  @return 1, if there are only contour-only shapes
 *  @return 2, if there are texts or properties to draw anywhere
 */
static int 
any_shapes (RedrawThread::MicroInstanceCache &cache, 
            const db::Layout &layout, db::Cell::cell_index_type cell_index, unsigned int layer, 
            unsigned long levels, const std::set<db::Cell::cell_index_type> *hidden_cells,
            const std::set<db::PropertiesRepository::properties_id_type> *prop_sel, bool inv_prop_sel)
{
  //  if the cell is "hidden", it does not need to be drawn
  if (hidden_cells && hidden_cells->find (cell_index) != hidden_cells->end ()) {
    return 0;
  }

  //  the cache contains all cells that do not have any shapes
  RedrawThread::MicroInstanceCache::const_iterator c = cache.find (std::make_pair (cell_index, levels));
  if (c != cache.end ()) {
    return c->second;
  }

  int ret = 0;

  const db::Cell &cell = layout.cell (cell_index);

  if (! cell.shapes (layer).begin (db::ShapeIterator::Texts).at_end ()) {

    ret = 2;

  } else if (! cell.shapes (layer).begin (db::ShapeIterator::AllWithProperties, prop_sel, inv_prop_sel).at_end ()) {

    //  shapes with properties count like text objects ..
    ret = 2;

  } else if (! cell.shapes (layer).begin (db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Boxes).at_end ()) {

    ret = 1;

  } else if (levels > 1) {
    
    for (db::Cell::const_iterator inst = cell.begin (); inst != cell.end (); ++inst) {
      int r = any_shapes (cache, layout, inst->object ().cell_index (), layer, levels - 1, hidden_cells, prop_sel, inv_prop_sel);
      if (r > 0) {
        ret = r;
        break;
      }
    }

  }

  cache.insert (std::make_pair (std::make_pair (cell_index, levels), ret));

  return ret;

}

void 
RedrawThread::draw_layer ()
{
  test_snapshot (); 

  //  do not draw, if there is nothing to draw
  if (mp_layout->cells () <= m_ci || m_vp.empty ()) {
    return;
  }
  if (mp_cell_var_cache) {
    if (mp_cell_var_cache->find (std::make_pair (m_trans, m_ci)) != mp_cell_var_cache->end ()) {
      ++m_cache_hits;
      return;
    }
    ++m_cache_misses;
    mp_cell_var_cache->insert (std::make_pair (m_trans, m_ci));
  }

  const db::Cell &cell = mp_layout->cell (m_ci);

  //  For small bboxes, the cell outline can be reduced ..
  db::Box bbox = cell.bbox (m_layer);
  if (! bbox.empty ()) {

    bool hidden = (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ());
    bool need_to_dive = (m_level + 1 < m_to_level) && ! hidden;

    //  draw this level
    if (m_level >= m_from_level && m_level < m_to_level && ! hidden) {

      //  configure renderer to text drawing ..
      m_r.draw_texts (m_text_visible);
      m_r.draw_properties (m_show_properties);
      m_r.default_text_size (db::Coord (m_default_text_size / mp_layout->dbu ()));
      m_r.set_font (db::Font (m_text_font));
      m_r.apply_text_trans (m_apply_text_trans);

      db::DBox dbbox_tot = m_trans * cell.bbox ();
      bool text_simplified = m_text_lazy_rendering && (dbbox_tot.width () <= 8.0 || dbbox_tot.height () <= 8.0);
          
      const db::Shapes &shapes = cell.shapes (m_layer);

      db::DBox dbbox = m_trans * bbox;
      if ((dbbox.width () < 2.5 && dbbox.height () < 1.5) || 
          (dbbox.width () < 1.5 && dbbox.height () < 2.5)) {

        //  Here we first analyze the cell to find out 
        //  how far we need to go while maintaining the precision:

        int a;
        if (int (m_hidden_cells.size ()) <= m_cv_index) {
          a = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, 0, mp_prop_sel, m_inv_prop_sel);
        } else {
          a = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, & m_hidden_cells [m_cv_index], mp_prop_sel, m_inv_prop_sel);
        }

        if (a > 0) {
          //  any shapes here: paint bbox for simplification
          m_r.clear ();
          m_r.insert (dbbox);
          m_r.render_contour (m_frame);
          m_r.render_vertices (m_vertex);
        }

        //  texts not visible or no texts below - do not dive further into hierarchy
        if (text_simplified || ! m_text_visible || a < 2) {
          need_to_dive = false;
        }

      } else {

        db::ShapeIterator shape (shapes.begin_touching (m_vp, db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths, mp_prop_sel, m_inv_prop_sel));
        while (! shape.at_end ()) {
          test_snapshot (); 
          m_r.draw (*shape, &mp_layout->properties_repository (), m_trans, &m_fill, &m_frame, &m_vertex, &m_text);
          ++shape;
        }

      }

      //  In lazy text rendering mode, all texts are only rendered if the cell is "reasonable large". 
      //  Empirically, a reasonable minimum dimension of 8x8 pixels was determined.
      //  otherwise just a few texts are rendered.
      if (m_text_visible) {

        //  this is the number of texts to draw in lazy text rendering mode
        size_t ntexts = 2;
        if (! text_simplified) {
          ntexts = std::numeric_limits <size_t>::max ();
        }

        db::ShapeIterator shape = shapes.begin_touching (m_vp, db::ShapeIterator::Texts, mp_prop_sel, m_inv_prop_sel);
        while (! shape.at_end () && ntexts > 0) {

          test_snapshot (); 

          m_r.draw (*shape, &mp_layout->properties_repository (), m_trans, &m_fill, &m_frame, &m_vertex, &m_text);
          ++shape;

          --ntexts;

        }

      }

    }

    db::box_convert <db::CellInst> bc (*mp_layout, m_layer);

    //  dive down into the hierarchy ..
    if (need_to_dive) {

      db::MagTrans trans = m_trans;
      db::Box vp = m_vp;
      db::CellGraph::cell_index_type ci = m_ci;
      ++m_level;

      db::Cell::touching_iterator inst = cell.begin_touching (vp); 
      while (! inst.at_end ()) {

        const db::CellInstArray &cell_inst = *inst;
        ++inst;

        m_ci = cell_inst.object ().cell_index ();
        bool hidden = (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ());

        db::Box cell_box = mp_layout->cell (m_ci).bbox (m_layer);
        if (! cell_box.empty () && ! hidden) {

          db::Point a, b;
          unsigned long amax, bmax; 
          bool simplify = false;
          bool anything = true;

          if (cell_inst.is_regular_array (a, b, amax, bmax)) {

            db::DBox inst_box = trans * (cell_box * cell_inst.mag ());

            if (((a.x () == 0 && b.y () == 0) || (a.y () == 0 && b.x () == 0)) && 
                inst_box.width () < 1.5 && inst_box.height () < 1.5 && 
                (amax <= 1 || trans.ctrans (a.distance ()) < 1.5) && 
                (bmax <= 1 || trans.ctrans (b.distance ()) < 1.5)) {

              int as;
              if (int (m_hidden_cells.size ()) <= m_cv_index) {
                as = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, 0, mp_prop_sel, m_inv_prop_sel);
              } else {
                as = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, & m_hidden_cells [m_cv_index], mp_prop_sel, m_inv_prop_sel);
              }
              if (as == 0) {
                anything = false;
              } else if ((m_text_visible && as == 1) || (! m_text_visible && as >= 1)) {
                //  only for simple shapes this array instance can be simplified ...
                simplify = true;
              }

            }

          }

          if (simplify) {

            //  The array can be simplified ..
            m_r.clear ();

            db::DBox array_box = trans * cell_inst.bbox (bc);
            m_r.insert (array_box);

            m_r.render_contour (m_frame);
            m_r.render_fill (m_frame);
            m_r.render_contour (m_vertex);
            m_r.render_fill (m_vertex);

          } else if (anything) {

            for (db::CellInstArray::iterator p = cell_inst.begin_touching (vp, bc); ! p.at_end (); ++p) {
          
              db::MagTrans t (*p, db::Mag (cell_inst.mag ()));
              m_trans = trans * t;
              m_vp = t.inverted () * db::DBox (vp);
              draw_layer ();
     
            }

          }

        }

      }

      --m_level;
      m_trans = trans;
      m_vp = vp;
      m_ci = ci;

    }

  }

}

void 
RedrawThread::draw_cell (const db::Box &box, const std::string &txt)
{
  m_r.clear ();

  db::DBox dbox = m_trans * box;
  m_r.insert (dbox);

  if (dbox.width () > m_min_size_for_label && dbox.height () > m_min_size_for_label) {
    m_r.insert (dbox, txt, 
                db::Font (m_box_font), 
                db::HAlignCenter, 
                db::VAlignCenter, 
                db::fixpoint_trans (m_box_text_transform ? m_trans.t1.rot () : db::fixpoint_trans::r0));
  }

  m_r.render_texts (m_frame);
  m_r.render_contour (m_frame);
}

void 
RedrawThread::test_snapshot ()
{
  if (m_request_stop) {
    throw std::exception ();
  }

  if (! m_synchroneous) {
    if (m_test_count == 0) {

      m_test_count = 100; // TODO: make variable?

      tl::Clock c = tl::Clock::current ();
      if ((c - m_clock).seconds () > update_interval * 0.001) {
        transfer ();
        if (m_initial_update) {
          m_initial_wait_lock.lock ();
          m_initial_wait_cond.wakeAll ();
          m_initial_wait_lock.unlock ();
          m_initial_update = false;
        } else {
          QApplication::postEvent (mp_canvas, new QEvent (QEvent::Type (QEvent::User + 1)));
        }
        m_clock = c;
      }

    } else {
      --m_test_count;
    }
  }
}

void RedrawThread::transfer ()
{
  if (m_buffers.size () > 0) {
    for (std::vector<std::pair<unsigned int, lay::Bitmap *> >::iterator b = m_buffers.begin (); b != m_buffers.end (); ++b) {
      mp_canvas->set_bitmap (b->first, *(b->second));
    }
  }
}

} // namespace lay
