
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_layRedrawThread
#define HDR_layRedrawThread

#include <vector>
#include <set>

#include <QtCore/QThread>
#include <QtCore/QMutex>
#include <QtCore/QWaitCondition>

#include "dbBox.h"
#include "dbTrans.h"
#include "dbLayout.h"
#include "layRenderer.h"
#include "layLayoutView.h"
#include "layBitmap.h"
#include "tlTimer.h"

namespace lay {

class LayoutCanvas;

//  update (snapshot) interval in ms
const int update_interval = 500;

struct RedrawLayerInfo
{
  RedrawLayerInfo (const lay::LayerProperties &lp);

  bool visible;
  unsigned int layer_index;
  int cellview_index;
  db::MagTrans trans;
  std::set <db::PropertiesRepository::properties_id_type> prop_sel;
  bool inverse_prop_sel;
};

class CellVariantCacheCompare
{
public:
  bool operator () (const std::pair<db::MagTrans, db::Layout::cell_index_type> &a,
                    const std::pair<db::MagTrans, db::Layout::cell_index_type> &b) const
  {
    if (a.second != b.second) {
      return a.second < b.second;
    }
    db::TransFuzzyLessF<db::MagTrans> less;
    return less (a.first, b.first);
  }
};


class RedrawThread 
  : public QThread
{
public:
  typedef std::map <std::pair <db::CellGraph::cell_index_type, unsigned long>, int> MicroInstanceCache;

  RedrawThread (lay::LayoutCanvas *canvas, lay::LayoutView *view);
  virtual ~RedrawThread ();

  void set_canvas (lay::LayoutCanvas *canvas);

  void start (bool clear, const std::set<int> &layers);

  virtual void run ();
  void stop ();

  static bool synchroneous ();
  static void set_synchroneous (bool sync);

private:
  QMutex m_main_lock;
  QMutex m_restart_lock;
  QWaitCondition m_restart_cond;
  QMutex m_initial_wait_lock;
  QWaitCondition m_initial_wait_cond;
  bool m_initial_update;
  const db::Layout *mp_layout;
  std::set <std::pair <db::MagTrans, db::Layout::cell_index_type>, lay::CellVariantCacheCompare> *mp_cell_var_cache;
  unsigned int m_cache_hits, m_cache_misses;
  int m_cv_index;
  lay::LayoutCanvas *mp_canvas;
  lay::LayoutView *mp_view;
  bool m_start_recursion_sentinel;
  bool m_request_stop;
  bool m_request_terminate;
  bool m_redraw;
  unsigned int m_test_count;
  int m_from_level, m_to_level;
  bool m_box_text_transform;
  unsigned int m_box_font;
  int m_min_size_for_label;
  unsigned int m_text_font;
  bool m_text_visible;
  bool m_text_lazy_rendering;
  bool m_show_properties;
  bool m_apply_text_trans;
  double m_default_text_size;
  std::vector <std::set <lay::LayoutView::cell_index_type> > m_hidden_cells;
  std::vector <RedrawLayerInfo> m_properties;
  std::set <std::pair <db::MagTrans, int> > m_box_variants;
  std::vector <lay::CellView> m_cellviews;

  unsigned int m_layer;
  int m_level;
  MicroInstanceCache m_mi_cache;
  db::MagTrans m_trans;
  const std::set<db::PropertiesRepository::properties_id_type> *mp_prop_sel;
  bool m_inv_prop_sel;
  db::Box m_vp;
  db::CellGraph::cell_index_type m_ci;
  lay::Renderer m_r;
  lay::Bitmap m_frame, m_fill, m_vertex, m_text;
  tl::Clock m_clock;
  std::vector<std::pair<unsigned int, lay::Bitmap *> > m_buffers;
  int m_width, m_height;

  std::vector <int> m_draw_queue;
  std::set <int> m_already_drawn;

  static bool m_synchroneous;

  void draw_layer ();
  void draw_boxes ();
  void draw_cell (const db::Box &box, const std::string &txt);
  void test_snapshot ();
  void transfer ();
  void iterate_variants (void (RedrawThread::*what) ());
  void iterate_variants_rec (void (RedrawThread::*what) ());
};

}

#endif

