
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QMouseEvent>

#include "layViewObject.h"
#include "layBitmap.h"
#include "tlException.h"

namespace lay
{

// ---------------------------------------------------------------
//  ViewObject implementation

ViewObject::ViewObject (ViewObjectCanvas *canvas, bool _static)
  : mp_canvas (canvas), m_static (_static), m_visible (true)
{
  if (canvas) {
    canvas->m_objects.insert (this);
    redraw ();
  }
}

ViewObject::ViewObject (const lay::ViewObject &d)
  : m_static (d.m_static)
{
  redraw ();
}

ViewObject &
ViewObject::operator= (const ViewObject &d)
{
  m_static = d.m_static;
  redraw ();
  return *this;
}

ViewObject::~ViewObject ()
{
  redraw ();
}

void 
ViewObject::visible (bool vis)
{
  if (vis != m_visible) {
    m_visible = vis;
    redraw ();
  }
}

void 
ViewObject::redraw ()
{
  if (canvas ()) {
    if (m_static) {
      canvas ()->touch ();
    } else {
      canvas ()->update ();
    }
  }
}

void
ViewObject::thaw ()
{
  if (canvas ()) {
    canvas ()->thaw (this);
  }
}

void
ViewObject::freeze ()
{
  if (canvas ()) {
    canvas ()->freeze (this);
  }
}

// ---------------------------------------------------------------
//  ViewService implementation

ViewService::ViewService (ViewObjectCanvas *canvas)
  : mp_canvas (canvas), m_abs_grab (false), m_enabled (true)
{
  if (canvas) {
    canvas->m_services.insert (this);
  }
}

void
ViewService::mouse_enable (bool en)
{
  m_enabled = en;
  if (! en) {
    canvas ()->ungrab_mouse (this);
  }
}

// ---------------------------------------------------------------
//  ViewObject implementation

ViewObjectCanvas::ViewObjectCanvas (QWidget *parent, const char *name)
  : QWidget (parent, name), 
    m_objects (this), m_services (this),
    m_needs_update_static (false),
    m_needs_update_bg (false),
    mp_active_service (0)
{
  setMouseTracking (true); 
}

ViewObjectCanvas::~ViewObjectCanvas ()
{
  //  release any mouse grabs now
  while (m_grabbed.begin () != m_grabbed.end ()) {
    ungrab_mouse (*m_grabbed.begin ());
  }
}

void
ViewObjectCanvas::activate (lay::ViewService *service)
{
  if (mp_active_service != service) {
    if (mp_active_service) {
      mp_active_service->deactivated ();
    }
    mp_active_service = service;
    if (mp_active_service) {
      mp_active_service->activated ();
    }
  }
}

void 
ViewObjectCanvas::mouseMoveEvent (QMouseEvent *e)
{
  m_mouse_pos = e->pos ();
  do_mouse_move (m_mouse_pos);
}

void
ViewObjectCanvas::do_mouse_move (const QPoint &qp)
{
  bool done = false; 

  db::DPoint p = trans ().inverted () * db::DPoint (qp.x (), height () - 1 - qp.y ());

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    done = ((*g)->m_enabled && (*g)->mouse_move_event (p, true));
    g = gg;
  }

  if (! done && mp_active_service) {
    done = (mp_active_service->m_enabled && mp_active_service->mouse_move_event (p, true));
  }

  service_iterator svc = begin_services ();
  while (svc != end_services () && !done) {
    service_iterator next = svc;
    ++next;
    done = (svc->m_enabled && svc->mouse_move_event (p, false));
    svc = next;
  }

  if (! done) {
    mouse_move_event (p);
  }
}

void 
ViewObjectCanvas::mousePressEvent (QMouseEvent *e)
{
  setFocus ();

  bool done = false; 

  m_mouse_pos = e->pos ();
  m_mouse_pressed = e->pos ();

  unsigned int buttons =
    ((e->button () == Qt::LeftButton) ? LeftButton : 0) |
    ((e->button () == Qt::MidButton) ? MidButton : 0) |
    ((e->button () == Qt::RightButton) ? RightButton : 0) |
    ((e->state () & Qt::ShiftButton) != 0 ? ShiftButton : 0) |
    ((e->state () & Qt::ControlButton) != 0 ? ControlButton : 0) |
    ((e->state () & Qt::AltButton) != 0 ? AltButton : 0);

  db::DPoint p = trans ().inverted () * db::DPoint (e->pos ().x (), height () - 1 - e->pos ().y ());

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    done = ((*g)->m_enabled && (*g)->mouse_press_event (p, buttons, true));
    g = gg;
  }

  if (! done && mp_active_service) {
    done = (mp_active_service->m_enabled && mp_active_service->mouse_press_event (p, buttons, true));
  }

  service_iterator svc = begin_services ();
  while (svc != end_services () && !done) {
    service_iterator next = svc;
    ++next;
    done = (svc->m_enabled && svc->mouse_press_event (p, buttons, false));
    svc = next;
  }

  if (! done) {
    mouse_press_event (p, buttons);
  }
}

void 
ViewObjectCanvas::mouseReleaseEvent (QMouseEvent *e)
{
  bool done = false; 
  bool click = (abs (e->pos ().x () - m_mouse_pressed.x ()) < 2 && abs (e->pos ().y () - m_mouse_pressed.y ()) < 2);

  m_mouse_pos = e->pos ();
  db::DPoint p = trans ().inverted () * db::DPoint (e->pos ().x (), height () - 1 - e->pos ().y ());

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    done = ((*g)->m_enabled && (*g)->mouse_release_event (p, click, true));
    g = gg;
  }

  if (! done && mp_active_service) {
    done = (mp_active_service->m_enabled && mp_active_service->mouse_release_event (p, click, true));
  }

  service_iterator svc = begin_services ();
  while (svc != end_services () && !done) {
    service_iterator next = svc;
    ++next;
    done = (svc->m_enabled && svc->mouse_release_event (p, click, false));
    svc = next;
  }

  if (! done) {
    mouse_release_event (p, click);
  }
}

void
ViewObjectCanvas::trans (const db::DMagTrans &trans)
{
  if (trans != m_trans) {

    m_trans = trans;

    //  call update on the objects
    for (object_iterator obj = begin_objects (); obj != end_objects (); ++obj) {
      obj->update ();
    }

    //  issue a move event in order to reposition the mouse in the new coordinate system
    do_mouse_move (m_mouse_pos);

  }
}

db::DBox 
ViewObjectCanvas::viewport () const
{
  db::DPoint p1 = trans ().inverted () * db::DPoint (0, 0);
  db::DPoint p2 = trans ().inverted () * db::DPoint (width () - 1, height () - 1);
  return db::DBox (p1, p2);
}
void 
ViewObjectCanvas::drag_cancel ()
{
  for (service_iterator svc = begin_services (); svc != end_services (); ++svc) {
    svc->drag_cancel ();
  }
}

void 
ViewObjectCanvas::do_render_bg (QImage &img)
{
  m_needs_update_bg = false;

  for (object_iterator obj = begin_objects (); obj != end_objects (); ++obj) {
    if (obj->is_visible ()) {
      obj->render_bg (img);
    }
  }
}

void 
ViewObjectCanvas::do_render (bool st)
{
  if (st) {
    m_needs_update_static = false;
  }

  for (object_iterator obj = begin_objects (); obj != end_objects (); ++obj) {
    if (obj->m_static == st && obj->is_visible ()) {
      obj->render ();
    }
  }
}

void 
ViewObjectCanvas::grab_mouse (ViewService *obj, bool a)
{
  obj->m_abs_grab = a;
  m_grabbed.push_front (obj);
  if (a) {
    grabMouse ();
  }
}

void 
ViewObjectCanvas::ungrab_mouse (ViewService *obj)
{
  bool any_abs_grabbed = false;

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    if (*g == obj) {
      m_grabbed.erase (g);
    } else if ((*g)->m_abs_grab) {
      any_abs_grabbed = true;
    }
    g = gg;
  }

  if (! any_abs_grabbed && mouseGrabber () == this) {
    releaseMouse ();
  }
}

void 
ViewObjectCanvas::freeze (ViewObject *obj)
{
  if (! obj->m_static) {
    obj->m_static = true;
    m_needs_update_static = true;
    //  no update needed since the display will not change through this.
  }
}

void 
ViewObjectCanvas::thaw (ViewObject *obj)
{
  if (obj->m_static) {
    obj->m_static = false;
    m_needs_update_static = true;
    //  no update needed since the display will not change through this.
  }
}

void
ViewObjectCanvas::touch ()
{
  m_needs_update_static = true;
  update ();
}

void
ViewObjectCanvas::touch_bg ()
{
  m_needs_update_bg = true;
  update ();
}

void 
ViewObjectCanvas::added (tl::CollectionBase *coll, tl::Collectable *)
{
  if (coll == & m_objects) {
    touch ();
  }
}

void 
ViewObjectCanvas::removed (tl::CollectionBase *coll, tl::Collectable *obj)
{
  if (coll == & m_objects) {
    touch ();
    update ();
  }
  if (coll == & m_services) {
    if (mp_active_service == obj) {
      mp_active_service = 0;
    }
  }
}

}

