
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layZoomBox.h"
#include "layRubberBox.h"
#include "layLayoutView.h"

namespace lay
{

// -------------------------------------------------------------
//  ZoomService implementation

ZoomService::ZoomService (lay::LayoutCanvas *canvas, lay::LayoutView *view)
  : lay::ViewService (canvas), 
    mp_view (view),
    mp_box (0)
{ }

ZoomService::~ZoomService ()
{
  if (mp_box) {
    delete mp_box;
    mp_box = 0;
  }
}

void 
ZoomService::set_color (QColor color)
{
  m_color = color.rgb ();
  if (mp_box) {
    mp_box->set_color (m_color);
  }
}

bool 
ZoomService::mouse_move_event (const db::DPoint &p, bool prio) 
{
  if (prio && mp_box) {
    m_p2 = p;
    mp_box->set_points (m_p1, m_p2);
  }
  return false;
}

bool 
ZoomService::mouse_press_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (! prio && ! mp_box && (buttons & lay::RightButton) != 0) {
    mp_view->stop_redraw (); // @@@ how to restart if zoom is aborted?
    begin (p);
    return true;
  }

  return false;
}

bool 
ZoomService::mouse_release_event (const db::DPoint &p, bool click, bool prio) 
{ 
  if (prio && mp_box) {

    canvas ()->ungrab_mouse (this);

    delete mp_box;
    mp_box = 0;

    db::DBox vp = canvas ()->viewport ();
    db::DMagTrans trans = canvas ()->trans ();

    if (mp_view && vp.contains (p) && vp.width () > 0 && vp.height () > 0) {

      //  compute the distance of the new box center to the old center
      double dx = (m_p1.x () + m_p2.x ()) * 0.5 - vp.center ().x ();
      double dy = (m_p1.y () + m_p2.y ()) * 0.5 - vp.center ().y ();

      if (! click) {

        // zoom
        double fx = fabs (m_p2.x () - m_p1.x ()) / vp.width ();
        double fy = fabs (m_p2.y () - m_p1.y ()) / vp.height ();
        double f = std::max (fx, fy);

        if (m_p1.x () > m_p2.x () || m_p1.y () < m_p2.y ()) {
          f = 1 / f;
        } 

        dx += vp.width () * (1 - f) * 0.5;
        dy += vp.height () * (1 - f) * 0.5;

        db::DMagTrans tnew (db::DTrans (trans.t1.rot (), trans.t1.disp () * (1.0 / f - 1.0) + trans * db::DPoint (-dx / f, -dy / f)), db::DMag (trans.t2.factor () / f));
        mp_view->zoom_to (tnew);

      } else {

        // pan: center the view to the mouse position
        db::DMagTrans trans = canvas ()->trans ();
        db::DMagTrans tnew (db::DTrans (trans.t1.rot (), trans * db::DPoint (-dx, -dy)), db::DMag (trans.t2.factor ()));
        mp_view->zoom_to (tnew);

      }

    }

  }

  return false;
}

void 
ZoomService::begin (const db::DPoint &pos)
{ 
  if (mp_box) {
    delete mp_box;
  }

  m_p1 = pos;
  m_p2 = pos;
  mp_box = new lay::RubberBox (canvas (), m_color, pos, pos);

  canvas ()->grab_mouse (this, true);
}

}

