
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_tlObserver
#define HDR_tlObserver

#include <list>

namespace tl
{

class Observed;

/**
 *  @brief A simple observer pattern implementation
 */
class Observer 
{
public:
  /**
   *  @brief Constructor
   */
  Observer ();

  /**
   *  @brief The destructor
   *
   *  Will unregister the observer from the observed object's observer list.
   */ 
  virtual ~Observer ();

  /**
   *  @brief The pointer to the observed object
   */
  Observed *observed () const
  {
    return m_observed;
  }

  /**
   *  @brief The simple signal even handler 
   */
  virtual void signal_observed ();

  /**
   *  @brief The simple signal even handler with an int argument
   */
  virtual void signal_observed_int (int arg);

  /**
   *  @brief The simple signal even handler with an void* argument
   */
  virtual void signal_observed_voidp (void *arg);

  /**
   *  @brief Query the state of the observation
   *
   *  @return true, if the observer does observe something
   */
  bool observes () const;

private:
  friend class Observed;

  std::list<Observer *>::iterator m_ptr;
  Observed *m_observed;

  std::list<Observer *>::iterator ptr () const;
  void observer_attach (Observed *observed, std::list<Observer *>::iterator ptr);
};

inline std::list<Observer *>::iterator 
Observer::ptr () const
{
  return m_ptr;
}

class Observed
{
public:
  /**
   *  @basic Destructor: will detach from all observers
   */
  virtual ~Observed ();

  /**
   *  @basic Observer management
   */
  void add_observer (Observer &observer);

  /**
   *  @remove An observer object
   */
  void remove_observer (Observer &observer);

  /**
   *  @brief Signal to all observers
   *
   *  Will call the signal method on all observers
   */
  void signal_observers () const;

  /**
   *  @brief Signal to all observers
   *
   *  Will call the signal(int) method on all observers
   */
  void signal_observers (int arg) const;

  /**
   *  @brief Signal to all observers
   *
   *  Will call the signal(void *) method on all observers
   */
  void signal_observers (void *arg) const;

  /**
   *  @brief Count the number of observers attached to this observed object
   */
  size_t observers () const;

private:
  std::list<Observer *> m_observers;
};

} // namespace tl

#endif
